@@ -35,54 +35,22 @@
 #include <vector>
 
 #include "APmain.hh"
+#include "Common.hh"
+#include "PrintOperator.hh"
 #include "Svar_DB.hh"
 #include "Svar_signals.hh"
 
 #ifdef DYNAMIC_LOG_WANTED
+extern bool LOG_startup;
 extern bool LOG_shared_variables;
+extern bool LOG_Svar_DB_signals;
+bool LOG_startup = false;
 bool LOG_shared_variables = false;
+bool LOG_Svar_DB_signals = false;
 #else
 #endif
 
-//-----------------------------------------------------------------------------
-class CERR_Out : public filebuf
-{
-public:
-   CERR_Out() : print_sema_held(false) {}
-
-   virtual int overflow(int c);
-
-   filebuf * use()   { used = true;   return this; }
-
-   bool print_sema_held;
-
-   static bool used;   // set when CERR is constructed
-
-} CERR_filebuf;
-bool CERR_Out::used = false;
-
-//-----------------------------------------------------------------------------
-int
-CERR_Out::overflow(int c)
-{
-   if (!print_sema_held)
-      {
-        Svar_DB::start_print(LOC);
-        print_sema_held = true;
-      }
-
-   cerr << (char)c;
-
-   if (c == '\n')
-      {
-        Svar_DB::end_print(LOC);
-        print_sema_held = false;
-      }
-
-   return 0;
-}
-
-ostream CERR(CERR_filebuf.use());
+extern const char * prog_name();
 
 //-----------------------------------------------------------------------------
 
@@ -92,8 +60,6 @@ int event_port = 0;
 const char * prog = 0;
 char pref[FILENAME_MAX + 20];
 
-pid_t parent_pid = 0;
-
 /// coupled variables
 vector<Coupled_var> coupled_vars;
 
@@ -107,7 +73,7 @@ AP_num3 ProcessorID::id(NO_AP, AP_NULL, AP_NULL);
 
 ostream & get_CERR()
 {
-   return CERR_Out::used ? CERR : cerr;
+   return cerr;
 }
 //-----------------------------------------------------------------------------
 uint64_t
@@ -147,17 +113,24 @@ print_vars(ostream & out)
    for (int c = 0; c < coupled_vars.size(); ++c)
        {
          Coupled_var & cv = coupled_vars[c];
-         CERR << "   key: 0x" << hex << cv.key << dec << " ";
-         offered_SVAR * svar = Svar_DB::find_var(cv.key);
-         if (svar)   svar->print_name(CERR) << endl;
-         else     CERR << "(unknown var)" << endl;
+         get_CERR() << "   key: 0x" << hex << cv.key << dec << " ";
+         const uint32_t * varname = Svar_DB::get_varname(cv.key);
+         if (varname)
+            {
+              while (*varname)   get_CERR() << (Unicode)(*varname++);
+            }
+         else
+            {
+              get_CERR() << "(unknown var)";
+            }
+         get_CERR() << endl;
        }
 }
 //-----------------------------------------------------------------------------
 void
 do_Assert(const char * cond, const char * fun, const char * file, int line)
 {
-   CERR << "Assertion '" << cond << "' failed at "
+   get_CERR() << "Assertion '" << cond << "' failed at "
         << file << ":" << line << endl;
 
    assert(0);
@@ -166,18 +139,13 @@ do_Assert(const char * cond, const char * fun, const char * file, int line)
 void
 control_C(int)
 {
-   if (verbose)   CERR << pref << " unregistering processor "
-                       << ProcessorID::get_id().proc << endl;
+AP_num3 this_proc = ProcessorID::get_id();
+   if (verbose)   get_CERR() << pref << " unregistering processor "
+                       << this_proc << endl;
 
-   {
-     Svar_partner this_proc;
-     this_proc.clear();
-     this_proc.id = ProcessorID::get_id();
-     this_proc.pid = getpid();
-     Svar_DB::unregister_processor(this_proc);
-   }
+const TCP_socket tcp = Svar_DB::get_DB_tcp();
 
-   if (verbose)   CERR << pref << " done (got ^C)" << endl;
+   if (verbose)   get_CERR() << pref << " done (got ^C)" << endl;
 
    exit(0);
 }
@@ -188,7 +156,7 @@ static struct sigaction new_ctl_C_action;
 //-----------------------------------------------------------------------------
 int usage()
 {
-   CERR
+   get_CERR()
 << "Usage:"                            << endl
 << prog << " [options]"                << endl
 <<                                        endl
@@ -196,7 +164,6 @@ int usage()
 << "    --auto        automatically started (exit after last retract)"  << endl
 << "    --id num      run as processor num"  << endl
 << "    --par num     run as child of num"  << endl
-<< "    --ppid num    terminate if process num has died" << endl
 << "    --gra num     run as grandchild of num"  << endl
 << "    -h, --help    print this help" << endl
 << "    -v            verbose"         << endl
@@ -205,15 +172,12 @@ int usage()
    return 1;
 }
 //-----------------------------------------------------------------------------
-void
-got_event(uint32_t event, SV_key key)
+TCP_socket tcp2 = NO_TCP_SOCKET;
+
+TCP_socket
+get_TCP_for_key(SV_key key)
 {
-   if (event_port)   // event timer is running
-      {
-        UdpClientSocket sock(LOC, event_port);
-        GOT_EVENT_c signal(sock, key, event);
-        event_port = 0;
-      }
+   return tcp2;
 }
 //-----------------------------------------------------------------------------
 int
@@ -223,6 +187,17 @@ bool need_help = false;
 bool auto_started = false;
 
    prog = argv[0];
+char bin_path[FILENAME_MAX];
+   strncpy(bin_path, prog, sizeof(bin_path));
+
+char * slash = strrchr(bin_path, '/');
+   if (slash)
+      {
+        prog = slash + 1;
+        *slash = 0;
+      }
+
+   Svar_DB::init(bin_path, prog, false, true);
 
    if (strrchr(prog, '/'))   prog = strrchr(prog, '/') + 1;
    snprintf(pref, sizeof(pref) - 1, "%s(%u) ", prog, getpid());
@@ -243,14 +218,12 @@ bool auto_started = false;
                           { ProcessorID::set_own_ID(AP_num(atoi(val))); ++a; }
          else if (!strcmp(opt, "--par") && val)
                          { ProcessorID::set_parent_ID(AP_num(atoi(val))); ++a; }
-         else if (!strcmp(opt, "--ppid") && val)
-                        { parent_pid = atoi(val); ++a; }
          else if (!strcmp(opt, "--gra") && val)
                          { ProcessorID::set_grand_ID(AP_num(atoi(val))); ++a; }
          else
             {
-              CERR << pref << ": Bad command line option "
-                   << argv[a] << endl;
+              get_CERR() << pref << ": Bad command line option '"
+                   << argv[a] << "'" << endl;
               need_help = true;
             }
        }
@@ -267,117 +240,108 @@ bool auto_started = false;
    // register as e.g. AP210-port. We do this BEFORE closing stdout so that
    // caller of our parent waits until we have closed stdout
    //
-UdpServerSocket sock(LOC, 0);
-   if (verbose)  sock.set_debug(CERR);
 
    memset(&new_ctl_C_action, 0, sizeof(struct sigaction));
    new_ctl_C_action.sa_handler = &control_C;
    sigaction(SIGTERM, &new_ctl_C_action, &old_ctl_C_action);
 
-Svar_partner this_proc;
-   this_proc.clear();
-   this_proc.id = ProcessorID::get_id();
-   this_proc.pid = getpid();
-   this_proc.port = sock.get_local_port();
+Svar_partner this_proc(ProcessorID::get_id(), NO_TCP_SOCKET);
 
-   Svar_DB::register_processor(this_proc);
+   // tcp is the "normal" connection to APserver as server
+   // tcp2 is a special connection on which APserver (as client) can 
+   // send events to this process.
+   //
+const TCP_socket tcp = Svar_DB::get_DB_tcp();
+
+   if (ProcessorID::get_id().proc < AP_FIRST_USER)
+      tcp2 = Svar_DB::connect_to_APserver(0, prog_name(), verbose);
+
+   if (tcp == NO_TCP_SOCKET || tcp2 == NO_TCP_SOCKET)
+      {
+        cerr << prog_name() << ":*** connection to APserver failed" << endl;
+        return 3;
+      }
+
+const int nfds = 1 + tcp2;
+
+string progname(prog_name());
+
+      { REGISTER_PROCESSOR_c request(tcp, this_proc.id.proc,
+                                          this_proc.id.parent,
+                                          this_proc.id.grand,
+                                          0, progname);
+      }
+
+   if (this_proc.id.proc < AP_FIRST_USER)
+      { progname += "-EV";
+         REGISTER_PROCESSOR_c request(tcp2, this_proc.id.proc,
+                                            this_proc.id.parent,
+                                            this_proc.id.grand,
+                                            1, progname);
+      }
 
    fclose(stdout);   // cause getc() of caller to return EOF !
 
    for (bool goon = true; goon;)
        {
-         uint8_t buffer[Signal_base::get_class_size()];
-         const Signal_base * signal = Signal_base::recv(sock, buffer, 10000);
+         uint8_t buff[MAX_SIGNAL_CLASS_SIZE + 40000];
+         char * del = 0;
+         const Signal_base * signal = Signal_base::recv_TCP(tcp2, (char *)buff,
+                                                         sizeof(buff), del, 0);
+
          if (signal == 0)   // no signal for 10 seconds
             {
-              // if our parent has died, then we are done
-              //
-              if (parent_pid && !Svar_partner::pid_alive(parent_pid))
-                 {
-                   goon = false;
-                   if (verbose)   CERR << AP_NAME
-                                       << " done (parent died)" << endl;
-                 }
+              goon = false;
+              if (verbose)   get_CERR() << AP_NAME
+                                        << " done (parent died)" << endl;
               continue;
             }
 
+#if 0
+#if AP_NUM == 0
+cerr << "APnnn got " << signal->get_sigName() << endl;
+#endif
+#endif
          switch(signal->get_sigID())
             {
-            case sid_DISCONNECT:   // master (local APL interpreter) disconnect
-                 if (verbose)   CERR << AP_NAME << " got DISCONNECT" << endl;
-                 goon = false;
-                 continue;
-
-            case sid_NEW_VARIABLE:     // a new (not yet matched) offer
-                 if (verbose)   CERR << AP_NAME << " got NEW_VARIABLE" << endl;
-                 {
-                   const SV_key key = signal->get__NEW_VARIABLE__key();
-                   offered_SVAR * svar = Svar_DB::find_var(key);
-                  if (svar == 0)
-                     {
-                       CERR << "Could not find svar for key "
-                            << key << " at " << LOC << endl;
-                       continue;
-                     }
-
-                   if (! is_valid_varname(svar->varname))
-                      {
-                       CERR << "Bad varname ";
-                       svar->print_name(CERR) << " at " << LOC << endl;
-                       continue;
-                      }
-
-                   add_var(key);
-                 }
-                 continue;
-
             case sid_MAKE_OFFER:        // a new offer from a peer
-                 if (verbose)   CERR << AP_NAME << " got MAKE_OFFER" << endl;
+                 if (verbose)   get_CERR() << AP_NAME
+                                           << " got MAKE_OFFER" << endl;
                  {
                    const SV_key key = signal->get__MAKE_OFFER__key();
-                   offered_SVAR * svar = Svar_DB::find_var(key);
-                  if (svar == 0)
+                   const uint32_t * varname = Svar_DB::get_varname(key);
+                  if (varname == 0)
                      {
-                       CERR << "Could not find svar for key "
+                       get_CERR() << "Could not find svar for key "
                             << key << " at " << LOC << endl;
-                       continue;
+                       break;
                      }
 
-                   if (! is_valid_varname(svar->varname))
+                   if (! is_valid_varname(varname))
                       {
-                       CERR << "Bad varname ";
-                       svar->print_name(CERR) << " at " << LOC << endl;
-                       continue;
+                       get_CERR() << "Bad varname ";
+                        while (*varname)   get_CERR() << (Unicode)(*varname++);
+                        get_CERR() << " at " << LOC << endl;
+                       break;
                       }
 
                    // make a counter offer (APs 100 and 210) or not (APnnn)
                    //
-                   if (!make_counter_offer(key))   continue;   // APnnn
+                   if (!make_counter_offer(key))   break;   // APnnn
 
-                   SV_Coupling coupling = NO_COUPLING;
-                   Svar_DB::match_or_make(svar->varname, svar->offering.id, 
-                                          this_proc, coupling);
+                   const AP_num3 offering_id = Svar_DB::find_offering_id(key);
+
+                   Svar_DB::match_or_make(varname, offering_id, this_proc);
 
                    add_var(key);
 
-                   svar->offering.flags &= ~OSV_OFFER_SENT;
-                   svar->set_state(true, LOC);
+                   Svar_DB::set_state(key, true, LOC);
                  }
-                 continue;
-
-            case sid_OFFER_MATCHED:     // our offer was matched
-                 if (verbose)   CERR << AP_NAME << " got OFFER_MATCHED" << endl;
-                 {
-                   const SV_key key = signal->get__OFFER_MATCHED__key();
-
-                   add_var(key);
-                   Svar_DB::add_event(SVE_OFFER_MATCHED,
-                                      ProcessorID::get_id(), key);
-                 }
-                 continue;
+                 break;
 
             case sid_RETRACT_OFFER:     // ⎕SVR varname
-                 if (verbose)   CERR << AP_NAME << " got RETRACT_OFFER" << endl;
+                 verbose && get_CERR() << AP_NAME
+                                       << " got RETRACT_OFFER" << endl;
                  {
                    const SV_key key = signal->get__RETRACT_OFFER__key();
                    for (int c = 0; c < coupled_vars.size(); ++c)
@@ -390,20 +354,21 @@ Svar_partner this_proc;
 
                               if (coupled_vars.size() == 0 && auto_started)
                                  {
-                                   if (verbose)      CERR << AP_NAME << " done"
+                                   if (verbose)      get_CERR() << AP_NAME << " done"
                                       " (last variable retracted)" << endl;
                                    goon  = false;
                                    break;
                                  }
                            }
                        }
-                   Svar_DB::add_event(SVE_OFFER_RETRACT,
-                                      ProcessorID::get_id(), key);
+                   Svar_DB::add_event(key, ProcessorID::get_id(),
+                                      SVE_OFFER_RETRACT);
                  }
-                 continue;
+                 break;
 
             case sid_GET_VALUE:
-                 if (verbose)   CERR << AP_NAME << " got GET_VALUE" << endl;
+                 if (verbose)   get_CERR() << AP_NAME
+                                           << " got GET_VALUE" << endl;
                  {
                    const SV_key key = signal->get__GET_VALUE__key();
                    APL_error_code error = E_VALUE_ERROR;
@@ -422,19 +387,20 @@ Svar_partner this_proc;
 
                    if (!found)
                       {
-                        CERR << "Key 0x" << hex << key << dec
+                        get_CERR() << "Key 0x" << hex << key << dec
                              << " not found. Variables are:"
                              << endl;
-                        print_vars(CERR);
+                        print_vars(get_CERR());
                         error_loc = LOC;   error = E_VALUE_ERROR;
                       }
 
-                   VALUE_IS_c response(sock, key, error, error_loc, data);
+                   VALUE_IS_c response(tcp2, key, error, error_loc, data);
                  }
-                 continue;
+                 break;
 
             case sid_ASSIGN_VALUE:
-                 if (verbose)   CERR << AP_NAME << " got ASSIGN_VALUE" << endl;
+                 verbose && get_CERR() << AP_NAME
+                                       << " got ASSIGN_VALUE" << endl;
                  {
                    const SV_key key = signal->get__ASSIGN_VALUE__key();
                    APL_error_code error = E_VALUE_ERROR;
@@ -446,61 +412,69 @@ Svar_partner this_proc;
                             {
                               found = true;
                               error_loc = LOC;   error = assign_value(cv,
-                                           signal->get__ASSIGN_VALUE__value());
+                                         signal->get__ASSIGN_VALUE__cdr_value());
                               break;
                            }
                        }
 
                    if (!found)
                       {
-                        CERR << "Key 0x" << hex << key << dec
+                        get_CERR() << "Key 0x" << hex << key << dec
                              << " not found. Variables are:"
                              << endl;
-                        print_vars(CERR);
+                        print_vars(get_CERR());
                         error_loc = LOC;   error = E_VALUE_ERROR;
                       }
 
-                   ASSIGNED_c response(sock, key, error, error_loc);
+                   SVAR_ASSIGNED_c response(tcp, key, error, error_loc);
                  }
-                 continue;
+                 break;
 
-            case sid_START_EVENT_REPORTING:
-                 if (verbose)
-                    CERR << AP_NAME << " got START_EVENT_REPORTING" << endl;
-
-                 event_port = signal->get__START_EVENT_REPORTING__event_port();
-                 {
-                   Svar_event events = SVE_NO_EVENTS;
-                   SV_key key = Svar_DB::get_events(events,
-                                                    ProcessorID::get_id());
-                   if (events || key)   got_event(events, key);
-                 }
-                 continue;
-
-            case sid_STOP_EVENT_REPORTING:
-                 if (verbose)
-                    CERR << AP_NAME << " got STOP_EVENT_REPORTING" << endl;
-                 event_port = 0;
-                 continue;
-
-            case sid_NEW_EVENT:
-                 if (verbose)   CERR << AP_NAME << " got NEW_EVENT "
-                    << event_name(Svar_event(signal->get__NEW_EVENT__event()))
-                    << endl;
-
-                 // someone has created a new event with Svar_DB::add_event()
-                 // We forward it to our interpreter (if ⎕SVE is running)
-                 //
-                 got_event(signal->get__NEW_EVENT__event(),
-                           signal->get__NEW_EVENT__key());
-                 continue;
-
-            default: CERR << pref << ": bad signal ID "
-                          << signal->get_sigID() << endl;
+            default: get_CERR() << pref << ": bad signal ID "
+                          << signal->get_sigID() << " ("
+                          << signal->get_sigName() << ")" << endl;
           }
+
+         if (del)   delete del;
        }
 
-   Svar_DB::unregister_processor(this_proc);
    return 0;
 }
 //-----------------------------------------------------------------------------
+ostream & operator << (ostream & out, const AP_num3 & ap3)
+{
+   return out << ap3.proc << "." << ap3.parent << "." << ap3.grand;
+}
+//-----------------------------------------------------------------------------
+ostream &
+operator << (ostream & os, Unicode uni)
+{
+   if (uni < 0x80)      return os << (char)uni;
+
+   if (uni < 0x800)     return os << (char)(0xC0 | (uni >> 6))
+                                  << (char)(0x80 | (uni & 0x3F));
+
+   if (uni < 0x10000)    return os << (char)(0xE0 | (uni >> 12))
+                                   << (char)(0x80 | (uni >>  6 & 0x3F))
+                                   << (char)(0x80 | (uni       & 0x3F));
+
+   if (uni < 0x200000)   return os << (char)(0xF0 | (uni >> 18))
+                                   << (char)(0x80 | (uni >> 12 & 0x3F))
+                                   << (char)(0x80 | (uni >>  6 & 0x3F))
+                                   << (char)(0x80 | (uni       & 0x3F));
+
+   if (uni < 0x4000000)  return os << (char)(0xF8 | (uni >> 24))
+                                   << (char)(0x80 | (uni >> 18 & 0x3F))
+                                   << (char)(0x80 | (uni >> 12 & 0x3F))
+                                   << (char)(0x80 | (uni >>  6 & 0x3F))
+                                   << (char)(0x80 | (uni       & 0x3F));
+
+   return os << (char)(0xFC | (uni >> 30))
+             << (char)(0x80 | (uni >> 24 & 0x3F))
+             << (char)(0x80 | (uni >> 18 & 0x3F))
+             << (char)(0x80 | (uni >> 12 & 0x3F))
+             << (char)(0x80 | (uni >>  6 & 0x3F))
+             << (char)(0x80 | (uni       & 0x3F));
+}
+//-----------------------------------------------------------------------------
+
