@@ -75,7 +75,8 @@ bool LOG_shared_variables = false;
 AP_num3 ProcessorID::id;
 
 //-----------------------------------------------------------------------------
-const char * prog_name()
+const char *
+prog_name()
 {
    return "APserver";
 }
@@ -616,7 +617,6 @@ AP3_fd * ap_fd = 0;
              }
              return;
 
-
         case sid_GET_OFFERED_VARS:
              {
                const AP_num to_proc = (AP_num)
@@ -627,7 +627,7 @@ AP3_fd * ap_fd = 0;
                db.get_offered_variables(to_proc, from_proc, varnames);
                string svars((const char *)varnames.data(),
                             varnames.size()*sizeof(uint32_t));
-               OFFERING_PROCS_ARE_c(fd, svars);
+               OFFERED_VARS_ARE_c(fd, svars);
              }
              return;
 
@@ -1063,7 +1063,7 @@ int janitor = 0;
         return argc;
       }
 
-   // enable ^C and ^\ when in debig mode
+   // enable ^C and ^\ when in debug mode
    //
    memset(&new_control_C_action, 0, sizeof(struct sigaction));
    memset(&new_control_BSL_action, 0, sizeof(struct sigaction));
