@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -32,10 +32,10 @@
 #include "CharCell.hh"
 #include "ComplexCell.hh"
 #include "Executable.hh"
-#include "IntCell.hh"
 #include "FloatCell.hh"
 #include "Function.hh"
 #include "IndexExpr.hh"
+#include "IntCell.hh"
 #include "LvalCell.hh"
 #include "NativeFunction.hh"
 #include "Output.hh"
@@ -739,8 +739,8 @@ CERR << "LVAL CELL in " << p << " at " LOC << endl;
 
    // save certain system variables
    //
-#define rw_sv_def(x) save_symbol(Workspace::get_v_ ## x());
-#define ro_sv_def(x) save_symbol(Workspace::get_v_ ## x());
+#define rw_sv_def(x, txt) save_symbol(Workspace::get_v_ ## x());
+#define ro_sv_def(x, txt) save_symbol(Workspace::get_v_ ## x());
 #include "SystemVariable.def"
 
    // save state indicator
@@ -776,7 +776,7 @@ CERR << "LVAL CELL in " << p << " at " LOC << endl;
    do_indent();
 
    // write closing tag and a few 0's so that string functions
-   // can be used on mmaped file.
+   // can be used on the mmaped file.
    //
    out << "</Workspace>" << endl
        << char(0) << char(0) <<char(0) <<char(0) << endl;
@@ -798,7 +798,8 @@ XML_Loading_Archive::XML_Loading_Archive(const char * _filename, int & dump_fd)
      protection(false),
      reading_vids(false),
      have_allowed_objects(false),
-     filename(_filename)
+     filename(_filename),
+     file_is_complete(false)
 {
    Log(LOG_archive)   CERR << "Loading file " << filename << endl;
 
@@ -872,7 +873,7 @@ XML_Loading_Archive::skip_to_tag(const char * tag)
    for (;;)
       {
          if (next_tag(LOC))   return true;
-         if (is_tag(tag))   break;
+         if (is_tag(tag))     break;
       }
 
    return false;
@@ -883,7 +884,7 @@ XML_Loading_Archive::read_vids()
 {
    reset();   // skips to <Workspace>
    reading_vids = true;
-   read_Workspace();
+   read_Workspace(true);
    reading_vids = false;
 
    reset();
@@ -1050,7 +1051,7 @@ again:
 }
 //-----------------------------------------------------------------------------
 void
-XML_Loading_Archive::read_Workspace()
+XML_Loading_Archive::read_Workspace(bool silent)
 {
    expect_tag("Workspace", LOC);
 
@@ -1067,11 +1068,17 @@ int sec   = find_int_attr("second",   false, 10);
    sec  += offset          % 60;
    min  += (offset /   60) % 60;
    hour += (offset / 3600) % 60;
-   if (sec  >= 60)   { sec  -= 60;   ++min;  }
-   if (min  >= 60)   { min  -= 60;   ++hour; }
-   if (hour >= 24)   { hour -= 24;   ++day;  }
+   if      (sec  >= 60)   { sec  -= 60;   ++min;  }
+   else if (sec  <  0)    { sec  += 60;   --min;  }
+
+   if      (min  >= 60)   { min  -= 60;   ++hour; }
+   else if (min  <   0)   { min  += 60;   --hour; }
+
+   if      (hour >= 24)   { hour -= 24;   ++day;  }
+   else if (hour <   0)   { hour += 24;   --day;  }
 
 bool next_month = false;
+bool prev_month = false;
    switch(day)
       {
         case 32: next_month = true;
@@ -1084,23 +1091,64 @@ bool next_month = false;
         case 30: if (mon == 2)   next_month = true;
                  break;
 
-        case 29: if (mon != 2)         break;               // not februry
-                 if (year & 3)         next_month = true;   // not leap yer
+        case 29: if (mon != 2)         break;               // not february
+                 if (year & 3)         next_month = true;   // not leap year
                  // the above fails if someone loads a workspace that
-                 // was saved around midnight on 2/28/2100. Bad luck!
+                 // was saved around midnight on 2/28/2100. Dont do that!
+                 break;
+
+        case 0:  prev_month = true;
                  break;
 
         default: break;
       }
 
-   if (next_month)   { day = 1; ++mon;  }
-   if (mon > 12)     { mon = 1; ++year; }
+   if      (next_month)   { day = 1; ++mon;  }
+   else if (prev_month)
+           {
+             day = 31;   --mon;
+             if (mon == 4 || mon == 6 || mon == 9 || mon == 11)   day = 30;
+             else if (mon == 2)                 day = (year & 3) ? 28 : 29;
+           }
+
+   if      (mon > 12)   { mon =  1; ++year; }
+   else if (mon <  1)   { mon = 12; --year; }
 
    Log(LOG_archive)   CERR << "read_Workspace() " << endl;
 
-const char * tag_order[] = { "Value",          "Ravel",
-                             "SymbolTable",    "Symbol",
-                             "StateIndicator", "/Workspace", 0 };
+   // quick check that the file is complete
+   //
+   for (const UTF8 * c = file_end - 12; (c > data) && (c > file_end - 200); --c)
+       {
+         if (!strncmp((const char *)c, "</Workspace>", 12))
+            {
+              file_is_complete = true;
+              break;
+            }
+       }
+
+   if (!file_is_complete && !copying)
+      {
+        CERR <<
+"*** workspace file " << filename << endl <<
+"    seems to be incomplete (possibly caused by a crash on )SAVE?)\n" 
+"    You may still be able to )COPY from it.\n"
+"\nNOT COPIED" << endl;
+        return;
+      }
+
+   // the order in which tags are written to the xml file
+   //
+const char * tag_order[] =
+{
+  "Value",
+  "Ravel",
+  "SymbolTable",
+  "Symbol",
+  "StateIndicator",
+  "/Workspace",
+  0
+};
 const char ** tag_pos = tag_order;
 
    for (;;)
@@ -1123,6 +1171,7 @@ const char ** tag_pos = tag_order;
          else if (is_tag("Ravel"))            read_Ravel();
          else if (is_tag("SymbolTable"))      read_SymbolTable();
          else if (is_tag("Symbol"))           read_Symbol();
+         else if (copying)                    break;
          else if (is_tag("StateIndicator"))   read_StateIndicator();
          else if (is_tag("/Workspace"))       break;
          else    /* complain */               expect_tag("UNEXPECTED", LOC);
@@ -1136,7 +1185,8 @@ const char ** tag_pos = tag_order;
    if (reading_vids)   return;
 
 const char * tz_sign = (offset < 0) ? "" : "+";
-   COUT << "SAVED "
+   if (!silent)   COUT
+        << "SAVED "
         << setfill('0') << year        << "-"
         << setw(2)      << mon         << "-"
         << setw(2)      << day         << " "
@@ -1224,13 +1274,14 @@ bool no_copy = false;   // assume the value is needed
       {
         Assert(vid == values.size());
 
-        Value_P val(new Value(sh_value, LOC));
+        Value_P val(sh_value, LOC);
         values.push_back(val);
       }
 }
 //-----------------------------------------------------------------------------
 void
-XML_Loading_Archive::read_Cells(Cell * & C, const UTF8 * & first)
+XML_Loading_Archive::read_Cells(Cell * & C, Value & C_owner,
+                                const UTF8 * & first)
 {
    while (*first <= ' ')   ++first;
 
@@ -1294,7 +1345,7 @@ const Unicode type = UTF8_string::toUni(first, len);
                const int vid = strtol((const char *)first, &end, 10);
                Assert(vid >= 0);
                Assert(vid < values.size());
-               new (C++) PointerCell(values[vid]);
+               new (C++) PointerCell(values[vid], C_owner);
                first = (const UTF8 *)end;
              }
              break;
@@ -1309,7 +1360,8 @@ const Unicode type = UTF8_string::toUni(first, len);
                Assert(*end == '[');   ++end;
                const ShapeItem offset = strtoll(end, &end, 16);
                Assert(*end == ']');   ++end;
-               new (C++) LvalCell(&values[vid]->get_ravel(offset));
+               new (C++) LvalCell(&values[vid]->get_ravel(offset),
+                                  values[vid].get());
                first = (const UTF8 *)end;
              }
              break;
@@ -1403,7 +1455,7 @@ Cell * end = C + count;
 
    while (C < end)
       {
-        read_Cells(C, cells);
+        read_Cells(C, val.getref(), cells);
         while (*cells <= ' ')   ++cells;   // skip trailing whitespace
       }
 
@@ -1442,10 +1494,10 @@ const unsigned int vid = find_int_attr("vid", false, 10);
    // some system variables are  saved for troubleshooting purposes, but
    // should not be loaded...
    //
-   if (symbol.is_readonly())                         return;
-   if (symbol.get_name().starts_iwith("⎕NLT"))       return;
-   if (symbol.get_name() == id_name(ID_Quad_SVE))    return;
-   if (symbol.get_name() == id_name(ID_Quad_SYL))    return;
+   if (symbol.is_readonly())                     return;
+   if (symbol.get_name().starts_iwith("⎕NLT"))   return;
+   if (symbol.get_Id() == ID::Quad_SVE)           return;
+   if (symbol.get_Id() == ID::Quad_SYL)           return;
 
    if (!values[vid])   return;   // value filtered out
 
@@ -1635,7 +1687,6 @@ bool no_copy = is_protected || (have_allowed_objects && !is_selected);
 
    // in a )COPY without dedicated objects only
    // ⎕CT, ⎕FC, ⎕IO, ⎕LX, ⎕PP, ⎕PR, and ⎕RL shall be copied
-   // shall be copied.
    //
    if (!have_allowed_objects       &&   // no dedicated object list
         copying                    &&   // )COPY
@@ -1655,6 +1706,18 @@ bool no_copy = is_protected || (have_allowed_objects && !is_selected);
         no_copy = true;
       }
 
+   // in a )LOAD silentlr ignore ⎕TZ
+   //
+   if (!copying                    &&   // )LOAD
+        Avec::is_quad(name_ucs[0]) &&   // ⎕xx
+        name_ucs.size() == 3       &&
+        ( name_ucs[1] == 'T' && name_ucs[2] == 'Z' )
+        )
+      {
+        skip_to_tag("/Symbol");
+        return;
+      }
+
    if (copying)
       {
         if (no_copy || (depth == 0))
@@ -1710,7 +1773,7 @@ XML_Loading_Archive::read_StateIndicator()
    if (copying)
       {
         skip_to_tag("/StateIndicator");
-          return;
+        return;
       }
 
    Log(LOG_archive)   CERR << "read_StateIndicator()" << endl;
@@ -1941,7 +2004,8 @@ const TokenTag tag = TokenTag(find_int_attr("tag", false, 16));
                if (fun_name == 0)   // primitive or Quad function
                   {
                     const int fun_id = find_int_attr("fun-id", false, 16);
-                    Function * sysfun = get_system_function(Id(fun_id));
+                    Function * sysfun = ID::get_system_function(ID::Id(fun_id));
+                    Assert1(sysfun);
                     new (&tloc.tok) Token(tag, sysfun);
                   }
                else
