@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -264,6 +264,7 @@ Avec::map_alternative_char(Unicode alt_av)
         case 0x2377: return AV_EPSILON_UBAR;     //  map ⍷ to ⋸
         case 0x25AF: return AV_Quad_Quad;        //  map ▯ to ⎕
         case 0x25E6: return AV_RING_OPERATOR;    //  map ◦ to ∘
+        case 0x2662: return AV_DIAMOND;          //  map ♢ to ◊
         case 0x26AA: return AV_CIRCLE;           //  map ⚪ to ○
         case 0x2A7D: return AV_LESS_OR_EQUAL;    //  map ⩽ to ≤
         case 0x2A7E: return AV_MORE_OR_EQUAL;    //  map ⩾ to ≥
@@ -295,8 +296,8 @@ Avec::is_known_char(Unicode av)
 bool
 Avec::need_UCS(Unicode uni)
 {
-   if (uni < UNI_ASCII_SPACE)    return true;   // ASCII control char
-   if (uni < UNI_ASCII_DELETE)   return false;   // printable ASCII char
+   if (is_control(uni))   return true;                      // ASCII control
+   if (uni >= 0 && uni < UNI_ASCII_DELETE)  return false;   // printable ASCII
 
 const CHT_Index idx = find_char(uni);
    if (idx == Invalid_CHT)   return true;           // char not in GNU APL's ⎕AV
@@ -328,7 +329,7 @@ Avec::is_digit(Unicode av)
 bool
 Avec::is_digit_or_space(Unicode av)
 {
-   return (av <= UNI_ASCII_SPACE || (av <= UNI_ASCII_9 && av >= UNI_ASCII_0));
+   return (is_white(av) || (av <= UNI_ASCII_9 && av >= UNI_ASCII_0));
 }
 //-----------------------------------------------------------------------------
 bool
