@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2014  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -29,7 +29,8 @@
 #include "Value.icc"
 #include "Workspace.hh"
                                             
-Bif_F12_FORMAT Bif_F12_FORMAT::fun;       // ⍕
+Bif_F12_FORMAT   Bif_F12_FORMAT::_fun;       // ⍕
+Bif_F12_FORMAT * Bif_F12_FORMAT::fun = &Bif_F12_FORMAT::_fun;
 
 //-----------------------------------------------------------------------------
 Token
@@ -41,7 +42,7 @@ Bif_F12_FORMAT::eval_B(Value_P B)
 
    if (B->is_empty())
       {
-        Value_P Z(new Value(B->get_shape(), LOC));
+        Value_P Z(B->get_shape(), LOC);
         new (&Z->get_ravel(0)) CharCell(UNI_ASCII_SPACE);
         return Token(TOK_APL_VALUE1, Z);
       }
@@ -62,7 +63,7 @@ Value_P Z;
         Assert(shape_Z.get_volume() == Z->element_count());
         Z->set_shape(shape_Z);
       }
-   else
+   else   // B->get_rank() is 0, 1, or 2
       {
         Z = monadic_format(B);
       }
@@ -73,7 +74,7 @@ Value_P Z;
    //
 const Depth depth = B->compute_depth();
 Shape sZ;
-   if (depth > 1)   // nested
+   if (depth > 1)   // B is nested ⊢ ⍴⍴R is 1 or 2
       {
         // the  examples in lrm contradict the text in lrm.
         //
@@ -97,9 +98,12 @@ Shape sZ;
              sZ = Z->get_shape();   // leave Z as is (2-dimensional)
            }
       }
-   else             // simple:                    result rank: 
+   else             // B is simple:                    result rank: ,1⌈⍴ρB
       {
-        if (B->get_rank() < 2)  sZ = Shape(Z->get_cols());
+        if (B->get_rank() < 2)
+           {
+             sZ = Shape(Z->get_cols());
+           }
         else
            {
              sZ = B->get_shape();
@@ -107,11 +111,11 @@ Shape sZ;
            }
       }
 
-        if (sZ != Z->get_shape())
-           {
-             Assert(sZ.get_volume() == Z->element_count());
-             Z->set_shape(sZ);
-           }
+   if (sZ != Z->get_shape())
+      {
+        Assert(sZ.get_volume() == Z->element_count());
+        Z->set_shape(sZ);
+      }
 
    Z->check_value(LOC);
    return Token(TOK_APL_VALUE1, Z);
@@ -160,8 +164,8 @@ const ShapeItem height = pb.get_height();
    // in lrm.
    //
 Value_P Z;
-   if (height == 1)   Z = Value_P(new Value(width, LOC));
-   else               Z = Value_P(new Value(Shape(height, width), LOC));
+   if (height == 1)   Z = Value_P(width, LOC);
+   else               Z = Value_P(Shape(height, width), LOC);
 
    loop(h, height)
    loop(w, width)
@@ -235,7 +239,7 @@ Shape shape_Z(B->get_shape());
    if (B->is_scalar())   shape_Z.add_shape_item(1);
    shape_Z.set_last_shape_item(format.size());
 
-Value_P Z(new Value(shape_Z, LOC));
+Value_P Z(shape_Z, LOC);
 
    try
       {
@@ -861,21 +865,20 @@ const ShapeItem len_A = A->is_scalar() ? 1 : A->get_shape_item(0);
    if (len_A != 1 && len_A != 2 && len_A != 2*cols_B)
       LENGTH_ERROR;
 
-const APL_Float qct = Workspace::get_CT();
    if (shape_B.get_volume() == 0)   // empty B
       {
         ShapeItem W = 0;
         loop(c, cols_B)
            {
-             if (len_A <= 2)   W += A->get_ravel(0).get_near_int(qct);
-             else              W += A->get_ravel(2*c).get_near_int(qct);
+             if (len_A <= 2)   W += A->get_ravel(0).get_near_int();
+             else              W += A->get_ravel(2*c).get_near_int();
            }
 
         Shape shape_Z = shape_B.without_axis(shape_B.get_rank() - 1);
         shape_Z.add_shape_item(W);
         const ShapeItem ec_Z = shape_Z.get_volume();
 
-        Value_P Z(new Value(shape_Z, LOC));
+        Value_P Z(shape_Z, LOC);
         loop(z, ec_Z)   new (Z->next_ravel()) CharCell(UNI_ASCII_SPACE);
 
         Z->set_default_Spc();
@@ -892,17 +895,17 @@ PrintBuffer pb;
          if (len_A == 1)
             {
               col_width = 0;
-              precision = A->get_ravel(0).get_near_int(qct);
+              precision = A->get_ravel(0).get_near_int();
             }
          else if (len_A == 2)
             {
-              col_width = A->get_ravel(0).get_near_int(qct);
-              precision = A->get_ravel(1).get_near_int(qct);
+              col_width = A->get_ravel(0).get_near_int();
+              precision = A->get_ravel(1).get_near_int();
             }
          else
             {
-              col_width = A->get_ravel(2*col)    .get_near_int(qct);
-              precision = A->get_ravel(2*col + 1).get_near_int(qct);
+              col_width = A->get_ravel(2*col)    .get_near_int();
+              precision = A->get_ravel(2*col + 1).get_near_int();
             }
 
          PrintBuffer pb_col(format_col_spec(col_width, precision,
@@ -919,7 +922,7 @@ const ShapeItem pb_h = pb.get_height();
 Shape shape_Z(shape_B);
    shape_Z.set_last_shape_item(pb_w);
 
-Value_P Z(new Value(shape_Z, LOC));
+Value_P Z(shape_Z, LOC);
 
    loop(h, pb_h)
    loop(w, pb_w)   new (Z->next_ravel()) CharCell(pb.get_char(w, h));
