@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -21,8 +21,11 @@
 #include "Bif_OPER1_REDUCE.hh"
 #include "Workspace.hh"
 
-Bif_OPER1_REDUCE   Bif_OPER1_REDUCE::fun;
-Bif_OPER1_REDUCE1  Bif_OPER1_REDUCE1::fun;
+Bif_OPER1_REDUCE    Bif_OPER1_REDUCE ::_fun;
+Bif_OPER1_REDUCE1   Bif_OPER1_REDUCE1::_fun;
+
+Bif_OPER1_REDUCE  * Bif_OPER1_REDUCE ::fun = &Bif_OPER1_REDUCE ::_fun;
+Bif_OPER1_REDUCE1 * Bif_OPER1_REDUCE1::fun = &Bif_OPER1_REDUCE1::_fun;
 
 //-----------------------------------------------------------------------------
 Token
@@ -40,7 +43,6 @@ Shape shape_B = B->get_shape();
    if (A->get_rank() > 1)             RANK_ERROR;
    if (axis >= shape_B.get_rank())    INDEX_ERROR;
 
-const APL_Float qct = Workspace::get_CT();
 const ShapeItem len_B = shape_B.get_shape_item(axis);
 ShapeItem len_A = A->element_count();
 ShapeItem len_Z = 0;
@@ -49,7 +51,7 @@ vector<ShapeItem> rep_counts;
    if (len_A == 1)   // single a -> a a ... a (len_B times)
       {
         len_A = len_B;
-        APL_Integer rep_A = A->get_ravel(0).get_near_int(qct);
+        APL_Integer rep_A = A->get_ravel(0).get_near_int();
         loop(a, len_A)   rep_counts.push_back(rep_A);
         if (rep_A > 0)        len_Z =  rep_A*len_B;
         else if (rep_A < 0)   len_Z = -rep_A*len_B;
@@ -59,7 +61,7 @@ vector<ShapeItem> rep_counts;
         ShapeItem geq_A = 0;   // number of items >= 0 in A
         loop(a, len_A)
            {
-             APL_Integer rep_A = A->get_ravel(a).get_near_int(qct);
+             APL_Integer rep_A = A->get_ravel(a).get_near_int();
              rep_counts.push_back(rep_A);
              if (rep_A > 0)        { len_Z += rep_A;   ++geq_A; }
              else if (rep_A < 0)   len_Z -= rep_A;
@@ -72,7 +74,7 @@ vector<ShapeItem> rep_counts;
 Shape shape_Z(shape_B);
    shape_Z.set_shape_item(axis, len_Z);
 
-Value_P Z(new Value(shape_Z, LOC));
+Value_P Z(shape_Z, LOC);
 
 const Shape3 shape_B3(shape_B, axis);
 
@@ -88,7 +90,8 @@ const Shape3 shape_B3(shape_B, axis);
                   loop(r, rep)
                   loop(l, shape_B3.l())
                      {
-                       Z->next_ravel()->init(B->get_ravel(shape_B3.hml(h, bm, l)));
+                       const ShapeItem src = shape_B3.hml(h, bm, l);
+                       Z->next_ravel()->init(B->get_ravel(src), Z.getref());
                      }
                   if (shape_B3.m() > 1)   ++bm;
                 }
@@ -97,7 +100,8 @@ const Shape3 shape_B3(shape_B, axis);
                   loop(r, -rep)
                   loop(l, shape_B3.l())
                      {
-                       Z->next_ravel()->init_type(B->get_ravel(shape_B3.hml(h, 0, l)));
+                       const ShapeItem src = shape_B3.hml(h, 0, l);
+                       Z->next_ravel()->init_type(B->get_ravel(src), Z.getref());
                      }
 
                   // cB is not incremented when fill item is used.
@@ -131,7 +135,7 @@ const Shape shape_Z = B->get_shape().without_axis(axis);
         return LO->eval_identity_fun(B, axis);
       }
 
-   if (m_len == 1)   return Bif_F12_RHO::do_reshape(shape_Z, B);
+   if (m_len == 1)   return Bif_F12_RHO::do_reshape(shape_Z, *B);
 
 const Shape3 B3(B->get_shape(), axis);
 const Shape3 Z3(B3.h(), 1, B3.l());
@@ -166,8 +170,8 @@ const int n_wise = A0 < 0 ? -A0 : A0;   // the number of items
       {
         Shape shape_B1 = B->get_shape().insert_axis(axis, 0);
         shape_B1.increment_shape_item(axis + 1);
-        Value_P val(new Value(shape_B1, LOC));
-        val->get_ravel(0).init(B->get_ravel(0));   // prototype
+        Value_P val(shape_B1, LOC);
+        val->get_ravel(0).init(B->get_ravel(0), val.getref());   // prototype
 
         Token result = LO->eval_identity_fun(val, axis);
         return result;
@@ -182,8 +186,8 @@ const int n_wise = A0 < 0 ? -A0 : A0;   // the number of items
         //
         Shape shape_Z = B->get_shape();
         shape_Z.set_shape_item(axis, 0);
-        Value_P Z(new Value(shape_Z, LOC));
-        Z->get_ravel(0).init(B->get_ravel(0));
+        Value_P Z(shape_Z, LOC);
+        Z->get_ravel(0).init(B->get_ravel(0), Z.getref());
         Z->check_value(LOC);
         return Token(TOK_APL_VALUE1, Z);
       }
@@ -191,7 +195,7 @@ const int n_wise = A0 < 0 ? -A0 : A0;   // the number of items
 Shape shape_Z(B->get_shape());
    shape_Z.set_shape_item(axis, shape_Z.get_shape_item(axis) - n_wise + 1);
 
-   if (n_wise == 1)   return Bif_F12_RHO::do_reshape(shape_Z, B);
+   if (n_wise == 1)   return Bif_F12_RHO::do_reshape(shape_Z, *B);
 
 const Shape3 Z3(shape_Z, axis);
    return do_reduce(shape_Z, Z3, A0, LO, axis, B, B->get_shape_item(axis));
@@ -203,94 +207,137 @@ Bif_REDUCE::do_reduce(const Shape & shape_Z, const Shape3 & Z3, ShapeItem a,
 {
    if (shape_Z.is_empty())   return LO->eval_identity_fun(B, axis);
 
-Value_P Z(new Value(shape_Z, LOC));
+Value_P Z(shape_Z, LOC);
 EOC_arg arg(Z, B);
 REDUCTION & _arg = arg.u.u_REDUCTION;
 
-   _arg.init(Z3, LO, &B->get_ravel(0), bm, a, 0);
+const bool scan = (a == -1);
+const bool reverse = (a < -1);
 
-Token tok(TOK_FIRST_TIME);
-   eoc_beam(tok, arg);
-   return tok;
+   _arg.LO         = LO;
+   _arg.scan         = scan;
+   _arg.len_L        = Z3.get_last_shape_item();
+   _arg.len_L_s       = scan ? 0 : _arg.len_L;
+   _arg.len_BML      = bm * _arg.len_L;
+   _arg.len_ZM       = Z3.m();
+   _arg.len_Z        = shape_Z.get_volume();
+   _arg.beam_len     = reverse ? -a : a;
+   _arg.dB           = reverse ? _arg.len_L : - _arg.len_L;
+   _arg.start_offset = reverse ? (_arg.beam_len - 1) * _arg.dB : 0;
+
+   _arg.todo_B = 0;   // new beam
+   _arg.z = -1;
+
+   return finish_REDUCE(arg, true);
+}
+//-----------------------------------------------------------------------------
+Token
+Bif_REDUCE::finish_REDUCE(EOC_arg & arg, bool first)
+{
+REDUCTION & _arg = arg.u.u_REDUCTION;
+
+   for (;;)
+      {
+        if (_arg.todo_B == 0)   // new beam
+           {
+             if (++_arg.z >= _arg.len_Z)   break;   // all beams done
+             
+             _arg.z_L =  _arg.z % _arg.len_L;
+             _arg.z_M = (_arg.z / _arg.len_L) % _arg.len_ZM;
+             _arg.z_H =  _arg.z / (_arg.len_L * _arg.len_ZM);
+
+             if (_arg.scan)   _arg.todo_B = 1 + _arg.z_M;
+             else             _arg.todo_B = _arg.beam_len;
+
+            _arg.b = _arg.z_L + _arg.z_H * _arg.len_BML   // start of row
+                   + _arg.z_M * _arg.len_L_s              // start of beam
+                   + (_arg.todo_B - 1) * _arg.len_L       // end of beam
+                   - _arg.start_offset;                   // or start (reverse)
+
+             // Z[z] ? B[b]
+             //
+             arg.Z->next_ravel()->init(arg.B->get_ravel(_arg.b),
+                                       arg.Z.getref());
+           }
+
+        // update position and check for end of beam
+        //
+        Assert(_arg.todo_B > 0);
+        _arg.b += _arg.dB;
+        --_arg.todo_B;
+        if (_arg.todo_B == 0)
+           {
+             continue;
+           }
+
+        // reduction step (one call of LO)
+        //
+        {
+          Value_P LO_A = arg.B->get_ravel(_arg.b).to_value(LOC);
+          Value_P LO_B = arg.Z->get_ravel(_arg.z).to_value(LOC);
+          Token result = _arg.LO->eval_AB(LO_A, LO_B);
+          arg.Z->get_ravel(_arg.z).release(LOC);
+
+          if (result.get_Class() == TC_VALUE)
+             {
+               Value_P ZZ = result.get_apl_val();
+               arg.Z->get_ravel(_arg.z).init_from_value(ZZ, arg.Z.getref(),LOC);
+               continue;
+             }
+
+          if (result.get_tag() == TOK_ERROR)   return result;
+
+          if (result.get_tag() == TOK_SI_PUSHED)
+             {
+               // LO was a user defined function
+               //
+               if (first)   // first call
+                  Workspace::SI_top()->add_eoc_handler(eoc_REDUCE, arg, LOC);
+               else           // subsequent call
+                  Workspace::SI_top()->move_eoc_handler(eoc_REDUCE, &arg, LOC);
+
+               return result;   // continue in user defined function...
+             }
+
+          Q1(result);   FIXME;
+        }
+      }
+
+   arg.Z->set_default(*arg.B.get());
+
+   arg.Z->check_value(LOC);
+   return Token(TOK_APL_VALUE1, arg.Z);
 }
 //-----------------------------------------------------------------------------
 bool
-Bif_REDUCE::eoc_beam(Token & token, EOC_arg & si_arg)
+Bif_REDUCE::eoc_REDUCE(Token & token)
 {
-   if (token.get_tag() == TOK_ERROR)   return false;   // stop it
+EOC_arg * arg = Workspace::SI_top()->remove_eoc_handlers();
+EOC_arg * next = arg->next;
+REDUCTION & _arg = arg->u.u_REDUCTION;
 
-EOC_arg arg = si_arg;
-REDUCTION & _arg = arg.u.u_REDUCTION;
+   if (token.get_Class() != TC_VALUE)  return false;   // stop it
 
-   if (token.get_tag() == TOK_FIRST_TIME)   // first call to eoc_beam()
-      {
-new_beam:
-        const Cell * cB = _arg.beam.next_B();
-
-        // we need a token with a free (erasable) value. If the value were
-        // nested, then we would get a double delete (from the original owner
-        // and from our result). We can't therefore call cB->to_value() for
-        // nested values.
-        //
-        if (cB->is_pointer_cell())
-           copy_1(token, Token(TOK_APL_VALUE1,
-                                cB->get_pointer_value()->clone(LOC)), LOC);
-        else
-           copy_1(token, Token(TOK_APL_VALUE1, cB->to_value(LOC)), LOC);
-      }
-
-again:
-Value_P BB = token.get_apl_val();
-
-   if (_arg.beam.done())   // last reduction in current beam
-      {
-        _arg.frame.next_hml();
-
-        // if beam has only one element, as for the first column in scan,
-        // then BB->clear_eoc() below is never reached and we have to do
-        // it here.
-        //
-        arg.Z->next_ravel()->init_from_value(BB, LOC);
-
-        if (_arg.frame.done())   // if last beam (final result complete)
-           {
-             arg.Z->check_value(LOC);
-             copy_1(token, Token(TOK_APL_VALUE1, arg.Z), LOC);
-             return false;   // stop it
-           }
-
-        if (_arg.frame.A0_inc)   _arg.beam.length = _arg.frame.m + 1;
-        const Cell * beam = _arg.frame.beam_start();
-        _arg.beam.reset(beam);
-        goto new_beam;
-      }
-
-   // pop context for previous eval_AB() call
+   // the user defined function has returned a value. Store it.
    //
-   if (_arg.need_pop)   Workspace::pop_SI(LOC);
+   {
+     Value_P ZZ = token.get_apl_val();
+     arg->Z->get_ravel(_arg.z).init_from_value(ZZ, arg->Z.getref(), LOC);
+   }
 
-const Cell * cA = _arg.beam.next_B();
-Value_P AA = cA->to_value(LOC);
-   copy_1(token, _arg.beam.LO->eval_AB(AA, BB), LOC);
-
-   // if token is an APL value, then LO was a primitive function and the last
-   // LO->eval_AB() succeeded. No SI entry was pushed, so we can loop locally.
+   // pop the SI unless this is the last reduction of the last 
+   // ravel element of Z to be computed
    //
-   if (token.get_Class() == TC_VALUE)   goto again;
+   if (_arg.z < (_arg.len_Z - 1) || _arg.todo_B > 1 )   Workspace::pop_SI(LOC);
 
-   // if eval_AB() returned an error then stop. This can happen for both
-   // primitive and user-defined LO.
-   //
-   if (token.get_tag() == TOK_ERROR)   return false;   // stop it
+   copy_1(token, finish_REDUCE(*arg, false), LOC);
+   if (token.get_tag() == TOK_SI_PUSHED)   return true;   // continue
 
-   // Otherwise LO must have been a user defined function
-   //
-   Assert(token.get_tag() == TOK_SI_PUSHED);
-   _arg.need_pop = true;
+   delete arg;
+   Workspace::SI_top()->set_eoc_handlers(next);
+   if (next)   return next->handler(token);
 
-   Workspace::SI_top()->set_eoc_handler(eoc_beam);
-   Workspace::SI_top()->get_eoc_arg() = arg;
-   return true;   // continue
+   return false;   // stop it
 }
 //-----------------------------------------------------------------------------
 Token
