@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -23,8 +23,11 @@
 #include "LvalCell.hh"
 #include "Workspace.hh"
 
-Bif_OPER1_SCAN     Bif_OPER1_SCAN::fun;
-Bif_OPER1_SCAN1    Bif_OPER1_SCAN1::fun;
+Bif_OPER1_SCAN    Bif_OPER1_SCAN ::_fun;
+Bif_OPER1_SCAN1   Bif_OPER1_SCAN1::_fun;
+
+Bif_OPER1_SCAN  * Bif_OPER1_SCAN ::fun = &Bif_OPER1_SCAN ::_fun;
+Bif_OPER1_SCAN1 * Bif_OPER1_SCAN1::fun = &Bif_OPER1_SCAN1::_fun;
 
 //-----------------------------------------------------------------------------
 Token
@@ -44,13 +47,12 @@ Shape shape_B = B->get_shape();
    if (shape_B.get_rank() <= axis)   RANK_ERROR;
 
 const ShapeItem ec_A = A->element_count();
-const APL_Float qct = Workspace::get_CT();
 ShapeItem ones_A = 0;
 vector<ShapeItem> rep_counts;
    rep_counts.reserve(ec_A);
    loop(a, ec_A)
       {
-        APL_Integer rep_A = A->get_ravel(a).get_near_int(qct);
+        APL_Integer rep_A = A->get_ravel(a).get_near_int();
         rep_counts.push_back(rep_A);
         if      (rep_A == 0)        ;
         else if (rep_A == 1)        ++ones_A;
@@ -59,7 +61,7 @@ vector<ShapeItem> rep_counts;
 
 Shape shape_Z(shape_B);
    shape_Z.set_shape_item(axis, ec_A);
-Value_P Z(new Value(shape_Z, LOC));
+Value_P Z(shape_Z, LOC);
 
    if (ec_A == 0)   // (⍳0)/B : 
       {
@@ -93,18 +95,21 @@ ShapeItem inc_2 = 0;              // increment after result m*l items
            {
              if (rep_counts[m] == 1)   // copy items from B
                 {
-                  loop(l, shape_Z3.l())   Z->next_ravel()->init(cB[l]);
+                  loop(l, shape_Z3.l())
+                      Z->next_ravel()->init(cB[l], Z.getref());
                   cB += inc_1;
                 }
              else                      // init items
                 {
                   if (lval)
                      {
-                       loop(l, shape_Z3.l())   new (Z->next_ravel()) LvalCell(0);
+                       loop(l, shape_Z3.l())
+                           new (Z->next_ravel()) LvalCell(0, 0);
                      }
                   else
                      {
-                       loop(l, shape_Z3.l())   Z->next_ravel()->init_type(fill[l]);
+                       loop(l, shape_Z3.l())
+                           Z->next_ravel()->init_type(fill[l], Z.getref());
                      }
                 }
            }
@@ -135,12 +140,12 @@ const ShapeItem m_len = B->get_shape_item(axis);
    if (m_len == 1)
       {
         const Shape shape_Z = B->get_shape().without_axis(axis);
-        return Bif_F12_RHO::do_reshape(shape_Z, B);
+        return Bif_F12_RHO::do_reshape(shape_Z, *B);
       }
 
 const Shape3 shape_Z3(B->get_shape(), axis);
 
-Value_P Z(new Value(B->get_shape(), LOC));
+Value_P Z(B->get_shape(), LOC);
 ErrorCode (Cell::*assoc_f2)(Cell *, const Cell *) const = LO->get_assoc();
 
    if (assoc_f2)
@@ -155,35 +160,30 @@ ErrorCode (Cell::*assoc_f2)(Cell *, const Cell *) const = LO->get_assoc();
               Cell * cZ = Z->next_ravel();
               if (m == 0)
                  {
-                   cZ->init(*cB++);
+                   cZ->init(*cB++, Z.getref());
                  }
               else
                  {
 
-                   Value_P AA(new Value(LOC));
-                   AA->next_ravel()->init(cZ[-shape_Z3.l()]);
+                   Value_P AA(LOC);
+                   AA->next_ravel()->init(cZ[-shape_Z3.l()], AA.getref());
 
-                   Value_P BB(new Value(LOC));
-                   BB->next_ravel()->init(*cB++);
+                   Value_P BB(LOC);
+                   BB->next_ravel()->init(*cB++, BB.getref());
 
                    Token tok = LO->eval_AB(AA, BB);
                    if (!tok.is_apl_val())   return tok;
 
                    Value_P ZZ = tok.get_apl_val();
-                   cZ->init(ZZ->get_ravel(0));
+                   cZ->init(ZZ->get_ravel(0), ZZ.getref());
                  }
             }
         return Token(TOK_APL_VALUE1, Z);
       }
 
-EOC_arg arg(Z, B);
-REDUCTION & _arg = arg.u.u_REDUCTION;
+const Shape3 Z3(B->get_shape(), axis);
 
-   _arg.init(shape_Z3, LO, &B->get_ravel(0), m_len, 1, 1);
-
-Token tok(TOK_FIRST_TIME);
-   Bif_REDUCE::eoc_beam(tok, arg);
-   return tok;
+   return Bif_REDUCE::do_reduce(B->get_shape(), Z3, -1, LO, axis, B, m_len);
 }
 //-----------------------------------------------------------------------------
 Token
