@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2014  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -23,7 +23,8 @@
 #include "PointerCell.hh"
 #include "Workspace.hh"
 
-Bif_OPER2_RANK     Bif_OPER2_RANK::fun;
+Bif_OPER2_RANK   Bif_OPER2_RANK::_fun;
+Bif_OPER2_RANK * Bif_OPER2_RANK::fun = &Bif_OPER2_RANK::_fun;
 
 /* general comment: we use the term 'chunk' instead of 'p-rank' to avoid
  * confusion with the rank of a value
@@ -65,69 +66,64 @@ Bif_OPER2_RANK::do_LyXB(Function * LO, const Shape * axes,
    // split shape of B into high and low shapes.
    //
 const Shape shape_Z = B->get_shape().high_shape(B->get_rank() - rk_chunk_B);
-Value_P Z(new Value(shape_Z, LOC));
+Value_P Z(shape_Z, LOC);
 EOC_arg arg(Z, B);
-RANK_LyXB & _arg = arg.u.u_RANK_LyXB;
+RANK & _arg = arg.u.u_RANK;
 
    _arg.get_sh_chunk_B() = B->get_shape().low_shape(rk_chunk_B);
    _arg.get_sh_frame() = shape_Z;
-   _arg.ec_frame = shape_Z.get_volume();
+   _arg.ec_frame   = shape_Z.get_volume();
+   if (!_arg.ec_frame)   _arg.ec_frame = 1;
    _arg.ec_chunk_B = _arg.get_sh_chunk_B().get_volume();
    _arg.cB = &B->get_ravel(0);
    _arg.LO = LO;
    _arg.axes_valid = (axes != 0);
    if (axes)   _arg.get_axes() = *axes;
-   _arg.how = 0;
+   _arg.h = -1;
 
-   return finish_LyXB(arg);
+   return finish_LyXB(arg, true);
 }
 //-----------------------------------------------------------------------------
 Token
-Bif_OPER2_RANK::finish_LyXB(EOC_arg & arg)
+Bif_OPER2_RANK::finish_LyXB(EOC_arg & arg, bool first)
 {
-RANK_LyXB & _arg = arg.u.u_RANK_LyXB;
+RANK & _arg = arg.u.u_RANK;
 
-   if (_arg.how == 1)   goto how_1;
+   while (++_arg.h < _arg.ec_frame)
+      {
+        Value_P BB(_arg.get_sh_chunk_B(), LOC);
+        Assert1(_arg.ec_chunk_B == _arg.get_sh_chunk_B().get_volume());
+        loop(l, _arg.ec_chunk_B)
+            BB->next_ravel()->init(*_arg.cB++, BB.getref());
+        BB->check_value(LOC);
 
-   Assert1(_arg.how == 0);
-   _arg.h = 0;
+        Token result = _arg.LO->eval_B(BB);
+        if (result.get_tag() == TOK_SI_PUSHED)
+           {
+             // LO was a user defined function or ⍎
+             //
+             if (first)   // first call
+                Workspace::SI_top()->add_eoc_handler(eoc_RANK, arg, LOC);
+             else           // subsequent call
+                Workspace::SI_top()->move_eoc_handler(eoc_RANK, &arg, LOC);
 
-loop_h:
-   {
-     Value_P BB(new Value(_arg.get_sh_chunk_B(), LOC));
-     Assert1(_arg.ec_chunk_B == _arg.get_sh_chunk_B().get_volume());
-     loop(l, _arg.ec_chunk_B)   BB->next_ravel()->init(*_arg.cB++);
-     BB->check_value(LOC);
+             return result;   // continue in user defined function...
+           }
 
-     Token result = _arg.LO->eval_B(BB);
-     if (result.get_tag() == TOK_SI_PUSHED)
-        {
-          // LO was a user defined function or ⍎
-          //
-          _arg.how = 1;
-          Workspace::SI_top()->set_eoc_handler(eoc_LyXB);
-          Workspace::SI_top()->get_eoc_arg() = arg;
-          return result;   // continue in user defined function...
-        }
+        if (result.get_tag() == TOK_ERROR)   return result;
 
-     if (result.get_tag() == TOK_ERROR)   return result;
+        if (result.get_Class() == TC_VALUE)
+           {
+             Value_P ZZ = result.get_apl_val();
+             Cell * cZ = arg.Z->next_ravel();
+             if (cZ == 0)           cZ = &arg.Z->get_ravel(0);   // empty Z
+             if (ZZ->is_scalar())   cZ->init(ZZ->get_ravel(0), arg.Z.getref());
+             else                   new (cZ)   PointerCell(ZZ, arg.Z.getref());
+             continue;
+           }
 
-     if (result.get_Class() == TC_VALUE)
-        {
-          Value_P ZZ = result.get_apl_val();
-          Cell * cZ = arg.Z->next_ravel();
-          if (cZ == 0)   cZ = &arg.Z->get_ravel(0);   // empty Z
-          if (ZZ->is_scalar())   cZ->init(ZZ->get_ravel(0));
-          else                   new (cZ)   PointerCell(ZZ);
-          goto next_h;
-       }
-
-     Q1(result);   FIXME;
-   }
-
-how_1:
-next_h:
-   if (++_arg.h < _arg.ec_frame)   goto loop_h;
+        Q1(result);   FIXME;
+      }
 
    arg.Z->check_value(LOC);
    if (!_arg.axes_valid)   return Bif_F12_PICK::disclose(arg.Z, true);
@@ -135,28 +131,6 @@ next_h:
    return Bif_F12_PICK::disclose_with_axis(_arg.get_axes(), arg.Z, true);
 }
 //-----------------------------------------------------------------------------
-bool
-Bif_OPER2_RANK::eoc_LyXB(Token & token, EOC_arg & si_arg)
-{
-EOC_arg arg = si_arg;
-RANK_LyXB & _arg = arg.u.u_RANK_LyXB;
-
-   if (token.get_Class() != TC_VALUE)  return false;   // stop it
-
-   // the user defined function has returned a value. Store it.
-   //
-Value_P ZZ = token.get_apl_val();
-   if (ZZ->is_scalar())   arg.Z->next_ravel()->init(ZZ->get_ravel(0));
-   else                   new (arg.Z->next_ravel())   PointerCell(ZZ);
-
-   if (_arg.h < (_arg.ec_frame - 1))   Workspace::pop_SI(LOC);
-
-   copy_1(token, finish_LyXB(arg), LOC);
-   if (token.get_tag() == TOK_SI_PUSHED)   return true;   // continue
-
-   return false;   // stop it
-}
-//-----------------------------------------------------------------------------
 Token
 Bif_OPER2_RANK::eval_ALRB(Value_P A, Token & LO, Token & y, Value_P B)
 {
@@ -225,9 +199,9 @@ Shape shape_Z;
         shape_Z = sh_B_frame;
       }
 
-Value_P Z(new Value(shape_Z, LOC));
+Value_P Z(shape_Z, LOC);
 EOC_arg arg(Z, B, A);
-RANK_ALyXB & _arg = arg.u.u_RANK_ALyXB;
+RANK & _arg = arg.u.u_RANK;
 
    _arg.get_sh_chunk_A() = A->get_shape().low_shape(rk_chunk_A);
    _arg.get_sh_chunk_B() = B->get_shape().low_shape(rk_chunk_B);
@@ -254,7 +228,8 @@ RANK_ALyXB & _arg = arg.u.u_RANK_ALyXB;
    // at this point sh_A_frame == sh_B_frame, so we can use either of them
    // to compute ec_chunk_A
    //
-   _arg.ec_frame   = _arg.get_sh_frame().get_volume();
+   _arg.ec_frame = _arg.get_sh_frame().get_volume();
+   if (!_arg.ec_frame)   _arg.ec_frame = 1;
    _arg.ec_chunk_A = _arg.get_sh_chunk_A().get_volume();
    _arg.ec_chunk_B = _arg.get_sh_chunk_B().get_volume();
 
@@ -263,63 +238,58 @@ RANK_ALyXB & _arg = arg.u.u_RANK_ALyXB;
    _arg.axes_valid = (axes != 0);
    if (axes)   _arg.get_axes() = *axes;
    _arg.cB = &B->get_ravel(0);
-   _arg.how = 0;
+   _arg.h = -1;
 
-   return finish_ALyXB(arg);
+   return finish_ALyXB(arg, true);
 }
 //-----------------------------------------------------------------------------
 Token
-Bif_OPER2_RANK::finish_ALyXB(EOC_arg & arg)
+Bif_OPER2_RANK::finish_ALyXB(EOC_arg & arg, bool first)
 {
-RANK_ALyXB & _arg = arg.u.u_RANK_ALyXB;
+RANK & _arg = arg.u.u_RANK;
 
-   if (_arg.how == 1)   goto how_1;
+   while (++_arg.h < _arg.ec_frame)
+      {
+        Value_P AA(_arg.get_sh_chunk_A(), LOC);
+        Value_P BB(_arg.get_sh_chunk_B(), LOC);
 
-   Assert1(_arg.how == 0);
-   _arg.h = 0;
+        loop(l, _arg.ec_chunk_A)  AA->get_ravel(l).init(*_arg.cA++,AA.getref());
+        loop(l, _arg.ec_chunk_B)  BB->get_ravel(l).init(*_arg.cB++,BB.getref());
 
-loop_h:
-   {
-     Value_P AA(new Value(_arg.get_sh_chunk_A(), LOC));
-     Value_P BB(new Value(_arg.get_sh_chunk_B(), LOC));
+        if (_arg.repeat_A)   _arg.cA = &arg.A->get_ravel(0);
+        if (_arg.repeat_B)   _arg.cB = &arg.B->get_ravel(0);
 
-     loop(l, _arg.ec_chunk_A)   AA->get_ravel(l).init(*_arg.cA++);
-     loop(l, _arg.ec_chunk_B)   BB->get_ravel(l).init(*_arg.cB++);
+        AA->check_value(LOC);
+        BB->check_value(LOC);
 
-     if (_arg.repeat_A)   _arg.cA = &arg.A->get_ravel(0);
-     if (_arg.repeat_B)   _arg.cB = &arg.B->get_ravel(0);
+        Token result = _arg.LO->eval_AB(AA, BB);
+        if (result.get_tag() == TOK_SI_PUSHED)
+           {
+             // LO was a user defined function or ⍎
+             //
+             if (first)   // first call
+                Workspace::SI_top()->add_eoc_handler(eoc_RANK, arg, LOC);
+             else           // subsequent call
+                Workspace::SI_top()->move_eoc_handler(eoc_RANK, &arg, LOC);
 
-     AA->check_value(LOC);
-     BB->check_value(LOC);
+             return result;   // continue in user defined function...
+           }
 
-     Token result = _arg.LO->eval_AB(AA, BB);
-     if (result.get_tag() == TOK_SI_PUSHED)
-        {
-          // LO was a user defined function or ⍎
-          //
-          _arg.how = 1;
-          Workspace::SI_top()->set_eoc_handler(eoc_ALyXB);
-          Workspace::SI_top()->get_eoc_arg() = arg;
-          return result;   // continue in user defined function...
-        }
+        if (result.get_tag() == TOK_ERROR)   return result;
 
-     if (result.get_tag() == TOK_ERROR)   return result;
+        if (result.get_Class() == TC_VALUE)
+           {
+             Value_P ZZ = result.get_apl_val();
+             Cell * cZ = arg.Z->is_empty() ? &arg.Z->get_ravel(0)
+                                           : arg.Z->next_ravel();
+             if (ZZ->is_scalar())   cZ->init(ZZ->get_ravel(0), arg.Z.getref());
+             else                   new (cZ) PointerCell(ZZ, arg.Z.getref());
 
-     if (result.get_Class() == TC_VALUE)
-        {
-          Value_P ZZ = result.get_apl_val();
-          if (ZZ->is_scalar())   arg.Z->next_ravel()->init(ZZ->get_ravel(0));
-          else                   new (arg.Z->next_ravel())   PointerCell(ZZ);
+             continue;
+          }
 
-          goto next_h;
-       }
-
-     Q1(result);   FIXME;
-   }
-
-how_1:
-next_h:
-   if (++_arg.h < _arg.ec_frame)   goto loop_h;
+        Q1(result);   FIXME;
+      }
 
    arg.Z->check_value(LOC);
    if (!_arg.axes_valid)   return Bif_F12_PICK::disclose(arg.Z, true);
@@ -328,24 +298,33 @@ next_h:
 }
 //-----------------------------------------------------------------------------
 bool
-Bif_OPER2_RANK::eoc_ALyXB(Token & token, EOC_arg & si_arg)
+Bif_OPER2_RANK::eoc_RANK(Token & token)
 {
-EOC_arg arg = si_arg;
-RANK_ALyXB & _arg = arg.u.u_RANK_ALyXB;
+EOC_arg * arg = Workspace::SI_top()->remove_eoc_handlers();
+EOC_arg * next = arg->next;
+RANK & _arg = arg->u.u_RANK;
 
    if (token.get_Class() != TC_VALUE)  return false;   // stop it
 
    // the user defined function has returned a value. Store it.
    //
 Value_P ZZ = token.get_apl_val();
-   if (ZZ->is_scalar())   arg.Z->next_ravel()->init(ZZ->get_ravel(0));
-   else                   new (arg.Z->next_ravel())   PointerCell(ZZ);
+   if (ZZ->is_scalar())
+      arg->Z->next_ravel()->init(ZZ->get_ravel(0), arg->Z.getref());
+   else
+      new (arg->Z->next_ravel())   PointerCell(ZZ, arg->Z.getref());
 
    if (_arg.h < (_arg.ec_frame - 1))   Workspace::pop_SI(LOC);
 
-   copy_1(token, finish_ALyXB(arg), LOC);
+   if (!arg->A)   copy_1(token, finish_LyXB (*arg, false), LOC);
+   else           copy_1(token, finish_ALyXB(*arg, false), LOC);
+
    if (token.get_tag() == TOK_SI_PUSHED)   return true;   // continue
 
+   delete arg;
+   Workspace::SI_top()->set_eoc_handlers(next);
+   if (next)   return next->handler(token);
+
    return false;   // stop it
 }
 //-----------------------------------------------------------------------------
@@ -375,18 +354,16 @@ const Rank rk_B = rank_B;
    // yM yA yB :        N/A        yM           yA       yB
    // ---------------------------------------------------------
 
-const APL_Float qct = Workspace::get_CT();
-
    switch(y123->element_count())
       {
-        case 1: rank_B = y123->get_ravel(0).get_near_int(qct);   break;
+        case 1: rank_B = y123->get_ravel(0).get_near_int();   break;
 
-        case 2:          y123->get_ravel(0).get_near_int(qct);
-                rank_B = y123->get_ravel(1).get_near_int(qct);   break;
+        case 2:          y123->get_ravel(0).get_near_int();
+                rank_B = y123->get_ravel(1).get_near_int();   break;
 
-        case 3: rank_B = y123->get_ravel(0).get_near_int(qct);
-                         y123->get_ravel(1).get_near_int(qct);
-                         y123->get_ravel(2).get_near_int(qct);   break;
+        case 3: rank_B = y123->get_ravel(0).get_near_int();
+                         y123->get_ravel(1).get_near_int();
+                         y123->get_ravel(2).get_near_int();   break;
 
              default: LENGTH_ERROR;
       }
@@ -427,19 +404,17 @@ const Rank rk_B = rank_B;
    // yM yA yB :        N/A        yM           yA       yB
    // ---------------------------------------------------------
 
-const APL_Float qct = Workspace::get_CT();
-
    switch(y123->element_count())
       {
-        case 1:  rank_A = y123->get_ravel(0).get_near_int(qct);
+        case 1:  rank_A = y123->get_ravel(0).get_near_int();
                  rank_B = rank_A;                            break;
 
-        case 2:  rank_A = y123->get_ravel(0).get_near_int(qct);
-                 rank_B = y123->get_ravel(1).get_near_int(qct);  break;
+        case 2:  rank_A = y123->get_ravel(0).get_near_int();
+                 rank_B = y123->get_ravel(1).get_near_int();  break;
 
-        case 3:           y123->get_ravel(0).get_near_int(qct);
-                 rank_A = y123->get_ravel(1).get_near_int(qct);
-                 rank_B = y123->get_ravel(2).get_near_int(qct);  break;
+        case 3:           y123->get_ravel(0).get_near_int();
+                 rank_A = y123->get_ravel(1).get_near_int();
+                 rank_B = y123->get_ravel(2).get_near_int();  break;
 
         default: LENGTH_ERROR;
       }
@@ -504,15 +479,15 @@ const ShapeItem length = y123_B->element_count();
                  }
               else
                  {
-                   B = Value_P(new Value(LOC));
-                   B->next_ravel()->init(B0);
+                   B = Value_P(LOC);
+                   B->next_ravel()->init(B0, B.getref());
                  }
             }
          else                    // vector B
             {
-              B = Value_P(new Value(length - 1, LOC));
+              B = Value_P(length - 1, LOC);
               loop(l, length - 1)
-                  B->next_ravel()->init(y123_B->get_ravel(l + 1));
+                  B->next_ravel()->init(y123_B->get_ravel(l + 1), B.getref());
             }
          return;
       }
@@ -520,12 +495,11 @@ const ShapeItem length = y123_B->element_count();
    // case 1. ruled out, so the first 1, 2, or 3 cells are j123.
    // see how many (at most)
    //
-const APL_Float qct = Workspace::get_CT();
 int y123_len = 0;
    loop(yy, 3)
       {
         if (yy >= length)   break;
-        if (y123_B->get_ravel(yy).is_near_int(qct))   ++y123_len;
+        if (y123_B->get_ravel(yy).is_near_int())   ++y123_len;
         else                                          break;
       }
    if (y123_len == 0)   LENGTH_ERROR;   // at least y1 is needed
@@ -541,19 +515,21 @@ int y123_len = 0;
    if (length == (y123_len + 1) &&
        y123_B->get_ravel(y123_len).is_pointer_cell())   // case 3. y123:⊂B
       {
-        y123 = Value_P(new Value(y123_len, LOC));
-        loop(yy, y123_len)   y123->next_ravel()->init(y123_B->get_ravel(yy));
+        y123 = Value_P(y123_len, LOC);
+        loop(yy, y123_len)
+            y123->next_ravel()->init(y123_B->get_ravel(yy), y123.getref());
         B = y123_B->get_ravel(y123_len).get_pointer_value();
         return;
       }
 
    // case 4: y123:B...
    //
-   y123 = Value_P(new Value(y123_len, LOC));
-   loop(yy, y123_len)   y123->next_ravel()->init(y123_B->get_ravel(yy));
+   y123 = Value_P(y123_len, LOC);
+   loop(yy, y123_len)
+       y123->next_ravel()->init(y123_B->get_ravel(yy), y123.getref());
 
-   B = Value_P(new Value(length - y123_len, LOC));
+   B = Value_P(length - y123_len, LOC);
    loop(bb, (length - y123_len))
-       B->next_ravel()->init(y123_B->get_ravel(bb + y123_len));
+       B->next_ravel()->init(y123_B->get_ravel(bb + y123_len), B.getref());
 }
 //-----------------------------------------------------------------------------
