@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -267,8 +267,8 @@ const uint32_t nelm = val.element_count();
              const Cell & cell = val.get_ravel(e);
              if (cell.is_character_cell() || cell.is_numeric())
                 {
-                  Value_P sub_val(new Value(LOC));
-                  sub_val->get_ravel(0).init(cell);
+                  Value_P sub_val(LOC);
+                  sub_val->get_ravel(0).init(cell, sub_val.getref());
 
                   const CDR_type sub_type = sub_val->get_CDR_type();
                   const int sub_len = sub_val->total_size_brutto(sub_type);
@@ -325,7 +325,7 @@ Shape shape;
         shape.add_shape_item(sh);
       }
 
-Value_P ret(new Value(shape, loc));
+Value_P ret(shape, loc);
 
 const uint8_t * ravel = data + 16 + 4*rank;
 
@@ -459,7 +459,7 @@ const uint8_t * ravel = data + 16 + 4*rank;
                    loop(r, sub_rank)
                       sh.add_shape_item(get_4_be(sub_data + 16));
 
-                   Value_P sub_val(new Value(sh, LOC));
+                   Value_P sub_val(sh, LOC);
                    const APL_Integer qio = Workspace::get_IO();
                    loop(v, sh.get_volume())
                        {
@@ -473,7 +473,7 @@ const uint8_t * ravel = data + 16 + 4*rank;
               CDR_string sub_cdr(sub_data, sub_cdr_len);
               Value_P sub_val = from_CDR(sub_cdr, LOC);
               Assert(!!sub_val);
-              new (&ret->get_ravel(n))   PointerCell(sub_val);
+              new (&ret->get_ravel(n))   PointerCell(sub_val, ret.getref());
             }
       }
    else
