@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -39,7 +39,7 @@ Cell::operator new(std::size_t s, void * pos)
 }
 //-----------------------------------------------------------------------------
 void
-Cell::init(const Cell & other)
+Cell::init(const Cell & other, Value & cell_owner)
 {
    Assert(&other);
    switch(other.get_cell_type())
@@ -67,12 +67,13 @@ Cell::init(const Cell & other)
              {
                Value_P Z = other.get_pointer_value()->clone(LOC);
 
-               new (this) PointerCell(Z);
+               new (this) PointerCell(Z, cell_owner);
              }
              return;
 
         case CT_CELLREF:
-             new (this) LvalCell(other.get_lval_value());
+             new (this) LvalCell(other.get_lval_value(),
+                                 other.get_cell_owner());
              return;
       }
 
@@ -80,15 +81,15 @@ Cell::init(const Cell & other)
 }
 //-----------------------------------------------------------------------------
 void
-Cell::init_from_value(Value_P value, const char * loc)
+Cell::init_from_value(Value_P value, Value & cell_owner, const char * loc)
 {
    if (value->is_scalar())
       {
-        init(value->get_ravel(0));
+        init(value->get_ravel(0), cell_owner);
       }
    else
       {
-        new (this) PointerCell(value);
+        new (this) PointerCell(value, cell_owner);
       }
 }
 //-----------------------------------------------------------------------------
@@ -102,8 +103,8 @@ Value_P ret;
       }
    else
       {
-        ret = Value_P(new Value(loc));
-        ret->get_ravel(0).init(*this);
+        ret = Value_P(loc);
+        ret->get_ravel(0).init(*this, ret.getref());
         ret->check_value(LOC);
       }
 
@@ -111,7 +112,7 @@ Value_P ret;
 }
 //-----------------------------------------------------------------------------
 Cell *
-Cell::init_type(const Cell & other)
+Cell::init_type(const Cell & other, Value & cell_owner)
 {
    // Note: this function changes the type of this cell, but the
    // compiler may not notice that and use the declaration for this
@@ -119,9 +120,14 @@ Cell::init_type(const Cell & other)
    //
    if (other.is_pointer_cell())
       {
-        new (this) PointerCell(other.get_pointer_value()->clone(LOC));
+        new (this) PointerCell(other.get_pointer_value()->clone(LOC),
+                               cell_owner);
         get_pointer_value()->to_proto();
       }
+   else if (other.is_lval_cell())
+      {
+        FIXME;
+      }
    else if (other.is_character_cell())
       {
         new (this) CharCell(UNI_ASCII_SPACE);
@@ -140,12 +146,12 @@ Cell::copy(Value & val, const Cell * & src, ShapeItem count)
    loop(c, count)
       {
         Assert1(val.more());
-        val.next_ravel()->init(*src++);
+        val.next_ravel()->init(*src++, val);
       }
 }
 //-----------------------------------------------------------------------------
 bool
-Cell::greater(const Cell * other, bool ascending) const
+Cell::greater(const Cell & other) const
 {
    CERR << "greater() called on object of class" << get_classname() << endl;
    Assert(0);
@@ -161,21 +167,45 @@ Cell::equal(const Cell & other, APL_Float qct) const
 bool
 Cell::tolerantly_equal(APL_Complex A, APL_Complex B, APL_Float C)
 {
-   if (A == B)                             return true;
-   if (A.real() < 0.0 && B.real() > 0.0)   return false;
-   if (A.real() > 0.0 && B.real() < 0.0)   return false;
-   if (A.imag() < 0.0 && B.imag() > 0.0)   return false;
-   if (A.imag() > 0.0 && B.imag() < 0.0)   return false;
+   // if A equals B, return true
+   //
+   if (A == B) return true;
 
-const APL_Float mag2_A  = A.real() * A.real() + A.imag() * A.imag();
-const APL_Float mag2_B  = B.real() * B.real() + B.imag() * B.imag();
-const APL_Float max_mag = sqrt((mag2_A > mag2_B) ? mag2_A : mag2_B);
+   // if A and B are not in the same half-plane, return false.
+   //
+   // Implementation: If A and B are in the same real half-plane then
+   //                 the product of their real parts is ≥ 0,
+   //
+   //                 If A and B are in the same imag half-plane then
+   //                 the product of their imag parts is ≥ 0,
+   //
+   //                 Otherwise: they are not in the same half-plane
+   //                 and we return false;
+   //
+   if (A.real() * B.real() < 0.0 &&
+       A.imag() * B.imag() < 0.0)   return false;
 
+   // If the distance-between A and B is ≤ C times the larger-magnitude
+   // of A and B, return true
+   //
+   // Implementation: Instead of mag(A-B)  ≤ C × max(mag(A),   mag(B))
+   // we compute                 mag²(A-B) ≤ C² × max(mag²(A), mag²(B))
+   //
+   // 1. compute max(mag²A, mag²B)
+   //
+const APL_Float mag2_A   = A.real() * A.real() + A.imag() * A.imag();
+const APL_Float mag2_B   = B.real() * B.real() + B.imag() * B.imag();
+const APL_Float mag2_max = mag2_A > mag2_B ? mag2_A : mag2_B;
+
+   // 2. compute mag²(A-B)
+   //
 const APL_Complex A_B = A - B;
-const APL_Float dist2_A_B = sqrt(A_B.real() * A_B.real()
-                               + A_B.imag() * A_B.imag());
+const APL_Float dist2_A_B = A_B.real() * A_B.real()
+                          + A_B.imag() * A_B.imag();
+   // compare
+   //
 
-   return (dist2_A_B < C*max_mag);
+   return dist2_A_B <= C*C*mag2_max;
 }
 //-----------------------------------------------------------------------------
 bool
@@ -190,94 +220,41 @@ Cell::tolerantly_equal(APL_Float A, APL_Float B, APL_Float C)
 
 APL_Float mag_A = A < 0 ? -A : A;
 APL_Float mag_B = B < 0 ? -B : B;
-APL_Float max_mag = (mag_A > mag_B) ? mag_A : mag_B;
+APL_Float mag_max = (mag_A > mag_B) ? mag_A : mag_B;
 
 const APL_Float dist_A_B = (A > B) ? (A - B) : (B - A);
 
-   return (dist_A_B < C*max_mag);
+   return (dist_A_B < C*mag_max);
 }
 //-----------------------------------------------------------------------------
 bool
-Cell::is_near_int(APL_Float value, APL_Float qct)
+Cell::is_near_int(APL_Float value)
 {
-   if (value > LARGE_INT)   return false;
-   if (value < SMALL_INT)   return false;
+   if (value > LARGE_INT)   return true;
+   if (value < SMALL_INT)   return true;
 
 const double result = nearbyint(value);
 const double diff = value - result;
-   if (diff > qct)    return false;
-   if (diff < -qct)   return false;
+   if (diff >= INTEGER_TOLERANCE)    return false;
+   if (diff <= -INTEGER_TOLERANCE)   return false;
 
    return true;
 }
 //-----------------------------------------------------------------------------
 APL_Integer
-Cell::near_int(APL_Float value, APL_Float qct)
+Cell::near_int(APL_Float value)
 {
 const double result = nearbyint(value);
 const double diff = value - result;
-   if (diff > qct)    DOMAIN_ERROR;
-   if (diff < -qct)   DOMAIN_ERROR;
+   if (diff > INTEGER_TOLERANCE)    DOMAIN_ERROR;
+   if (diff < -INTEGER_TOLERANCE)   DOMAIN_ERROR;
 
    if (result > 0)   return   APL_Integer(0.3 + result);
    else              return - APL_Integer(0.3 - result);
 }
 //-----------------------------------------------------------------------------
-void 
-Cell::make_heap(const Cell ** a, ShapeItem heapsize, int64_t i, bool ascending,
-                const void * comp_arg, greater_fun greater)
-{
-const int64_t l = 2*i + 1;   // left  child of i.
-const int64_t r = l + 1;     // right child of i.
-int64_t max = i;   // assume parent is the max.
-
-    if ((l < heapsize) && (*greater)(a[l], a[max], ascending, comp_arg))
-       max = l;   // left child is larger
-    if ((r < heapsize) && (*greater)(a[r], a[max], ascending, comp_arg))
-       max = r;   // right child is larger
-    if (max != i)   // parent was not the max: exchange it.
-    {
-        const Cell * t = a[max];   a[max] = a[i];   a[i] = t;
-        make_heap(a, heapsize, max, ascending, comp_arg, greater);
-    }
-}
-//-----------------------------------------------------------------------------
-void 
-Cell::init_heap(const Cell ** a, ShapeItem heapsize, bool ascending,
-                const void * comp_arg, greater_fun greater)
-{
-   for (int64_t p = heapsize/2 - 1; p >= 0; --p)
-       make_heap(a, heapsize, p, ascending, comp_arg, greater);
-
-   // here a is a heap, i.e. a[i >= a[2i+1], a[2i+2]
-}
-//-----------------------------------------------------------------------------
-void
-Cell::heapsort(const Cell ** a, ShapeItem heapsize, bool ascending,
-               const void * comp_arg, greater_fun greater)
-{
-   // Sort a[] into a heap.
-   //
-   init_heap(a, heapsize, ascending, comp_arg, greater);
-
-   // here a[] is a heap, with a[0] being the largest node.
-
-   for (int k = heapsize - 1; k > 0; k--)
-       {
-         // The root a[0] is the largest element (in a[0]..a[k]).
-         // Store the root, replace it by another element.
-         //
-         const Cell * t = a[k];   a[k] = a[0];   a[0] = t;
-
-         // Sort a[] into a heap again.
-         //
-         make_heap(a, k, 0, ascending, comp_arg, greater);
-       }
-}
-//-----------------------------------------------------------------------------
 bool
-Cell::greater_vec(const Cell * ca, const Cell * cb, bool ascending, 
-                  const void * comp_arg)
+Cell::greater_vec(const Cell * ca, const Cell * cb, const void * comp_arg)
 {
 const ShapeItem comp_len = *(const ShapeItem *)comp_arg;
 const APL_Float qct = Workspace::get_CT();
@@ -288,7 +265,7 @@ const APL_Float qct = Workspace::get_CT();
       {
         const bool equal = ca[0].equal(cb[0], qct);
         if (equal)   return ca > cb;
-        const bool result = ca[0].greater(&cb[0], ascending);
+        const bool result = ca[0].greater(cb[0]);
         return result;
       }
 
@@ -296,13 +273,40 @@ const APL_Float qct = Workspace::get_CT();
       {
         const bool equal = ca[c].equal(cb[c], qct);
         if (equal)   continue;
-        const bool result = ca[c].greater(cb + c, ascending);
+        const bool result = ca[c].greater(cb[c]);
         return result;
       }
 
    return ca > cb;   // a and b are equal: sort by position
 }
 //-----------------------------------------------------------------------------
+bool
+Cell::smaller_vec(const Cell * ca, const Cell * cb, const void * comp_arg)
+{
+const ShapeItem comp_len = *(const ShapeItem *)comp_arg;
+const APL_Float qct = Workspace::get_CT();
+
+   // most frequently comp_len is 1, so we optimize for this case.
+   //
+   if (comp_len == 1)
+      {
+        const bool equal = ca[0].equal(cb[0], qct);
+        if (equal)   return ca > cb;
+        const bool result = ca[0].greater(cb[0]);
+        return !result;
+      }
+
+   loop(c, comp_len)
+      {
+        const bool equal = ca[c].equal(cb[c], qct);
+        if (equal)   continue;
+        const bool result = ca[c].greater(cb[c]);
+        return !result;
+      }
+
+   return ca > cb;   // a and b are equal: sort by position
+}
+//-----------------------------------------------------------------------------
 ostream & 
 operator <<(ostream & out, const Cell & cell)
 {
@@ -314,56 +318,44 @@ UCS_string ucs(pb, 0, Workspace::get_PrintContext().get_PW());
 ErrorCode
 Cell::bif_equal(Cell * Z, const Cell * A) const
 {
-APL_Integer eq = 0;   // assume incompatible cell types ( == not equal)
+   // incompatible types ?
+   //
+   if (is_character_cell() != A->is_character_cell())   return IntCell::z0(Z);
 
-   if (is_character_cell() == A->is_character_cell())   // compatible cell types
-      {
-       eq = equal(*A, Workspace::get_CT());
-      }
-
-   new (Z) IntCell(eq);
-   return E_NO_ERROR;
+   return IntCell::zv(Z, equal(*A, Workspace::get_CT()));
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 Cell::bif_not_equal(Cell * Z, const Cell * A) const
 {
-APL_Integer neq = 1;   // assume incompatible cell types ( == not equal)
+   // incompatible types ?
+   //
+   if (is_character_cell() != A->is_character_cell())   return IntCell::z1(Z);
 
-   if (is_character_cell() == A->is_character_cell())   // compatible cell types
-      {
-       neq = !equal(*A, Workspace::get_CT());
-      }
-
-   new (Z) IntCell(neq);
-   return E_NO_ERROR;
+   return IntCell::zv(Z, !equal(*A, Workspace::get_CT()));
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 Cell::bif_greater_than(Cell * Z, const Cell * A) const
 {
-   new (Z) IntCell((A->compare(*this) == COMP_GT) ? 1 : 0);
-   return E_NO_ERROR;
+   return IntCell::zv(Z, (A->compare(*this) == COMP_GT) ? 1 : 0);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 Cell::bif_less_eq(Cell * Z, const Cell * A) const
 {
-   new (Z) IntCell((A->compare(*this) != COMP_GT) ? 1 : 0);
-   return E_NO_ERROR;
+   return IntCell::zv(Z, (A->compare(*this) != COMP_GT) ? 1 : 0);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 Cell::bif_less_than(Cell * Z, const Cell * A) const
 {
-   new (Z) IntCell((A->compare(*this) == COMP_LT) ? 1 : 0);
-   return E_NO_ERROR;
+   return IntCell::zv(Z, (A->compare(*this) == COMP_LT) ? 1 : 0);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 Cell::bif_greater_eq(Cell * Z, const Cell * A) const
 {
-   new (Z) IntCell((A->compare(*this) != COMP_LT) ? 1 : 0);
-   return E_NO_ERROR;
+   return IntCell::zv(Z, (A->compare(*this) != COMP_LT) ? 1 : 0);
 }
 //-----------------------------------------------------------------------------
