@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,6 +18,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+#include <dirent.h>
 #include <dirent.h>
 #include <errno.h>
 #include <limits.h>
@@ -30,10 +31,8 @@
 #include "FloatCell.hh"
 #include "IndexExpr.hh"
 #include "IntCell.hh"
-#include "Input.hh"
 #include "IO_Files.hh"
-#include "LibPaths.hh"
-#include "main.hh"
+#include "LineInput.hh"
 #include "Nabla.hh"
 #include "Output.hh"
 #include "Parser.hh"
@@ -60,14 +59,23 @@ vector<Command::user_command> Command::user_commands;
 void
 Command::process_line()
 {
-UCS_string line = Input::get_line();   // get_line() removes leading whitespace
+UCS_string line;
+bool eof = false;
+   InputMux::get_line(LIM_ImmediateExecution, Workspace::get_prompt(),
+                      line, eof, LineInput::get_history());
+   if (eof)
+      {
+        CERR << "EOF at " << LOC << endl;
+        return;
+      }
+
    process_line(line);
 }
 //-----------------------------------------------------------------------------
 void
 Command::process_line(UCS_string & line)
 {
-   line.remove_leading_and_trailing_whitespaces();
+   line.remove_leading_whitespaces();
    if (line.size() == 0)   return;   // empty input line
 
    switch(line[0])
@@ -118,7 +126,7 @@ vector<UCS_string> args;
         Workspace::more_error().clear();
       }
 
-#define cmd_def(cmd_str, code, _arg) \
+#define cmd_def(cmd_str, code, _arg, _hint) \
    if (cmd.starts_iwith(cmd_str)) { code; return true; }
 #include "Command.def"
 
@@ -149,16 +157,14 @@ Executable * statements = 0;
       }
    catch (Error err)
       {
-        COUT << "SYNTAX ERROR";
-        if (Workspace::more_error().size())   COUT << "+";
-        COUT << endl;
+        UERR << Error::error_name(err.error_code) << endl;
         if (err.get_error_line_2().size())
            {
              COUT << "      " << err.get_error_line_2() << endl
                   << "      " << err.get_error_line_3() << endl;
            }
 
-        err.print(CERR);
+        err.print(UERR);
         return;
       }
    catch (...)
@@ -220,15 +226,14 @@ Executable * statements = 0;
 
 // Q(token)
 
-         // start over if execution has pushed a new context
+         // start over if execution has pushed a new SI entry
          //
          if (token.get_tag() == TOK_SI_PUSHED)   continue;
 
          // maybe call EOC handler and repeat if true returned
          //
        check_EOC:
-         if (Workspace::SI_top()->call_eoc_handler(token))
-            continue;
+         if (Workspace::SI_top()->call_eoc_handler(token))   continue;
 
          // the far most frequent cases are TC_VALUE and TOK_VOID
          // so we handle them first.
@@ -317,6 +322,16 @@ Executable * statements = 0;
 
          if (token.get_tag() == TOK_ERROR)
             {
+              if (token.get_int_val() == E_COMMAND_PUSHED)
+                 {
+                   Workspace::pop_SI(LOC);
+                   UCS_string pushed_command = Workspace::get_pushed_Command();
+                   process_line(pushed_command);
+                   pushed_command.clear();
+                   Workspace::push_Command(pushed_command);   // clear in
+                   return;
+                 }
+
               // clear attention and interrupt flags
               //
               attention_raised = false;
@@ -380,25 +395,58 @@ Executable * statements = 0;
 }
 //-----------------------------------------------------------------------------
 void 
+Command::cmd_XTERM(ostream & out, const UCS_string & arg)
+{
+   const char * term = getenv("TERM");
+   if (!strncmp(term, "dumb", 4) && arg.starts_iwith("ON"))
+      {
+        out << "impossible on dumb terminal" << endl;
+      }
+   else if (arg.starts_iwith("OFF") || arg.starts_iwith("ON"))
+      {
+        Output::toggle_color(arg);
+      }
+   else if (arg.size() == 0)
+      {
+        out << "]COLOR/XTERM ";
+        if (Output::color_enabled()) out << "ON"; else out << "OFF";
+        out << endl;
+      }
+   else
+      {
+        out << "BAD COMMAND" << endl;
+        return;
+      }
+}
+//-----------------------------------------------------------------------------
+void 
 Command::cmd_BOXING(ostream & out, const UCS_string & arg)
 {
 int format = arg.atoi();
 
+   if (arg.size() == 0)
+      {
+        
+        out << "]BOXING ";
+        if (boxing_format == -1) out << "OFF"; else out << boxing_format;
+        out << endl;
+        return;
+      }
+
    if (arg.starts_iwith("OFF"))   format = -1;
 
    switch (format)
       {
         case -1:
-        case  2:
-        case  3:
-        case  4:
-        case  7:
-        case  8:
-        case  9: boxing_format = format;
+        case  2: case   3: case   4:
+        case  7: case   8: case   9:
+        case 20: case  21: case  22:
+        case 23: case  24: case  25:
+                 boxing_format = format;
                  return;
 
         default:   out << "Bad ]BOXING parameter " << arg
-                       << " (valid values are: OFF, 2, 3, 4, 7, 8, and 9)"
+                       << " (valid values are: OFF, 2...4, 7...9, and 20...25)"
                        << endl;
       }
 }
@@ -426,6 +474,7 @@ Command::cmd_CHECK(ostream & out)
           char cc[200];
           snprintf(cc, sizeof(cc), "ERROR   - %d stale values", stale);
           out << cc << endl;
+          IO_Files::apl_error(LOC);
         }
      else out << "OK      - no stale values" << endl;
    }
@@ -436,6 +485,7 @@ Command::cmd_CHECK(ostream & out)
           char cc[200];
           snprintf(cc, sizeof(cc), "ERROR   - %d stale indices", stale);
           out << cc << endl;
+          IO_Files::apl_error(LOC);
         }
      else out << "OK      - no stale indices" << endl;
    }
@@ -457,7 +507,6 @@ Command::cmd_DROP(ostream & out, const vector<UCS_string> & lib_ws)
    // )DROP wsname
    // )DROP libnum wsname
 
-
    // check number of arguments (1 or 2)
    //
    if (lib_ws.size() == 0)   // missing argument
@@ -497,17 +546,36 @@ int result = unlink(filename.c_str());
 void
 Command::cmd_ERASE(ostream & out, vector<UCS_string> & args)
 {
-   Workspace::erase_symbols(CERR, args);
+   Workspace::erase_symbols(out, args);
 }
 //-----------------------------------------------------------------------------
 void 
-Command::cmd_KEYB(ostream & out, const UCS_string & arg)
+Command::cmd_KEYB(ostream & out)
 {
-const int layout = arg.atoi();
-
-   switch(layout)
+   // maybe print user-supplied keyboard layout file
+   //
+   if (uprefs.keyboard_layout_file.size())
       {
-        case 0:
+        FILE * layout = fopen(uprefs.keyboard_layout_file.c_str(), "r");
+        if (layout == 0)
+           {
+             out << "Could not open " << uprefs.keyboard_layout_file
+                 << ": " << strerror(errno) << endl
+                 << "Showing default layout instead" << endl;
+           }
+        else
+           {
+             out << "User-defined Keyboard Layout:\n";
+             for (;;)
+                 {
+                    const int cc = fgetc(layout);
+                    if (cc == EOF)   break;
+                    out << (char)cc;
+                 }
+             out << endl;
+             return;
+           }
+      }
 
    out << "US Keyboard Layout:\n"
                               "\n"
@@ -525,51 +593,104 @@ const int layout = arg.atoi();
 "║  SHIFT      ║ z⊂ ║ x⊃ ║ c∩ ║ v∪ ║ b⊥ ║ n⊤ ║ m| ║ ,⍝ ║ .⍀ ║ /⌿ ║  SHIFT   ║\n"
 "╚═════════════╩════╩════╩════╩════╩════╩════╩════╩════╩════╩════╩══════════╝\n"
    << endl;
-        break;
+}
+//-----------------------------------------------------------------------------
+void 
+Command::cmd_PSTAT(ostream & out, const UCS_string & arg)
+{
+#ifndef PERFORMANCE_COUNTERS_WANTED
+   out << "\n"
+<< "Command ]PSTAT is not available, since performance counters were not\n"
+"configured for this APL interpreter. To enable performance counters (which\n"
+"will slightly decrease performance), recompile the interpreter as follows:"
 
-        case 1:
-   out << "US Keyboard Layout 1:\n"
-                                "\n"
-"╔════╦════╦════╦════╦════╦════╦════╦════╦════╦════╦════╦════╦════╦═════════╗\n"
-"║ ~  ║ !⌶ ║ @⍫ ║ #⍒ ║ $⍋ ║ %⌽ ║ ^⍉ ║ &⊖ ║ *⍟ ║ (⍱ ║ )⍲ ║ _! ║ +⌹ ║         ║\n"
-"║ `◊ ║ 1¨ ║ 2¯ ║ 3< ║ 4≤ ║ 5= ║ 6≥ ║ 7> ║ 8≠ ║ 9∨ ║ 0∧ ║ -× ║ =÷ ║ BACKSP  ║\n"
-"╠════╩══╦═╩══╦═╩══╦═╩══╦═╩══╦═╩══╦═╩══╦═╩══╦═╩══╦═╩══╦═╩══╦═╩══╦═╩══╦══════╣\n"
-"║       ║ Q  ║ W⍹ ║ E⍷ ║ R  ║ T⍨ ║ Y¥ ║ U  ║ I⍸ ║ O⍥ ║ P⍣ ║ {⍞ ║ }⍬ ║  |⊣  ║\n"
-"║  TAB  ║ q? ║ w⍵ ║ e∈ ║ r⍴ ║ t∼ ║ y↑ ║ u↓ ║ i⍳ ║ o○ ║ p⋆ ║ [← ║ ]→ ║  \\⊢  ║\n"
-"╠═══════╩═╦══╩═╦══╩═╦══╩═╦══╩═╦══╩═╦══╩═╦══╩═╦══╩═╦══╩═╦══╩═╦══╩═╦══╩══════╣\n"
-"║ (CAPS   ║ A⍶ ║ S  ║ D  ║ F  ║ G  ║ H⍙ ║ J⍤ ║ K  ║ L⌷ ║ :≡ ║ \"≢ ║         ║\n"
-"║  LOCK)  ║ a⍺ ║ s⌈ ║ d⌊ ║ f_ ║ g∇ ║ h∆ ║ j∘ ║ k' ║ l⎕ ║ ;⍎ ║ '⍕ ║ RETURN  ║\n"
-"╠═════════╩═══╦╩═══╦╩═══╦╩═══╦╩═══╦╩═══╦╩═══╦╩═══╦╩═══╦╩═══╦╩═══╦╩═════════╣\n"
-"║             ║ Z  ║ Xχ ║ C¢ ║ V  ║ B£ ║ N  ║ M  ║ <⍪ ║ >⍙ ║ ?⍠ ║          ║\n"
-"║  SHIFT      ║ z⊂ ║ x⊃ ║ c∩ ║ v∪ ║ b⊥ ║ n⊤ ║ m| ║ ,⍝ ║ .⍀ ║ /⌿ ║  SHIFT   ║\n"
-"╚═════════════╩════╩════╩════╩════╩════╩════╩════╩════╩════╩════╩══════════╝\n"
-   << endl;
-        break;
+<< "\n\n"
+"   ./configure PERFORMANCE_COUNTERS_WANTED=yes (... "
+<< "other configure options"
+<< ")\n"
+"   make\n"
+"   make install (or try: src/apl)\n"
+"\n"
 
-        default: out << "*** unknown keyboard layout " << layout << endl;
+<< "above the src directory."
+<< "\n";
+
+   return;
+#endif
+
+   if (arg.starts_iwith("CLEAR"))
+      {
+        out << "Performance counters cleared" << endl;
+        Performance::reset_all();
+        return;
       }
+
+   if (arg.starts_iwith("SAVE"))
+      {
+        const char * filename = "./PerformanceData.def";
+        ofstream outf(filename, ofstream::out);
+        if (!outf.is_open())
+           {
+             out << "opening " << filename
+                 << " failed: " << strerror(errno) << endl;
+             return;
+           }
+
+        out << "Writing performance data to file " << filename << endl;
+        Performance::save_data(out, outf);
+        return;
+      }
+
+Pfstat_ID iarg = PFS_ALL;
+   if (arg.size() > 0)   iarg = (Pfstat_ID)(arg.atoi());
+
+   Performance::print(iarg, out);
 }
 //-----------------------------------------------------------------------------
 void 
 Command::cmd_HELP(ostream & out)
 {
-   out << "Commands are:" << endl;
-#define cmd_def(cmd_str, _cod, arg) \
-   CERR << "      " cmd_str " " #arg << endl;
+   out << "APL Commands:" << endl;
+#define cmd_def(cmd_str, _cod, arg, _hint) \
+   out << "      " cmd_str " " arg << endl;
 #include "Command.def"
-   out << endl;
 
-   if (user_commands.size())
-      {
-        out << "User defined commands:" << endl;
-        loop(u, user_commands.size())
+  if (user_commands.size())
+     {
+       out << endl << "User defined commands:" << endl;
+       loop(u, user_commands.size())
            {
-             out << "      " << user_commands[u].prefix << " → ";
-             if (user_commands[u].mode)   out << "A ";
-             out << user_commands[u].apl_function << " B"
-                 << " (mode " << user_commands[u].mode << ")" << endl;
+             out << "      " << user_commands[u].prefix << " [args]  calls:  ";
+             if (user_commands[u].mode)   out << "tokenized-args ";
+ 
+             out << user_commands[u].apl_function << " quoted-args" << endl;
            }
-      }
+     }
+
+   out << endl << "System variables:" << endl;
+#define ro_sv_def(x, txt)                                           \
+   { const UCS_string & ucs = Workspace::get_v_ ## x().get_name();  \
+        out << "      " << setw(8) << ucs << #txt << endl; }
+#define rw_sv_def(x, txt)                                           \
+   { const UCS_string & ucs = Workspace::get_v_ ## x().get_name();  \
+        out << "      " << setw(8) << ucs << #txt << endl; }
+#include "SystemVariable.def"
+
+   out << endl << "System functions:" << endl;
+#define ro_sv_def(x, txt)
+#define rw_sv_def(x, txt)
+#define sf_def(q, txt) { const char * qu = #q;                      \
+   if (!strncmp(qu, "Quad_", 5))                                    \
+        out << "      ⎕" << setw(8) << UCS_string(qu + 5) << #txt << endl; }
+#include "SystemVariable.def"
+}
+//-----------------------------------------------------------------------------
+void 
+Command::cmd_HISTORY(ostream & out, const UCS_string & arg)
+{
+   if (arg.size() == 0)                  LineInput::print_history(out);
+   else if (arg.starts_iwith("CLEAR"))   LineInput::clear_history(out);
+   else                                  out << "BAD COMMAND" << endl;
 }
 //-----------------------------------------------------------------------------
 void
@@ -672,18 +793,36 @@ transfer_context tctx(protection);
 }
 //-----------------------------------------------------------------------------
 void 
-Command::cmd_LIBS(ostream & out, const vector<UCS_string> & lib_ref)
+Command::cmd_LIBS(ostream & out, const vector<UCS_string> & args)
 {
    // Command is:
    //
-   // )LIB path           (set root to path)
+   // )LIB N path         (set libdir N to path)
+   // )LIB path           (set libroot to path)
    // )LIB                (display root and path states)
    //
-   if (lib_ref.size() > 0)   // set path
+   if (args.size() >= 2)   // set individual dir
       {
-        UTF8_string utf(lib_ref[0]);
+        const UCS_string & libref_ucs = args[0];
+        const int libref = libref_ucs.atoi();
+        if (libref_ucs.size() != 1 || libref < 0 || libref > 9)
+           {
+             CERR << "Invalid library referenc " << libref_ucs << "'" << endl;
+             return;
+           }
+
+        UTF8_string path(args[1]);
+        LibPaths::set_lib_dir((LibRef)libref, path.c_str(),
+                               LibPaths::LibDir::CSRC_CMD);
+        out << "LIBRARY REFERENCE " << libref << " SET TO " << path << endl;
+        return;
+      }
+
+   if (args.size() == 1)   // set root
+      {
+        UTF8_string utf(args[0]);
         LibPaths::set_APL_lib_root(utf.c_str());
-        out << "LIBRARY ROOT SET TO " << lib_ref[0] << endl;
+        out << "LIBRARY ROOT SET TO " << args[0] << endl;
         return;
       }
 
@@ -703,12 +842,13 @@ Command::cmd_LIBS(ostream & out, const vector<UCS_string> & lib_ref)
           out << " " << d << " ";
           switch(LibPaths::get_cfg_src((LibRef)d))
              {
-                case LibPaths::LibDir::CS_NONE:      out << "NONE" << endl;
+                case LibPaths::LibDir::CSRC_NONE:      out << "NONE" << endl;
                                                      continue;
-                case LibPaths::LibDir::CS_ENV:       out << "ENV   ";   break;
-                case LibPaths::LibDir::CS_ARGV0:     out << "BIN   ";   break;
-                case LibPaths::LibDir::CS_PREF_SYS:  out << "PSYS  ";   break;
-                case LibPaths::LibDir::CS_PREF_HOME: out << "PUSER ";   break;
+                case LibPaths::LibDir::CSRC_ENV:       out << "ENV   ";   break;
+                case LibPaths::LibDir::CSRC_ARGV0:     out << "BIN   ";   break;
+                case LibPaths::LibDir::CSRC_PREF_SYS:  out << "PSYS  ";   break;
+                case LibPaths::LibDir::CSRC_PREF_HOME: out << "PUSER ";   break;
+                case LibPaths::LibDir::CSRC_CMD:       out << "CMD   ";   break;
              }
 
         out << left << setw(52) << path.c_str();
@@ -817,10 +957,11 @@ vector<UCS_string> directories;
          dirent * entry = readdir(dir);
          if (entry == 0)   break;   // directory done
 
-         UCS_string filename(entry->d_name);
+         UTF8_string filename_utf8(entry->d_name);
+         UCS_string filename(filename_utf8);
          if (is_directory(entry, path))
             {
-              if (filename[0] == '.') continue;
+              if (filename_utf8[0] == '.')   continue;
               filename.append(UNI_ASCII_SLASH);
               directories.push_back(filename);
               continue;
@@ -829,15 +970,12 @@ vector<UCS_string> directories;
          const int dlen = strlen(entry->d_name);
          if (variant == 1)
             {
-              if (filename[dlen - 1] != 'l')   continue;   // not .apl or .xml
-              if (filename[dlen - 4] != '.')   continue;   // not .apl or .xml
-
-              if (filename[dlen - 3] == 'a' && filename[dlen - 2] == 'p')
-                {
+              if (filename_utf8.ends_with(".apl"))
+                 {
                    filename.shrink(filename.size() - 4);   // skip extension
                    apl_files.push_back(filename);
                  }
-              else if (filename[dlen - 3] == 'x' && filename[dlen - 2] == 'm')
+              else if (filename_utf8.ends_with(".xml"))
                  {
                    filename.shrink(filename.size() - 4);   // skip extension
                    xml_files.push_back(filename);
@@ -994,7 +1132,7 @@ Command::cmd_MORE(ostream & out)
 void 
 Command::cmd_OFF(int exit_val)
 {
-   cleanup();
+   cleanup(true);
    COUT << endl;
    if (!uprefs.silent)   COUT << "Goodbye." << endl;
    exit(exit_val);
@@ -1090,16 +1228,24 @@ void
 Command::cmd_USERCMD(ostream & out, const UCS_string & cmd,
                      vector<UCS_string> & args)
 {
+   // ]USERCMD
    // ]USERCMD REMOVE-ALL
    // ]USERCMD REMOVE        ]existing-command
    // ]USERCMD ]new-command  APL-fun
    // ]USERCMD ]new-command  APL-fun  mode
    //
-   if (args.size() < 2)
+   if (args.size() == 0)
       {
-        out << "BAD COMMAND" << endl;
-        Workspace::more_error() =
-                   UCS_string("too few parameters in command ]USERCMD");
+        if (user_commands.size())
+           {
+             loop(u, user_commands.size())
+                {
+                  out << user_commands[u].prefix << " → ";
+                  if (user_commands[u].mode)   out << "A ";
+                  out << user_commands[u].apl_function << " B"
+                      << " (mode " << user_commands[u].mode << ")" << endl;
+                }
+           }
         return;
       }
 
@@ -1126,7 +1272,7 @@ Command::cmd_USERCMD(ostream & out, const UCS_string & cmd,
                 }
            }
 
-       out << "BAD COMMAND" << endl;
+       out << "BAD COMMAND+" << endl;
        Workspace::more_error() = UCS_string(
                 "user command in command ]USERCMD REMOVE does not exist");
        return;
@@ -1134,7 +1280,7 @@ Command::cmd_USERCMD(ostream & out, const UCS_string & cmd,
 
    if (args.size() > 3)
       {
-        out << "BAD COMMAND" << endl;
+        out << "BAD COMMAND+" << endl;
         Workspace::more_error() =
                    UCS_string("too many parameters in command ]USERCMD");
         return;
@@ -1143,9 +1289,9 @@ Command::cmd_USERCMD(ostream & out, const UCS_string & cmd,
 const int mode = (args.size() == 3) ? args[2].atoi() : 0;
    if (mode < 0 || mode > 1)
       {
-        out << "BAD COMMAND" << endl;
+        out << "BAD COMMAND+" << endl;
         Workspace::more_error() =
-                   UCS_string("unsupported mode in command ]USERCMD");
+           UCS_string("unsupported mode in command ]USERCMD (0 or 1 expected)");
         return;
       }
 
@@ -1158,7 +1304,7 @@ const int mode = (args.size() == 3) ? args[2].atoi() : 0;
         else          error = error || !Avec::is_symbol_char(args[0][c]);
         if (error)
            {
-             out << "BAD COMMAND" << endl;
+             out << "BAD COMMAND+" << endl;
              Workspace::more_error() =
                    UCS_string("bad user command name in command ]USERCMD");
              return;
@@ -1167,7 +1313,7 @@ const int mode = (args.size() == 3) ? args[2].atoi() : 0;
 
    // check conflicts with existing commands
    //
-#define cmd_def(cmd_str, _cod, arg) \
+#define cmd_def(cmd_str, _cod, _arg, _hint) \
    if (check_name_conflict(out, cmd_str, args[0]))   return;
 #include "Command.def"
    if (check_redefinition(out, args[0], args[1], mode))
@@ -1183,7 +1329,7 @@ const int mode = (args.size() == 3) ? args[2].atoi() : 0;
       {
         if (!Avec::is_symbol_char(args[1][c]))
            {
-             out << "BAD COMMAND" << endl;
+             out << "BAD COMMAND+" << endl;
              Workspace::more_error() =
                    UCS_string("bad APL function name in command ]USERCMD");
              return;
@@ -1222,7 +1368,10 @@ Command::do_USERCMD(ostream & out, UCS_string & apl_cmd,
 void
 Command::log_control(const UCS_string & arg)
 {
-   if (arg.size() == 0 || arg[0] == UNI_ASCII_QUESTION)  // no arg or '?'
+vector<UCS_string> args;
+   arg.copy_black_list(args);
+
+   if (args.size() == 0 || arg[0] == UNI_ASCII_QUESTION)  // no arg or '?'
       {
         for (LogId l = LID_MIN; l < LID_MAX; l = LogId(l + 1))
             {
@@ -1237,23 +1386,19 @@ Command::log_control(const UCS_string & arg)
         return;
       }
 
-LogId val = LogId(0);
-bool skip_leading_ws = true;
-
-   loop(a, arg.size())
-      {
-        if ((arg[a] <= ' ') && skip_leading_ws)   continue;
-        skip_leading_ws = false;
-        if (arg[a] < UNI_ASCII_0)   break;
-        if (arg[a] > UNI_ASCII_9)   break;
-        val = LogId(10*val + arg[a] - UNI_ASCII_0);
-      }
+const LogId val = LogId(args[0].atoi());
+int on_off = -1;
+   if      (args.size() > 1 && args[1].starts_iwith("ON"))    on_off = 1;
+   else if (args.size() > 1 && args[1].starts_iwith("OFf"))   on_off = 0;
 
    if (val >= LID_MIN && val <= LID_MAX)
       {
         const char * info = Log_info(val);
         Assert(info);
-        const bool new_status = !Log_status(val);
+        bool new_status = !Log_status(val);   // toggle
+        if (on_off == 0)        new_status = false;
+        else if (on_off == 1)   new_status = true;
+
         Log_control(val, new_status);
         CERR << "    Log facility '" << info << "' is now "
              << (new_status ? "ON " : "OFF") << endl;
@@ -1417,7 +1562,7 @@ Token_string tos;
  
    if (tos.size() != shape.get_volume())   return;
 
-Value_P val(new Value(shape, LOC));
+Value_P val(shape, LOC);
    new (&val->get_ravel(0)) IntCell(0);   // prototype
 
 const ShapeItem ec = val->element_count();
@@ -1467,7 +1612,7 @@ Symbol * sym = 0;
         CERR << "'" << endl;
       }
 
-Value_P val(new Value(shape, LOC));
+Value_P val(shape, LOC);
    new (&val->get_ravel(0)) CharCell(UNI_ASCII_SPACE);   // prototype
 
 const ShapeItem ec = val->element_count();
@@ -1657,3 +1802,499 @@ Command::is_lib_ref(const UCS_string & lib)
    return false;
 }
 //-----------------------------------------------------------------------------
+ExpandResult
+Command::expand_tab(UCS_string & user, int & replace_count)
+{
+   replace_count = user.size();
+
+const bool have_trailing_blank = replace_count && user.last() == ' ';
+
+   // skip leading and trailing blanks
+   //
+   user.remove_leading_and_trailing_whitespaces();
+      
+   if (user.size() == 0 || Avec::is_first_symbol_char(user[0]))
+      {
+        // nothing or user-defined name: expand names
+        //
+        const int symbol_count = Workspace::symbols_allocated();
+        DynArray(Symbol *, all_symbols, symbol_count);
+        Workspace::get_all_symbols(&all_symbols[0], symbol_count);
+
+        vector<UCS_string>matches;
+        loop(s, symbol_count)
+            {
+             const Symbol * sym = all_symbols[s];
+             if (sym->is_erased())   continue;
+
+              const UCS_string & sym_name = sym->get_name();
+              if (!sym_name.starts_with(user))   continue;
+              matches.push_back(sym_name);
+            }
+
+        if (matches.size() == 0)   return ER_IGNORE;   // no match
+        if (matches.size() > 1)   // multiple names match user input
+           {
+             const int user_len = user.size();
+             user.clear();
+             return show_alternatives(user, user_len, matches);
+           }
+
+        // unique match
+        //
+        if (user.size() < matches[0].size())
+           {
+             // the name is longer than user, so we expand it.
+             //
+             user = matches[0];
+             return ER_REPLACE;
+           }
+        return ER_IGNORE;
+      }
+
+   if (user[0] == ')' || user[0] == ']')   // APL command
+      {
+        ExpandHint ehint = EH_NO_PARAM;
+        const char * shint = 0;
+        vector<UCS_string>matches;
+        UCS_string cmd = user;
+        UCS_string arg;
+        cmd.split_ws(arg);
+
+#define cmd_def(cmd_str, code, arg, hint)  \
+   { UCS_string ustr(cmd_str);             \
+     if (ustr.starts_iwith(cmd))           \
+        { matches.push_back(ustr); ehint = hint; shint = arg; } }
+#include "Command.def"
+
+        // if no match was found then ignore the TAB
+        //
+        if (matches.size() == 0)   return ER_IGNORE;
+
+        // if we have multiple matches but the user has provided a command
+        // argument then some matches were wrong. For example:
+        //
+        // LIB 3 matches LIB and LIBS
+        //
+        // remove wrong matches
+        //
+        if (matches.size() > 1 && arg.size() > 0)
+           {
+             again:
+             if (matches.size() > 1)
+                {
+                  loop(m, matches.size())
+                     {
+                       if (matches[m].size() != cmd.size())   // wrong match
+                          {
+                             matches[m].swap(matches.back());
+                             matches.pop_back();
+                             goto again;
+                          }
+                     }
+                }
+           }
+
+        // if multiple matches were found then either expand the common part
+        // or list all matches
+        //
+        if (matches.size() > 1)   // multiple commands match cmd
+           {
+             user.clear();
+             return show_alternatives(user, cmd.size(), matches);
+           }
+
+        // unique match
+        //
+        if (cmd.size() < matches[0].size())
+           {
+             // the command is longer than user, so we expand it.
+             //
+             user = matches[0];
+             if (ehint != EH_NO_PARAM)   user.append(UNI_ASCII_SPACE);
+             return ER_REPLACE;
+           }
+
+        if (cmd.size() == matches[0].size() &&
+            ehint != EH_NO_PARAM            &&
+            !arg.size()                     &&
+            !have_trailing_blank)   // no blank entered
+           {
+             // the entire command was entered but without a blank. If the
+             // command has arguments then append a space to indicate that.
+             // Otherwiese fall throught to expand_command_arg();
+             //
+             user = matches[0];
+             user.append(UNI_ASCII_SPACE);
+             return ER_REPLACE;
+           }
+        return expand_command_arg(user, have_trailing_blank,
+                                  ehint, shint, matches[0], arg);
+      }
+
+   // not an APL command.
+   //
+   user.remove_leading_and_trailing_whitespaces();
+int qpos = -1;
+   loop(e, 5)
+      {
+        if (e >= user.size())   break;
+        if (user[user.size() - e - 1] == UNI_Quad_Quad)
+           {
+             qpos = user.size() - e;
+             break;
+           }
+      }
+
+   if (qpos != -1)   // ⎕xxx at end
+      {
+        UCS_string qxx(user, qpos, user.size() - qpos);
+        vector<UCS_string>matches;
+
+#define ro_sv_def(q, txt) { const char * qu = #q;                  \
+   if (!strncmp(qu, "Quad_", 5)) { UCS_string ustr(qu + 5);   \
+        if (ustr.starts_iwith(qxx)) matches.push_back(ustr); } }
+
+#define rw_sv_def(q, txt) { const char * qu = #q;                  \
+   if (!strncmp(qu, "Quad_", 5)) { UCS_string ustr(qu + 5);   \
+        if (ustr.starts_iwith(qxx)) matches.push_back(ustr); } }
+
+#define sf_def(q, txt) { const char * qu = #q;                  \
+   if (!strncmp(qu, "Quad_", 5)) { UCS_string ustr(qu + 5);   \
+        if (ustr.starts_iwith(qxx)) matches.push_back(ustr); } }
+
+#include "SystemVariable.def"
+
+        if (matches.size() == 0)   return ER_IGNORE;
+        if (matches.size() > 1)
+           {
+            UCS_string::sort_names(matches);
+
+            const int common_len = compute_common_length(qxx.size(), matches);
+            if (common_len == qxx.size())
+               {
+                 // qxx is already the common part of all matching ⎕xx
+                 // display matching ⎕xx
+                 //
+                 CIN << endl;
+                 loop(m, matches.size())
+                     {
+                        CERR << "⎕" << matches[m] << " ";
+                     }
+                 CERR << endl;
+                 return ER_AGAIN;
+               }
+            else
+               {
+                 // qxx is a prefix of the common part of all matching ⎕xx.
+                 // expand to common part.
+                 //
+                 user = matches[0];
+                 user.shrink(common_len);
+                 return ER_REPLACE;
+               }
+           }
+
+        // unique match
+        //
+        user = UCS_string(UNI_Quad_Quad);
+        user.append(matches[0]);
+        return ER_REPLACE;
+      }
+
+   return ER_IGNORE;
+}
+//-----------------------------------------------------------------------------
+ExpandResult
+Command::expand_command_arg(UCS_string & user, bool have_trailing_blank,
+                            ExpandHint ehint, const char * shint,
+                            const UCS_string cmd, const UCS_string arg)
+{
+   switch(ehint)
+      {
+        case EH_NO_PARAM:  return ER_IGNORE;
+
+        case EH_oWSNAME:
+        case EH_oLIB_WSNAME:
+        case EH_oLIB_oPATH:
+        case EH_FILENAME:
+        case EH_DIR_OR_LIB:
+        case EH_WSNAME:    return expand_filename(user, have_trailing_blank,
+                                                  ehint, shint, cmd, arg);
+
+        default:
+             CIN << endl;
+             CERR << cmd << " " << shint << endl;
+             return ER_AGAIN;
+      }
+}
+//-----------------------------------------------------------------------------
+ExpandResult
+Command::expand_filename(UCS_string & user, bool have_trailing_blank,
+                         ExpandHint ehint, const char * shint,
+                         const UCS_string cmd, UCS_string arg)
+{
+   if (arg.size() == 0)
+      {
+        // the user has entered a command but nothing else.
+        // If the command accepts a library number then we propose one of
+        // the existing libraries.
+        //
+        if (ehint == EH_oLIB_WSNAME || ehint == EH_DIR_OR_LIB)
+           {
+             // the command accepts a library reference number
+             //
+             UCS_string libs_present;
+             loop(lib, 10)
+                 {
+                    if (!LibPaths::is_present((LibRef)lib))   continue;
+                       
+                    libs_present.append(UNI_ASCII_SPACE);
+                    libs_present.append((Unicode)(UNI_ASCII_0 + lib));
+                 }
+             if (libs_present.size() == 0)   goto nothing;
+
+             CIN << endl;
+             CERR << cmd << libs_present << " <workspace-name>" << endl;
+             return ER_AGAIN;
+           }
+        
+        goto nothing;
+      }
+
+   if (arg[0] >= '0' && arg[0] <= '9')
+      {
+        // library number 0-9. EH_DIR_OR_LIB is complete already so
+        // EH_oLIB_WSNAME is the only expansion case left
+        //
+        if (ehint != EH_oLIB_WSNAME)   goto nothing;
+
+        const LibRef lib = (LibRef)(arg[0] - '0');
+
+        if (arg.size() == 1 && !have_trailing_blank)   // no space yet
+           {
+             user = cmd;
+             user.append(UNI_ASCII_SPACE);
+             user.append((Unicode)(arg[0]));
+             user.append(UNI_ASCII_SPACE);
+             return ER_REPLACE;
+           }
+
+         // discard library reference number
+         //
+         if (arg.size() == 1)   arg.drop_leading(1);
+         else                   arg.drop_leading(2);
+         return expand_wsname(user, cmd, lib, arg);
+      }
+
+   {
+     UCS_string dir_ucs;
+
+     if (arg[0] == '/')
+        {
+          dir_ucs = arg;
+        }
+     else if (arg[0] == '.')
+        {
+          const char * pwd = getenv("PWD");
+          if (pwd == 0)   goto nothing;
+          dir_ucs = UCS_string(pwd);
+          dir_ucs.append(UNI_ASCII_SLASH);
+          dir_ucs.append(arg);
+        }
+     else goto nothing;
+
+     UTF8_string dir_utf = UTF8_string(dir_ucs);
+
+     const char * dir_dirname = dir_utf.c_str();
+     const char * dir_basename = strrchr(dir_dirname, '/');
+     if (dir_basename == 0)   goto nothing;
+
+     UTF8_string base_utf = UTF8_string(++dir_basename);
+     dir_utf.shrink(dir_basename - dir_dirname);
+     dir_ucs = UCS_string(dir_utf);
+ 
+     DIR * dir = opendir(dir_utf.c_str());
+     if (dir == 0)   goto nothing;
+
+     vector<UCS_string>matches;
+     read_matching_filenames(dir, dir_utf, base_utf, ehint, matches);
+     closedir(dir);
+
+     if (matches.size() == 0)
+        {
+          CIN << endl;
+          CERR << "  no matching filesnames" << endl;
+          return ER_AGAIN;
+        }
+
+     if (matches.size() > 1)
+        {
+          UCS_string prefix(base_utf);
+          user = cmd;
+          user.append(UNI_ASCII_SPACE);
+          user.append(dir_ucs);
+          user.append(prefix);
+          return show_alternatives(user, prefix.size(), matches);
+        }
+
+     // unique match
+     //
+     user = cmd;
+     user.append(UNI_ASCII_SPACE);
+     user.append(dir_ucs);
+     user.append(matches[0]);
+     return ER_REPLACE;
+   }
+
+nothing:
+   CIN << endl;
+   CERR << cmd << " " << shint << endl;
+   return ER_AGAIN;
+}
+//-----------------------------------------------------------------------------
+ExpandResult
+Command::expand_wsname(UCS_string & user, const UCS_string cmd,
+                       LibRef lib, const UCS_string filename)
+{
+UTF8_string path = LibPaths::get_lib_dir(lib);
+   if (path.size() == 0)
+      {
+        CIN << endl;
+        CERR << "Invalib library reference " << lib << endl;
+        return ER_AGAIN;
+      }
+
+DIR * dir = opendir(path.c_str());
+   if (dir == 0)
+      {
+        CIN << endl;
+        CERR
+<< "  library reference " << lib
+<< " is a valid number, but the corresponding directory " << endl
+<< "  " << path << " does not exist" << endl
+<< "  or is not readable. " << endl
+<< "  The relation between library reference numbers and filenames" << endl
+<< "  (aka. paths) can be configured file 'preferences'." << endl << endl
+<< "  At this point, you can use a path instead of the optional" << endl
+<< "  library reference number and the workspace name." << endl;
+
+        user = cmd;
+        user.append(UNI_ASCII_SPACE);
+        return ER_REPLACE;
+      }
+
+vector<UCS_string>matches;
+UTF8_string arg_utf(filename);
+   read_matching_filenames(dir, path, arg_utf, EH_oLIB_WSNAME, matches);
+   closedir(dir);
+
+   if (matches.size() == 0)   goto nothing;
+   if (matches.size() > 1)
+      {
+        user = cmd;
+        user.append(UNI_ASCII_SPACE);
+        user.append_number(lib);
+        user.append(UNI_ASCII_SPACE);
+        return show_alternatives(user, filename.size(), matches);
+      }
+
+   // unique match
+   //
+   user = cmd;
+   user.append(UNI_ASCII_SPACE);
+   user.append_utf8(path.c_str());
+   user.append(UNI_ASCII_SLASH);
+   user.append(matches[0]);
+   return ER_REPLACE;
+
+nothing:
+   CIN << endl;
+   CERR << cmd << " " << lib << " '" << filename << "'" << endl;
+   return ER_AGAIN;
+}
+//-----------------------------------------------------------------------------
+int
+Command::compute_common_length(int len, const vector<UCS_string> & matches)
+{
+   // we assume that all matches have the same case
+
+   for (;; ++len)
+       {
+         loop(m, matches.size())
+            {
+              if (len >= matches[m].size())   return matches[m].size();
+              if (matches[0][len] != matches[m][len])    return len;
+            }
+       }
+}
+//-----------------------------------------------------------------------------
+void
+Command::read_matching_filenames(DIR * dir, UTF8_string dirname,
+                                 UTF8_string prefix, ExpandHint ehint,
+                                 vector<UCS_string> & matches)
+{
+const bool only_workspaces = (ehint == EH_oLIB_WSNAME) ||
+                             (ehint == EH_WSNAME     ) ||
+                             (ehint == EH_oWSNAME    );
+
+   for (;;)
+       {
+          struct dirent * dent = readdir(dir);
+          if (dent == 0)   break;
+
+          const size_t dlen = strlen(dent->d_name);
+          if (dlen == 1 && dent->d_name[0] == '.')   continue;
+          if (dlen == 2 && dent->d_name[0] == '.'
+                        && dent->d_name[1] == '.')   continue;
+
+          if (strncmp(dent->d_name, prefix.c_str(), prefix.size()))   continue;
+
+          UCS_string name(dent->d_name);
+
+          const bool is_dir = is_directory(dent, dirname);
+          if (is_dir)   name.append(UNI_ASCII_SLASH);
+          else if (only_workspaces)
+             {
+               const UTF8_string filename(dent->d_name);
+               bool is_wsname = false;
+               if (filename.ends_with(".apl"))   is_wsname = true;
+               if (filename.ends_with(".xml"))   is_wsname = true;
+               if (!is_wsname)   continue;
+             }
+
+          matches.push_back(name);
+       }
+}
+//-----------------------------------------------------------------------------
+ExpandResult
+Command::show_alternatives(UCS_string & user, int prefix_len,
+                           vector<UCS_string>matches)
+{
+const int common_len = compute_common_length(prefix_len, matches);
+
+   if (common_len == prefix_len)
+      {
+        // prefix is already the common part of all matching files
+        // display matching items
+        //
+        CIN << endl;
+        loop(m, matches.size())
+            {
+              CERR << matches[m] << " ";
+            }
+        CERR << endl;
+        return ER_AGAIN;
+      }
+   else
+      {
+        // prefix is a prefix of the common part of all matching files.
+        // expand to common part.
+        //
+        const int usize = user.size();
+        user.append(matches[0]);
+        user.shrink(usize + common_len);
+        return ER_REPLACE;
+      }
+}
+//-----------------------------------------------------------------------------
