@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -21,17 +21,118 @@
 #include <fcntl.h>
 #include <stdio.h>
 #include <string.h>
-#include <sys/types.h>
+#include <sys/resource.h>
 #include <sys/stat.h>
 #include <sys/time.h>
+#include <sys/types.h>
 
+#include "CharCell.hh"
 #include "Common.hh"
+#include "ComplexCell.hh"
 #include "FloatCell.hh"
+#include "LineInput.hh"
+#include "LibPaths.hh"
+#include "NativeFunction.hh"
 #include "Output.hh"
+#include "ProcessorID.hh"
+#include "Svar_DB.hh"
 #include "Value.icc"
+#include "UserPreferences.hh"
+#include "ValueHistory.hh"
 
 uint64_t total_memory = 0;
 
+APL_time_us interrupt_when = 0;
+uint64_t interrupt_count = 0;
+bool interrupt_raised = false;
+
+bool attention_raised = false;
+uint64_t attention_count = 0;
+
+//-----------------------------------------------------------------------------
+void
+init_1(const char * argv0, bool log_startup)
+{
+rlimit rl;
+   getrlimit(RLIMIT_AS, &rl);
+   total_memory = rl.rlim_cur;
+
+   if (log_startup)
+      CERR << "sizeof(Svar_record) is    " << sizeof(Svar_record) << endl
+           << "sizeof(Svar_partner) is   " << sizeof(Svar_partner)
+           << endl;
+
+   // CYGWIN does not have RLIMIT_NPROC
+   //
+#ifdef RLIMIT_NPROC
+   getrlimit(RLIMIT_NPROC, &rl);
+   if (log_startup)
+      CERR << "increasing rlimit RLIMIT_NPROC from " <<  rl.rlim_cur
+           << " to infinity" << endl;
+   rl.rlim_cur = RLIM_INFINITY;
+   setrlimit(RLIMIT_NPROC, &rl);
+#endif
+
+   Avec::init();
+   LibPaths::init(argv0, log_startup);
+   Value::init();
+   VH_entry::init();
+}
+//-----------------------------------------------------------------------------
+/// initialize subsystems that depend on argv[]
+void
+init_2(bool log_startup)
+{
+   Output::init(log_startup);
+   Svar_DB::init(LibPaths::get_APL_bin_path(),
+                 LibPaths::get_APL_bin_name(),
+                 log_startup, uprefs.system_do_svars);
+
+   LineInput::init(true);
+
+   Parallel::init(log_startup || LOG_Parallel);
+}
+//-----------------------------------------------------------------------------
+/// the opposite of init()
+void
+cleanup(bool soft)
+{
+   if (soft)   // proper clean-up
+      {
+        ProcessorID::disconnect();
+
+        NativeFunction::cleanup();
+
+        // write line history
+        //
+        LineInput::close(false);
+
+        Output::reset_colors();
+      }
+   else        // minimal clean-up
+      {
+        LineInput::close(true);
+        Output::reset_colors();
+      }
+}
+//-----------------------------------------------------------------------------
+void
+control_C(int)
+{
+APL_time_us when = now();
+
+   CIN << "^C";
+
+   attention_raised = true;
+   ++attention_count;
+   if ((when - interrupt_when) < 1000000)   // second ^C within 1 second
+      {
+        interrupt_raised = true;
+        ++interrupt_count;
+      }
+
+   interrupt_when = when;
+}
 //-----------------------------------------------------------------------------
 
 // Probes...
@@ -115,11 +216,10 @@ operator << (ostream & out, const Function_PC2 & ft)
 }
 //-----------------------------------------------------------------------------
 ostream &
-print_flags (ostream & out, ValueFlags flags)
+print_flags(ostream & out, ValueFlags flags)
 {
    return out << ((flags & VF_marked)   ?  "M" : "-")
-              << ((flags & VF_complete) ?  "C" : "-")
-              << ((flags & VF_forever)  ?  "∞" : "-");
+              << ((flags & VF_complete) ?  "C" : "-");
 }
 //-----------------------------------------------------------------------------
 int
@@ -162,7 +262,7 @@ sixbit(Unicode uni)
 Value_P
 IntScalar(APL_Integer val, const char * loc)
 {
-Value_P Z(new Value(loc));
+Value_P Z(loc);
    new (Z->next_ravel())   IntCell(val);
    return Z;
 }
@@ -170,31 +270,39 @@ Value_P Z(new Value(loc));
 Value_P
 FloatScalar(APL_Float val, const char * loc)
 {
-Value_P Z(new Value(loc));
+Value_P Z(loc);
    new (Z->next_ravel())   FloatCell(val);
    return Z;
 }
 //-----------------------------------------------------------------------------
 Value_P
+ComplexScalar(APL_Complex val, const char * loc)
+{
+Value_P Z(loc);
+   new (Z->next_ravel()) ComplexCell(val);
+   return Z;
+}
+//-----------------------------------------------------------------------------
+Value_P
+CharScalar(Unicode uni, const char * loc)
+{
+Value_P Z(loc);
+   new (Z->next_ravel()) CharCell(uni);
+   return Z;
+}
+//-----------------------------------------------------------------------------
+Value_P
 Idx0(const char * loc)
 {
-Value_P Z(new Value((ShapeItem)0, loc));
+Value_P Z((ShapeItem)0, loc);
    new (&Z->get_ravel(0))   IntCell(0);
    return Z;
 }
 //-----------------------------------------------------------------------------
 Value_P
-Spc(const char * loc)
-{
-Value_P Z(new Value(loc));
-   new (Z->next_ravel())   CharCell(UNI_ASCII_SPACE);
-   return Z;
-}
-//-----------------------------------------------------------------------------
-Value_P
 Str0(const char * loc)
 {
-Value_P Z(new Value((ShapeItem)0, loc));
+Value_P Z((ShapeItem)0, loc);
    new (&Z->get_ravel(0))   CharCell(UNI_ASCII_SPACE);
    return Z;
 }
@@ -203,7 +311,7 @@ Value_P
 Str0_0(const char * loc)
 {
 Shape sh((ShapeItem)0, (ShapeItem)0);
-Value_P Z(new Value(sh, loc));
+Value_P Z(sh, loc);
    new (&Z->get_ravel(0))   CharCell(UNI_ASCII_SPACE);
    return Z;
 }
