@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -34,125 +34,81 @@
 //-----------------------------------------------------------------------------
 ComplexCell::ComplexCell(APL_Complex c)
 {
-   value.cpxp = new APL_Complex(c);
-   Log(LOG_delete)
-      CERR << "new    " << (const void *)value.cpxp << " at " LOC << endl;
+   value.cval_r  = c.real();
+   value2.cval_i = c.imag();
 }
 //-----------------------------------------------------------------------------
 ComplexCell::ComplexCell(APL_Float r, APL_Float i)
 {
-   value.cpxp = new APL_Complex(r, i);
-   Log(LOG_delete)
-      CERR << "new    " << (const void *)value.cpxp << " at " LOC << endl;
-}
-//-----------------------------------------------------------------------------
-ComplexCell::~ComplexCell()
-{
-   Log(LOG_delete)   CERR << "delete " HEX(value.cpxp) << " at " LOC << endl;
-   delete value.cpxp;
+   value.cval_r  = r;
+   value2.cval_i = i;
 }
 //-----------------------------------------------------------------------------
 APL_Float
 ComplexCell::get_real_value() const
 {
-   return value.cpxp->real();
+   return value.cval_r;
 }
 //-----------------------------------------------------------------------------
 APL_Float
 ComplexCell::get_imag_value() const
 {
-   return value.cpxp->imag();
+   return value2.cval_i;
 }
 //-----------------------------------------------------------------------------
 bool
-ComplexCell::is_near_int(APL_Float qct) const
+ComplexCell::is_near_int() const
 {
-   return Cell::is_near_int(value.cpxp->real(), qct) &&
-          Cell::is_near_int(value.cpxp->imag(), qct);
-}
-//-----------------------------------------------------------------------------
-void
-ComplexCell::demote_complex_to_real(APL_Float qct)
-{
-const bool small_real = value.cpxp->real() <  qct &&
-                        value.cpxp->real() >  -qct;
-
-const bool small_imag = value.cpxp->imag() <  qct &&
-                        value.cpxp->imag() >  -qct;
-
-const APL_Float qct1 = 100000*qct;
-const bool large_real = value.cpxp->real() >  qct1 ||
-                        value.cpxp->real() <  -qct1;
-
-const bool large_imag = value.cpxp->imag() >  qct1 ||
-                        value.cpxp->imag() <  -qct1;
-
-   if (small_imag && large_real)
-      {
-        // imag is close to zero, so we demote this cell either
-        // to a FloatCell, or to an IntCell
-        //
-        const double val = value.cpxp->real();
-        delete value.cpxp;
-
-        if (Cell::is_near_int(val, qct))
-           {
-             if (val < 0)   new (this)   IntCell(APL_Integer(val - 0.3));
-             else           new (this)   IntCell(APL_Integer(val + 0.3));
-           }
-        else
-           {
-             new (this) FloatCell(val);
-           }
-      }
-   else if (small_real && large_imag)
-      {
-        // real is close to zero, so we round it down.
-        //
-        *value.cpxp = APL_Complex(0.0, value.cpxp->imag());
-      }
+   return Cell::is_near_int(value.cval_r) &&
+          Cell::is_near_int(value2.cval_i);
 }
 //-----------------------------------------------------------------------------
 bool
-ComplexCell::is_near_zero(APL_Float qct) const
+ComplexCell::is_near_zero() const
 {
-   if (value.cpxp->real() >  qct)   return false;
-   if (value.cpxp->real() < -qct)   return false;
-   if (value.cpxp->imag() >  qct)   return false;
-   if (value.cpxp->imag() < -qct)   return false;
+   if (value.cval_r  >=  INTEGER_TOLERANCE)   return false;
+   if (value.cval_r  <= -INTEGER_TOLERANCE)   return false;
+   if (value2.cval_i >=  INTEGER_TOLERANCE)   return false;
+   if (value2.cval_i <= -INTEGER_TOLERANCE)   return false;
    return true;
 }
 //-----------------------------------------------------------------------------
 bool
-ComplexCell::is_near_one(APL_Float qct) const
+ComplexCell::is_near_one() const
 {
-   if (value.cpxp->real() >  (1.0 + qct))   return false;
-   if (value.cpxp->real() <  (1.0 - qct))   return false;
-   if (value.cpxp->imag() >  qct)           return false;
-   if (value.cpxp->imag() < -qct)           return false;
+   if (value.cval_r >  (1.0 + INTEGER_TOLERANCE))   return false;
+   if (value.cval_r <  (1.0 - INTEGER_TOLERANCE))   return false;
+   if (value2.cval_i >  INTEGER_TOLERANCE)           return false;
+   if (value2.cval_i < -INTEGER_TOLERANCE)           return false;
    return true;
 }
 //-----------------------------------------------------------------------------
 bool
+ComplexCell::is_near_real() const
+{
+const APL_Float B = REAL_TOLERANCE*REAL_TOLERANCE;
+const double I = value2.cval_i * value2.cval_i;
+
+   if (I < B)     return true;   // I is absolutely small
+
+const double R = value.cval_r * value.cval_r;
+   if (I < R*B)   return true;   // I is relatively small
+   return false;
+}
+//-----------------------------------------------------------------------------
+bool
 ComplexCell::equal(const Cell & A, APL_Float qct) const
 {
    if (!A.is_numeric())   return false;
    return tolerantly_equal(A.get_complex_value(), get_complex_value(), qct);
 }
 //-----------------------------------------------------------------------------
-void
-ComplexCell::release(const char * loc)
-{
-   delete value.cpxp;
-   new (this) Cell;
-}
-//-----------------------------------------------------------------------------
 // monadic build-in functions...
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_factorial(Cell * Z) const
 {
-   if (is_near_real(Workspace::get_CT()))
+   if (is_near_real())
       {
         const FloatCell fc(get_real_value());
         fc.bif_factorial(Z);
@@ -171,55 +127,61 @@ const APL_Complex z(zr, zi);
 ErrorCode
 ComplexCell::bif_conjugate(Cell * Z) const
 {
-   new (Z) ComplexCell(conj(*value.cpxp));
+   new (Z) ComplexCell(conj(cval()));
    return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_negative(Cell * Z) const
 {
-   new (Z) ComplexCell(- *value.cpxp);
+   new (Z) ComplexCell(-cval());
    return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_direction(Cell * Z) const
 {
-const APL_Complex val = *value.cpxp; 
-const APL_Float mag = abs(val);
+const APL_Float mag = abs(cval());
    if (mag == 0.0)   new (Z) IntCell(0);
-   else              new (Z) ComplexCell(val/mag);
+   else              new (Z) ComplexCell(cval()/mag);
    return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_magnitude(Cell * Z) const
 {
-   new (Z) FloatCell(abs(*value.cpxp));
+   new (Z) FloatCell(abs(cval()));
    return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_reciprocal(Cell * Z) const
 {
-const APL_Float qct = Workspace::get_CT();
-   if (is_near_zero(qct))  return E_DOMAIN_ERROR;
+   if (is_near_zero())  return E_DOMAIN_ERROR;
 
-   if (is_near_real(qct))   new (Z) FloatCell(1.0/value.cpxp->real());
-   else                     new (Z) ComplexCell(APL_Complex(1.0)/ *value.cpxp);
-   return E_NO_ERROR;
+   if (is_near_real())
+      {
+        const APL_Float z = 1.0/value.cval_r;
+        if (!isfinite(z))   return E_DOMAIN_ERROR;
+        return FloatCell::zv(Z, z);
+      }
+
+const APL_Complex z = 1.0 / cval();
+   if (!isfinite(z.real()))   return E_DOMAIN_ERROR;
+   if (!isfinite(z.imag()))   return E_DOMAIN_ERROR;
+   return zv(Z, z);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_roll(Cell * Z) const
 {
 const APL_Integer qio = Workspace::get_IO();
-   if (!is_near_int(Workspace::get_CT()))   return E_DOMAIN_ERROR;
+   if (!is_near_int())   return E_DOMAIN_ERROR;
 
 const APL_Integer set_size = get_checked_near_int();
    if (set_size <= 0)   return E_DOMAIN_ERROR;
 
-const uint64_t rnd = Workspace::get_RL();
+const uint64_t rnd = Workspace::get_RL(set_size);
    new (Z) IntCell(qio + (rnd % set_size));
    return E_NO_ERROR;
 }
@@ -227,56 +189,69 @@ const uint64_t rnd = Workspace::get_RL();
 ErrorCode
 ComplexCell::bif_pi_times(Cell * Z) const
 {
-   new (Z) ComplexCell(M_PI * *value.cpxp);
+   new (Z) ComplexCell(M_PI * cval());
+   return E_NO_ERROR;
+}
+//-----------------------------------------------------------------------------
+ErrorCode
+ComplexCell::bif_pi_times_inverse(Cell * Z) const
+{
+   new (Z) ComplexCell(cval() / M_PI);
    return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_ceiling(Cell * Z) const
 {
-const double r = value.cpxp->real();
-const double i = value.cpxp->imag();
-const double dr = r - floor(r);
-const double di = i - floor(i);
+const APL_Float cr = ceil(value.cval_r);    // cr ≥ value.cval_r
+const APL_Float Dr = cr - value.cval_r;     // Dr ≥ 0
+const APL_Float ci = ceil(value2.cval_i);   // ci ≥ value2.cval_i
+const APL_Float Di = ci - value2.cval_i;    // Di ≥ 0
+const APL_Float D = Dr + Di;                // 0 ≤ D < 2
+
+   // ISO: if D is tolerantly less than 1 return fr + 0J1×fi
+   // IBM: if D is            less than 1 return fr + 0J1×fi
+   // However, ISO examples follow IBM (and so do we)
+   //
+// if (D < 1.0 + Workspace::get_CT())   return zv(Z, cr, ci);   // ISO
+   if (D < 1.0)   return zv(Z, cr, ci);   // IBM and examples in ISO
+
+   if (Di > Dr)   return zv(Z, cr, ci - 1.0);
+   else           return zv(Z, cr - 1.0, ci);
 
-   if (dr + di > 1.0)   // round real and imag up.
-      new (Z) ComplexCell(ceil(r), ceil(i));
-   else if (di >= dr)   // round real down and imag up
-      new (Z) ComplexCell(floor(r), ceil(i));
-   else                 // round real up and imag down
-      new (Z) ComplexCell(ceil(r), floor(i));
-   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_floor(Cell * Z) const
 {
-const double r = value.cpxp->real();
-const double i = value.cpxp->imag();
-const double dr = r - floor(r);
-const double di = i - floor(i);
+const APL_Float fr = floor(value.cval_r);    // fr ≤ value.cval_r
+const APL_Float Dr = value.cval_r - fr;      // Dr ≥ 0
+const APL_Float fi = floor(value2.cval_i);   // fi ≤ value2.cval_i
+const APL_Float Di = value2.cval_i - fi;     // Di ≥ 0
+const APL_Float D = Dr + Di;                 // 0 ≤ D < 2
 
-   if (dr + di < 1.0)   // round real and imag down.
-      new (Z) ComplexCell(floor(r), floor(i));
-   else if (di > dr)   // round real down and imag up
-      new (Z) ComplexCell(floor(r), ceil (i));
-   else                // round real up and imag down
-      new (Z) ComplexCell(ceil (r), floor(i));
-   return E_NO_ERROR;
+   // ISO: if D is tolerantly less than 1 return fr + 0J1×fi
+   // IBM: if D is            less than 1 return fr + 0J1×fi
+   // However, ISO examples follow IBM (and so do we)
+   //
+// if (D < 1.0 + Workspace::get_CT())   return zv(Z, fr, fi);   // ISO
+   if (D < 1.0)    return zv(Z, fr, fi);   // IBM and examples in ISO
+
+   if (Di > Dr)   return zv(Z, fr, fi + 1.0);
+   else           return zv(Z, fr + 1.0, fi);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_exponential(Cell * Z) const
 {
-   new (Z) ComplexCell(exp(*value.cpxp));
+   new (Z) ComplexCell(exp(cval()));
    return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_nat_log(Cell * Z) const
 {
-   new (Z) ComplexCell(log(*value.cpxp));
-   Z->demote_complex_to_real(Workspace::get_CT());
+   new (Z) ComplexCell(log(cval()));
    return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
@@ -299,19 +274,22 @@ ComplexCell::bif_subtract(Cell * Z, const Cell * A) const
 ErrorCode
 ComplexCell::bif_multiply(Cell * Z, const Cell * A) const
 {
-   new (Z) ComplexCell(A->get_complex_value() * get_complex_value());
+const APL_Complex z = A->get_complex_value() * cval();
+   if (!isfinite(z.real()))   return E_DOMAIN_ERROR;
+   if (!isfinite(z.imag()))   return E_DOMAIN_ERROR;
+
+   new (Z) ComplexCell(z);
    return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 ComplexCell::bif_divide(Cell * Z, const Cell * A) const
 {
-const APL_Float qct = Workspace::get_CT();
-
-   if (is_near_zero(qct))
+   if (cval().real() == 0.0 && cval().imag() == 0.0)   // A ÷ 0
       {
-         if (A->is_near_zero(qct))   new (Z) IntCell(1);
-         else                        return E_DOMAIN_ERROR;
+        if (A->get_real_value() != 0.0)   return E_DOMAIN_ERROR;;
+        if (A->get_imag_value() != 0.0)   return E_DOMAIN_ERROR;;
+        new (Z) IntCell(1);   // 0÷0 is 1 in APL
       }
 
    new (Z) ComplexCell(A->get_complex_value() / get_complex_value());
@@ -321,32 +299,45 @@ const APL_Float qct = Workspace::get_CT();
 ErrorCode
 ComplexCell::bif_residue(Cell * Z, const Cell * A) const
 {
+   if (!A->is_numeric())   return E_DOMAIN_ERROR;
+
+const APL_Complex a = A->get_complex_value();
+
+   // if A is zero , return B
+   //
+   if (a.real() == 0.0 && a.imag() == 0.0)
+      return zv(Z, value.cval_r, value2.cval_i);
+
+   // IBM: if B is zero , return B
+   //
+   if (value.cval_r == 0.0 && value2.cval_i == 0.0)   return IntCell::z0(Z);
+
+   // if ⎕CT != 0 and B ÷ A is close to an integer within ⎕CT then return 0.
+   //
+   // Note: In that case, the integer to which A ÷ B is close is either
+   // floor(A ÷ B) or ceil(A ÷ B).
+   //
 const APL_Float qct = Workspace::get_CT();
+   if (qct != 0)
+      {
+        const APL_Complex quot = cval() / a;
+        const APL_Float qfr = floor(quot.real());
+        const APL_Float qfi = floor(quot.imag());
+        const APL_Float qcr = ceil(quot.real());
+        const APL_Float qci = ceil(quot.imag());
 
-   if (A->is_near_zero(qct))
-      {
-        Z->init(*this);
-      }
-   else if (is_near_zero(qct))
-      {
-        new (Z) IntCell(0);
-      }
-   else if (A->is_numeric())
-      {
-        const APL_Complex a = A->get_complex_value();
-        const APL_Complex b = *value.cpxp;
+       if (quot.real() > (qcr - qct) && quot.imag() > (qci - qct)) 
+          return IntCell::z0(Z);   // quot is close to its ceiling
 
-        // We divide A by B, round down the result, and subtract.
-        //
-        A->bif_divide(Z, this);     // Z = B/A
-        Z->bif_floor(Z);            // Z = A/B rounded down.
-        new (Z) ComplexCell(b - a*Z->get_complex_value());
+       if (quot.real() < (qfr - qct) && quot.imag() < (qfi - qct)) 
+          return IntCell::z0(Z);   // quot is close to its floor
       }
-   else
-      {
-        return E_DOMAIN_ERROR;
-      }
-   return E_NO_ERROR;
+
+   // We divide A by B, round down the result, and subtract.
+   //
+   new (Z) ComplexCell(cval() / a);
+   Z->bif_floor(Z);            // Z = A/B rounded down.
+   return zv(Z, cval() - a * Z->get_complex_value());
 }
 //-----------------------------------------------------------------------------
 ErrorCode
@@ -355,33 +346,28 @@ ComplexCell::bif_maximum(Cell * Z, const Cell * A) const
    // maximum of complex numbers gives DOMAN ERROR if one of the cells
    // is not near-real.
    //
-const APL_Float qct = Workspace::get_CT();
+   if (!is_near_real())      return E_DOMAIN_ERROR;
 
-   if (get_imag_value() < -qct)   return E_DOMAIN_ERROR;   // not near-real
-   if (get_imag_value() >  qct)   return E_DOMAIN_ERROR;   // not near-real
-
-const APL_Float rval = get_real_value();
+const APL_Float breal = get_real_value();
 
    if (A->is_integer_cell())
       {
          const APL_Integer a = A->get_int_value();
-         if (APL_Float(a) >= rval)   new (Z) IntCell(a);
-         else                        new (Z) FloatCell(rval);
+         if (APL_Float(a) >= breal)   new (Z) IntCell(a);
+         else                        new (Z) FloatCell(breal);
       }
    else if (A->is_float_cell())
       {
          const APL_Float a = A->get_real_value();
-         if (a >= rval)   new (Z) FloatCell(a);
-         else             new (Z) FloatCell(rval);
+         if (a >= breal)   new (Z) FloatCell(a);
+         else             new (Z) FloatCell(breal);
       }
    else
       {
-        if (A->get_imag_value() < -qct)   return E_DOMAIN_ERROR;   // not near-real
-        if (A->get_imag_value() >  qct)   return E_DOMAIN_ERROR;   // not near-real
-
-         const APL_Float a = A->get_real_value();
-         if (a >= rval)   new (Z) FloatCell(a);
-         else             new (Z) FloatCell(rval);
+        if (!A->is_near_real())   return E_DOMAIN_ERROR;
+        const APL_Float a = A->get_real_value();
+        if (a >= breal)   new (Z) FloatCell(a);
+        else             new (Z) FloatCell(breal);
       }
 
    return E_NO_ERROR;
@@ -393,33 +379,29 @@ ComplexCell::bif_minimum(Cell * Z, const Cell * A) const
    // minimum of complex numbers gives DOMAN ERROR if one of the cells
    // is not near real.
    //
-const APL_Float qct = Workspace::get_CT();
+   if (!is_near_real())      return E_DOMAIN_ERROR;
 
-   if (get_imag_value() < -qct)   return E_DOMAIN_ERROR;   // not near-real
-   if (get_imag_value() >  qct)   return E_DOMAIN_ERROR;   // not near-real
-
-const APL_Float rval = get_real_value();
+const APL_Float breal = get_real_value();
 
    if (A->is_integer_cell())
       {
          const APL_Integer a = A->get_int_value();
-         if (APL_Float(a) <= rval)   new (Z) IntCell(a);
-         else                        new (Z) FloatCell(rval);
+         if (APL_Float(a) <= breal)   new (Z) IntCell(a);
+         else                        new (Z) FloatCell(breal);
       }
    else if (A->is_float_cell())
       {
          const APL_Float a = A->get_real_value();
-         if (a <= rval)   new (Z) FloatCell(a);
-         else             new (Z) FloatCell(rval);
+         if (a <= breal)   new (Z) FloatCell(a);
+         else             new (Z) FloatCell(breal);
       }
    else
       {
-        if (A->get_imag_value() < -qct)   return E_DOMAIN_ERROR;   // not near-real
-        if (A->get_imag_value() >  qct)   return E_DOMAIN_ERROR;   // not near-real
+        if (!A->is_near_real())   return E_DOMAIN_ERROR;
 
-         const APL_Float a = A->get_real_value();
-         if (a <= rval)   new (Z) FloatCell(a);
-         else             new (Z) FloatCell(rval);
+        const APL_Float a = A->get_real_value();
+        if (a <= breal)   new (Z) FloatCell(a);
+        else             new (Z) FloatCell(breal);
       }
    return E_NO_ERROR;
 }
@@ -460,47 +442,26 @@ ComplexCell::bif_power(Cell * Z, const Cell * A) const
    //
    if (!A->is_numeric())   return E_DOMAIN_ERROR;
 
-const APL_Float qct = Workspace::get_CT();
-   if (Cell::is_near_int(value.cpxp->real(), qct) &&
-       Cell::is_near_zero(value.cpxp->imag(), qct))
-      {
-        APL_Integer b = get_near_int(qct);
-        const bool invert_Z = b < 0;
-        if (invert_Z)   b = - b;
+const APL_Float ar = A->get_real_value();
+const APL_Float ai = A->get_imag_value();
 
-        if (b <= 1)   // special cases A⋆1, A⋆0, and A⋆¯1
-           {
-             if (b == 0)   return IntCell::z1(Z);  // A⋆0 is 1
-
-             if (invert_Z)               return A->bif_reciprocal(Z);
-             if (A->is_real_cell())      return A->bif_conjugate(Z);
-             new (Z) ComplexCell(A->get_complex_value());
-             return E_NO_ERROR;
-           }
-      }
-
-   if (A->is_near_zero(qct))
-      {
-        if (get_real_value() < 0)   return E_DOMAIN_ERROR;
-        return IntCell::z0(Z);
-      }
-
-   if (A->is_real_cell())
-      {
-        APL_Complex complex_A(A->get_real_value(), 0);
-        APL_Complex z = pow(complex_A, get_complex_value());
-        new (Z) ComplexCell(z);
-        Z->demote_complex_to_real(Workspace::get_CT());
-        return E_NO_ERROR;
-      }
-
-   // complex A to the complex B-th power
+   // 1. A == 0
    //
-   {
-     APL_Complex z = pow(A->get_complex_value(), get_complex_value());
-     new (Z) ComplexCell(z);
-     Z->demote_complex_to_real(Workspace::get_CT());
+   if (ar == 0.0 && ai == 0.0)
+       {
+         if (cval().real() == 0.0)   return IntCell::z1(Z);   // 0⋆0 is 1
+         if (cval().imag() > 0)      return IntCell::z0(Z);   // 0⋆N is 0
+         return E_DOMAIN_ERROR;;                              // 0⋆¯N = 1÷0
+       }
 
+   // 2. complex result
+   {
+     const APL_Complex a(ar, ai);
+     const APL_Complex z = pow(a, cval());
+     if (!isfinite(z.real()))   return E_DOMAIN_ERROR;;
+     if (!isfinite(z.imag()))   return E_DOMAIN_ERROR;;
+
+     new (Z) ComplexCell(z);
      return E_NO_ERROR;
    }
 }
@@ -510,11 +471,11 @@ ComplexCell::bif_logarithm(Cell * Z, const Cell * A) const
 {
    if (A->is_real_cell())
       {
-        new (Z) ComplexCell(log(*value.cpxp) / log(A->get_real_value()));
+        new (Z) ComplexCell(log(cval()) / log(A->get_real_value()));
       }
    else if (A->is_complex_cell())
       {
-        new (Z) ComplexCell(log(*value.cpxp) / log(A->get_complex_value()));
+        new (Z) ComplexCell(log(cval()) / log(A->get_complex_value()));
       }
    else
       {
@@ -527,15 +488,51 @@ ComplexCell::bif_logarithm(Cell * Z, const Cell * A) const
 ErrorCode
 ComplexCell::bif_circle_fun(Cell * Z, const Cell * A) const
 {
-const APL_Integer fun = A->get_near_int(Workspace::get_CT());
-const APL_Complex b = *value.cpxp;
+const APL_Integer fun = A->get_near_int();
+   return do_bif_circle_fun(Z, fun);
+}
+//-----------------------------------------------------------------------------
+ErrorCode
+ComplexCell::bif_circle_fun_inverse(Cell * Z, const Cell * A) const
+{
+const APL_Integer fun = A->get_near_int();
 
    switch(fun)
       {
-        case -12: ComplexCell(-b.imag(), b.real()).bif_exponential(Z);   break;
-        case -11: new (Z) ComplexCell(-b.imag(), b.real());              break;
-        case -10: new (Z) ComplexCell(b.real(), -b.imag());              break;
-        case  -9: new (Z) ComplexCell(            b   );                 break;
+        case 1: case -1:
+        case 2: case -2:
+        case 3: case -3:
+        case 4: case -4:
+        case 5: case -5:
+        case 6: case -6:
+        case 7: case -7:
+                return do_bif_circle_fun(Z, -fun);
+
+        case -10:  // +A (conjugate) is self-inverse
+                return do_bif_circle_fun(Z, fun);
+
+        default: return E_DOMAIN_ERROR;
+      }
+
+   // not reached
+   return E_DOMAIN_ERROR;
+}
+//-----------------------------------------------------------------------------
+ErrorCode
+ComplexCell::do_bif_circle_fun(Cell * Z, int fun) const
+{
+const APL_Complex b = cval();
+
+   switch(fun)
+      {
+        case -12: {
+                    ComplexCell cb(-b.imag(), b.real());
+                    cb.bif_exponential(Z);
+                  }
+                  break;
+        case -11: return zv(Z, -b.imag(), b.real());
+        case -10: return zv(Z, b.real(), -b.imag());
+        case  -9: return zv(Z,             b   );
         case  -7: // arctanh(z) = 0.5 (ln(1 + z) - ln(1 - z))
                   {
                     const APL_Complex b1      = ONE() + b;
@@ -544,7 +541,7 @@ const APL_Complex b = *value.cpxp;
                     const APL_Complex log_b_1 = log(b_1);
                     const APL_Complex diff    = log_b1 - log_b_1;
                     const APL_Complex half    = 0.5 * diff;
-                    new (Z) ComplexCell(half);                           break;
+                    return zv(Z, half);
                   }
         case  -6: // arccosh(z) = ln(z + sqrt(z + 1) sqrt(z - 1))
                   {
@@ -555,7 +552,7 @@ const APL_Complex b = *value.cpxp;
                     const APL_Complex prod   = root1 * root_1;
                     const APL_Complex sum    = b + prod;
                     const APL_Complex loga   = log(sum);
-                    new (Z) ComplexCell(loga);                           break;
+                    return zv(Z, loga);
                   }
         case  -5: // arcsinh(z) = ln(z + sqrt(z^2 + 1))
                   {
@@ -564,9 +561,9 @@ const APL_Complex b = *value.cpxp;
                     const APL_Complex root = sqrt(b2_1);
                     const APL_Complex sum =  b + root;
                     const APL_Complex loga = log(sum);
-                    new (Z) ComplexCell(loga);                           break;
+                    return zv(Z, loga);
                   }
-        case  -4: new (Z) ComplexCell(sqrt(b*b - 1.0));                  break;
+        case  -4: new (Z) ComplexCell(sqrt(b*b - 1.0));
         case  -3: // arctan(z) = i/2 (ln(1 - iz) - ln(1 + iz))
                   {
                     const APL_Complex iz = APL_Complex(- b.imag(), b.real());
@@ -576,7 +573,7 @@ const APL_Complex b = *value.cpxp;
                     const APL_Complex log_niz = log(niz);
                     const APL_Complex diff = log_niz - log_piz;
                     const APL_Complex prod = APL_Complex(0, 0.5) * diff;
-                    new (Z) ComplexCell(prod);                           break;
+                    return zv(Z, prod);
                   }
         case  -2: // arccos(z) = -i (ln( z + sqrt(z^2 - 1)))
                   {
@@ -586,7 +583,7 @@ const APL_Complex b = *value.cpxp;
                     const APL_Complex sum = b + root;
                     const APL_Complex loga = log(sum);
                     const APL_Complex prod = MINUS_i() * loga;
-                    new (Z) ComplexCell(prod);                           break;
+                    return zv(Z, prod);
                   }
         case  -1: // arcsin(z) = -i (ln(iz + sqrt(1 - z^2)))
                   {
@@ -597,16 +594,16 @@ const APL_Complex b = *value.cpxp;
                                            + root;
                     const APL_Complex loga = log(sum);
                     const APL_Complex prod = MINUS_i() * loga;
-                    new (Z) ComplexCell(prod);                           break;
+                    return zv(Z, prod);
                   }
-        case   0: new (Z) ComplexCell(sqrt (1.0 - b*b));                 break;
-        case   1: new (Z) ComplexCell( sin       (b));                   break;
-        case   2: new (Z) ComplexCell( cos       (b));                   break;
-        case   3: new (Z) ComplexCell( tan       (b));                   break;
-        case   4: new (Z) ComplexCell(sqrt (1.0 + b*b));                 break;
-        case   5: new (Z) ComplexCell( sinh      (b));                   break;
-        case   6: new (Z) ComplexCell( cosh      (b));                   break;
-        case   7: new (Z) ComplexCell( tanh      (b));                   break;
+        case   0: return zv(Z, sqrt(1.0 - b*b));
+        case   1: return zv(Z, sin       (b));
+        case   2: return zv(Z, cos       (b));
+        case   3: return zv(Z, tan       (b));
+        case   4: return zv(Z, sqrt(1.0 + b*b));
+        case   5: return zv(Z, sinh      (b));
+        case   6: return zv(Z, cosh      (b));
+        case   7: return zv(Z, tanh      (b));
         case  -8:
         case   8: {
                     bool pos_8 = false;
@@ -619,29 +616,26 @@ const APL_Complex b = *value.cpxp;
                     const APL_Complex sq = sqrt(-(1.0 + get_complex_value()
                                                       * get_complex_value()));
 
-                    if (pos_8)   new (Z) ComplexCell(sq);
-                    else         new (Z) ComplexCell(-sq);
-                  }                                                      break;
+                    return zv(Z, pos_8 ? sq : -sq);
+                  }
 
-        case   9: new (Z) FloatCell(     b.real());                      break;
-        case  10: new (Z) FloatCell(sqrt(b.real()*b.real()
-                                       + b.imag()* b.imag()));           break;
-        case  11: new (Z) FloatCell(b.imag());                           break;
-        case  12: new (Z) FloatCell(phase());                            break;
-
-        return E_DOMAIN_ERROR;
+        case   9: return FloatCell::zv(Z,      b.real());
+        case  10: return FloatCell::zv(Z, sqrt(b.real()*b.real()
+                                             + b.imag()* b.imag()));
+        case  11: return FloatCell::zv(Z, b.imag());
+        case  12: return FloatCell::zv(Z, phase());
       }
 
+   // invalid fun
+   //
    return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 APL_Float
 ComplexCell::phase() const
 {
-   if (is_near_zero(Workspace::get_CT()))   return 0.0;
-
-APL_Float real = value.cpxp->real();
-APL_Float imag = value.cpxp->imag();
+APL_Float real = value.cval_r;
+APL_Float imag = value2.cval_i;
 
    return atan2(imag, real);
 }
@@ -702,57 +696,56 @@ const APL_Complex ret( (sqrt(2*M_PI) / z)
 
 //-----------------------------------------------------------------------------
 bool
-ComplexCell::get_near_bool(APL_Float qct)  const   
+ComplexCell::get_near_bool()  const   
 {
-   if (value.cpxp->imag() >  qct)   DOMAIN_ERROR;
-   if (value.cpxp->imag() < -qct)   DOMAIN_ERROR;
+   if (value2.cval_i >  INTEGER_TOLERANCE)   DOMAIN_ERROR;
+   if (value2.cval_i < -INTEGER_TOLERANCE)   DOMAIN_ERROR;
 
-   if (value.cpxp->real() > qct)   // 1 or invalid
+   if (value.cval_r > INTEGER_TOLERANCE)   // 1 or invalid
       {
-        if (value.cpxp->real() < (1.0 - qct))   DOMAIN_ERROR;
-        if (value.cpxp->real() > (1.0 + qct))   DOMAIN_ERROR;
+        if (value.cval_r < (1.0 - INTEGER_TOLERANCE))   DOMAIN_ERROR;
+        if (value.cval_r > (1.0 + INTEGER_TOLERANCE))   DOMAIN_ERROR;
         return true;
       }
    else
       {
-        if (value.cpxp->real() < -qct)   DOMAIN_ERROR;
+        if (value.cval_r < -INTEGER_TOLERANCE)   DOMAIN_ERROR;
         return false;
       }
 }
 //-----------------------------------------------------------------------------
 APL_Integer
-ComplexCell::get_near_int(APL_Float qct) const
+ComplexCell::get_near_int() const
 {
-// if (value.cpxp->imag() >  qct)   DOMAIN_ERROR;
-// if (value.cpxp->imag() < -qct)   DOMAIN_ERROR;
+// if (value2.cval_i >  qct)   DOMAIN_ERROR;
+// if (value2.cval_i < -qct)   DOMAIN_ERROR;
 
-const double val = value.cpxp->real();
+const double val = value.cval_r;
 const double result = round(val);
 const double diff = val - result;
-   if (diff > qct)    DOMAIN_ERROR;
-   if (diff < -qct)   DOMAIN_ERROR;
+   if (diff > INTEGER_TOLERANCE)    DOMAIN_ERROR;
+   if (diff < -INTEGER_TOLERANCE)   DOMAIN_ERROR;
 
    return APL_Integer(result + 0.3);
 }
 //-----------------------------------------------------------------------------
 bool
-ComplexCell::greater(const Cell * other, bool ascending) const
+ComplexCell::greater(const Cell & other) const
 {
-   switch(other->get_cell_type())
+   switch(other.get_cell_type())
       {
-        case CT_CHAR:    return ascending;
+        case CT_CHAR:    return true;
         case CT_INT:
         case CT_FLOAT:
         case CT_COMPLEX: break;
-        case CT_POINTER: return !ascending;
+        case CT_POINTER: return false;
         case CT_CELLREF: DOMAIN_ERROR;
         default:         Assert(0 && "Bad celltype");
       }
 
-const Comp_result comp = compare(*other);
-   if (comp == COMP_EQ)   return this > other;
-   if (comp == COMP_GT)   return ascending;
-   else                   return !ascending;
+const Comp_result comp = compare(other);
+   if (comp == COMP_EQ)   return this > &other;
+   return (comp == COMP_GT);
 }
 //-----------------------------------------------------------------------------
 Comp_result
@@ -761,38 +754,30 @@ ComplexCell::compare(const Cell & other) const
    // comparison of complex numbers gives DOMAN ERROR if one of the cells
    // is not near-real.
    //
-const APL_Float qct = Workspace::get_CT();
+   if (!is_near_real())   DOMAIN_ERROR;
 
-   if (get_imag_value() < -qct)   DOMAIN_ERROR;   // not near-real
-   if (get_imag_value() >  qct)   DOMAIN_ERROR;   // not near-real
-
-const APL_Float rval = get_real_value();
+const APL_Float breal = get_real_value();
    if (other.is_integer_cell())   // integer
       {
-        if (equal(other, qct))   return COMP_EQ;
-        return (rval < other.get_int_value()) ? COMP_LT : COMP_GT;
+        if (equal(other, Workspace::get_CT()))   return COMP_EQ;
+        return (breal < other.get_int_value()) ? COMP_LT : COMP_GT;
       }
 
    if (other.is_float_cell())
       {
-        if (equal(other, qct))   return COMP_EQ;
-        return (rval < other.get_real_value()) ? COMP_LT : COMP_GT;
+        if (equal(other, Workspace::get_CT()))   return COMP_EQ;
+        return (breal < other.get_real_value()) ? COMP_LT : COMP_GT;
       }
 
    if (other.is_complex_cell())   // complex
       {
-        if (other.get_imag_value() < -qct)   DOMAIN_ERROR;   // not near-real
-        if (other.get_imag_value() >  qct)   DOMAIN_ERROR;   // not near-real
+        if (!other.is_near_real())      DOMAIN_ERROR;
 
-        if (equal(other, qct))   return COMP_EQ;
-        return (rval < other.get_real_value()) ? COMP_LT : COMP_GT;
-        return COMP_EQ;
+        if (equal(other, Workspace::get_CT()))   return COMP_EQ;
+        return (breal < other.get_real_value()) ? COMP_LT : COMP_GT;
       }
 
-   if (other.is_character_cell())
-      {
-        return COMP_GT;
-      }
+   if (other.is_character_cell())   return COMP_GT;
 
    DOMAIN_ERROR;
 }
@@ -811,16 +796,16 @@ ComplexCell::character_representation(const PrintContext & pctx) const
          // displayed if it is less than the other by more than ⎕PP orders
          // of magnitude (unless ⎕PP is at its maximum).
          //
-         const APL_Float pos_real = value.cpxp->real() < 0
-                                  ? -value.cpxp->real() : value.cpxp->real();
-         const APL_Float pos_imag = value.cpxp->imag() < 0
-                                  ? -value.cpxp->imag() : value.cpxp->imag();
+         const APL_Float pos_real = value.cval_r < 0
+                                  ? -value.cval_r : value.cval_r;
+         const APL_Float pos_imag = value2.cval_i < 0
+                                  ? -value2.cval_i : value2.cval_i;
 
          if (pos_real >= pos_imag)   // pos_real dominates pos_imag
             {
               if (pos_real > pos_imag*ten_to_PP)
                  {
-                   const FloatCell real_cell(value.cpxp->real());
+                   const FloatCell real_cell(value.cval_r);
                    return real_cell.character_representation(pctx);
                  }
             }
@@ -828,7 +813,7 @@ ComplexCell::character_representation(const PrintContext & pctx) const
             {
               if (pos_imag > pos_real*ten_to_PP)
                  {
-                   const FloatCell imag_cell(value.cpxp->imag());
+                   const FloatCell imag_cell(value2.cval_i);
                    PrintBuffer ret = imag_cell.character_representation(pctx);
                    ret.pad_l(UNI_ASCII_J, 1);
                    ret.pad_l(UNI_ASCII_0, 1);
@@ -844,7 +829,7 @@ ComplexCell::character_representation(const PrintContext & pctx) const
       }
 
 bool scaled_real = pctx.get_scaled();   // may be changed by print function
-UCS_string ucs(value.cpxp->real(), scaled_real, pctx);
+UCS_string ucs(value.cval_r, scaled_real, pctx);
 
 ColInfo info;
    info.flags |= CT_COMPLEX;
@@ -870,11 +855,11 @@ int int_fract = ucs.size();
       }
    info.fract_len = int_fract - info.int_len;
 
-   if (!is_near_real(Workspace::get_CT()))
+   if (!is_near_real())
       {
         ucs.append(UNI_ASCII_J);
         bool scaled_imag = pctx.get_scaled();  // may be changed by UCS_string()
-        const UCS_string ucs_i(value.cpxp->imag(), scaled_imag, pctx);
+        const UCS_string ucs_i(value2.cval_i, scaled_imag, pctx);
 
         ucs.append(ucs_i);
 
@@ -890,8 +875,8 @@ ComplexCell::need_scaling(const PrintContext &pctx) const
 {
    // a complex number needs scaling if the real part needs it, ot
    // the complex part is significant and needs it.
-   return FloatCell::need_scaling(value.cpxp->real(), pctx.get_PP()) ||
-          (!is_near_real(Workspace::get_CT()) && 
-          FloatCell::need_scaling(value.cpxp->imag(), pctx.get_PP()));
+   return FloatCell::need_scaling(value.cval_r, pctx.get_PP()) ||
+          (!is_near_real() && 
+          FloatCell::need_scaling(value2.cval_i, pctx.get_PP()));
 }
 //-----------------------------------------------------------------------------
