@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -29,7 +29,7 @@
 //=============================================================================
 DerivedFunction::DerivedFunction(Token & lfun, Function * dyop, Token & rfun,
                                  const char * loc)
-   : Function(ID_USER_SYMBOL, TOK_FUN2),
+   : Function(ID::USER_SYMBOL, TOK_FUN2),
      left_fun(lfun),
      oper(dyop),
      right_fun(rfun),
@@ -46,7 +46,7 @@ DerivedFunction::DerivedFunction(Token & lfun, Function * dyop, Token & rfun,
 //-----------------------------------------------------------------------------
 DerivedFunction::DerivedFunction(Token & lfun, Function * dyop, Value_P X,
                                  Token & rval, const char * loc)
-   : Function(ID_USER_SYMBOL, TOK_FUN2),
+   : Function(ID::USER_SYMBOL, TOK_FUN2),
      left_fun(lfun),
      oper(dyop),
      right_fun(rval),
@@ -56,7 +56,7 @@ DerivedFunction::DerivedFunction(Token & lfun, Function * dyop, Value_P X,
 //-----------------------------------------------------------------------------
 DerivedFunction::DerivedFunction(Token & lfun, Function * monop,
                                  const char * loc)
-   : Function(ID_USER_SYMBOL, TOK_FUN2),
+   : Function(ID::USER_SYMBOL, TOK_FUN2),
      left_fun(lfun),
      oper(monop),
      right_fun(TOK_VOID),
@@ -73,7 +73,7 @@ DerivedFunction::DerivedFunction(Token & lfun, Function * monop,
 //-----------------------------------------------------------------------------
 DerivedFunction::DerivedFunction(Token & lfun, Function * monop,
                                  Value_P X, const char * loc)
-   : Function(ID_USER_SYMBOL, TOK_FUN2),
+   : Function(ID::USER_SYMBOL, TOK_FUN2),
      left_fun(lfun),
      oper(monop),
      right_fun(TOK_VOID),
