@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -21,47 +21,54 @@
 #include "Avec.hh"
 #include "Common.hh"
 #include "DiffOut.hh"
-#include "Input.hh"
 #include "InputFile.hh"
 #include "IO_Files.hh"
-#include "main.hh"
 #include "Output.hh"
-#include "Svar_DB.hh"
+#include "Performance.hh"
 #include "UTF8_string.hh"
 
 //-----------------------------------------------------------------------------
 int
 DiffOut::overflow(int c)
 {
+PERFORMANCE_START(cout_perf)
    Output::set_color_mode(errout ? Output::COLM_UERROR : Output::COLM_OUTPUT);
    cout << char(c);
 
    if (c != '\n')   // not end of line
       {
         aplout.append(c);
+        PERFORMANCE_END(fs_COUT_B, cout_perf, 1)
         return 0;
       }
 
-   if (!InputFile::is_validating())   return 0;
+   if (!InputFile::is_validating())
+      {
+        PERFORMANCE_END(fs_COUT_B, cout_perf, 1)
+        return 0;
+      }
 
 ofstream & rep = IO_Files::get_current_testreport();
    Assert(rep.is_open());
 
 const char * apl = aplout.c_str();
-const UTF8 * ref = Input::read_file_line();
-   if (ref == 0)   // nothing in current_testfile
+UTF8_string ref;
+bool eof = false;
+   IO_Files::read_file_line(ref, eof);
+   if (eof)   // nothing in current_testfile
       {
         rep << "extra: " << apl << endl;
+        PERFORMANCE_END(fs_COUT_B, cout_perf, 1)
         return 0;
       }
 
    // print common part.
    //
-   if (different((const UTF8 *)apl, ref))   // different
+   if (different((const UTF8 *)apl, (const UTF8 *)ref.c_str()))   // different
       {
         IO_Files::diff_error();
         rep << "apl: ⋅⋅⋅" << apl << "⋅⋅⋅" << endl
-            << "ref: ⋅⋅⋅" << ref << "⋅⋅⋅" << endl;
+            << "ref: ⋅⋅⋅" << ref.c_str() << "⋅⋅⋅" << endl;
       }
    else                    // same
       {
@@ -69,6 +76,7 @@ const UTF8 * ref = Input::read_file_line();
       }
 
    aplout.clear();
+   PERFORMANCE_END(fs_COUT_B, cout_perf, 1)
    return 0;
 }
 //-----------------------------------------------------------------------------
