@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -189,6 +189,21 @@ Error error(code, loc);
 }
 //-----------------------------------------------------------------------------
 void
+throw_tokenize_error(ErrorCode code, const UCS_string & more_error_info)
+{
+   Log(LOG_error_throw)
+      CERR << endl
+           << "throwing " << Error::error_name(code)
+           << " in  Tokenizer" << endl;
+
+   Workspace::more_error() = more_error_info;
+
+Error error(code, LOC);
+
+   throw error;
+}
+//-----------------------------------------------------------------------------
+void
 throw_parse_error(ErrorCode code, const char * par_loc, const char *loc)
 {
    Log(LOG_error_throw)
