@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -136,7 +136,17 @@ Token_string in;
 const Parser parser(get_parse_mode(), loc);
    {
      ErrorCode ec = parser.parse(ucs_line, in);
+     if (ec)
+        {
+          if (ec == E_NO_TOKEN)
+             {
+               Error error(ec, LOC);
+               throw error;
+
+             }
+
      if (ec != E_NO_ERROR)   throw_parse_error(ec, LOC, LOC);
+        }
    }
 
    parse_body_line(line, in, trace, loc);
@@ -222,7 +232,7 @@ Token_string out;
 
    Log(LOG_UserFunction__set_line)
       {
-        CERR << "[non-reverse " << line << "] ";
+        CERR << "[final line " << line << "] ";
         out.print(CERR);
       } 
 
@@ -440,6 +450,16 @@ Executable::unmark_all_values() const
         Value_P value = tok.get_apl_val();
         if (!!value)   value->unmark();
       }
+
+   loop (n, named_lambdas.size())
+      {
+        named_lambdas[n]->unmark_all_values();
+      }
+
+   loop (u, unnamed_lambdas.size())
+      {
+        unnamed_lambdas[u]->unmark_all_values();
+      }
 }
 //-----------------------------------------------------------------------------
 int
@@ -465,150 +485,205 @@ int count = 0;
 void
 Executable::setup_lambdas()
 {
+   // quick check if this body contains any lambdas. Parser::match_par_bra()
+   // should have checked for unbalanced { } so we simply look for } (which
+   // is the first to occur since the body is reversed.
+   //
+bool have_curly = false;
+   loop(b, body.size())
+       {
+         if (body[b].get_tag() == TOK_R_CURLY)
+            {
+              have_curly = true;
+              break;
+            }
+       }
+
+   if (!have_curly)   return;   // no { ... } in this body
+
    // setup_lambdas() may be called multiple times for the same executable.
    // remove old lambdas...
    //
    clear_lambdas();
 
+   // undo the token reversion of the body so that the body token and the
+   // text run in the same direction.
+   //
+
+   reverse_statement_token(body);
+
    loop(b, body.size())
        {
-         // body is in reverse order, so } comes before { (and optional ← NAME)
-         //
-         if (body[b].get_tag() != TOK_R_CURLY)   continue;   // not {
+         if (body[b].get_tag() != TOK_L_CURLY)   continue;   // not {
 
-         // a named lambda has the form Q←{ ... }
-         //
-         // however, eg. Q←{ ... } / ... is not a named lambda but an unnamed
-         // lambda for / whose result is assigned to Q. We need to distinguish
-         // the two cases and remember if the token after } is the end of
-         // the statement or not.
-         //
-         const bool maybe_named = b == 0 || (body[b-1].get_tag() == TOK_DIAMOND
-                                         || body[b-1].get_tag() == TOK_END
-                                         || body[b-1].get_tag() == TOK_ENDL);
-
-         body[b++].clear(LOC);   // invalidate }
-         Token_string rev_lambda_body;
-
-         int signature = SIG_NONE;
-         int body_nesting = 1;   /// the TOK_R_CURLY above
-
-         for (bool goon = true; goon;)
-             {
-               if (b >= body.size())   SYNTAX_ERROR;
-               Token t;
-               move_1(t, body[b], LOC);
-               body[b].clear(LOC);   // invalidate in main body
-
-               // figure the signature by looking for ⍺, ⍶, ⍵, ⍹, and χ
-               // and compain about ◊ and →
-               switch(t.get_tag())
-                  {
-                    case TOK_L_CURLY:
-                         if (body_nesting == 1)  { goon = false;     continue; }
-                         else                    { --body_nesting;   break;    }
-
-                    case TOK_ALPHA:    signature |= SIG_A;    /* no break */
-                    case TOK_OMEGA:    signature |= SIG_B;    break;
-                    case TOK_CHI:      signature |= SIG_X;    break;
-                    case TOK_OMEGA_U:  signature |= SIG_RO;   /* no break */
-                    case TOK_ALPHA_U:  signature |= SIG_LO;   break;
-
-                    case TOK_R_CURLY:  ++body_nesting;        break;
-                    case TOK_DIAMOND:  DEFN_ERROR;
-                    case TOK_BRANCH:   DEFN_ERROR;
-                    case TOK_ESCAPE:   DEFN_ERROR;
-
-                    default: break;
-                  }
-
-               rev_lambda_body.append(t, LOC);
-               ++b;
-             }
-
-         // if the lambda is { } then we can't assign anything to λ
-         // and make the lambda result-less.
-         //
-         if (rev_lambda_body.size())   signature |= SIG_Z;
-
-         Token_string forw_lambda_body;
-         if (signature & SIG_Z)
-              {
-                Token ret_lambda(TOK_RETURN_SYMBOL, &Workspace::get_v_LAMBDA());
-                forw_lambda_body.append(ret_lambda, LOC);
-
-                const int64_t tr = 0;
-                forw_lambda_body.append(Token(TOK_ENDL, tr));
-
-                Symbol * sym_Z = &Workspace::get_v_LAMBDA();
-                forw_lambda_body.append(Token(TOK_LAMBDA, sym_Z), LOC);
-                forw_lambda_body.append(Token(TOK_ASSIGN), LOC);
-
-                const ShapeItem body_len = rev_lambda_body.size();
-                loop(r, body_len)
-                    forw_lambda_body.append(
-                                    rev_lambda_body[body_len - r - 1]);
-              }
-           else
-              {
-                Token ret_void(TOK_RETURN_VOID);
-                forw_lambda_body.append(ret_void, LOC);
-              }
-
-         // at this point {} was copied from body to forw_lambda_body and
-         // b is at the (now invalidated) { token.
-         // check if lambda is named, i.e. Name ← { ... }
-         //
-         UCS_string lambda_name;
-         Symbol * lambda_sym = 0;
-         if (   maybe_named && (b + 2) < body.size()
-             && body[b + 1].get_tag() == TOK_ASSIGN
-             && body[b + 2].get_tag() == TOK_LSYMB)
-            {
-              // named lambda
-              //
-              lambda_sym = body[b + 2].get_sym_ptr();
-              lambda_name = lambda_sym->get_name();
-              signature |= SIG_FUN;
-            }
-         else
-            {
-              lambda_name.append(UNI_LAMBDA);
-              lambda_name.append_number(unnamed_lambdas.size() + 1);
-            }
-
-         const UCS_string lambda_text = extract_lambda_text(
-                                (Fun_signature)signature,
-                                unnamed_lambdas.size() + named_lambdas.size());
-
-         UserFunction * ufun = new UserFunction(Fun_signature(signature),
-                                                lambda_name, lambda_text,
-                                                forw_lambda_body);
-
-         if (signature & SIG_FUN)   // named lambda
-            {
-              named_lambdas.push_back(ufun);
-              move_2(body[b], ufun->get_token(), LOC);
-
-              // UserFunction::UserFunction has bound the lambda body to
-              // the lambda name. However, we want that to happen when
-              // V←{ ...} is executed, not now. We therefore unbind it here
-              //
-              lambda_sym->set_nc(NC_UNUSED_USER_NAME, 0);
-            }
-         else                       // unnamed lambda
-            {
-              unnamed_lambdas.push_back(ufun);
-              move_2(body[b], ufun->get_token(), LOC);
-            }
+         b = setup_one_lambda(b) - 1;   // -1 due to ++b in loop(b) above
        }
 
+   // redo the token reversion of the body so that the body token and the
+   // text run in the same direction.
+   //
+   reverse_statement_token(body);
+
+   adjust_line_starts();
    Parser::remove_void_token(body);
+   Parser::match_par_bra(body, true);
+}
+//-----------------------------------------------------------------------------
+ShapeItem
+Executable::setup_one_lambda(ShapeItem b)
+{
+const ShapeItem bend = b + body[b].get_int_val2();   // the corresponding }
+   Assert(bend < body.size());
+   Assert(body[bend].get_tag() == TOK_R_CURLY);
+
+   // a named lambda has the form Q←{ ... }
+   //
+   // however, eg. Q←{ ... } / ... is NOT a named lambda but an unnamed
+   // lambda for / whose result is assigned to Q. We need to distinguish
+   // the two cases and remember if the token after } is the start of
+   // the next statement or not.
+   //
+const bool right_curly_at_end = (bend >= body.size() - 1)
+                             ||  body[bend + 1].get_tag() == TOK_DIAMOND
+                             ||  body[bend + 1].get_tag() == TOK_END
+                             ||  body[bend + 1].get_tag() == TOK_ENDL;
+
+const bool is_named = right_curly_at_end                  && 
+                      b >=   2                            &&
+                      body[b - 1].get_tag() == TOK_ASSIGN &&
+                      body[b - 2].get_tag() == TOK_LSYMB;
+
+Symbol * lambda_sym = is_named ? body[b - 2].get_sym_ptr() : 0;
+
+   body[b++].clear(LOC);    // invalidate {
+   body[bend].clear(LOC);   // invalidate }
+
+Token_string lambda_body;
+
+const Fun_signature signature =
+      compute_lambda_body(lambda_body, b, bend, is_named);
+
+   // at this point { ... } was copied from body to lambda_body and
+   // bend is at the (now invalidated) { token.
+   // check if lambda is named, i.e. Name ← { ... }
+   //
+UCS_string lambda_name;
+   if (is_named)   // named lambda
+      {
+        lambda_name = lambda_sym->get_name();
+      }
+   else            // unnamed lambda
+      {
+        lambda_name.append(UNI_LAMBDA);
+        lambda_name.append_number(unnamed_lambdas.size() + 1);
+      }
+
+const UCS_string lambda_text = extract_lambda_text(signature);
+
+   reverse_statement_token(lambda_body);
+   reverse_all_token(lambda_body);
+UserFunction * ufun = new UserFunction(signature, lambda_name,
+                                       lambda_text, lambda_body);
+
+   if (signature & SIG_FUN)   // named lambda
+      {
+        named_lambdas.push_back(ufun);
+
+        // put a token for the lambda at thr place where the { was.
+        // That means we replace (in forward notation) e.g.:
+        //
+        // A←{ ... }   by:
+        // A←UFUN      UFUN being a user-defined function with body { ... }
+        //
+        // This is to bind UFUN to A when A←UFUN is executed
+        //
+        move_2(body[bend], ufun->get_token(), LOC);
+
+        // UserFunction::UserFunction has bound the lambda body to
+        // the lambda name. However, we want that to happen when
+        // V←{ ...} is executed, not now. We therefore unbind it here
+        //
+        lambda_sym->set_nc(NC_UNUSED_USER_NAME, 0);
+      }
+   else                       // unnamed lambda
+      {
+        unnamed_lambdas.push_back(ufun);
+        move_2(body[bend], ufun->get_token(), LOC);
+      }
+
+   return bend + 1;
+}
+//-----------------------------------------------------------------------------
+Fun_signature
+Executable::compute_lambda_body(Token_string & lambda_body,
+                                ShapeItem b, ShapeItem bend, bool is_named)
+{
+int signature = is_named ? SIG_FUN : SIG_NONE;
+int level = 0;
+
+   for (; b < bend; ++b)
+       {
+         Token t;
+         move_1(t, body[b], LOC);
+         body[b].clear(LOC);   // invalidate in main body
+
+         // figure the signature by looking for ⍺, ⍶, ⍵, ⍹, and χ
+         // and complain about ◊ and →
+         switch(t.get_tag())
+            {
+              case TOK_ALPHA:   if (!level)   signature |= SIG_A;   // no break
+              case TOK_OMEGA:   if (!level)   signature |= SIG_B;         break;
+              case TOK_CHI:     if (!level)   signature |= SIG_X;         break;
+              case TOK_ALPHA_U: if (!level)   signature |= SIG_LO;  // no break
+              case TOK_OMEGA_U: if (!level)   signature |= SIG_RO;        break;
+
+              case TOK_DIAMOND: DEFN_ERROR;
+              case TOK_BRANCH:  DEFN_ERROR;
+              case TOK_ESCAPE:  DEFN_ERROR;
+
+              case TOK_L_CURLY: ++level;   break;
+              case TOK_R_CURLY: --level;   break;
+              default: break;
+            }
+
+         if (lambda_body.size() == 0)
+            {
+              // first token: prepend λ ← tokens
+              //
+              Symbol * sym_Z = &Workspace::get_v_LAMBDA();
+              lambda_body.append(Token(TOK_LAMBDA, sym_Z), LOC);
+              lambda_body.append(Token(TOK_ASSIGN), LOC);
+
+              signature |= SIG_Z;
+            }
+
+         lambda_body.append(t, LOC);
+       }
+
+   // if the lambda has at least one token, then it returns λ.
+   // Otherwise the lambda is result-less
+   //
+   if (signature & SIG_Z)
+      {
+        const int64_t trace = 0;
+        lambda_body.append(Token(TOK_ENDL, trace));
+
+        Token ret_lambda(TOK_RETURN_SYMBOL, &Workspace::get_v_LAMBDA());
+        lambda_body.append(ret_lambda, LOC);
+      }
+   else
+      {
+        Token ret_void(TOK_RETURN_VOID);
+        lambda_body.append(ret_void, LOC);
+      }
+
+   return (Fun_signature)signature;
 }
 //-----------------------------------------------------------------------------
 UCS_string
-Executable::extract_lambda_text(Fun_signature signature, int skip) const
+Executable::extract_lambda_text(Fun_signature signature) const
 {
+int skip = unnamed_lambdas.size() + named_lambdas.size();
 UCS_string lambda_text;
    if (signature & SIG_Z)   lambda_text.append_utf8("λ←");
 
@@ -624,9 +699,7 @@ int tcol = 0;
        {
          if (tidx >= text.size())
             {
-              Q(copying)
-              Q(tidx)
-              FIXME;
+              Q(copying)   Q(tidx)   FIXME;
             }
 
          const UCS_string & line = text[tidx];
@@ -634,7 +707,6 @@ int tcol = 0;
          if (tcol >= line.size())   // end of line: wrap to next line
             {
               ++tidx;
-
               tcol = 0;
               continue;
             }
@@ -662,9 +734,48 @@ int tcol = 0;
    lambda_text.pop();   // the last }
    return lambda_text;
 }
+//-----------------------------------------------------------------------------
+void
+Executable::reverse_statement_token(Token_string & tos)
+{
+ShapeItem from = 0;
+
+   for (ShapeItem to = from + 1; to < tos.size(); ++to)
+       {
+         if (tos[to].get_Class() == TC_END)
+            {
+              Token * t1 = &tos[from];
+              Token * t2 = &tos[to - 1];
+              for (;t1 < t2; ++t1, --t2)
+                  {
+                    char tt[sizeof(Token)];
+                    memcpy(tt, t1, sizeof(Token));
+                    memcpy(t1, t2,   sizeof(Token));
+                    memcpy(t2, tt, sizeof(Token));
+                  }
+              from = to + 1;
+            }
+       }
+
+}
+//-----------------------------------------------------------------------------
+void
+Executable::reverse_all_token(Token_string & tos)
+{
+Token * t1 = &tos[0];
+Token * t2 = &tos[tos.size() - 1];
+
+   for (;t1 < t2; ++t1, --t2)
+       {
+         char tt[sizeof(Token)];
+         memcpy(tt, t1, sizeof(Token));
+         memcpy(t1, t2,   sizeof(Token));
+         memcpy(t2, tt, sizeof(Token));
+       }
+}
 //=============================================================================
 ExecuteList *
-ExecuteList::fix(const UCS_string & data, const char * loc)
+ExecuteList::fix(const UCS_string & data, bool is_cmd, const char * loc)
 {
    // clear errors that may have occured before
    {
@@ -672,6 +783,15 @@ ExecuteList::fix(const UCS_string & data, const char * loc)
      if (err) err->error_code = E_NO_ERROR;
    }
 
+   if (is_cmd)
+      {
+         UCS_string empty;
+         ExecuteList * fun = new ExecuteList(empty, loc);
+         Assert(fun);
+         fun->text.push_back(data);
+         return fun;
+      }
+
 ExecuteList * fun = new ExecuteList(data, loc);
 
    Log(LOG_UserFunction__fix)
