@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -40,73 +40,54 @@ FloatCell::equal(const Cell & A, APL_Float qct) const
 }
 //-----------------------------------------------------------------------------
 bool
-FloatCell::greater(const Cell * other, bool ascending) const
+FloatCell::greater(const Cell & other) const
 {
 const APL_Float this_val  = get_real_value();
 
-   switch(other->get_cell_type())
+   switch(other.get_cell_type())
       {
         case CT_INT:
              {
-               const APL_Integer other_val = other->get_int_value();
-               if (this_val == other_val)   return this > other;
-               return this_val > other_val ? ascending : !ascending;
+               const APL_Integer other_val = other.get_int_value();
+               if (this_val == other_val)   return this > &other;
+               return this_val > other_val;
              }
 
         case CT_FLOAT:
              {
-               const APL_Float other_val = other->get_real_value();
-               if (this_val == other_val)   return this > other;
-               return this_val > other_val ? ascending : !ascending;
+               const APL_Float other_val = other.get_real_value();
+               if (this_val == other_val)   return this > &other;
+               return this_val > other_val;
              }
 
-        case CT_CHAR:    return ascending;
+        case CT_CHAR:    return true;
         case CT_COMPLEX: break;
-        case CT_POINTER: return !ascending;
+        case CT_POINTER: return false;
         case CT_CELLREF: DOMAIN_ERROR;
         default:         Assert(0 && "Bad celltype");
       }
 
-const Comp_result comp = compare(*other);
-   if (comp == COMP_EQ)   return this > other;
-   if (comp == COMP_GT)   return ascending; 
-   else                   return !ascending;
+const Comp_result comp = compare(other);
+   if (comp == COMP_EQ)   return this > &other;
+   return (comp == COMP_GT);
 }
 //-----------------------------------------------------------------------------
 bool
-FloatCell::get_near_bool(APL_Float qct)  const
+FloatCell::get_near_bool()  const
 {
-   if (value.fval > qct)   // maybe 1
+   if (value.fval > INTEGER_TOLERANCE)   // maybe 1
       {
-        if (value.fval > (1.0 + qct))   DOMAIN_ERROR;
-        if (value.fval < (1.0 - qct))   DOMAIN_ERROR;
+        if (value.fval > (1.0 + INTEGER_TOLERANCE))   DOMAIN_ERROR;
+        if (value.fval < (1.0 - INTEGER_TOLERANCE))   DOMAIN_ERROR;
         return true;
       }
 
    // maybe 0. We already know that value.fval ≤ qct
    //
-   if (value.fval < -qct)   DOMAIN_ERROR;
+   if (value.fval < -INTEGER_TOLERANCE)   DOMAIN_ERROR;
    return false;
 }
 //-----------------------------------------------------------------------------
-void
-FloatCell::demote_float_to_int(APL_Float qct)
-{
-   if (Cell::is_near_int(value.fval, qct))
-      {
-        if (value.fval < 0)
-           {
-             APL_Integer ival(0.3 - value.fval);
-             new (this) IntCell(-ival);
-           }
-        else
-           {
-             APL_Integer ival(0.3 + value.fval);
-             new (this) IntCell(ival);
-           }
-      }
-}
-//-----------------------------------------------------------------------------
 Comp_result
 FloatCell::compare(const Cell & other) const
 {
@@ -134,21 +115,11 @@ FloatCell::compare(const Cell & other) const
 ErrorCode
 FloatCell::bif_factorial(Cell * Z) const
 {
-const APL_Float qct = Workspace::get_CT();
-
-   if (is_near_int(qct))
-      {
-        const APL_Integer iv = get_near_int(qct);
-        IntCell icell(iv);
-        icell.bif_factorial(Z);
-        return E_NO_ERROR;
-      }
-
    // max N! that fits into double is about 170
    //
    if (value.fval > 170)   return E_DOMAIN_ERROR;
 
-const double arg = value.fval + 1.0;
+const APL_Float arg = value.fval + 1.0;
    new (Z) FloatCell(tgamma(arg));
    return E_NO_ERROR;
 }
@@ -156,10 +127,7 @@ const double arg = value.fval + 1.0;
 ErrorCode
 FloatCell::bif_conjugate(Cell * Z) const
 {
-   if (is_near_int(Workspace::get_CT()))
-      new (Z) IntCell(get_checked_near_int());
-   else
-      new (Z) FloatCell(value.fval);
+   new (Z) FloatCell(value.fval);
    return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
@@ -192,7 +160,8 @@ FloatCell::bif_magnitude(Cell * Z) const
 ErrorCode
 FloatCell::bif_reciprocal(Cell * Z) const
 {
-   if (is_near_zero(Workspace::get_CT()))  return E_DOMAIN_ERROR;
+const APL_Float z = 1.0/value.fval;
+   if (!isfinite(z))   return E_DOMAIN_ERROR;;
 
    new (Z) FloatCell(1.0/value.fval);
    return E_NO_ERROR;
@@ -202,12 +171,12 @@ ErrorCode
 FloatCell::bif_roll(Cell * Z) const
 {
 const APL_Integer qio = Workspace::get_IO();
-   if (!is_near_int(Workspace::get_CT()))   return E_DOMAIN_ERROR;
+   if (!is_near_int())   return E_DOMAIN_ERROR;
 
 const APL_Integer set_size = get_checked_near_int();
    if (set_size <= 0)   return E_DOMAIN_ERROR;
 
-const uint64_t rnd = Workspace::get_RL();
+const uint64_t rnd = Workspace::get_RL(set_size);
    new (Z) IntCell(qio + (rnd % set_size));
    return E_NO_ERROR;
 }
@@ -215,34 +184,45 @@ const uint64_t rnd = Workspace::get_RL();
 ErrorCode
 FloatCell::bif_pi_times(Cell * Z) const
 {
-   new (Z) FloatCell(M_PI * value.fval);
-   return E_NO_ERROR;
+   return zv(Z, value.fval * M_PI);
+}
+//-----------------------------------------------------------------------------
+ErrorCode
+FloatCell::bif_pi_times_inverse(Cell * Z) const
+{
+   return zv(Z, value.fval / M_PI);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 FloatCell::bif_ceiling(Cell * Z) const
 {
-   // we want values slightly above an int to be rounded down rather than up
-   //
 const APL_Float qct = Workspace::get_CT();
-const double d_ret = ceil(value.fval - Workspace::get_CT());
+const APL_Float c = ceil(value.fval);   // c >= value.fval
 
-   if (Cell::is_near_int(d_ret, qct))   new (Z) IntCell(llrint(d_ret));
-   else                                 new (Z) FloatCell(d_ret);
-   return E_NO_ERROR;
+   // if the distance between value.fval and c is close to 1 (within ⎕CT)
+   // then value.fval is rounded down rather than up. In that decision we
+   // compute:  tolerantly_equal(c - value.fval, 1)
+   // and NOT   tolerantly_equal(value.fval, c - 1) !
+   //
+   if (tolerantly_equal(c - value.fval, 1.0, qct))   return zv(Z, c - 1);
+
+   return zv(Z, c);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 FloatCell::bif_floor(Cell * Z) const
 {
-   // we want values slightly below an int to be rounded up rather than down
-   //
 const APL_Float qct = Workspace::get_CT();
-const double d_ret = floor(value.fval + qct);
+const APL_Float f = floor(value.fval);   // f <= value.fval
 
-   if (Cell::is_near_int(d_ret, qct))   new (Z) IntCell(llrint(d_ret)); 
-   else                                 new (Z) FloatCell(d_ret);
-   return E_NO_ERROR;
+   // if the distance between value.fval and f is close to 1 (within ⎕CT)
+   // then value.fval is rounded up rather than down. In that decision we
+   // compute:  tolerantly_equal(value.fval - f, 1)
+   // and NOT   tolerantly_equal(value.fval, f + 1) !
+   //
+   if (tolerantly_equal(value.fval - f, 1.0, qct))   return zv(Z, f + 1);
+
+   return zv(Z, f);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
@@ -263,8 +243,8 @@ FloatCell::bif_nat_log(Cell * Z) const
       }
    else if (value.fval < 0)
       {
-        const double real = log(-value.fval);
-        const double imag = M_PI;   // argz(-1)
+        const APL_Float real = log(-value.fval);
+        const APL_Float imag = M_PI;   // argz(-1)
         new (Z) ComplexCell(real, imag);
       }
    else
@@ -287,7 +267,6 @@ FloatCell::bif_add(Cell * Z, const Cell * A) const
    if (A->is_real_cell())
       {
         new (Z) FloatCell(A->get_real_value() + get_real_value());
-        Z->demote_float_to_int(Workspace::get_CT());
         return E_NO_ERROR;
       }
 
@@ -302,7 +281,6 @@ FloatCell::bif_subtract(Cell * Z, const Cell * A) const
    if (A->is_real_cell())
       {
         new (Z) FloatCell(A->get_real_value() - get_real_value());
-        Z->demote_float_to_int(Workspace::get_CT());
         return E_NO_ERROR;
       }
 
@@ -314,42 +292,61 @@ FloatCell::bif_subtract(Cell * Z, const Cell * A) const
 ErrorCode
 FloatCell::bif_multiply(Cell * Z, const Cell * A) const
 {
-   if (A->is_real_cell())
-      { 
-        new (Z) FloatCell(A->get_real_value() * get_real_value());
-        Z->demote_float_to_int(Workspace::get_CT());
+   if (!A->is_numeric())   return E_DOMAIN_ERROR;
+
+const APL_Float ar = A->get_real_value();
+const APL_Float ai = A->get_imag_value();
+
+   if (ai == 0.0)   // real result
+      {
+        const APL_Float z = ar * value.fval;
+        if (!isfinite(z))   return E_DOMAIN_ERROR;;
+        new (Z) FloatCell(z);
         return E_NO_ERROR;
       } 
 
-   // delegate to A
+   // complex result
    //
-   return A->bif_multiply(Z, this);
+const APL_Complex a(ar, ai);
+const APL_Complex z = a * value.fval;
+   if (!isfinite(z.real()))   return E_DOMAIN_ERROR;;
+   if (!isfinite(z.imag()))   return E_DOMAIN_ERROR;;
+   new (Z) ComplexCell(z);
+   return E_NO_ERROR;
 }     
 //-----------------------------------------------------------------------------
 ErrorCode
 FloatCell::bif_divide(Cell * Z, const Cell * A) const
 {
-const APL_Float qct = Workspace::get_CT();
+   if (!A->is_numeric())   return E_DOMAIN_ERROR;
 
-   if (this->is_near_zero(qct))   // divide by 0
+const APL_Float ar = A->get_real_value();
+const APL_Float ai = A->get_imag_value();
+
+   if (value.fval == 0.0)   // A ÷ 0
       {
-         if (!A->is_near_zero(qct))   return E_DOMAIN_ERROR;
-         new (Z) IntCell(1);   // 0/0 is 1 in APL
-         return E_NO_ERROR;
+         if (ar != 0.0)   return E_DOMAIN_ERROR;
+         if (ai != 0.0)   return E_DOMAIN_ERROR;
+
+         return IntCell::z1(Z);   // 0÷0 is 1 in APL
       }
 
 
-   if (A->is_real_cell())
+   if (ai == 0.0)   // real result
       {
-        const APL_Float real = A->get_real_value() / get_real_value();
-        if (Cell::is_near_int(real, qct))   new (Z) IntCell(real, qct);
-        else                                new (Z) FloatCell(real);
+        const APL_Float real = ar / value.fval ;
+        new (Z) FloatCell(real);
          return E_NO_ERROR;
       }
 
-   // delegate to A
+   // complex result
    //
-   return A->bif_divide(Z, this);
+const APL_Complex a(ar, ai);
+const APL_Complex z = a / value.fval;
+   if (!isfinite(z.real()))   return E_DOMAIN_ERROR;;
+   if (!isfinite(z.imag()))   return E_DOMAIN_ERROR;;
+   new (Z) ComplexCell(z);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 ErrorCode
@@ -359,63 +356,43 @@ FloatCell::bif_power(Cell * Z, const Cell * A) const
    //
    if (!A->is_numeric())   return E_DOMAIN_ERROR;
 
-const APL_Float qct = Workspace::get_CT();
-   if (is_near_int(qct))
+const APL_Float ar = A->get_real_value();
+const APL_Float ai = A->get_imag_value();
+
+   // 1. A == 0
+   //
+   if (ar == 0.0 && ai == 0.0)
+       {
+         if (value.fval == 0.0)   return IntCell::z1(Z);   // 0⋆0 is 1
+         if (value.fval > 0)      return IntCell::z0(Z);   // 0⋆N is 0
+         return E_DOMAIN_ERROR;;                           // 0⋆¯N = 1÷0
+       }
+
+   // 2. real A > 0   (real result)
+   //
+   if (ai == 0.0)
       {
-        APL_Integer b = get_near_int(qct);
-        const bool invert_Z = b < 0;
-        if (invert_Z)   b = - b;
-
-        if (b <= 1)   // special cases A⋆1, A⋆0, and A⋆¯1
+        if (ar  == 1.0)   return IntCell::z1(Z);   // 1⋆b = 1
+        if (ar > 0)
            {
-             if (b == 0)   return IntCell::z1(Z);  // A⋆0 is 1
-
-             if (invert_Z)               return A->bif_reciprocal(Z);
-             if (A->is_real_cell())      return A->bif_conjugate(Z);
-             new (Z) ComplexCell(A->get_complex_value());
+             const APL_Float z = pow(ar, value.fval);
+             if (!isfinite(z))   return E_DOMAIN_ERROR;;
+             new (Z) FloatCell(z);
              return E_NO_ERROR;
            }
+
+        /* fall through */
       }
 
-   if (A->is_near_zero(qct))
-      {
-        if (get_real_value() < 0)   return E_DOMAIN_ERROR;
-        return IntCell::z0(Z);
-      }
-          
-   if (A->is_real_cell())
-      {
-        if (A->get_real_value() < 0)
-           {
-             if (is_near_int(qct))
-                {
-                  const APL_Integer b = get_near_int(qct);
-                  const APL_Float z = pow(A->get_real_value(), (APL_Float)b);
-                  new (Z) FloatCell(z);
-                }
-             else
-                {
-                  const APL_Complex z = pow(A->get_complex_value(),
-                                            get_complex_value());
-                  new (Z) ComplexCell(z);
-                  Z->demote_complex_to_real(Workspace::get_CT());
-                }
-           }
-        else
-           {
-             new (Z) FloatCell(pow(A->get_real_value(),
-                               get_real_value()));
-           }
-        return E_NO_ERROR;
-      }
-
-   // complex A to the B-th power
+   // 3. complex result
    //
    {
-     APL_Complex z = pow(A->get_complex_value(), get_real_value());
+     const APL_Complex a(ar, ai);
+     const APL_Complex z = pow(a, value.fval);
+     if (!isfinite(z.real()))   return E_DOMAIN_ERROR;;
+     if (!isfinite(z.imag()))   return E_DOMAIN_ERROR;;
 
      new (Z) ComplexCell(z);
-     Z->demote_complex_to_real(Workspace::get_CT());
      return E_NO_ERROR;
    }
 }
@@ -423,51 +400,61 @@ const APL_Float qct = Workspace::get_CT();
 ErrorCode
 FloatCell::bif_residue(Cell * Z, const Cell * A) const
 {
-   if (A->is_complex_cell())
+   if (!A->is_numeric())   return E_DOMAIN_ERROR;
+
+   if (A->get_imag_value() != 0)   // complex A
       {
         ComplexCell B(get_real_value(), 0);
         return B.bif_residue(Z, A);
       }
 
-   if (A->is_numeric())
+   // if A is zero , return B
+   //
+const APL_Float a = A->get_real_value();
+const APL_Float b = value.fval;
+   if (a == 0.0)   return zv(Z, b);
+
+   // IBM: if B is zero , return B
+   //
+   if (b == 0.0)   return IntCell::z0(Z);
+
+   // if ⎕CT != 0 and B ÷ A is close to an integer within ⎕CT then return 0.
+   //
+   // Note: In that case, the integer to which A ÷ B is close is either
+   // floor(A ÷ B) or ceil(A ÷ B).
+   //
+const APL_Float quot = b / a;
+const APL_Float qf = floor(quot);
+APL_Float qct = Workspace::get_CT();
+
+   if (qct != 0)
       {
-        const APL_Float qct = Workspace::get_CT();
+        const APL_Float qc = ceil(quot);
 
-        // if A is zero , return B
-        //
-        if (A->is_near_zero(0))
-           {
-             Z->init(*this);
-             return E_NO_ERROR;
-           }
-
-        // if ⎕CT != 0 and B÷A is close to an integer within ⎕CT then return 0.
-        // Otherwise return B mod A
-        //
-        const APL_Float remainder = fmod(get_real_value(), A->get_real_value()); 
-        if (qct == 0.0)   // ⎕CT is 0
-           {
-             new (Z) FloatCell(remainder);
-             return E_NO_ERROR;
-           }
-
-        if (Cell::is_near_zero(remainder, qct))
-           {
-             new (Z) IntCell(0);
-             return E_NO_ERROR;
-           }
-
-        if (Cell::is_near_zero(A->get_real_value() - remainder, qct))
-           {
-             new (Z) IntCell(0);
-             return E_NO_ERROR;
-           }
-
-        new (Z) FloatCell(remainder);
-        return E_NO_ERROR;
+        if (quot < qf + qct)   return IntCell::z0(Z);
+        if (quot > qc - qct)   return IntCell::z0(Z);
       }
 
-   return E_DOMAIN_ERROR;
+   // Otherwise return B mod A
+   //
+   // ISO: R←B-(×B)×|A×⌊|B÷A and return R if (×A)=×B or R+A otherwise
+   // IBM: if A > 0  then 0 ≤ Z < A
+   //                else 0 ≥ Z > A
+   //
+APL_Float z = b - a * qf;
+   if (a < 0)   // Z ≤ 0
+      {
+         while (z > 0)    z += a;
+         while (z <= a)   z -= a;
+      }
+   else         // Z ≥ 0
+      {
+         while (z < 0)    z += a;
+         while (z >= a)   z -= a;
+      }
+
+   Assert(z*a >= 0);
+   return zv(Z, z);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
@@ -561,7 +548,7 @@ int int_fract = ucs.size();
 bool
 FloatCell::is_big(APL_Float val, int quad_pp)
 {
-static const double big[MAX_Quad_PP + 1] =
+static const APL_Float big[MAX_Quad_PP + 1] =
 {
                   1ULL, // not used since MIN_Quad_PP == 1
                  10ULL,
