@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -19,9 +19,20 @@
 */
 
 #include "Avec.hh"
+#include "Bif_F12_FORMAT.hh"
+#include "Bif_F12_SORT.hh"
+#include "Bif_OPER1_COMMUTE.hh"
+#include "Bif_OPER1_EACH.hh"
+#include "Bif_OPER1_REDUCE.hh"
+#include "Bif_OPER1_SCAN.hh"
+#include "Bif_OPER2_INNER.hh"
+#include "Bif_OPER2_OUTER.hh"
+#include "Bif_OPER2_POWER.hh"
+#include "Bif_OPER2_RANK.hh"
 #include "Common.hh"
 #include "Id.hh"
 #include "Output.hh"
+#include "PrimitiveFunction.hh"
 #include "PrintOperator.hh"
 #include "QuadFunction.hh"
 #include "Quad_FX.hh"
@@ -32,24 +43,42 @@
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
-ostream &
-operator << (ostream & out, Id id)
-{
-   return out << id_name(id);
-}
-//-----------------------------------------------------------------------------
-Function *
-get_system_function(Id id)
+const char *
+ID::name(ID::Id id)
 {
    switch(id)
       {
-#define av(x) case ID_ ## x: return &Bif_F12_ROLL::fun;
-#define pp(x)
-#define qf(x) case ID_Quad_ ## x: return &Quad_ ## x::fun;
-#define qv(x)
-#define st(x) 
+#define pp(i, _u, _v) case i:          return #i;
+#define qf(i,  u, _v)  case Quad_ ## i: return u;
+#define qv(i,  u, _v)  case Quad_ ## i: return u;
+#define sf(i,  u, _v)  case i:          return u;
+#define st(i,  u, _v)  case i:          return u;
+
+#include "Id.def"
+
+        default: break;
+      }
+
+   return "Unknown-ID";
+}
+//-----------------------------------------------------------------------------
+ostream &
+operator << (ostream & out, ID::Id id)
+{
+   return out << ID::name(id);
+}
+//-----------------------------------------------------------------------------
+Function *
+ID::get_system_function(ID::Id id)
+{
+   switch(id)
+      {
+#define pp(i, _u, _v)
+#define qf(i, _u, _v) case ID::Quad_ ## i: return Quad_ ## i::fun;
+#define qv(i, _u, _v)
+#define sf(i, _u, _v) case ID:: i:        return Bif_ ## i::fun;
+#define st(i, _u, _v)
 
-#define id_def(_id, _uni, _val, _mac) _mac(_id)
 #include "Id.def"
 
         default: break;
@@ -59,17 +88,17 @@ get_system_function(Id id)
 }
 //-----------------------------------------------------------------------------
 Symbol *
-get_system_variable(Id id)
+ID::get_system_variable(ID::Id id)
 {
    switch(id)
       {
-#define av(x)
-#define pp(x)
-#define qf(x)
-#define qv(x) case ID_Quad_ ## x:return &Workspace::get_v_Quad_ ## x();
-#define st(x) 
+#define pp(_i, _u, _v)
+#define qf(_i, _u, _v)
+#define qv( i, _u, _v)   case ID::Quad_ ## i: \
+                              return &Workspace::get_v_Quad_ ## i();
+#define sf(_i, _u, _v) 
+#define st(_i, _u, _v) 
 
-#define id_def(_id, _uni, _val, _mac) _mac(_id)
 #include "Id.def"
 
         default: break;
