@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -27,13 +27,18 @@
 //-----------------------------------------------------------------------------
 IndexExpr::IndexExpr(Assign_state astate, const char * loc)
    : DynamicObject(loc),
-     quad_ct(Workspace::get_CT()),
      quad_io(Workspace::get_IO()),
      rank(0),
      assign_state(astate)
 {
 }
 //-----------------------------------------------------------------------------
+IndexExpr::~IndexExpr()
+{
+// Q(this)
+// Backtrace::show(__FILE__, __LINE__);
+}
+//-----------------------------------------------------------------------------
 Value_P
 IndexExpr::extract_value(Rank rk)
 {
@@ -65,7 +70,7 @@ IndexExpr::check_range(const Shape & shape) const
         const ShapeItem max_idx = shape.get_shape_item(r) + quad_io;
         loop(i, ival->element_count())
            {
-             const APL_Integer idx = ival->get_ravel(i).get_near_int(quad_ct);
+             const APL_Integer idx = ival->get_ravel(i).get_near_int();
              if (idx < quad_io)    return true;
              if (idx >= max_idx)   return true;
            }
@@ -79,18 +84,17 @@ IndexExpr::get_axis(Rank max_axis) const
 {
    if (rank != 1)   INDEX_ERROR;
 
-const APL_Float   qct = Workspace::get_CT();
 const APL_Integer qio = Workspace::get_IO();
 
 Value_P I = values[0];
    if (!I->is_scalar_or_len1_vector())     INDEX_ERROR;
 
-   if (!I->get_ravel(0).is_near_int(qct))   INDEX_ERROR;
+   if (!I->get_ravel(0).is_near_int())   INDEX_ERROR;
 
    // if axis becomes (signed) negative then it will be (unsigned) too big.
    // Therefore we need not test for < 0.
    //
-Rank axis = I->get_ravel(0).get_near_int(qct) - qio;
+Rank axis = I->get_ravel(0).get_near_int() - qio;
    if (axis >= max_axis)   INDEX_ERROR;
 
    return axis;
@@ -116,12 +120,11 @@ Value_P vx = values[0];
    if (!vx)                  INDEX_ERROR;
 
 const ShapeItem xlen = vx->element_count();
-const APL_Float qct = Workspace::get_CT();
 const APL_Integer qio = Workspace::get_IO();
  
 Shape shape;
      loop(x, xlen)
-        shape.add_shape_item(vx->get_ravel(x).get_near_int(qct) - qio);
+        shape.add_shape_item(vx->get_ravel(x).get_near_int() - qio);
 
    return shape;
 }
