@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -405,11 +405,10 @@ IntCell::bif_roll(Cell * Z) const
 {
 const APL_Integer qio = Workspace::get_IO();
 const APL_Integer set_size = value.ival;
-   if (set_size < 0)   return E_DOMAIN_ERROR;
+   if (set_size <= 0)   return E_DOMAIN_ERROR;
 
-const uint64_t rnd = Workspace::get_RL();
-   if (set_size == 0)   new (Z) FloatCell((1.0*rnd)/0x8000000000000000ULL);
-   else                 new (Z) IntCell(qio + (rnd % set_size));
+const uint64_t rnd = Workspace::get_RL(set_size);
+   new (Z) IntCell(qio + (rnd % set_size));
    return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
@@ -421,17 +420,22 @@ IntCell::bif_pi_times(Cell * Z) const
 }
 //-----------------------------------------------------------------------------
 ErrorCode
+IntCell::bif_pi_times_inverse(Cell * Z) const
+{
+   new (Z) FloatCell(value.ival / M_PI);
+   return E_NO_ERROR;
+}
+//-----------------------------------------------------------------------------
+ErrorCode
 IntCell::bif_ceiling(Cell * Z) const
 {
-   new (Z) IntCell(value.ival);
-   return E_NO_ERROR;
+   return zv(Z, value.ival);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
 IntCell::bif_floor(Cell * Z) const
 {
-   new (Z) IntCell(value.ival);
-   return E_NO_ERROR;
+   return zv(Z, value.ival);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
@@ -516,6 +520,8 @@ IntCell::bif_subtract(Cell * Z, const Cell * A) const
 ErrorCode
 IntCell::bif_divide(Cell * Z, const Cell * A) const
 {
+   if (!A->is_numeric())   return E_DOMAIN_ERROR;
+
    if (A->is_integer_cell())
       {
         // both cells are integers.
@@ -523,20 +529,17 @@ IntCell::bif_divide(Cell * Z, const Cell * A) const
         const APL_Integer a = A->get_int_value();
         const APL_Integer b =    get_int_value();
 
-        if (b == 0)   // allowed if a == 0 as well
+        if (b == 0)   // a ÷ 0 is allowed iff a == 0
            {
              if (a != 0)   return E_DOMAIN_ERROR;
-             new (Z) IntCell(1);   // 0 ÷ 0 is defined as 1
-             return E_NO_ERROR;
+             return IntCell::z1(Z);   // 0÷0 is 1 in APL
            }
 
         const APL_Float i_quot = a / b;
         const APL_Float r_quot = a / (APL_Float)b;
 
-        if (r_quot > LARGE_INT ||
-            r_quot < SMALL_INT)     new (Z) FloatCell(r_quot);
-        else if (a != i_quot * b)   new (Z) FloatCell(r_quot);
-   else                             new (Z) IntCell(i_quot);
+        if (a != i_quot * b)   new (Z) FloatCell(r_quot);
+        else                   new (Z) IntCell(i_quot);
         return E_NO_ERROR;
       }
 
@@ -702,22 +705,41 @@ const bool invert_Z = b < 0;
 ErrorCode
 IntCell::bif_residue(Cell * Z, const Cell * A) const
 {
-const APL_Float qct = Workspace::get_CT();
+   if (!A->is_numeric())   return E_DOMAIN_ERROR;
 
-   // special cases
+   if (A->get_imag_value() != 0)   // complex A
+      {
+        ComplexCell B(value.ival, 0);
+        return B.bif_residue(Z, A);
+      }
+
+   // if A is zero , return B
    //
-   if (A->is_near_zero(qct))   // 0∣B is B
+const APL_Float a = A->get_real_value();
+   if (a == 0.0)   return zv(Z, value.ival);
+
+   // IBM: if B is zero , return B
+   //
+   if (value.ival == 0)   return IntCell::z0(Z);
+
+   // if ⎕CT != 0 and B ÷ A is close to an integer within ⎕CT then return 0.
+   //
+const APL_Float quot = value.ival / a;
+const APL_Float qf = floor(quot);
+const APL_Float qct = Workspace::get_CT();
+   if (qct != 0)
       {
-        Z->init(*this);
-        return E_NO_ERROR;
-      }
-
-   if (value.ival == 0)   // B∣0 is 0
-      {
-        new (Z) IntCell(0);
-        return E_NO_ERROR;
+        const APL_Float qc = ceil(quot);
+        if (quot > qc - qct)   return IntCell::z0(Z);
+        if (quot < qf + qct)   return IntCell::z0(Z);
       }
 
+   // Otherwise return B mod A
+   //
+   // ISO: R←B-(×B)×|A×⌊|B÷A and return R if (×A)=×B or R+A otherwise
+   // IBM: if A > 0  then 0 ≤ Z < A
+   //                else 0 ≥ Z > A
+   //
    if (A->is_integer_cell())
       {
         APL_Integer rest = value.ival % A->get_int_value();
@@ -729,53 +751,18 @@ const APL_Float qct = Workspace::get_CT();
            {
               if (rest < 0)   rest += A->get_int_value();
            }
-        new (Z) IntCell(rest);
-        return E_NO_ERROR;
+        return zv(Z, rest);
       }
 
-   if (A->is_real_cell())
-      {
-        const APL_Float valA = A->get_real_value();
-        const APL_Float f_quot = value.ival / valA;
-        APL_Integer i_quot = floor(f_quot);
+APL_Float z = value.ival - a * qf;
+   if (Cell::is_near_zero(z))   return zv(Z, 0);
 
-        // compute i_quot with respect to ⎕CT
-        //
-        if (tolerantly_equal(i_quot + 1, f_quot, qct))  ++i_quot;
+   // make ×Z == ×A
+   //
+   if      (a < 0 && z > 0)   z += a;   // += since a < 0
+   else if (a > 0 && z < 0)   z += a;
 
-        APL_Float rest = value.ival - valA * i_quot;
-        if (Cell::is_near_zero(rest, qct))
-          {
-            new (Z) IntCell(0);
-            return E_NO_ERROR;
-          }
-
-        if (valA < 0)   // A negative -> Z negative
-           {
-             if (rest > 0)   rest += valA;   // += since A < 0
-           }
-        else
-           {
-             if (rest < 0)   rest += valA;
-           }
-        new (Z) FloatCell(rest);
-        return E_NO_ERROR;
-      }
-
-   if (A->is_complex_cell())   // i.e. complex
-      {
-        const APL_Complex a = A->get_complex_value();
-        const APL_Complex b(value.ival);
-
-        // We divide A by B, round down the result, and subtract.
-        //
-        A->bif_divide(Z, this);      // Z = B/A
-        Z->bif_floor(Z);             // Z = A/B rounded down.
-        new (Z) ComplexCell(b - a*Z->get_complex_value());
-        return E_NO_ERROR;
-      }
-
-   return E_DOMAIN_ERROR;
+   return FloatCell::zv(Z, z);
 }
 //-----------------------------------------------------------------------------
 ErrorCode
@@ -835,37 +822,36 @@ IntCell::compare(const Cell & other) const
 }
 //-----------------------------------------------------------------------------
 bool
-IntCell::greater(const Cell * other, bool ascending) const
+IntCell::greater(const Cell & other) const
 {
 const APL_Integer this_val  = get_int_value();
 
-   switch(other->get_cell_type())
+   switch(other.get_cell_type())
       {
         case CT_INT:
              {
-               const APL_Integer other_val = other->get_int_value();
-               if (this_val == other_val)   return this > other;
-               return this_val > other_val ? ascending : !ascending;
+               const APL_Integer other_val = other.get_int_value();
+               if (this_val == other_val)   return this > &other;
+               return this_val > other_val;
              }
 
         case CT_FLOAT:
              {
-               const APL_Float other_val = other->get_real_value();
-               if (this_val == other_val)   return this > other;
-               return this_val > other_val ? ascending : !ascending;
+               const APL_Float other_val = other.get_real_value();
+               if (this_val == other_val)   return this > &other;
+               return this_val > other_val;
              }
 
         case CT_COMPLEX: break;
-        case CT_CHAR:    return ascending;   // never greater
-        case CT_POINTER: return !ascending;
+        case CT_CHAR:    return true;    // int is always greater than char
+        case CT_POINTER: return false;   // int is always smaller than nested
         case CT_CELLREF: DOMAIN_ERROR;
         default:         Assert(0 && "Bad celltype");
       }
 
-const Comp_result comp = compare(*other);
-   if (comp == COMP_EQ)   return this > other;
-   if (comp == COMP_GT)   return ascending;
-   else                   return !ascending;
+const Comp_result comp = compare(other);
+   if (comp == COMP_EQ)   return this > &other;
+   return (comp == COMP_GT);
 }
 //-----------------------------------------------------------------------------
 PrintBuffer
