@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2013  Dr. Jürgen Sauermann
+    Copyright (C) 2013-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,6 +18,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+#include <dirent.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -53,9 +54,9 @@ LibPaths::init(const char * argv0, bool logit)
 
    loop(d, LIB_MAX)
       {
-        if (root_from_env)        lib_dirs[d].cfg_src = LibDir::CS_ENV;
-        else if (root_from_pwd)   lib_dirs[d].cfg_src = LibDir::CS_ARGV0;
-        else                      lib_dirs[d].cfg_src = LibDir::CS_NONE;
+        if (root_from_env)        lib_dirs[d].cfg_src = LibDir::CSRC_ENV;
+        else if (root_from_pwd)   lib_dirs[d].cfg_src = LibDir::CSRC_ARGV0;
+        else                      lib_dirs[d].cfg_src = LibDir::CSRC_NONE;
       }
 }
 //-----------------------------------------------------------------------------
@@ -225,14 +226,14 @@ LibPaths::get_lib_dir(LibRef libref)
 {
    switch(lib_dirs[libref].cfg_src)
       {
-        case LibDir::CS_NONE: return UTF8_string();
+        case LibDir::CSRC_NONE:      return UTF8_string();
 
-        case LibDir::CS_ENV:
-        case LibDir::CS_ARGV0:     break;   // continue below
-
-        case LibDir::CS_PREF_SYS:
-        case LibDir::CS_PREF_HOME: return lib_dirs[libref].dir_path;
+        case LibDir::CSRC_ENV:
+        case LibDir::CSRC_ARGV0:     break;   // continue below
 
+        case LibDir::CSRC_PREF_SYS:
+        case LibDir::CSRC_PREF_HOME:
+        case LibDir::CSRC_CMD:     return lib_dirs[libref].dir_path;
       }
 
 UTF8_string ret(APL_lib_root);
@@ -367,4 +368,17 @@ UTF8_string filename = get_lib_dir(lib);
       }
 }
 //-----------------------------------------------------------------------------
+bool
+LibPaths::is_present(LibRef lib)
+{
+   // try to open dir and return false if that fails
+   //
+UTF8_string path = LibPaths::get_lib_dir(lib);
+DIR * dir = opendir(path.c_str());
+   if (!dir)   return false;
+
+   closedir(dir);
+   return true;
+}
+//-----------------------------------------------------------------------------
 
