@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -22,6 +22,12 @@
 #include "PrintOperator.hh"
 #include "UTF8_string.hh"
 
+//-----------------------------------------------------------------------------
+LvalCell::LvalCell(Cell * cell, Value * cell_owner)
+{
+   value.lval = cell;
+   value2.owner = cell_owner;
+}
 //-----------------------------------------------------------------------------
 Cell *
 LvalCell::get_lval_value()  const
