@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -21,14 +21,15 @@
 
 #include "Avec.hh"
 #include "Command.hh"
-#include "Input.hh"
 #include "InputFile.hh"
+#include "LineInput.hh"
 #include "Logging.hh"
 #include "Nabla.hh"
 #include "Output.hh"
 #include "PrintOperator.hh"
 #include "Symbol.hh"
 #include "UserFunction.hh"
+#include "UserPreferences.hh"
 #include "Workspace.hh"
 
 #define NABLA_ERROR throw_edit_error(LOC)
@@ -48,6 +49,7 @@ Nabla::Nabla(const UCS_string & cmd)
      edit_from(-1),
      edit_to(-1),
      function_existed(false),
+     modified(false),
      do_close(false),
      locked(false),
      current_line(1),
@@ -92,10 +94,21 @@ int control_D_count = 0;
    Log(LOG_nabla)   UERR << "Nabla(" << fun_header << ")..." << endl;
    while (!do_close)
        {
-         const UCS_string prompt = current_line.print_prompt();
-         const char * line = Input::get_user_line_nabla(&prompt);
+         const UCS_string prompt = current_line.print_prompt(0);
+         bool eof = false;
+         UCS_string line;
+         if (uprefs.raw_cin)
+            {
+              LineHistory lh(10);
+              InputMux::get_line(LIM_Nabla, prompt, line, eof, lh);
+            }
+         else
+            {
+              LineHistory lh(*this);
+              InputMux::get_line(LIM_Nabla, prompt, line, eof, lh);
+            }
 
-         if (line == 0)   // end-of-input (^D) pressed
+         if (eof)   // end-of-input (^D) pressed
             {
               ++control_D_count;
               if (control_D_count < 5)
@@ -104,12 +117,10 @@ int control_D_count = 0;
                     continue;
                  }
                COUT << endl << "      *** end of input" << endl;
-               Command::cmd_OFF(2);
+               Command::cmd_OFF(5);
             }
 
-         const UTF8_string utf(line);
-         UCS_string ucs(utf);
-         if (const char * loc = parse_oper(ucs, false))
+         if (const char * loc = parse_oper(line, false))
             {
               UERR << "??? " << loc << endl;
               continue;
@@ -132,6 +143,41 @@ UCS_string fun_text;
         fun_text.append(UNI_ASCII_LF);
       }
 
+   // maybe copy function into the history
+   //
+   if (!uprefs.raw_cin && !InputFile::running_script() &&
+       ((uprefs.nabla_to_history ==  /* always     */ 2) ||
+        ((uprefs.nabla_to_history == /* if changed */ 1) && modified)))
+      {
+        // create a history entry that can be re-entered and replace
+        // the last history line (which contained some ∇foo ...)
+        //
+        {
+          UCS_string line_0("    ");
+          line_0.append(UNI_NABLA);
+          line_0.append(lines[0].text);
+          line_0.remove_trailing_whitespaces();
+          LineInput::replace_history_line(line_0);
+
+        }
+
+        for (int l = 1; l < lines.size(); ++l)
+            {
+              UCS_string line_l("[");
+              line_l.append_number(l);
+              line_l.append_utf8("]  ");
+              while (line_l.size() < 6)   line_l.append(UNI_ASCII_SPACE);
+              line_l.append(lines[l].text);
+              line_l.remove_trailing_whitespaces();
+
+              LineInput::add_history_line(line_l);
+           }
+
+        UCS_string line_N("   ");
+        line_N.append(UNI_NABLA);
+        LineInput::add_history_line(line_N);
+      }
+
 int error_line = 0;
 UCS_string creator(InputFile::current_filename());
    creator.append(UNI_ASCII_COLON);
@@ -178,7 +224,7 @@ UCS_string::iterator c(first_command.begin());
 
    // skip leading spaces
    //
-   while (c.more() && c.get() <= UNI_ASCII_SPACE)   c.next();
+   while (c.more() && Avec::is_white(c.get()))   c.next();
 
    // skip leading nabla.
    //
@@ -186,7 +232,7 @@ UCS_string::iterator c(first_command.begin());
 
    // skip leading spaces
    //
-   while (c.more() && c.get() <= UNI_ASCII_SPACE)   c.next();
+   while (c.more() && Avec::is_white(c.get()))   c.next();
 
    // function header.
    //
@@ -201,7 +247,7 @@ UCS_string::iterator c(first_command.begin());
       {
         for (int off = 1; ; ++off)
             {
-              if (c.get(off) <= UNI_ASCII_SPACE)   continue; 
+              if (Avec::is_white(c.get(off)))   continue; 
               if (Avec::is_first_symbol_char(c.get(off)))   // axis
                  {
                    fun_header.append(c.next());   //  copy the [
@@ -239,6 +285,7 @@ UserFunction_header hdr(fun_header);
       {
         case NC_UNUSED_USER_NAME:   // open a new function
              function_existed = false;
+             modified = true;
              {
                // a new function must not have a command
                //
@@ -282,7 +329,6 @@ UserFunction_header hdr(fun_header);
    // at this point the (new or existing) function was successfully opened.
    // That means that at least the header is present (lines.size() > 0)
 
-
    // immediate close (only show command is allowed here),
    // e.g. ∇fun[⎕]∇
    //
@@ -309,19 +355,19 @@ Nabla::parse_oper(UCS_string & oper, bool initial)
 
    // skip trailing spaces
    //
-   while (oper.size() > 0 && oper.last() < ' ')   oper.pop();
+   while (oper.size() > 0 && oper.last() <= ' ')   oper.pop();
    if (oper.size() > 0 && oper.last() == UNI_NABLA)
       {
         do_close = true;
         oper.pop();
-        while (oper.size() > 0 && oper.last() < ' ')   oper.pop();
+        while (oper.size() > 0 && oper.last() <= ' ')   oper.pop();
       }
    else if (oper.size() > 0 && oper.last() == UNI_DEL_TILDE)
       {
         do_close = true;
         locked = true;
         oper.pop();
-        while (oper.size() > 0 && oper.last() < ' ')   oper.pop();
+        while (oper.size() > 0 && oper.last() <= ' ')   oper.pop();
       }
 
    current_text.clear();
@@ -416,7 +462,7 @@ again:
    // [from ∆ to]
    // [from]
 
-   while (c.get() <= UNI_ASCII_SPACE)   c.next();
+   while (Avec::is_white(c.get()))   c.next();
 
    if (c.get() == UNI_ASCII_L_BRACK)   // another command: ignore previous
       {
@@ -684,6 +730,8 @@ Nabla::execute_delete()
 const int idx_to = find_line(LineLabel(edit_to));
    if (idx_to == -1)   return "Bad line number N in [M∆N] ";
 
+   modified = true;
+
    if (edit_from == -1)   // delete single line
       {
         lines.erase(lines.begin() + idx_to);
@@ -747,7 +795,7 @@ const UCS_string & new_name = header.FUN()->get_name();
         // can simply forget it and continue with the new function
         //
         // UserFunction_header::UserFunction_header() should have triggered
-        // the creation of header.FUN() so we can assume tht it exists. We
+        // the creation of header.FUN() so we can assume that it exists. We
         // need to check that it is not a variable or an existing function.
         //
         if (header.FUN()->get_nc() != NC_UNUSED_USER_NAME)
@@ -758,6 +806,8 @@ const UCS_string & new_name = header.FUN()->get_name();
            }
       }
 
+   modified = true;
+
    lines[0].text = current_text;
    current_line.next();
    return 0;
@@ -776,6 +826,8 @@ ErrorCode ec = parser.parse(current_text, in);
         return 0;
       }
 
+   modified = true;
+
 const int idx_from = find_line(edit_from);
 
    Assert(lines.size() > 0);
@@ -850,7 +902,7 @@ UCS_string ucs("[");
 }
 //-----------------------------------------------------------------------------
 UCS_string
-LineLabel::print_prompt() const
+LineLabel::print_prompt(int min_size) const
 {
 UCS_string ret("[");
    ret.append_number(ln_major);
@@ -861,8 +913,8 @@ UCS_string ret("[");
         loop(s, ln_minor.size())   ret.append(Unicode(char(ln_minor[s])));
       }
 
-   ret.append(Unicode(']'));
-   ret.append(Unicode(' '));
+   ret.append_utf8("] ");
+   while (ret.size() < min_size)   ret.append(UNI_ASCII_SPACE);
    return ret;
 }
 //-----------------------------------------------------------------------------
@@ -908,4 +960,21 @@ operator <<(ostream & out, const LineLabel & lab)
    return out;
 }
 //-----------------------------------------------------------------------------
+UCS_string
+Nabla::FunLine::get_label_and_text() const
+{
+UCS_string ret = label.print_prompt(6);
+   ret.append(text);
+
+   return ret;
+}
+//-----------------------------------------------------------------------------
+UCS_string
+Nabla::get_label_and_text(int line, bool & is_current) const
+{
+const FunLine & fl = lines[line];
+   is_current = fl.label == current_line;
+   return fl.get_label_and_text();
+}
+//-----------------------------------------------------------------------------
 
