@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -35,7 +35,7 @@ vector<NativeFunction *> NativeFunction::valid_functions;
 //-----------------------------------------------------------------------------
 NativeFunction::NativeFunction(const UCS_string & so_name,
                                const UCS_string & apl_name)
-   : Function(ID_USER_SYMBOL, TOK_FUN2),
+   : Function(ID::USER_SYMBOL, TOK_FUN2),
      handle(0),
      name(apl_name),
      original_so_path(so_name),
