@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -65,7 +65,7 @@ Parser::parse(const Token_string & input, Token_string & tos) const
 
    // split input into statements.
    //
-vector<Token_string *> statements;
+Simple_string<Token_string *> statements;
    {
      Source<Token> src(input);
      Token_string  * stat = new Token_string();
@@ -74,7 +74,7 @@ vector<Token_string *> statements;
           const Token tok = src.get();
           if (tok.get_tag() == TOK_DIAMOND)
              {
-               statements.push_back(stat);
+               statements.append(stat);
                stat = new Token_string();
              }
           else
@@ -82,14 +82,22 @@ vector<Token_string *> statements;
                stat->append(tok, LOC);
              }
         }
-     statements.push_back(stat);
+     statements.append(stat);
    }
 
    loop(s, statements.size())
       {
         Token_string * stat = statements[s];
         ErrorCode err = parse_statement(*stat);
-        if (err)   return err;
+        if (err)
+           {
+             while (s < statements.size())
+               {
+                 stat = statements[s++];
+                 delete stat;
+               }
+             return err;
+           }
 
         if (s)   tos.append(Token(TOK_DIAMOND));
 
@@ -105,7 +113,7 @@ vector<Token_string *> statements;
 }
 //-----------------------------------------------------------------------------
 ErrorCode
-Parser::parse_statement(Token_string & tos) const
+Parser::parse_statement(Token_string & tos)
 {
    // 1. convert (X) into X and ((X...)) into (X...)
    //
@@ -163,11 +171,22 @@ Parser::parse_statement(Token_string & tos) const
         tos.print(CERR);
       }
 
+   // 6. update distances between (), [], and {}
+   //
+   {
+     const ErrorCode ec = match_par_bra(tos, false);
+     if (ec != E_NO_ERROR)
+        {
+          loop(t, tos.size())   tos[t].clear(LOC);
+          return ec;
+        }
+   }
+
    return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 void
-Parser::collect_constants(Token_string & tos) const
+Parser::collect_constants(Token_string & tos)
 {
    Log(LOG_collect_constants)
       {
@@ -237,7 +256,7 @@ Parser::collect_constants(Token_string & tos) const
 }
 //-----------------------------------------------------------------------------
 bool
-Parser::collect_groups(Token_string & tos) const
+Parser::collect_groups(Token_string & tos)
 {
    Log(LOG_collect_constants)
       {
@@ -453,12 +472,14 @@ Parser::degrade_scan_reduce(Token_string & tos)
 {
    loop(src, tos.size())
        {
-         if (tos[src].get_Id() == ID_OPER1_REDUCE  ||
-             tos[src].get_Id() == ID_OPER1_REDUCE1 ||
-             tos[src].get_Id() == ID_OPER1_SCAN    ||
-             tos[src].get_Id() == ID_OPER1_SCAN1)
+         if (tos[src].get_Id() == ID::OPER1_REDUCE  ||
+             tos[src].get_Id() == ID::OPER1_REDUCE1 ||
+             tos[src].get_Id() == ID::OPER1_SCAN    ||
+             tos[src].get_Id() == ID::OPER1_SCAN1)
             {
-              const bool is_function = src == 0 ||
+              // tos[src] is / ⌿ \ or ⍀
+              //
+              const bool is_function = src == 0 ||   // (will SYNTAX_ERROR)
                          check_if_value(tos, src - 1);
               if (is_function)
                  {
@@ -474,41 +495,46 @@ Parser::degrade_scan_reduce(Token_string & tos)
 bool
 Parser::check_if_value(const Token_string & tos, int pos)
 {
-   // figure if token at pos is the end of a function (and then return false)
-   // or the end of a value (and then return true).
+   // figure if tos[pos] (the token left og /. ⌿. \. or ⍀ is the end of a
+   // function (and then return false) or the end of a value (and then
+   // return true).
    //
    switch(tos[pos].get_Class())
       {
-        case TC_ASSIGN:
-        case TC_R_ARROW:
-        case TC_L_BRACK:
-        case TC_END:
-        case TC_L_PARENT:
-        case TC_VALUE:
-        case TC_RETURN:
-             return true;
+        case TC_ASSIGN:     // e.g. ←/       (actually syntax error)
+        case TC_R_ARROW:    // e.g. →/       (actually syntax error)
+        case TC_L_BRACK:    // e.g. [/       (actually syntax error)
+        case TC_END:        // e.g.  /       (actually syntax error)
+        case TC_L_PARENT:   // e.g. (/       (actually syntax error)
+        case TC_VALUE:      // e.g. 5/
+        case TC_RETURN:     // e.g.  /       (actually syntax error)
+             return true;   // tos[pos] is at the end of a value
 
-        case TC_R_BRACK:
+        case TC_R_BRACK:    // e.g. +[1]/2 or 2/[1]2
              {
                const int pos1 = find_opening_bracket(tos, pos);
                if (pos1 == 0)   return true;   // this is a syntax error
                return check_if_value(tos, pos1 - 1);
              }
 
-        case TC_R_PARENT:
+        case TC_R_PARENT:   // e.g. (2+3)/2 or 1 2 3 (2+3)/2
              {
-               const int pos1 = find_opening_parent(tos, pos);
-               if (pos1 == 0)   return true;
-               return check_if_value(tos, pos1 - 1);
+               if (pos == 0)   return true;   // (actually syntax error)
+               return check_if_value(tos, pos - 1);
              }
 
-        case TC_SYMBOL:
+        case TC_SYMBOL:     // e.g. A/2 or FOO/2
+             if (tos[pos].get_tag() == TOK_ALPHA)     return true;
+             if (tos[pos].get_tag() == TOK_CHI)       return true;
+             if (tos[pos].get_tag() == TOK_OMEGA)     return true;
+             if (tos[pos].get_tag() == TOK_ALPHA_U)   return false;
+             if (tos[pos].get_tag() == TOK_OMEGA_U)   return false;
              return (tos[pos].get_tag() == TOK_P_SYMB);   // if value
 
         default: break;
       }
 
-   return false;
+   return false;   // tos[pos] is at the end of a function
 }
 //-----------------------------------------------------------------------------
 void
@@ -526,8 +552,56 @@ int dst = 0;
    tos.shrink(dst);
 }
 //-----------------------------------------------------------------------------
+ErrorCode
+Parser::match_par_bra(Token_string & tos, bool backwards)
+{
+Simple_string<ShapeItem> stack;
+   loop(s, tos.size())
+       {
+         const ShapeItem t = backwards ? (tos.size() - 1) - s : s;
+         ErrorCode ec;
+         TokenClass tc_peer;
+         switch(tos[t].get_Class())
+           {
+             case TC_L_BRACK:  if (tos[t].get_tag() != TOK_L_BRACK)   continue;
+             case TC_L_PARENT:
+             case TC_L_CURLY:  stack.append(t);
+                               continue;
+
+             case TC_R_BRACK:  ec = E_UNBALANCED_BRACKET;
+                               tc_peer = TC_L_BRACK;
+                               break;
+
+             case TC_R_PARENT: ec = E_UNBALANCED_PARENT;
+                               tc_peer = TC_L_PARENT;
+                               break;
+
+             case TC_R_CURLY:  ec = E_UNBALANCED_CURLY;
+                               tc_peer = TC_L_CURLY;
+                               break;
+
+             default:          continue;
+           }
+
+          // at this point, a closing ), ], or } was detected
+          //
+          if (stack.size() == 0)   return ec;
+
+           const ShapeItem t1 = stack.last();
+           stack.pop();
+
+          if (tos[t1].get_Class() != tc_peer)   return ec;
+
+          const ShapeItem diff = (t > t1) ? t - t1 : t1 - t;
+          tos[t].set_int_val2(diff);
+          tos[t1].set_int_val2(diff);
+       }
+
+   return E_NO_ERROR;
+}
+//-----------------------------------------------------------------------------
 void
-Parser::create_value(Token_string & tos, int pos, int count) const
+Parser::create_value(Token_string & tos, int pos, int count)
 {
    Log(LOG_create_value)
       {
@@ -548,13 +622,13 @@ Parser::create_value(Token_string & tos, int pos, int count) const
 }
 //-----------------------------------------------------------------------------
 void
-Parser::create_scalar_value(Token & output) const
+Parser::create_scalar_value(Token & output)
 {
    switch(output.get_tag())
       {
         case TOK_CHARACTER:
              {
-               Value_P scalar(new Value(LOC));
+               Value_P scalar(LOC);
 
                new (&scalar->get_ravel(0))  CharCell(output.get_char_val());
                scalar->check_value(LOC);
@@ -565,7 +639,7 @@ Parser::create_scalar_value(Token & output) const
 
         case TOK_INTEGER:
              {
-               Value_P scalar(new Value(LOC));
+               Value_P scalar(LOC);
 
                new (&scalar->get_ravel(0))   IntCell(output.get_int_val());
                scalar->check_value(LOC);
@@ -576,7 +650,7 @@ Parser::create_scalar_value(Token & output) const
 
         case TOK_REAL:
              {
-               Value_P scalar(new Value(LOC));
+               Value_P scalar(LOC);
 
                new (&scalar->get_ravel(0))  FloatCell(output.get_flt_val());
                scalar->check_value(LOC);
@@ -587,7 +661,7 @@ Parser::create_scalar_value(Token & output) const
 
         case TOK_COMPLEX:
              {
-               Value_P scalar(new Value(LOC));
+               Value_P scalar(LOC);
 
                new (&scalar->get_ravel(0))   ComplexCell(output.get_cpx_real(),
                                                          output.get_cpx_imag());
@@ -610,9 +684,9 @@ Parser::create_scalar_value(Token & output) const
 //-----------------------------------------------------------------------------
 void
 Parser::create_vector_value(Token_string & tos,
-                            int pos, int count) const
+                            int pos, int count)
 {
-Value_P vector(new Value(count, LOC));
+Value_P vector(count, LOC);
 
    loop(l, count)
        {
@@ -644,7 +718,7 @@ Value_P vector(new Value(count, LOC));
 
             case TOK_APL_VALUE1:
             case TOK_APL_VALUE3:
-                 new (addr) PointerCell(tok.get_apl_val());
+                 new (addr) PointerCell(tok.get_apl_val(), vector.getref());
                  tok.clear(LOC);   // invalidate token
                  break;
 
