@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -25,9 +25,24 @@
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
-PointerCell::PointerCell(Value_P val)
+PointerCell::PointerCell(Value_P sub_val, Value & cell_owner)
 {
-   new (&value._valp()) Value_P(val);
+   new (&value._valp()) Value_P(sub_val);
+   value2.owner = &cell_owner;
+
+   Assert(value2.owner != sub_val.get());   // typical cut-and-paste error
+   cell_owner.increment_pointer_cell_count();
+   cell_owner.add_subcount(sub_val->nz_element_count());
+}
+//-----------------------------------------------------------------------------
+void
+PointerCell::release(const char * loc)
+{
+   value2.owner->decrement_pointer_cell_count();
+   value2.owner->add_subcount(-get_pointer_value()->nz_element_count());
+
+   ptr_clear(value._valp(), loc);
+   new (this) Cell;
 }
 //-----------------------------------------------------------------------------
 bool
@@ -48,14 +63,14 @@ const ShapeItem count = A->element_count();
 }
 //-----------------------------------------------------------------------------
 bool
-PointerCell::greater(const Cell * other, bool ascending) const
+PointerCell::greater(const Cell & other) const
 {
-   switch(other->get_cell_type())
+   switch(other.get_cell_type())
       {
         case CT_CHAR:
         case CT_INT:
         case CT_FLOAT:
-        case CT_COMPLEX: return ascending;
+        case CT_COMPLEX: return true;
         case CT_POINTER: break;   // continue below
         case CT_CELLREF: DOMAIN_ERROR;
         default:         Assert(0 && "Bad celltype");
@@ -64,19 +79,19 @@ PointerCell::greater(const Cell * other, bool ascending) const
    // at this point both cells are pointer cells.
    //
 Value_P v1 = get_pointer_value();
-Value_P v2 = other->get_pointer_value();
+Value_P v2 = other.get_pointer_value();
 
    // compare ranks
    //
-   if (v1->get_rank() > v2->get_rank())   return  ascending;
-   if (v2->get_rank() > v1->get_rank())   return !ascending;
+   if (v1->get_rank() > v2->get_rank())   return true;
+   if (v2->get_rank() > v1->get_rank())   return false;
 
    // same rank, compare shapes
    //
    loop(r, v1->get_rank())
       {
-        if (v1->get_shape_item(r) > v2->get_shape_item(r))   return  ascending;
-        if (v2->get_shape_item(r) > v1->get_shape_item(r))   return !ascending;
+        if (v1->get_shape_item(r) > v2->get_shape_item(r))   return true;
+        if (v2->get_shape_item(r) > v1->get_shape_item(r))   return false;
       }
 
    // same rank and shape, compare ravel
@@ -86,20 +101,13 @@ const Cell * C2 = &v2->get_ravel(0);
    loop(e, v1->nz_element_count())
       {
         const Comp_result comp = C1++->compare(*C2++);
-        if (comp == COMP_GT)   return   ascending;
-        if (comp == COMP_LT)   return  !ascending;
+        if (comp == COMP_GT)   return   true;
+        if (comp == COMP_LT)   return  false;
       }
 
    // everthing equal
    //
-   return this > other;
-}
-//-----------------------------------------------------------------------------
-void
-PointerCell::release(const char * loc)
-{
-   ptr_clear(value._valp(), loc);
-   new (this) Cell;
+   return this > &other;
 }
 //-----------------------------------------------------------------------------
 Value_P
@@ -195,29 +203,31 @@ PrintBuffer ret(*val, pctx, 0);
              if (sh.get_volume() == 1)   // one prototype
                 {
                   ret = PrintBuffer(*proto, pctx, 0);
-                  ret.add_frame(PrintStyle(style), proto->get_rank(),
+                  ret.add_frame(PrintStyle(style), proto->get_shape(),
                                 proto->compute_depth());
                 }
              else                           // several prototypes
                 {
-                  Value_P proto_reshaped(new Value(sh, LOC));
+                  Value_P proto_reshaped(sh, LOC);
                   Cell * c = &proto_reshaped->get_ravel(0);
                   const ShapeItem len = proto_reshaped->element_count();
 
                   // store proto in the first ravel item, and copies of proto in
                   // the subsequent ravel items
                   //
-                  new (c++) PointerCell(proto);
-                  loop(rv, len - 1)   new (c++) PointerCell(proto->clone(LOC));
+                  new (c++) PointerCell(proto, proto_reshaped.getref());
+                  loop(rv, len - 1)
+                      new (c++) PointerCell(proto->clone(LOC),
+                                            proto_reshaped.getref());
 
                   ret = PrintBuffer(*proto_reshaped, pctx, 0);
-                  ret.add_frame(PrintStyle(style), proto_reshaped->get_rank(),
+                  ret.add_frame(PrintStyle(style), proto_reshaped->get_shape(),
                                 proto_reshaped->compute_depth());
                 }
            }
        else
            {
-             ret.add_frame(pctx.get_style(), val->get_rank(),
+             ret.add_frame(pctx.get_style(), val->get_shape(),
                            val->compute_depth());
            }
       }
