@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -23,7 +23,6 @@
 #include "DerivedFunction.hh"
 #include "Executable.hh"
 #include "IndexExpr.hh"
-#include "main.hh"
 #include "Prefix.hh"
 #include "StateIndicator.hh"
 #include "Symbol.hh"
@@ -44,33 +43,22 @@ Prefix::Prefix(StateIndicator & _si, const Token_string & _body)
 }
 //-----------------------------------------------------------------------------
 void
-Prefix::cleanup()
+Prefix::clean_up()
 {
-
-   //
-   // there is still an unresolved issue with double deletion of values
-   // by this function and Prefix.cc:948 below (line->erase(LOC); in
-   // function reduce_END_GOTO_B_(). Until resolved we do not erase here!
-   //
-   return;
-
-   if (size() == 0)   return;
-
-CERR << "Warning: discarding non-empty stack["
-     << si.get_level() << "]" << " with " << size() << " items " << endl;
-
-   while (size())
+   loop(s, size())
       {
-        Token_loc tl = pop();
-        if (tl.tok.get_Class() == TC_VALUE)
+        Token tok = at(s).tok;
+        if (tok.get_Class() == TC_VALUE)
            {
-             Value_P val = tl.tok.get_apl_val();
-             if (!val)   continue;
-
-CERR << "    Value is " << (const void *)val.get() << " " << *val << " at " LOC;
-// val->print_properties(CERR,  0);
+             tok.extract_apl_val(LOC);
+           }
+        else if (tok.get_ValueType() == TV_INDEX)
+           {
+             tok.get_index_val().extract_all();
            }
       }
+
+   put = 0;
 }
 //-----------------------------------------------------------------------------
 void
@@ -113,38 +101,89 @@ Prefix::syntax_error(const char * loc)
 }
 //-----------------------------------------------------------------------------
 bool
+Prefix::is_value_parent(int pc) const
+{
+   // we have ) XXX with XXX on the stack and need to know if the evaluation
+   // of (... ) will be a value as in e.g. (1 + 1) or a function as in (+/).
+   //
+   Assert1(body[pc].get_Class() == TC_R_PARENT);
+
+   ++pc;
+   if (pc >= body.size())   return true;   // syntax error
+
+TokenClass next = body[pc].get_Class();
+
+   if (next == TC_R_BRACK)   // skip [ ... ]
+      {
+        const int offset = body[pc].get_int_val2();
+        pc += offset;
+        Assert1(body[pc].get_Class() == TC_L_BRACK);   // opening [
+        if (pc >= body.size())   return true;   // syntax error
+        next = body[pc].get_Class();
+      }
+
+   if (next == TC_SYMBOL)   // resolve symbol if necessary
+      {
+        const Symbol * sym = body[pc].get_sym_ptr();
+        const NameClass nc = sym->get_nc();
+                      
+        if (nc == NC_FUNCTION)   return false;
+        if (nc == NC_OPERATOR)   return false;
+        return true;
+      }
+
+   if (next == TC_OPER1)   return false;
+   if (next == TC_OPER2)   return false;
+   if (next == TC_FUN12)   return false;
+
+   if (next == TC_L_PARENT)   // )) XXX
+      {
+        ++pc;
+        if (!is_value_parent(pc))   return false;   // (fun)) XXX
+        const int offset = body[pc].get_int_val();
+        pc += offset;
+        if (pc >= body.size())   return true;   // syntax error
+        next = body[pc].get_Class();
+        Assert1(next == TC_L_PARENT);   // opening (
+        ++pc;
+        if (pc >= body.size())   return true;   // syntax error
+
+        //   (val)) XXX
+        //  ^
+        //  pc
+        //
+        // result is a value unless (val) is the right function operand
+        // of a dyadic operator
+        //
+        next = body[pc].get_Class();
+        if (next == TC_OPER2)   return false;
+        if (next == TC_SYMBOL)   // resolve symbol if necessary
+           {
+             const Symbol * sym = body[pc].get_sym_ptr();
+             const Function * fun = sym->get_function();
+             return ! (fun && fun->is_operator() &&
+                       fun->get_oper_valence() == 2);
+           }
+        return true;
+      }
+
+   return true;
+}
+//-----------------------------------------------------------------------------
+bool
 Prefix::is_value_bracket() const
 {
    Assert1(body[PC - 1].get_Class() == TC_R_BRACK);
+const int offset = body[PC - 1].get_int_val();
+   Assert1(body[PC + offset - 1].get_Class() == TC_L_BRACK);   // opening [
 
-   // find opening [...
-   //
-int pc1 = PC;
-int br_count = 1;   // the bracket at PC
+const Token & tok1 = body[PC + offset];
+   if (tok1.get_Class() == TC_VALUE)    return true;
+   if (tok1.get_Class() != TC_SYMBOL)   return false;
 
-   for (; pc1 < body.size(); ++pc1)
-       {
-         const Token & tok = body[pc1];
-         if (tok.get_Class() == TC_R_BRACK)   ++br_count;   // another ]
-         else if (tok.get_tag() == TOK_L_BRACK)             // a [
-            {
-              --br_count;
-              if (br_count == 0)   // matching [
-                 {
-                   const Token & tok1 = body[pc1 + 1];
-                   if (tok1.get_Class() == TC_VALUE)    return true;
-                   if (tok1.get_Class() != TC_SYMBOL)   return false;
-
-                   Symbol * sym = tok1.get_sym_ptr();
-                   const bool left_sym = get_assign_state() == ASS_arrow_seen;
-                   return sym->resolve_class(left_sym) == TC_VALUE;
-                 }
-            }
-       }
-
-   // not reached since the tokenizer checks [ ] balance.
-   //
-   FIXME;
+Symbol * sym = tok1.get_sym_ptr();
+const bool left_sym = get_assign_state() == ASS_arrow_seen;
+   return sym->resolve_class(left_sym) == TC_VALUE;
 }
 //-----------------------------------------------------------------------------
 int
@@ -248,36 +287,20 @@ Prefix::value_expected()
    //
    if (saved_lookahead.tok.get_ValueType() == TV_INDEX)   return true;
 
-int brack_count = 0;
    for (int pc = PC; pc < body.size();)
       {
         const Token & tok = body[pc++];
         switch(tok.get_Class())
            {
-               case TC_R_BRACK:  // skip to opening [
+               case TC_R_BRACK:   // skip over [...] (func axis or value index)
                     //
-                    brack_count = 1;   // this [...]
-                    for (;pc < body.size(); ++pc)
-                        {
-                          if (body[pc].get_Class() == TC_R_BRACK)   // another ]
-                             {
-                               ++brack_count;
-                             }
-                          else if (body[pc].get_Class() == TC_L_BRACK)
-                             {
-                               --brack_count;
-                               if (brack_count == 0)   break;
-                               ++pc;
-                               continue;   // inner for (.,,)
-                             }
-                        }
-                    ++pc;
-                    continue;   // outer for (.,,)
+                    pc += tok.get_int_val2();
+                    continue;
 
-               case TC_END:      goto done;
+               case TC_END:     return false;   // syntax error
 
-               case TC_FUN0:   return true;   // niladic function is a value
-               case TC_FUN12:  return false;  // function
+               case TC_FUN0:    return true;   // niladic function is a value
+               case TC_FUN12:   return false;  // function
 
                case TC_SYMBOL:
                     {
@@ -289,18 +312,15 @@ int brack_count = 0;
                       return true;   // value
                     }
              
-               case TC_RETURN:   goto done;
+               case TC_RETURN:  return false;   // syntax error
                case TC_VALUE:   return true;
 
                default: continue;
            }
-
-         break;  // for (;;)
       }
 
    // this is a syntax error.
    //
-done:
    return false;
 }
 //-----------------------------------------------------------------------------
@@ -385,6 +405,12 @@ grow:
                if (left_sym)   set_assign_state(ASS_var_seen);
              }
 
+          Log(LOG_prefix_parser)
+             {
+               CERR << "   resolved symbol " << sym->get_name()
+                    << " to " << tl.tok.get_Class() << endl;
+             }
+
           if (tl.tok.get_tag() == TOK_SI_PUSHED)
             {
               // Quad_Quad::resolve() calls ⍎ which returns TOK_SI_PUSHED.
@@ -447,7 +473,8 @@ again:
 
 found_prefix:
 
-   // found prefix. see if the next token class binds stronger than best->prio
+   // found a reducible prefix. See if the next token class binds stronger
+   // than best->prio
    //
    {
      TokenClass next = TC_INVALID;
@@ -456,7 +483,7 @@ found_prefix:
           const Token & tok = body[PC];
 
           next = tok.get_Class();
-          if (tok.get_Class() == TC_SYMBOL)
+          if (next == TC_SYMBOL)
              {
                Symbol * sym = tok.get_sym_ptr();
                const bool left_sym = get_assign_state() == ASS_arrow_seen;
@@ -493,27 +520,11 @@ found_prefix:
 
 //   Q(next) Q(at0())
 
-     bool more = false;
-     switch (at0().get_Class())
-        {
-          case TC_VALUE:
-               more = ((best->prio < BS_VAL_VAL) &&
-                        (  next == TC_VALUE            // A B
-                        || next == TC_R_PARENT         // ) B
-                        )) ||
-
-                      (best->prio < BS_OP_RO && next == TC_OPER2);
-
-               break;
-
-          case TC_FUN12:
-               more = best->prio < BS_OP_RO && next == TC_OPER2;
-               break;
-
-          default: break;
-        }
-
-     if (more)
+     // we could reduce, but we could also shift. Compute more, which is true
+     // if we should shift.
+     //
+     const bool shift = dont_reduce(next);
+     if (shift)
         {
            Log(LOG_prefix_parser)  CERR
                << "   phrase #" << (best - hash_table)
@@ -582,6 +593,42 @@ found_prefix:
 }
 //-----------------------------------------------------------------------------
 bool
+Prefix::dont_reduce(TokenClass next) const
+{
+   if (at0().get_Class() == TC_VALUE)
+      {
+        if (next == TC_OPER2)           // DOP B
+           {
+             return best->prio < BS_OP_RO;
+           }
+        else if (next == TC_VALUE)      // A B
+           {
+             return best->prio < BS_VAL_VAL;
+           }
+        else if (next == TC_R_PARENT)   // ) B
+           {
+             if (is_value_parent(PC))     // e.g. (X+Y) B
+                {
+                  return best->prio < BS_VAL_VAL;
+                }
+              else                      // e.g. (+/) B
+                {
+                  return false;
+                }
+           }
+      }
+   else if (at0().get_Class() == TC_FUN12)
+      {
+        if (next == TC_OPER2)
+           {
+             return best->prio < BS_OP_RO;
+           }
+      }
+
+   return false;
+}
+//-----------------------------------------------------------------------------
+bool
 Prefix::replace_AB(Value_P old_value, Value_P new_value)
 {
    Assert(old_value);
@@ -614,7 +661,7 @@ Token * Prefix::locate_L()
    return 0;
 }
 //-----------------------------------------------------------------------------
-Token *
+Value_P *
 Prefix::locate_X()
 {
    // expect at least B X (so we have at0() and at1() and at2()
@@ -624,11 +671,20 @@ Prefix::locate_X()
    // either at0() (for monadic f X B) or at1() (for dyadic A f X B) must
    // be a function or operator
    //
-   for (int x = put - 2; x >= put - prefix_len; --x)
+   for (int x = put - 1; x >= put - prefix_len; --x)
        {
-         if (content[x].tok.get_Class() == TC_INDEX)   // found X
+         if (content[x].tok.get_ValueType() == TV_FUN)
             {
-              return &content[x].tok;
+              Function * fun = content[x].tok.get_function();
+              if (fun)
+                 {
+                   Value_P * X = fun->locate_X();
+                   if (X)   return  X;   // only for derived function
+                 }
+            }
+         else if (content[x].tok.get_Class() == TC_INDEX)   // maybe found X ?
+            {
+              return content[x].tok.get_apl_valp();
             }
        }
 
@@ -683,7 +739,7 @@ Prefix::reduce_LPAR_B_RPAR_()
 {
    Assert1(prefix_len == 3);
 
-Token result = at1();   // B
+Token result = at1();   // B or F
    if (result.get_tag() == TOK_APL_VALUE3)   result.ChangeTag(TOK_APL_VALUE1);
    pop_args_push_result(result);
    set_action(result);
@@ -696,6 +752,23 @@ Prefix::reduce_LPAR_F_RPAR_()
 }
 //-----------------------------------------------------------------------------
 void
+Prefix::reduce_LPAR_F_C_RPAR()
+{
+   Assert1(prefix_len == 4);
+
+   // C should be an axis and not a [;;] index
+   //
+   if (at2().get_ValueType() != TV_VAL)   SYNTAX_ERROR;
+   if (!at2().get_apl_val())              SYNTAX_ERROR;
+
+   move_1(at3(), at2(), LOC);
+   move_1(at2(), at1(), LOC);
+   pop();    // pop C
+   pop();    // pop RPAR
+   action = RA_CONTINUE;
+}
+//-----------------------------------------------------------------------------
+void
 Prefix::reduce_N___()
 {
    Assert1(prefix_len == 1);
@@ -710,7 +783,6 @@ Token result = at0().get_function()->eval_();
       }
 
    pop_args_push_result(result);
-   result.SET_temp();
    set_action(result);
 }
 //-----------------------------------------------------------------------------
@@ -742,7 +814,6 @@ Token result = at0().get_function()->eval_B(at1().get_apl_val());
       }
 
    pop_args_push_result(result);
-   result.SET_temp();
    set_action(result);
 }
 //-----------------------------------------------------------------------------
@@ -764,7 +835,14 @@ Prefix::reduce_MISC_F_C_B()
            }
       }
 
-   if (at1().get_ValueType() != TV_VAL)   SYNTAX_ERROR;
+   if (at1().get_ValueType() != TV_VAL)   // [i1;i2...] instead of [axis]
+      {
+        IndexExpr * idx = &at1().get_index_val();
+        Log(LOG_delete)   CERR << "delete " HEX(idx) << " at " LOC << endl;
+        delete idx;
+         at1().clear(LOC);
+         SYNTAX_ERROR;
+      }
    if (!at1().get_apl_val())              SYNTAX_ERROR;
 
 Token result = at0().get_function()->eval_XB(at1().get_apl_val(),
@@ -778,7 +856,6 @@ Token result = at0().get_function()->eval_XB(at1().get_apl_val(),
       }
 
    pop_args_push_result(result);
-   result.SET_temp();
    set_action(result);
 }
 //-----------------------------------------------------------------------------
@@ -798,7 +875,6 @@ Token result = at1().get_function()->eval_AB(at0().get_apl_val(),
       }
 
    pop_args_push_result(result);
-   result.SET_temp();
    set_action(result);
 }
 //-----------------------------------------------------------------------------
@@ -830,7 +906,6 @@ Token result = at1().get_function()->eval_AXB(at0().get_apl_val(),
       }
 
    pop_args_push_result(result);
-   result.SET_temp();
    set_action(result);
 }
 //-----------------------------------------------------------------------------
@@ -852,12 +927,12 @@ DerivedFunction * derived =
 }
 //-----------------------------------------------------------------------------
 void
-Prefix::reduce_F_C_M_()
+Prefix::reduce_F_M_C_()
 {
 DerivedFunction * derived =
    Workspace::SI_top()->fun_oper_cache.get(LOC);
-   new (derived) DerivedFunction(at0(), at2().get_function(),
-                                 at1().get_axes(), LOC);
+   new (derived) DerivedFunction(at0(), at1().get_function(),
+                                 at2().get_axes(), LOC);
 
    pop_args_push_result(Token(TOK_FUN2, derived));
    action = RA_CONTINUE;
@@ -880,7 +955,7 @@ Prefix::reduce_F_D_B_()
 {
    // same as G2, except for ⍤
    //
-   if (at1().get_function()->get_Id() != ID_OPER2_RANK)
+   if (at1().get_function()->get_Id() != ID::OPER2_RANK)
       {
          reduce_F_D_G_();
          return;
@@ -927,6 +1002,27 @@ Token result = Token(TOK_FUN2, derived);
 void
 Prefix::reduce_F_D_G_()
 {
+   // reduce, except if another dyadic operator is coming. In that case
+   // F belongs to the other operator and we simply continue.
+   //
+   if (PC < body.size())
+        {
+          const Token & tok = body[PC];
+          TokenClass next =  tok.get_Class();
+          if (next == TC_SYMBOL)
+             {
+               Symbol * sym = tok.get_sym_ptr();
+               const bool left_sym = get_assign_state() == ASS_arrow_seen;
+               next = sym->resolve_class(left_sym);
+             }
+
+          if (next == TC_OPER2)
+             {
+               action = RA_PUSH_NEXT;
+               return;
+             }
+        }
+
 DerivedFunction * derived =
    Workspace::SI_top()->fun_oper_cache.get(LOC);
    new (derived) DerivedFunction(at0(), at1().get_function(), at2(), LOC);
@@ -938,6 +1034,26 @@ DerivedFunction * derived =
 void
 Prefix::reduce_F_D_C_B()
 {
+   // reduce, except if another dyadic operator is coming. In that case
+   // F belongs to the other operator and we simply continue.
+   //
+   if (PC < body.size())
+        {
+          const Token & tok = body[PC];
+          TokenClass next =  tok.get_Class();
+          if (next == TC_SYMBOL)
+             {
+               Symbol * sym = tok.get_sym_ptr();
+               const bool left_sym = get_assign_state() == ASS_arrow_seen;
+               next = sym->resolve_class(left_sym);
+             }
+
+          if (next == TC_OPER2)
+             {
+               action = RA_PUSH_NEXT;
+               return;
+             }
+        }
    // we have f ⍤ [X] y_B with y_B glued beforehand. Unglue it.
    //
 Value_P y123;
@@ -982,12 +1098,35 @@ void
 Prefix::reduce_A_C__()
 {
 Value_P A = at0().get_apl_val();
-Value_P Z = A->index(at1());
+Value_P Z;
+
+   if (at1().get_tag() == TOK_AXES)
+      {
+        Z = A->index(at1().get_apl_val());
+      }
+   else
+      {
+        IndexExpr * idx =  &at1().get_index_val();
+        try
+           {
+             Z = A->index(*idx);
+             Log(LOG_delete)   CERR << "delete " HEX(idx) << " at " LOC << endl;
+             delete idx;
+           }
+        catch (Error err)
+           {
+             Token result = Token(TOK_ERROR, err.error_code);
+             Log(LOG_delete)   CERR << "delete " HEX(idx) << " at " LOC << endl;
+             delete idx;
+             pop_args_push_result(result);
+             set_action(result);
+             return;
+           }
+      }
 
 Token result = Token(TOK_APL_VALUE1, Z);
    pop_args_push_result(result);
 
-   at0().SET_temp();
    set_action(result);
 }
 //-----------------------------------------------------------------------------
@@ -1015,20 +1154,23 @@ Value_P B = at3().get_apl_val();
              set_action(result);
              return;
            }
-
       }
    else                               // [a;...]
       {
+        IndexExpr * idx = &at1().get_index_val();
         try
            {
-             V->assign_indexed(at1().get_index_val(), B);
+             V->assign_indexed(*idx, B);
+             Log(LOG_delete)   CERR << "delete " HEX(idx) << " at " LOC << endl;
+             delete idx;
            }
         catch (Error err)
            {
              Token result = Token(TOK_ERROR, err.error_code);
-             delete &at1().get_index_val();
              at1().clear(LOC);
              at3().clear(LOC);
+             Log(LOG_delete)   CERR << "delete " HEX(idx) << " at " LOC << endl;
+             delete idx;
              pop_args_push_result(result);
              set_assign_state(ASS_none);
              set_action(result);
@@ -1099,7 +1241,7 @@ Function * F = at2().get_function();
 Symbol * V = at0().get_sym_ptr();
    V->assign_named_lambda(F, LOC);
 
-Value_P Z(new Value(V->get_name(), LOC));
+Value_P Z(V->get_name(), LOC);
 Token result = Token(TOK_APL_VALUE2, Z);
    pop_args_push_result(result);
 
@@ -1140,7 +1282,11 @@ Prefix::reduce_RBRA___()
    // A[IDX}←B resolves IDX properly. assign_state is restored when the
    // index is complete.
    //
-   new (&at0()) Token(TOK_PINDEX, *new IndexExpr(get_assign_state(), LOC));
+IndexExpr * idx = new IndexExpr(get_assign_state(), LOC);
+   Log(LOG_delete)
+      CERR << "new    " << (const void *)idx << " at " LOC << endl;
+
+   new (&at0()) Token(TOK_PINDEX, *idx);
    set_assign_state(ASS_none);
    action = RA_CONTINUE;
 }
@@ -1161,6 +1307,8 @@ const bool last_index = (at0().get_tag() == TOK_L_BRACK);
         Token result = Token(TOK_INDEX, idx);
         pop_args_push_result(result);
         action = RA_CONTINUE;
+        Log(LOG_delete)   CERR << "delete " HEX(&idx) << " at " LOC << endl;
+        delete &idx;
         return;
       }
 
@@ -1207,6 +1355,8 @@ const bool last_index = (at0().get_tag() == TOK_L_BRACK);   // ; vs. [
              Value_P X = idx.extract_value(0);
              Assert1(!!X);
              move_2(I, Token(TOK_AXES, X), LOC);
+             Log(LOG_delete)
+                CERR << "delete " HEX(&idx) << " at " LOC << endl;
              delete &idx;
            }
         else
@@ -1232,7 +1382,6 @@ Token result;
    Value::glue(result, at0(), at1(), LOC);
    pop_args_push_result(result);
 
-   at0().SET_temp();
    set_action(result);
 }
 //-----------------------------------------------------------------------------
@@ -1353,12 +1502,11 @@ const bool trace = at0().get_Class() == TC_END &&
 Value_P line = at2().get_apl_val();
    if (trace && line->element_count() > 0)
       {
-        const int64_t line_num = line->get_line_number(Workspace::get_CT());
+        const int64_t line_num = line->get_line_number();
         Token bra(TOK_BRANCH, line_num);
         si.statement_result(bra, true);
       }
 
-
 const Token result = si.jump(line);
 
    if (result.get_tag() == TOK_BRANCH)   // branch back into a function
@@ -1379,7 +1527,26 @@ const Token result = si.jump(line);
    //
    reset(LOC);
 
-   if (result.get_tag() == TOK_VOID)   // branch not taken, e.g. →''
+   if (result.get_tag() == TOK_NOBRANCH)   // branch not taken, e.g. →⍬
+      {
+        if (trace)
+           {
+             Token bra(TOK_NOBRANCH);
+             si.statement_result(bra, true);
+           }
+
+        action = RA_PUSH_NEXT;
+        if (attention_raised && end_of_line)
+           {
+             attention_raised = false;
+             interrupt_raised = false;
+             ATTENTION;
+           }
+
+        return;
+      }
+
+   if (result.get_tag() == TOK_VOID)   // branch taken, e.g. →N
       {
         action = RA_PUSH_NEXT;
         if (attention_raised && end_of_line)
