@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -22,13 +22,22 @@
 #include <string.h>
 
 #include "Function.hh"
-#include "main.hh"
 #include "Output.hh"
+#include "Performance.hh"
 #include "PointerCell.hh"
 #include "PrintBuffer.hh"
 #include "PrintOperator.hh"
 #include "Value.icc"
 
+/// max sizes for arrays on the stack. Larger values are allocated with new()
+enum
+{
+   PB_MAX_COLS        = 200,
+   PB_MAX_ROWS        = 100,
+   PB_MAX_ITEMS       = PB_MAX_COLS * PB_MAX_ROWS,
+   PB_MAX_BREAKPOINTS = 200,
+};
+
 //-----------------------------------------------------------------------------
 PrintBuffer::PrintBuffer()
    : complete(true)
@@ -46,7 +55,9 @@ PrintBuffer::PrintBuffer(const Value & value, const PrintContext & _pctx,
                          ostream * out)
    : complete(false)
 {
-   // Note: if ostream is non-0 then this value may be inccomplete
+PERFORMANCE_START(start_0)
+
+   // Note: if ostream is non-0 then this value may be incomplete
    // (as indicated by member complete if it is huge. This is to speed
    // up printing if the value is discarded after having been printed
 
@@ -56,11 +67,10 @@ PrintContext pctx(_pctx);
    pctx.set_style(PrintStyle(outer_style &~ PST_CS_OUTER));
 
 const ShapeItem ec = value.element_count();
-const uint64_t ii_count = interrupt_count;
-const bool huge = out && ec > 10000;
 
    if (value.is_scalar())
       {
+        PERFORMANCE_START(start_1)
         PrintContext pctx1(pctx);
         if (value.get_ravel(0).need_scaling(pctx))   pctx1.set_scaled();
 
@@ -74,104 +84,108 @@ const bool huge = out && ec > 10000;
            }
 
         add_outer_frame(outer_style);
-        goto maybe_print_it;
+
+        if (out)
+           {
+             UCS_string ucs(*this, value.get_rank(), _pctx.get_PW());
+             if (ucs.size())   *out << ucs << endl;
+            }
+        complete = true;
+        PERFORMANCE_END(fs_PrintBuffer1_B, start_1, ec)
+        return;
       }
 
+   if (ec == 0)   // empty value of any dimension
+      {
+        pb_empty(value, pctx, outer_style);
+        if (out)
+           {
+             UCS_string ucs(*this, value.get_rank(), _pctx.get_PW());
+             if (ucs.size())   *out << ucs << endl;
+            }
+        complete = true;
+        return;
+      }
+
+   if (pctx.get_style() == PR_APL_FUN)
+      {
+        pb_for_function(value, pctx, outer_style);
+        if (out)
+           {
+             UCS_string ucs(*this, value.get_rank(), _pctx.get_PW());
+             if (ucs.size())   *out << ucs << endl;
+            }
+        complete = true;
+        return;
+      }
+
+   // non-trivial PrintBuffer
+   //
+   const ShapeItem cols = value.get_last_shape_item();
+   if (cols <= PB_MAX_COLS)   // few columns
+      {
+        DynArray(bool, scaling, cols);
+        DynArray(PrintBuffer, pcols, cols);
+        if (ec <= PB_MAX_ITEMS)
+           {
+             DynArray(PrintBuffer, item_matrix, ec);
+             do_PrintBuffer(value, pctx, out, outer_style,
+                            scaling, pcols, item_matrix);
+           }
+        else
+           {
+             PrintBuffer * item_matrix = new PrintBuffer[ec];
+             do_PrintBuffer(value, pctx, out, outer_style,
+                            scaling, pcols, item_matrix);
+             delete [] item_matrix;
+           }
+      }
+   else                      // many columns
+      {
+        bool * scaling = new bool[cols];
+        PrintBuffer * pcols = new PrintBuffer[cols];
+        if (ec <= PB_MAX_ITEMS)
+           {
+             DynArray(PrintBuffer, item_matrix, ec);
+             do_PrintBuffer(value, pctx, out, outer_style,
+                            scaling, pcols, item_matrix);
+           }
+        else
+           {
+             PrintBuffer * item_matrix = new PrintBuffer[ec];
+             do_PrintBuffer(value, pctx, out, outer_style,
+                            scaling, pcols, item_matrix);
+             delete [] item_matrix;
+           }
+        delete [] pcols;
+        delete [] scaling;
+      }
+
+   PERFORMANCE_END(fs_PrintBuffer_B, start_0, ec)
+}
+//-----------------------------------------------------------------------------
+void
+PrintBuffer::do_PrintBuffer(const Value & value, const PrintContext & pctx,
+                         ostream * out, PrintStyle outer_style,
+                         bool * scaling, PrintBuffer * pcols,
+                         PrintBuffer * item_matrix)
+{
+const bool framed = outer_style & (PST_CS_MASK | PST_CS_OUTER);
+const ShapeItem ec = value.element_count();
+const uint64_t ii_count = interrupt_count;
+const bool huge = out && ec > 10000;
+
    {
-     if (ec == 0)   // empty value of any dimension
-        {
-           if (value.get_rank() == 1)   // vector: 1 line
-              {
-                if (pctx.get_style() == PR_APL_FUN)
-                   {
-                     if (value.get_ravel(0).is_character_cell())   // ''
-                        {
-                          UCS_string ucs("''");
-                          ColInfo ci;
-                          *this = PrintBuffer(ucs, ci);
-                          add_outer_frame(outer_style);
-                          goto maybe_print_it;
-                        }
-
-                     if (value.get_ravel(0).is_numeric())   // ⍬
-                        {
-                          UCS_string ucs("⍬");
-                          ColInfo ci;
-                          *this = PrintBuffer(ucs, ci);
-                          add_outer_frame(outer_style);
-                          goto maybe_print_it;
-                        }
-                   }
-
-                UCS_string ucs;   // empty
-                append_ucs(ucs);
-                add_outer_frame(outer_style);
-                goto maybe_print_it;   // 1 rows
-              }
-
-           const Shape sh = value.get_shape().without_axis(value.get_rank()-1);
-           if (sh.get_volume() <= 1)   // empty vector
-              {
-                add_outer_frame(outer_style);
-                goto maybe_print_it;   // 0 rows
-              }
-
-           // value has > 0 rows. Compute how many lines we need.
-           //
-           ShapeItem lines = sh.get_volume();
-           loop(s, sh.get_rank())
-               lines += s * (sh.get_shape_item(sh.get_rank() - s - 1));
-
-           buffer.resize(lines);
-           add_outer_frame(outer_style);
-           goto maybe_print_it;
-        }
-
-     const ShapeItem cols = value.get_last_shape_item();
-     const ShapeItem rows = ec/cols;
-
-     if (pctx.get_style() == PR_APL_FUN)
-        {
-          UCS_string ucs;
-
-          if (value.is_char_vector())
-             {
-               ucs.append(UNI_SINGLE_QUOTE);
-               loop(e, ec)
-                  {
-                    const Unicode uni = value.get_ravel(e).get_char_value();
-                    ucs.append(uni);
-                    if (uni == UNI_SINGLE_QUOTE)   ucs.append(uni);   // ' -> ''
-                  }
-               ucs.append(UNI_SINGLE_QUOTE);
-             }
-          else
-             {
-               loop(e, ec)
-                  {
-                    PrintBuffer pb = value.get_ravel(e)
-                                          .character_representation(pctx);
-                    if (e)   ucs.append(UNI_ASCII_SPACE);
-                    ucs.append(UCS_string(pb, 0, pctx.get_PW()));
-                  }
-             }
-
-          ColInfo ci;
-          *this = PrintBuffer(ucs, ci);
-          add_outer_frame(outer_style);
-          goto maybe_print_it;
-        }
-
      // 1. create a vector with a bool per column that tells if the
      // column needs scaling (i.e. exponential format) or not
      //
-     vector<bool> scaling;
-     {
-       scaling.reserve(cols);
-       loop(x, cols)
-          {
-            bool need_scaling = false;
-            loop(y, rows)
+     const ShapeItem cols = value.get_last_shape_item();
+     const ShapeItem rows = ec/cols;
+
+     loop(x, cols)
+         {
+           bool need_scaling = false;
+           loop(y, rows)
                {
                  if (huge && (ii_count != interrupt_count))   goto interrupted;
                  if (value.get_ravel(x + y*cols).need_scaling(pctx))
@@ -180,144 +194,157 @@ const bool huge = out && ec > 10000;
                       break;
                     }
                }
-            scaling.push_back(need_scaling);
-          }
-     }
+           scaling[x] = need_scaling;
+         }
 
      // 2. create a matrix of items.
      //    An item of the matrix is a top-level cell (possibly nested),
      //    therefore we have (⍴,value) items. Items are rectangular.
      //
-     typedef vector<PrintBuffer> PB_row;
-     vector<PB_row> item_matrix;
-
-     item_matrix.reserve(rows);
+     PERFORMANCE_START(start_2)
      loop(y, rows)
         {
-          vector<PrintBuffer> item_row;
-          item_row.reserve(cols);
-          ShapeItem row_height = 0;
+          ShapeItem max_row_height = 0;
 
           loop(x, cols)
               {
                 if (huge && (ii_count != interrupt_count))   goto interrupted;
 
+                PrintBuffer & item = item_matrix[y*cols + x];
                 PrintContext pctx1 = pctx;
                 if (scaling[x])   pctx1.set_scaled();
                 const Cell & cell = value.get_ravel(x + y*cols);
-                PrintBuffer item = cell.character_representation(pctx1);
+                item = cell.character_representation(pctx1);
 
-                if (row_height < item.get_height())
-                   row_height = item.get_height();
+                if (max_row_height < item.get_height())
+                   max_row_height = item.get_height();
 
-                Assert(item.is_rectangular());
-                item_row.push_back(item);
+                Assert1(item.is_rectangular());
               }
 
           // pad all items to the same height
+          //
           loop(x, cols)
              {
+                PrintBuffer & item = item_matrix[y*cols + x];
                 if (huge && (ii_count != interrupt_count))   goto interrupted;
 
-               item_row[x].pad_height(UNI_iPAD_U6, row_height);
-               Assert(item_row[x].is_rectangular());
+               item.pad_height(UNI_iPAD_U6, max_row_height);
+               Assert1(item.is_rectangular());
              }
-          item_matrix.push_back(item_row);
         }
-
+     PERFORMANCE_END(fs_PrintBuffer2_B, start_2, ec)
 
      // 3. align all columns (which pads them to the same width).
      //
+     PERFORMANCE_START(start_3)
      loop(x, cols)
         {
           ColInfo col_info_x;
           loop(y, rows)
               {
                 if (huge && (ii_count != interrupt_count))   goto interrupted;
-                col_info_x.consider(item_matrix[y][x].get_info());
+                PrintBuffer * item_row = item_matrix + y*cols;
+                col_info_x.consider(item_row[x].get_info());
               }
 
           loop(y, rows)
               {
                 if (huge && (ii_count != interrupt_count))   goto interrupted;
-                item_matrix[y][x].align(col_info_x);
+                PrintBuffer * item_row = item_matrix + y*cols;
+                item_row[x].align(col_info_x);
               }
         }
+     PERFORMANCE_END(fs_PrintBuffer3_B, start_3, ec)
 
      // 4. combine colums, then rows.
      //
-     int last_spacing = 0;
-     bool last_notchar = false;
+     PERFORMANCE_START(start_4)
 
-     vector<PrintBuffer> pcols;
-     pcols.reserve(cols);
+     int last_col_spacing = 0;    // the col_spacing of the previous column
+     bool last_notchar = false;   // the notchar property of the previous column
 
      loop(x, cols)
         {
-          PrintBuffer pcol;
+          // merge the different items in column x of item_matrix into one
+          // PrintBuffer pcol. Insert separator rows as needed.
+          //
+          PrintBuffer & dest = pcols[x];
+          Assert(&dest);
+          new (&dest) PrintBuffer;
+
+          // compute the final height of dest and reserve enough rows as to
+          // avoid unnecessary copies
+          //
+          {
+            ShapeItem dest_height = 0;
+            loop(y, rows)
+               {
+                 dest_height += separator_rows(y, value.get_shape());
+                 dest_height += item_matrix[y*cols + x].get_height();
+               }
+            dest.buffer.reserve(dest_height);
+          }
 
           loop(y, rows)
-           {
-             if (huge && (ii_count != interrupt_count))   goto interrupted;
+              {
+                if (huge && (ii_count != interrupt_count))   goto interrupted;
 
-             // separator row(s)
-             //
-             if (y)
-                {
-                  const UCS_string sepa_row(pcol.get_width(0), UNI_iPAD_L0);
-                  ShapeItem prod = 1;
-                  loop(r, value.get_rank() - 2)
-                     {
-                       prod *= value.get_shape_item(value.get_rank() - r - 2);
-                       if (y % prod == 0)   pcol.append_ucs(sepa_row);
-                       else                 break;
-                     }
-                }
+                // insert separator row(s)
+                //
+                const ShapeItem srows = separator_rows(y, value.get_shape());
+                if (srows)   // unless first row
+                   {
+                    const UCS_string sepa_row(dest.get_width(0), UNI_iPAD_L0);
+                    loop(r, srows)   dest.append_ucs(sepa_row);
+                   }
 
-               const PB_row & row = item_matrix[y];
-               const PrintBuffer & item = row[x];
-               if (y)   pcol.add_row(item);
-               else     pcol = item;
-           }
+                const PrintBuffer & src = item_matrix[y*cols + x];
+                dest.add_row(src);
+              }
 
           bool not_char = false;   // determined by get_col_spacing()
 	  const int col_spacing = value.get_col_spacing(not_char, x, framed);
 
-          const int32_t max_spacing = (col_spacing > last_spacing) 
-                                    ?  col_spacing : last_spacing;
-          int32_t no_char_spacing = 0;
+          const int32_t max_spacing = (col_spacing > last_col_spacing) 
+                                    ?  col_spacing : last_col_spacing;
+          int32_t not_char_spaces = 0;
 
           if (huge && (ii_count != interrupt_count))   goto interrupted;
 
-          if (x)   // not first column
+          if (x)   // subsequent column
              {
-                if (last_notchar)
-                   {
-                     pcols.back().pad_r(UNI_iPAD_U2, 1);
-                     ++no_char_spacing;
-                   }
-                else if (not_char)
-                   {
-                     pcol.pad_l(UNI_iPAD_U3, 1);
-                     ++no_char_spacing;
-                   }
+               if (last_notchar)
+                  {
+                    // the previous column was 'notchar' so we add one pad
+                    // char to it.
+                    //
+                    pcols[x - 1].pad_r(UNI_iPAD_U2, 1);
+                    ++not_char_spaces;
+                  }
+               else if (not_char)
+                  {
+                    // the current column is 'notchar' so we add one pad
+                    // char to it.
+                    //
+                    dest.pad_l(UNI_iPAD_U3, 1);
+                    ++not_char_spaces;
+                  }
 
-                // we want a total spacing of 'max_spacing'
-                // but we deduct the 'no_char_spacing' chars ² and ³
-                // that were appended above.
-                //
-                const int u7_pad_len = max_spacing - no_char_spacing;
-                if (u7_pad_len)
-                   {
-                     pcols.back().pad_r(UNI_iPAD_U7, u7_pad_len);
-                   }
+               // we want a total spacing of 'max_spacing'
+               // but we deduct the 'not_char_spaces' chars ² and ³
+               // that were appended above.
+               //
+               const int u7_pad_len = max_spacing - not_char_spaces;
+               if (u7_pad_len)
+                  {
+                     pcols[x - 1].pad_r(UNI_iPAD_U7, u7_pad_len);
+                  }
              }
 
-          pcols.push_back(pcol);
-
           if (huge && (ii_count != interrupt_count))   goto interrupted;
 
-          last_spacing = col_spacing;
+          last_col_spacing = col_spacing;
           last_notchar = not_char;
         }
 
@@ -336,36 +363,42 @@ const bool huge = out && ec > 10000;
           pad_r(UNI_iPAD_U9, 1);
         }
 
-     if (!is_rectangular())
+     if (!is_rectangular())   // should not happen
         {
           Q1(get_height())
           loop(h, get_height())   CERR << "w=" << get_width(h) << "*" << endl;
-          loop(h, get_height())   CERR << "*" << get_line(h) << "*" << endl;
+          loop(h, get_height())   CERR << "*"  << get_line(h) << "*" << endl;
         }
 
      Assert(is_rectangular());
      add_outer_frame(outer_style);
+
+     PERFORMANCE_END(fs_PrintBuffer4_B, start_4, ec)
+   }
+
+   {
+     PERFORMANCE_START(start_5)
+
+     if (huge)   // ergo: out
+        {
+          print_interruptible(*out, value.get_rank(), pctx.get_PW());
+        }
+     else if (out)
+        {
+          UCS_string ucs(*this, value.get_rank(), pctx.get_PW());
+          if (ucs.size())   *out << ucs << endl;
+        }
+     complete = true;
+
+     PERFORMANCE_END(fs_PrintBuffer5_B, start_5, ec)
    }
 
-maybe_print_it:
-   if (huge)   // ergo: out
-      {
-        print_interruptible(*out, value.get_rank(), _pctx.get_PW());
-      }
-   else if (out)
-      {
-        UCS_string ucs(*this, value.get_rank(), _pctx.get_PW());
-        if (ucs.size() == 0)   return;
-        *out << ucs << endl;
-      }
-   complete = true;
    return;
 
 interrupted:
    attention_raised = false;
    interrupt_raised = false;
    *out << endl << "INTERRUPT" << endl;
-   return;
 }
 //-----------------------------------------------------------------------------
 void
@@ -374,10 +407,14 @@ PrintBuffer::print_interruptible(ostream & out, Rank rank, int quad_PW)
    if (get_height() == 0)   return;      // empty PrintBuffer
 
 const int total_width = get_width(0);
+const int max_breaks = 2*total_width/quad_PW;
 
-   // breakpoints can get rather large, so we reserve enough space
-vector<int> breakpoints;
-   breakpoints.reserve(2*total_width/quad_PW);
+ShapeItem * del = 0;
+ShapeItem __breakpoints[PB_MAX_BREAKPOINTS];
+ShapeItem * breakpoints = __breakpoints;
+   if (max_breaks >= PB_MAX_BREAKPOINTS)
+      breakpoints = del = new ShapeItem[max_breaks];
+ShapeItem bp_len = 0;
 
    // print rows, breaking at breakpoints
    //
@@ -393,9 +430,9 @@ vector<int> breakpoints;
               if (row == 0)   // first row: set up breakpoints
                  {
                    chunk_len = get_line(0).compute_chunk_length(quad_PW, col);
-                   breakpoints.push_back(chunk_len);
+                   breakpoints[bp_len++] = chunk_len;
                  }
-              else
+              else            // subsequent row: re-use breakpoints
                  {
                    chunk_len = breakpoints[b++];
                  }
@@ -417,25 +454,116 @@ vector<int> breakpoints;
                    out << endl << "INTERRUPT" << endl;
                    attention_raised = false;
                    interrupt_raised = false;
+                   delete del;
                    return;
                  }
             }
        }
 
    out << endl;
+   delete del;
 }
 //-----------------------------------------------------------------------------
-UCS_string
-PrintBuffer::remove_last_line()
+void
+PrintBuffer::pb_empty(const Value & value, PrintContext pctx, 
+                             PrintStyle outer_style)
 {
-   Assert(get_height() > 0);
+   if (value.get_rank() == 1)   // vector: 1 line
+      {
+        if (pctx.get_style() == PR_APL_FUN)
+           {
+             if (value.get_ravel(0).is_character_cell())   // ''
+                {
+                  UCS_string ucs("''");
+                  ColInfo ci;
+                  *this = PrintBuffer(ucs, ci);
+                  add_outer_frame(outer_style);
+                  return;
+                }
 
-UCS_string ucs = buffer[get_height() - 1];
-   buffer.resize(get_height() - 1);
+             if (value.get_ravel(0).is_numeric())   // ⍬
+                {
+                  UCS_string ucs("⍬");
+                  ColInfo ci;
+                  *this = PrintBuffer(ucs, ci);
+                  add_outer_frame(outer_style);
+                  return;
+                }
+           }
 
-   return ucs;
+        UCS_string ucs;   // empty
+        append_ucs(ucs);
+        add_outer_frame(outer_style);
+        return;   // 1 row
+      }
+
+const Shape sh = value.get_shape().without_axis(value.get_rank()-1);
+   if (sh.get_volume() <= 1)   // empty vector
+      {
+        add_outer_frame(outer_style);
+        return;   // 0 rows
+      }
+
+   // value has > 0 rows. Compute how many lines we need.
+   //
+ShapeItem lines = sh.get_volume();
+   loop(s, sh.get_rank())
+      lines += s * (sh.get_shape_item(sh.get_rank() - s - 1));
+
+   buffer.resize(lines);
+   add_outer_frame(outer_style);
 }
+//-----------------------------------------------------------------------------
+void
+PrintBuffer::pb_for_function(const Value & value, PrintContext pctx, 
+                             PrintStyle outer_style)
+{
+const ShapeItem ec = value.element_count();
+UCS_string ucs;
 
+   if (value.is_char_vector())
+      {
+        ucs.append(UNI_SINGLE_QUOTE);
+        loop(e, ec)
+           {
+             const Unicode uni = value.get_ravel(e).get_char_value();
+             ucs.append(uni);
+             if (uni == UNI_SINGLE_QUOTE)   ucs.append(uni);   // ' -> ''
+           }
+        ucs.append(UNI_SINGLE_QUOTE);
+      }
+   else
+      {
+        loop(e, ec)
+           {
+             PrintBuffer pb = value.get_ravel(e)
+                                          .character_representation(pctx);
+             if (e)   ucs.append(UNI_ASCII_SPACE);
+             ucs.append(UCS_string(pb, 0, pctx.get_PW()));
+           }
+      }
+
+ColInfo ci;
+   *this = PrintBuffer(ucs, ci);
+   add_outer_frame(outer_style);
+}
+//-----------------------------------------------------------------------------
+ShapeItem
+PrintBuffer::separator_rows(ShapeItem y, const Shape & shape)
+{
+   if (y == 0)   return 0;
+
+ShapeItem prod = 1;
+ShapeItem ret = 0;
+   loop(r, shape.get_rank() - 2)
+       {
+         prod *= shape.get_shape_item(shape.get_rank() - r - 2);
+         if (y % prod == 0)   ++ret;
+         else                 break;
+       }
+
+   return ret;
+}
 //-----------------------------------------------------------------------------
 Unicode
 PrintBuffer::get_char(uint32_t x, uint32_t y) const
@@ -532,7 +660,7 @@ PrintBuffer::get_frame_chars(PrintStyle pst,
 }
 //-----------------------------------------------------------------------------
 void
-PrintBuffer::add_frame(PrintStyle style, uint32_t rank, uint32_t depth)
+PrintBuffer::add_frame(PrintStyle style, const Shape & shape, uint32_t depth)
 {
    Assert(is_rectangular());
 
@@ -555,15 +683,21 @@ Unicode HORI, VERT, NW, NE, SE, SW;
         return;
       }
 
-   // draw a bar left and right
+   // draw | left and right
    //
    loop(y, get_height())
       {
         buffer[y].insert(0, 1, VERT);
         buffer[y].append(VERT);
+
+        // change internal pad characters to ASCII_SPACE so that they will
+        // not be removed later and the frame is printed correctly
+        //
+        buffer[y].map_pad();
+
       }
 
-   // draw a bar on top and bottom.
+   // draw - on top and bottom.
    //
 UCS_string hori(get_width(0), HORI);
 
@@ -579,8 +713,38 @@ UCS_string hori(get_width(0), HORI);
 
    // draw other decorators
    //
-   if (rank > 0)   set_char(1, 0, UNI_RIGHT_ARROW);
-   if (rank > 1)   set_char(0, 1, UNI_DOWN_ARROW);
+   if (shape.get_rank() > 0)
+      {
+        set_char(1, 0, UNI_RIGHT_ARROW);
+        if (style & PST_NARS)
+           {
+             UCS_string ucs;
+             ucs.append_number(shape.get_last_shape_item());
+             if (ucs.size() < (get_width(0) - 2))
+                {
+                  loop(u, ucs.size())   set_char(u + 1, 0, ucs[u]);
+                }
+           }
+      }
+
+   if (shape.get_rank() > 1)
+      {
+        set_char(0, 1, UNI_DOWN_ARROW);
+        if (style & PST_NARS)
+           {
+             UCS_string ucs;
+             loop(r, shape.get_rank() - 1)
+                {
+                  if (r)   ucs.append(VERT);
+                  ucs.append_number(shape.get_shape_item(r));
+                }
+             if (ucs.size() < (get_height() - 2))
+                {
+                  loop(u, ucs.size())   set_char(0, u + 1, ucs[u]);
+                }
+           }
+      }
+
    if (depth)
       {
         loop(d, depth - 1)
@@ -632,6 +796,11 @@ Unicode HORI, VERT, NW, NE, SE, SW;
       {
         buffer[y].insert(0, 1, VERT);
         buffer[y].append(VERT);
+
+        // change internal pad characters to ASCII_SPACE so that they will
+        // not be removed later and the frame is printed correctly
+        //
+        buffer[y].map_pad();
       }
 
    // draw a bar on top and bottom.
@@ -1008,7 +1177,8 @@ UCS_string new_buf(buffer[0], 0, col_info.int_len);
       // then we fill with spaces. Otherwise fill with '0', possibly inserting
       // a decimal point.
       //
-      loop(f, col_info.fract_len)   new_buf.append(buffer[0][col_info.int_len + f]);
+      loop(f, col_info.fract_len)
+          new_buf.append(buffer[0][col_info.int_len + f]);
       if (!want_expo)
          {
            loop(d, diff)   new_buf.append(UNI_iPAD_L4);
