@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -28,9 +28,9 @@ uint64_t Quad_RL::state = 0;
 
 //=============================================================================
 Quad_RL::Quad_RL()
-   : SystemVariable(ID_Quad_RL)
+   : SystemVariable(ID::Quad_RL)
 {
-Value_P value(new Value(LOC));
+Value_P value(LOC);
 
 const unsigned int seed = reset_seed();
 
@@ -52,7 +52,7 @@ Quad_RL::assign(Value_P value, const char * loc)
 const Cell & cell = value->get_ravel(0);
 const APL_Integer val = cell.get_int_value();
 
-   srandom(val);
+   state = val;
    Symbol::assign(value, LOC);
 }
 //-----------------------------------------------------------------------------
