@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -43,14 +43,34 @@ extern char **environ;
 #define PATH_MAX 4096
 #endif
 
-Quad_SVC Quad_SVC::fun;
-Quad_SVO Quad_SVO::fun;
-Quad_SVQ Quad_SVQ::fun;
-Quad_SVR Quad_SVR::fun;
-Quad_SVS Quad_SVS::fun;
+// shared variable function instances
+//
+Quad_SVC Quad_SVC::_fun;
+Quad_SVO Quad_SVO::_fun;
+Quad_SVQ Quad_SVQ::_fun;
+Quad_SVR Quad_SVR::_fun;
+Quad_SVS Quad_SVS::_fun;
+
+// shared variable function pointers
+//
+Quad_SVC * Quad_SVC::fun = & Quad_SVC::_fun;
+Quad_SVO * Quad_SVO::fun = & Quad_SVO::_fun;
+Quad_SVQ * Quad_SVQ::fun = & Quad_SVQ::_fun;
+Quad_SVR * Quad_SVR::fun = & Quad_SVR::_fun;
+Quad_SVS * Quad_SVS::fun = & Quad_SVS::_fun;
 
 APL_time_us Quad_SVE::timer_end = 0;
 
+//=============================================================================
+// there is a different prog_name() function around (in APserver) so we
+// declare the one for the APL interpreter here
+const char *
+prog_name()
+{
+   return "apl";
+}
+
+//=============================================================================
 TCP_socket get_TCP_for_key(SV_key key)
 {
    return Svar_DB::get_DB_tcp();
@@ -149,11 +169,12 @@ const ShapeItem var_count = B->get_rows();
 vector<UCS_string> vars(var_count);
    B->to_varnames(vars, false);
 
-const APL_Float qct = Workspace::get_CT();
 const Cell * cA = &A->get_ravel(0);
 
-const Shape shZ(var_count, 4);
-Value_P Z(var_count > 1 ? new Value(shZ, LOC) : new Value(4, LOC));
+Shape sh_Z;
+   if (var_count > 1)   sh_Z.add_shape_item(var_count);
+   sh_Z.add_shape_item(4);
+Value_P Z(sh_Z, LOC);
 
    loop(z, var_count)
       {
@@ -162,22 +183,22 @@ Value_P Z(var_count > 1 ? new Value(shZ, LOC) : new Value(4, LOC));
         int ctl = 0;
         if (A->is_scalar())
            {
-             const bool val = A->get_ravel(0).get_near_bool(qct);
+             const bool val = A->get_ravel(0).get_near_bool();
               ctl = val ? ALL_SVAR_CONTROLS : NO_SVAR_CONTROL;
            }
         else if (A->is_vector())
            {
-             if (A->get_ravel(0).get_near_bool(qct))   ctl |= SET_BY_1;
-             if (A->get_ravel(1).get_near_bool(qct))   ctl |= SET_BY_2;
-             if (A->get_ravel(2).get_near_bool(qct))   ctl |= USE_BY_1;
-             if (A->get_ravel(3).get_near_bool(qct))   ctl |= USE_BY_2;
+             if (A->get_ravel(0).get_near_bool())   ctl |= SET_BY_1;
+             if (A->get_ravel(1).get_near_bool())   ctl |= SET_BY_2;
+             if (A->get_ravel(2).get_near_bool())   ctl |= USE_BY_1;
+             if (A->get_ravel(3).get_near_bool())   ctl |= USE_BY_2;
            }
         else // matrix
            {
-             if (cA++->get_near_bool(qct))   ctl |= SET_BY_1;
-             if (cA++->get_near_bool(qct))   ctl |= SET_BY_2;
-             if (cA++->get_near_bool(qct))   ctl |= USE_BY_1;
-             if (cA++->get_near_bool(qct))   ctl |= USE_BY_2;
+             if (cA++->get_near_bool())   ctl |= SET_BY_1;
+             if (cA++->get_near_bool())   ctl |= SET_BY_2;
+             if (cA++->get_near_bool())   ctl |= USE_BY_1;
+             if (cA++->get_near_bool())   ctl |= USE_BY_2;
            }
 
         // set control.
@@ -208,8 +229,10 @@ const ShapeItem var_count = B->get_rows();
 vector<UCS_string> vars(var_count);
    B->to_varnames(vars, false);
 
-const Shape shZ(var_count, 4);
-Value_P Z(var_count > 1 ? new Value(shZ, LOC) : new Value(4, LOC));
+Shape sh_Z;
+   if (var_count > 1)   sh_Z.add_shape_item(var_count);
+   sh_Z.add_shape_item(4);
+Value_P Z(sh_Z, LOC);
 
    loop(z, var_count)
       {
@@ -230,7 +253,7 @@ Value_P Z(var_count > 1 ? new Value(shZ, LOC) : new Value(4, LOC));
 }
 //=============================================================================
 Quad_SVE::Quad_SVE()
-   : NL_SystemVariable(ID_Quad_SVE)
+   : NL_SystemVariable(ID::Quad_SVE)
 {
    Symbol::assign(IntScalar(0, LOC), LOC);
 }
@@ -289,7 +312,7 @@ const APL_Float remaining = timer_end - now();
 
    if (remaining < 0)   return IntScalar(0, LOC);
 
-Value_P Z(new Value(LOC));
+Value_P Z(LOC);
    new (&Z->get_ravel(0))   FloatCell(0.000001 * remaining);
    return Z;
 }
@@ -312,7 +335,9 @@ vector<UCS_string> surrogates(var_count);
 
    if (A->get_rank() == 1 && A->element_count() != var_count)   LENGTH_ERROR;
 
-Value_P Z(var_count > 1 ? new Value(var_count, LOC) : new Value(LOC));
+Shape sh_Z;
+   if (var_count > 1)   sh_Z.add_shape_item(var_count);
+Value_P Z(sh_Z, LOC);
 
    loop(z, var_count)
       {
@@ -445,7 +470,9 @@ const ShapeItem var_count = B->get_rows();
 vector<UCS_string> vars(var_count);
    B->to_varnames(vars, false);
 
-Value_P Z(var_count > 1 ? new Value(var_count, LOC) : new Value(LOC));
+Shape sh_Z;
+   if (var_count > 1)   sh_Z.add_shape_item(var_count);
+Value_P Z(sh_Z, LOC);
 
    loop(z, var_count)
       {
@@ -600,7 +627,7 @@ vector<int32_t> sorted;
             }
       }
 
-Value_P Z(new Value(sorted.size(), LOC));
+Value_P Z(sorted.size(), LOC);
    loop(z, sorted.size())   new (&Z->get_ravel(z)) IntCell(sorted[z]);
    return Z;
 }
@@ -631,7 +658,7 @@ ShapeItem max_len = 0;
        if (max_len < (var_lengths[z] - 1))   max_len = var_lengths[z] - 1;
 
 const Shape shZ(count, max_len);
-Value_P Z(new Value(shZ, LOC));
+Value_P Z(shZ, LOC);
 int v = 0;
 
    loop(z, count)
@@ -661,7 +688,9 @@ const ShapeItem var_count = B->get_rows();
 vector<UCS_string> vars(var_count);
    B->to_varnames(vars, false);
 
-Value_P Z(var_count > 1 ? new Value(var_count, LOC) : new Value(LOC));
+Shape sh_Z;
+   if (var_count > 1)   sh_Z.add_shape_item(var_count);
+Value_P Z(sh_Z, LOC);
 
    loop(z, var_count)
       {
@@ -685,8 +714,10 @@ const ShapeItem var_count = B->get_rows();
 vector<UCS_string> vars(var_count);
    B->to_varnames(vars, false);
 
-const Shape shZ(var_count, 4);
-Value_P Z(var_count > 1 ? new Value(shZ, LOC) : new Value(4, LOC));
+Shape sh_Z;
+   if (var_count > 1)   sh_Z.add_shape_item(var_count);
+   sh_Z.add_shape_item(4);
+Value_P Z(sh_Z, LOC);
 
    loop(z, var_count)
       {
