@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -30,7 +30,8 @@
 #include "Tokenizer.hh"
 #include "Workspace.hh"
 
-Quad_TF  Quad_TF::fun;
+Quad_TF   Quad_TF::_fun;
+Quad_TF * Quad_TF::fun = &Quad_TF::_fun;
 
 //-----------------------------------------------------------------------------
 Token
@@ -56,7 +57,7 @@ Value_P Z;
       {
         const bool inverse = is_inverse(symbol_name);
         if (!inverse)   return tf2(symbol_name);
-        Z = Value_P(new Value(tf2_inv(symbol_name), LOC));
+        Z = Value_P(tf2_inv(symbol_name), LOC);
       }
    else if (mode == 3)
       {
@@ -161,7 +162,7 @@ const ShapeItem ec = val->element_count();
                 {
                   ucs.append_number(cell.get_int_value());
                 }
-             else // if (cell.is_near_real(qct))
+             else // if (cell.is_near_real())
                 {
                   PrintContext pctx(PR_APL_MIN, MAX_Quad_PP, MAX_Quad_PW);
                   const APL_Float value = cell.get_real_value();
@@ -172,7 +173,7 @@ const ShapeItem ec = val->element_count();
            }
       }
 
-   return Value_P(new Value(ucs, LOC));
+   return Value_P(ucs, LOC);
 }
 //-----------------------------------------------------------------------------
 Value_P
@@ -199,7 +200,7 @@ UCS_string ucs;
        loop(c, max_len - lines[l].size())   ucs.append(UNI_ASCII_SPACE);
       }
 
-   return Value_P(new Value(ucs, LOC));
+   return Value_P(ucs, LOC);
 }
 //-----------------------------------------------------------------------------
 Value_P
@@ -264,19 +265,19 @@ const size_t data_chars = len - idx;
 
         if (data_chars != shape.get_volume())   return Value_P();
 
-        Value_P new_val(new Value(shape, LOC));
+        Value_P new_val(shape, LOC);
         loop(d, data_chars)
            new (&new_val->get_ravel(d)) CharCell(ravel[idx + d]);
         new_val->check_value(LOC);
 
-         Token t = Quad_FX::fun.eval_B(new_val);
+         Token t = Quad_FX::fun->eval_B(new_val);
       }
    else if (mode == UNI_ASCII_C)   // char array
       {
         if (data_chars != shape.get_volume())   return Value_P();
         if (nc != NC_UNUSED_USER_NAME && nc != NC_VARIABLE)   return Value_P();
 
-        Value_P new_val(new Value(shape, LOC));
+        Value_P new_val(shape, LOC);
         loop(d, data_chars)
            new (&new_val->get_ravel(d)) CharCell(ravel[idx + d]);
 
@@ -310,7 +311,7 @@ const size_t data_chars = len - idx;
 
         // at this point, we have a valid inverse 1 ⎕TF.
         //
-        Value_P new_val(new Value(shape, LOC));
+        Value_P new_val(shape, LOC);
 
         loop(t, tos.size())
            {
@@ -333,7 +334,7 @@ const size_t data_chars = len - idx;
       }
    else Assert(0);   // since checked above
    
-   return Value_P(new Value(name, LOC));
+   return Value_P(name, LOC);
 }
 //-----------------------------------------------------------------------------
 Token
@@ -349,7 +350,7 @@ const Function * function = obj->get_function();
         if (obj->is_user_defined())
            {
              UCS_string ucs = tf2_fun(name, *function);
-             Value_P Z(new Value(ucs, LOC));
+             Value_P Z(ucs, LOC);
              Z->check_value(LOC);
              return Token(TOK_APL_VALUE1, Z);
            }
@@ -382,7 +383,7 @@ const bool error = tf2_value(0, ucs, value);
    //
    if (error)   return Token(TOK_APL_VALUE1, Str0(LOC));
 
-Value_P Z(new Value(ucs, LOC));
+Value_P Z(ucs, LOC);
    Z->check_value(LOC);
    return Token(TOK_APL_VALUE1, Z);
 }
@@ -448,7 +449,7 @@ Token_string tos;
         Value_P fname   =  tos[0].get_apl_val();
         Value_P so_path =  tos[2].get_apl_val();
 
-        const Token tok = Quad_FX::fun.eval_AB(so_path, fname);
+        const Token tok = Quad_FX::fun->eval_AB(so_path, fname);
         if (tok.get_Class() == TC_VALUE)   // ⎕FX successful
            {
              Value_P val = tok.get_apl_val();
@@ -464,7 +465,7 @@ Token_string tos;
 
    {
      Value_P function_text =  tos[1].get_apl_val();
-     const Token tok = Quad_FX::fun.eval_B(function_text);
+     const Token tok = Quad_FX::fun->eval_B(function_text);
 
    if (tok.get_Class() == TC_VALUE)   // ⎕FX successful
       {
@@ -717,7 +718,7 @@ int d = 0;
                 }
              else
                 {
-                  move_2(tos[d], Bif_F12_RHO::do_reshape(sh, bval), LOC);
+                  move_2(tos[d], Bif_F12_RHO::do_reshape(sh, *bval), LOC);
                   progress = true;
                   ++d;
                   s += 2;
@@ -924,7 +925,7 @@ Symbol * symbol = obj->get_symbol();
              CDR_string cdr;
              CDR::to_CDR(cdr, value);
 
-             return Value_P(new Value(cdr, LOC));
+             return Value_P(cdr, LOC);
            }
 
         /* not a variable: fall through */
