@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -23,7 +23,6 @@
 #include "Command.hh"
 #include "Executable.hh"
 #include "IndexExpr.hh"
-#include "Input.hh"
 #include "Output.hh"
 #include "Prefix.hh"
 #include "StateIndicator.hh"
@@ -35,8 +34,7 @@
 StateIndicator::StateIndicator(Executable * exec, StateIndicator * _par)
    : executable(exec),
      safe_execution(false),
-     eoc_handler(0),
-     eoc_arg(Value_P()),
+     eoc_handlers(0),
      level(_par ? 1 + _par->get_level() : 0),
      error(E_NO_ERROR, LOC),
      current_stack(*this, exec->get_body()),
@@ -49,7 +47,7 @@ StateIndicator::~StateIndicator()
    // flush the FIFO. Do that before delete executable so that values
    // copied directly from the body of the executable are not killed.
    //
-   current_stack.cleanup();
+   current_stack.clean_up();
 
    // if executable is a user defined function then pop its local vars.
    // otherwise delete the body token
@@ -65,6 +63,20 @@ StateIndicator::~StateIndicator()
          delete executable;
          executable = 0;
       }
+
+   // clean up EOC handlers
+   //
+   if (eoc_handlers)
+      {
+        Log(LOG_EOC_handlers)
+           CERR << "StateIndicator::~StateIndicator() cleans up eoc_handlers"
+                << endl << "    deleting EOC handler "
+                << (const void *)eoc_handlers << endl;
+
+        eoc_handlers->loc = "(deleted)";
+        delete eoc_handlers;
+        eoc_handlers = 0;
+      }
 }
 //-----------------------------------------------------------------------------
 void
@@ -344,21 +356,22 @@ StateIndicator::jump(Value_P value)
    // perform a jump. We either remain in the current function (and then
    // return TOK_VOID), or we (want to) jump into back into the calling
    // function (and then return TOK_BRANCH.). The jump itself (if any)
-   // is executed in Command.cc.
+   // is executed in Prefix.cc.
    //
    if (value->get_rank() > 1)   DOMAIN_ERROR;
 
    if (value->element_count() == 0)     // →''
       {
-        // →'' in immediate execution means resume (retry) suspended function
+        // →⍬ in immediate execution means resume (retry) suspended function
+        // →⍬ on ⍎ or defined functions means do nothing
         //
         if (get_executable()->get_parse_mode() == PM_STATEMENT_LIST)
            return Token(TOK_BRANCH, int64_t(Function_Retry));
 
-        return Token(TOK_VOID);           // stay in context
+        return Token(TOK_NOBRANCH);           // stay in context
       }
 
-const Function_Line line = value->get_line_number(Workspace::get_CT());
+const Function_Line line = value->get_line_number();
 
 const UserFunction * ufun = get_executable()->get_ufun();
 
@@ -376,10 +389,21 @@ const UserFunction * ufun = get_executable()->get_ufun();
 void
 StateIndicator::escape()  
 {
-   if (eoc_handler)
+EOC_arg * e;
+   while ((e = eoc_handlers))
       {
+        Log(LOG_EOC_handlers)
+           CERR << "StateIndicator::escape() cleans up eoc_handler from "
+                << e->loc << endl;
+
         Token tok(TOK_ESCAPE);
-        eoc_handler(tok, eoc_arg);
+        e->handler(tok);
+        Log(LOG_EOC_handlers)
+           CERR << "StateIndicator::~StateIndicator() deletes "
+                << (const void *)e << endl;
+         
+        eoc_handlers = e->next;
+        delete e;
       }
 }
 //-----------------------------------------------------------------------------
@@ -447,8 +471,8 @@ Token * tok_R = current_stack.locate_R();
 Value_P
 StateIndicator::get_X()
 {
-Token * tok_X = current_stack.locate_X();
-   if (tok_X)   return tok_X->get_apl_val();
+Value_P * X = current_stack.locate_X();
+   if (X)   return *X;
    return Value_P();
 }
 //-----------------------------------------------------------------------------
@@ -475,19 +499,97 @@ Value_P old_value = tok_R->get_apl_val();   // so that
 void
 StateIndicator::set_X(Value_P new_value)
 {
-Token * tok_X = current_stack.locate_X();
-   if (tok_X == 0)   return;
+Value_P * X = current_stack.locate_X();
+   if (X)   *X = new_value;
+}
+//-----------------------------------------------------------------------------
+void
+StateIndicator::add_eoc_handler(EOC_HANDLER handler, EOC_arg & arg,
+                                const char * loc)
+{
+EOC_arg * new_e =  new EOC_arg(handler, arg, loc);
+
+   if (EOC_arg * e = eoc_handlers)
+      {
+        while (e->next)   e = e->next;
+        e->next = new_e;
+      }
+   else   // first EOC handler
+      {
+        eoc_handlers = new_e;
+      }
+
+   Log(LOG_EOC_handlers)
+      {
+        int count = 0;
+        for (EOC_arg * e = eoc_handlers; e; e = e->next)   ++count;
+        CERR << "SI[" << level << "] added new EOC handler(" << new_e
+             << " from " << loc << " len=" << count << endl;
+      }
+}
+//-----------------------------------------------------------------------------
+EOC_arg *
+StateIndicator::remove_eoc_handlers()
+{
+EOC_arg * ret = eoc_handlers;
+
+   Log(LOG_EOC_handlers)
+      {
+         CERR << "SI[" << level << "] remove_eoc_handler(" << eoc_handlers
+              << endl;
+      }
+
+   eoc_handlers = 0;
+   return ret;
+}
+//-----------------------------------------------------------------------------
+void
+StateIndicator::move_eoc_handler(EOC_HANDLER handler, EOC_arg * old_arg,
+                                 const char * loc)
+{
+   eoc_handlers = old_arg;
+   Assert(eoc_handlers);
+
+   Log(LOG_EOC_handlers)
+      {
+         CERR << "SI[" << level << "] move_eoc_handler(" << eoc_handlers
+              << " from " << eoc_handlers->loc
+              << " to " << loc << ")" << endl;
+      }
+
+   eoc_handlers->handler = handler;
+   eoc_handlers->loc = loc;
+}
+
+//-----------------------------------------------------------------------------
+bool
+StateIndicator::call_eoc_handler(Token & token)
+{
+   if (!eoc_handlers)     return false;   // no eoc_handler
+
+   Log(LOG_EOC_handlers)
+       CERR << "SI[" << level << "] call_eoc_handler(" << eoc_handlers
+            << " from " << eoc_handlers->loc << ")" << endl;
+
+
+const bool goon = eoc_handlers->handler(token);
+   return goon;   // false: success (done), true: goon
 
-Value_P old_value = tok_X->get_apl_val();   // so that 
-   move_2(*tok_X, Token(tok_X->get_tag(), new_value), LOC);
 }
 //-----------------------------------------------------------------------------
 Function_Line
 StateIndicator::get_line() const
 {
-   return executable->get_line(Function_PC(get_PC() - 1));
+int pc = get_PC();
+   if (pc)   --pc;
+   return executable->get_line(Function_PC(pc));
 }
 //-----------------------------------------------------------------------------
+#ifdef WANT_LIBAPL
+typedef int (*result_callback)(const Value * result, int committed);
+extern "C" result_callback res_callback;
+result_callback res_callback = 0;
+#endif
 void
 StateIndicator::statement_result(Token & result, bool trace)
 {
@@ -521,7 +623,19 @@ Value_P B(result.get_apl_val());
 
    // print TOK_APL_VALUE and TOK_APL_VALUE1, but not TOK_APL_VALUE2
    //
-   if (tag != TOK_APL_VALUE1 && tag != TOK_APL_VALUE3)   return;
+bool print_value = tag == TOK_APL_VALUE1 || tag == TOK_APL_VALUE3;
+
+#ifdef WANT_LIBAPL
+   if (res_callback)   // callback installed
+      {
+        // the callback decides whether the value shall be printed (even
+        // if it was committed)
+        //
+        print_value = res_callback(B.get(), !print_value);
+      }
+#endif
+
+   if (!print_value)   return;
 
    Quad_QUOTE::done(false, LOC);
 
