@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2014  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -40,10 +40,20 @@
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
-Symbol::Symbol(const UCS_string & ucs, Id id)
+Symbol::Symbol(ID::Id id)
    : NamedObject(id),
      next(0),
-     symbol(ucs),
+     name(UCS_string(UTF8_string(ID::name(id)))),
+     erased(false),
+     monitor_callback(0)
+{
+   push();
+}
+//-----------------------------------------------------------------------------
+Symbol::Symbol(const UCS_string & ucs, ID::Id id)
+   : NamedObject(id),
+     next(0),
+     name(ucs),
      erased(false),
      monitor_callback(0)
 {
@@ -53,7 +63,7 @@ Symbol::Symbol(const UCS_string & ucs, Id id)
 ostream &
 Symbol::print(ostream & out) const
 {
-   return out << symbol;
+   return out << name;
 }
 //-----------------------------------------------------------------------------
 ostream &
@@ -174,7 +184,7 @@ const ShapeItem max_idx = A->element_count();
             {
               Cell & dest = A->get_ravel(a);
               dest.release(LOC);   // free sub-values etc (if any)
-              dest.init(src);
+              dest.init(src, A.getref());
             }
         if (monitor_callback)   monitor_callback(*this, SEV_ASSIGNED);
         return;
@@ -183,7 +193,6 @@ const ShapeItem max_idx = A->element_count();
 const ShapeItem ec_B = B->element_count();
 const ShapeItem ec_X = X->element_count();
 const APL_Integer qio = Workspace::get_IO();
-const APL_Float qct = Workspace::get_CT();
 const int incr_B = (ec_B == 1) ? 0 : 1;   // maybe scalar extend B
 const Cell * cX = &X->get_ravel(0);
 const Cell * cB = &B->get_ravel(0);
@@ -192,12 +201,12 @@ const Cell * cB = &B->get_ravel(0);
 
    loop(x, ec_X)
       {
-        const ShapeItem idx = cX++->get_near_int(qct) - qio;
+        const ShapeItem idx = cX++->get_near_int() - qio;
         if (idx < 0)          INDEX_ERROR;
         if (idx >= max_idx)   INDEX_ERROR;
         Cell & dest = A->get_ravel(idx);
         dest.release(LOC);   // free sub-values etc (if any)
-        dest.init(*cB);
+        dest.init(*cB, A.getref());
 
          cB += incr_B;
       }
@@ -217,7 +226,7 @@ Symbol::assign_indexed(IndexExpr & IX, Value_P B)   // A[IX;...] ← B
    // see Value::index() for comments.
 
 Value_P A = get_apl_value();
-   if (A->get_rank() != IX.value_count())   INDEX_ERROR;
+   if (A->get_rank() != IX.value_count())   RANK_ERROR;   // ISO p. 159
 
    // B must either be a scalar (and is then scalar extended to the size
    // of the updated area, or else have the shape of the concatenated index
@@ -279,7 +288,7 @@ const int incr_B = (ec_B == 1) ? 0 : 1;
         const ShapeItem offset_A = mult.next();
         Cell & dest = A->get_ravel(offset_A);
         dest.release(LOC);   // free sub-values etc (if any)
-        dest.init(*cB);
+        dest.init(*cB, A.getref());
         cB += incr_B;
      }
 
@@ -309,7 +318,12 @@ ValueStackItem & vs = value_stack.back();
 void
 Symbol::pop()
 {
-   Assert(value_stack.size());
+   if (value_stack.size() == 0)
+      {
+       CERR << "Symbol is: '" << get_name() << "'" << endl;
+        Assert(0 && LOC);
+      }
+
 const ValueStackItem & vs = value_stack.back();
 
    if (vs.name_class == NC_VARIABLE)
@@ -317,7 +331,7 @@ const ValueStackItem & vs = value_stack.back();
         Value_P ret = vs.apl_val;
         Log(LOG_SYMBOL_push_pop)
            {
-             CERR << "-pop-value " << symbol
+             CERR << "-pop-value " << name
                   << " flags " << ret->get_flags() << " ";
              if (value_stack.size() == 0)   CERR << " (last)";
              CERR << " addr " << (const void *)ret.get() << endl;
@@ -331,7 +345,7 @@ const ValueStackItem & vs = value_stack.back();
       {
         Log(LOG_SYMBOL_push_pop)
            {
-             CERR << "-pop " << symbol
+             CERR << "-pop " << name
                   << " name_class " << vs.name_class << " ";
              if (value_stack.size() == 0)   CERR << " (last)";
              CERR << endl;
@@ -346,7 +360,7 @@ Symbol::push()
 {
    Log(LOG_SYMBOL_push_pop)
       {
-        CERR << "+push " << symbol;
+        CERR << "+push " << name;
         if (value_stack.size() == 0)   CERR << " (initial)";
         CERR << endl;
       }
@@ -360,7 +374,7 @@ Symbol::push_label(Function_Line label)
 {
    Log(LOG_SYMBOL_push_pop)
       {
-        CERR << "+push_label " << symbol;
+        CERR << "+push_label " << name;
         if (value_stack.size() == 0)   CERR << " (initial)";
         CERR << endl;
       }
@@ -374,7 +388,7 @@ Symbol::push_function(Function * function)
 {
    Log(LOG_SYMBOL_push_pop)
       {
-        CERR << "+push_function " << symbol << " " << (const void *)function;
+        CERR << "+push_function " << name << " " << (const void *)function;
         if (value_stack.size() == 0)   CERR << " (initial)";
         CERR << endl;
       }
@@ -397,7 +411,7 @@ ValueStackItem vs;
 
    Log(LOG_SYMBOL_push_pop)
       {
-        CERR << "+push-value " << symbol << " flags ";
+        CERR << "+push-value " << name << " flags ";
         print_flags(CERR, get_value()->get_flags()) << " ";
         if (value_stack.size() == 0)   CERR << " (initial)";
         CERR << " addr " << (const void *)get_value().get() << endl;
@@ -439,7 +453,7 @@ const char *
 Symbol::cant_be_defined() const
 {
 // if (value_stack.size() > 1)         return "symbol was localized";
-   if (Workspace::is_called(symbol))   return "function is called";
+   if (Workspace::is_called(name))   return "function is called";
 
    if (value_stack.back().name_class == NC_UNUSED_USER_NAME)   return 0;   // OK
    if (value_stack.back().name_class == NC_FUNCTION)           return 0;   // OK
@@ -457,6 +471,14 @@ Symbol::get_apl_value() const
    return value_stack.back().apl_val;
 }
 //-----------------------------------------------------------------------------
+const Cell *
+Symbol::get_first_cell() const
+{
+   Assert(value_stack.size() > 0);
+   if (value_stack.back().name_class != NC_VARIABLE)   return 0;
+   return &value_stack.back().apl_val->get_ravel(0);
+}
+//-----------------------------------------------------------------------------
 bool
 Symbol::can_be_assigned() const
 {
@@ -620,7 +642,7 @@ Symbol::resolve(Token & tok, bool left_sym)
 
              {
                IntCell lab(value_stack.back().sym_val.label);
-               Value_P value(new Value(lab, LOC));
+               Value_P value(lab, LOC);
                Token t(TOK_APL_VALUE1, value);
                move_1(tok, t, LOC);
              }
@@ -638,7 +660,8 @@ Symbol::resolve(Token & tok, bool left_sym)
 
         case NC_FUNCTION:
         case NC_OPERATOR:
-             move_2(tok, value_stack.back().sym_val.function->get_token(), LOC);
+             move_2(tok, value_stack.back().sym_val.function
+                                                    ->get_token(), LOC);
              return;
 
         case NC_SHARED_VAR:
@@ -831,9 +854,9 @@ ostream &
 Symbol::list(ostream & out)
 {
    out << "   ";
-   loop(s, symbol.size())   out << symbol[s];
+   loop(s, name.size())   out << name[s];
 
-   for (int s = symbol.size(); s < 32; ++s)   out << " ";
+   for (int s = name.size(); s < 32; ++s)   out << " ";
 
    if (erased)   out << "   ERASED";
    Assert(value_stack.size());
@@ -936,19 +959,7 @@ Symbol::unmark_all_values() const
                      // ufun can be 0 for example if F is a function argument
                      // of a defined operator and F is a primitive function
                      //
-                     if (ufun == 0)   break;
-
-                     const Token_string & body = ufun->get_body();
-                     loop(b, body.size())
-                        {
-                          const Token & tok = body[b];
-                          if (tok.get_ValueType() == TV_VAL)
-                             {
-                               Value_P val = tok.get_apl_val();
-                               Assert(val);
-                               val->unmark();
-                             }
-                        }
+                     if (ufun)   ufun->unmark_all_values();
                    }
                    break;
 
@@ -1017,8 +1028,8 @@ const Cell * cV = &values->get_ravel(0);
            }
         else
            {
-             Value_P val(new Value(LOC));
-             val->get_ravel(0).init(*cV);
+             Value_P val(LOC);
+             val->get_ravel(0).init(*cV, val.getref());
              sym->assign(val, LOC);
            }
 
@@ -1204,10 +1215,12 @@ const ErrorCode ec = ErrorCode(response->get__SVAR_ASSIGNED__error());
                   << endl;
            }
 
+        delete response;
         if (del)   delete del;
         throw_apl_error(ec, loc);
       }
 
+   delete response;
    if (del)   delete del;
 }
 //-----------------------------------------------------------------------------
@@ -1273,6 +1286,7 @@ const TCP_socket tcp = Svar_DB::get_DB_tcp();
         const string & data = response->get__WSWS_VALUE_IS__cdr_value();
         if (data.size() == 0)
            {
+             delete response;
              if (del)   delete del;
              CERR << "no data in signal WSWS_VALUE_IS" << endl;
              VALUE_ERROR;
@@ -1280,6 +1294,7 @@ const TCP_socket tcp = Svar_DB::get_DB_tcp();
 
         CDR_string cdr;
         loop(d, data.size())   cdr.append(data[d]);
+        delete response;
         if (del)   delete del;
 
         Value_P value = CDR::from_CDR(cdr, LOC);
@@ -1313,6 +1328,7 @@ const ErrorCode err(ErrorCode(response->get__VALUE_IS__error()));
                   << response->get__VALUE_IS__error_loc() << endl;
            }
 
+        delete response;
         if (del)   delete del;
         throw_apl_error(err, response->get__VALUE_IS__error_loc().c_str());
       }
@@ -1320,6 +1336,7 @@ const ErrorCode err(ErrorCode(response->get__VALUE_IS__error()));
 const string & data = response->get__VALUE_IS__cdr_value();
 CDR_string cdr;
    loop(d, data.size())   cdr.append(data[d]);
+   delete response;
    if (del)   delete del;
 
 Value_P value = CDR::from_CDR(cdr, LOC);
