@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -100,7 +100,7 @@ Symbol * sp = symbol_table[hash];
              CERR << "Symbol " << sym_name << " has hash " << HEX(hash) << endl;
            }
 
-        Symbol * new_symbol = new Symbol(sym_name, ID_USER_SYMBOL);
+        Symbol * new_symbol = new Symbol(sym_name, ID::USER_SYMBOL);
         symbol_table[hash] =  new_symbol;
         return new_symbol;
       }
@@ -130,7 +130,7 @@ int pos = 0;
          if (sym->is_erased())   // override an erased symbol.
             {
               sym->set_erased(false);
-              sym->symbol = sym_name;
+              sym->name = sym_name;
 
               Log(LOG_SYMBOL_lookup_symbol)
                  {
@@ -145,7 +145,7 @@ int pos = 0;
               if (pos >= 255)
                  throw_apl_error(E_SYSTEM_LIMIT_SYMTAB, LOC);
 
-              sym->next = new Symbol(sym_name, ID_USER_SYMBOL);
+              sym->next = new Symbol(sym_name, ID::USER_SYMBOL);
 
               Log(LOG_SYMBOL_lookup_symbol)
                  {
@@ -168,20 +168,20 @@ Token_string tos;
         Tokenizer tokenizer(PM_STATEMENT_LIST, LOC);
         if (tokenizer.tokenize(buf, tos) != E_NO_ERROR)
            {
-             CERR << "parse error 1 in )SYMBOL command." << endl;
+             CERR << "invalid token" << endl;
              return out;
            }
       }
 
    if (tos.size() == 0)   // empty argument
       {
-        CERR << "parse error 2 in )SYMBOL command" << endl;
+        CERR << "no symbol" << endl;
         return out;
       }
 
    if (tos[0].get_ValueType() != TV_SYM)
       {
-        CERR << "parse error 3 in )SYMBOL command" << endl;
+        CERR << "not a symbol" << endl;
         return out;
       }
 
@@ -376,29 +376,34 @@ SymbolTable::clear(ostream & out)
    //
    Assert(Workspace::SI_entry_count() == 0);
 
-   loop(hash, MAX_SYMBOL_COUNT)
-      {
-        Symbol * sym = symbol_table[hash];
-        if (sym == 0)   continue;   // no symbol a this hash
+   loop(hash, MAX_SYMBOL_COUNT)   clear_slot(out, hash);
+}
+//-----------------------------------------------------------------------------
+void
+SymbolTable::clear_slot(ostream & out, int hash)
+{
+Symbol * sym = symbol_table[hash];
+   if (sym == 0)   return;   // no symbol with this hash
 
-        const bool user_defined = sym->is_user_defined();
-        for (; sym; sym = sym->next)
+   symbol_table[hash] = 0;
+
+Symbol * next;
+   for (; sym; sym = next)
+       {
+         next = sym->next;
+
+         // keep not user-defined symbols ??? ARE THERE ANY ???
+         //
+         if (!sym->is_user_defined())
             {
-              if (sym->is_erased())               continue;
-              if (sym->value_stack.size() == 0)   continue;
-
-              Assert(!Workspace::is_called(sym->get_name()));
-              if (sym->value_stack.size() != 1)
-                {
-                  Q1(sym->value_stack.size())
-                  Q1(sym->symbol)
-                  Assert(0);
-                }
-              sym->clear_vs();
+              sym->next = symbol_table[hash];
+              symbol_table[hash] = sym;
+              continue;
             }
 
-        if (user_defined)   symbol_table[hash] = 0;
-      } 
+         delete sym;
+       }
+
 }
 //-----------------------------------------------------------------------------
 bool
