@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2014  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -34,10 +34,12 @@
 #include "Common.hh"
 #include "FloatCell.hh"
 #include "IndexExpr.hh"
-#include "Input.hh"
 #include "IntCell.hh"
+#include "LineInput.hh"
 #include "Output.hh"
+#include "Parallel.hh"
 #include "PointerCell.hh"
+#include "Prefix.hh"
 #include "ProcessorID.hh"
 #include "StateIndicator.hh"
 #include "SystemVariable.hh"
@@ -123,9 +125,9 @@ SystemVariable::get_attributes(int mode, Cell * dest) const
 }
 //=============================================================================
 Quad_AV::Quad_AV()
-   : RO_SystemVariable(ID_Quad_AV)
+   : RO_SystemVariable(ID::Quad_AV)
 {
-Value_P AV(new Value(MAX_AV, LOC));
+Value_P AV(MAX_AV, LOC);
    loop(cti, MAX_AV)
        {
          const int av_pos = Avec::get_av_pos(CHT_Index(cti));
@@ -147,7 +149,7 @@ Quad_AV::indexed_at(uint32_t pos)
 }
 //=============================================================================
 Quad_AI::Quad_AI()
-   : RO_SystemVariable(ID_Quad_AI),
+   : RO_SystemVariable(ID::Quad_AI),
      session_start(now()),
      user_wait(0)
 {
@@ -160,7 +162,7 @@ Quad_AI::get_apl_value() const
 const int total_ms = (now() - session_start)/1000;
 const int user_ms  = user_wait/1000;
 
-Value_P ret(new Value(4, LOC));
+Value_P ret(4, LOC);
    new (&ret->get_ravel(0))   IntCell(ProcessorID::get_own_ID());
    new (&ret->get_ravel(1))   IntCell(total_ms - user_ms);
    new (&ret->get_ravel(2))   IntCell(total_ms);
@@ -171,7 +173,7 @@ Value_P ret(new Value(4, LOC));
 }
 //=============================================================================
 Quad_ARG::Quad_ARG()
-   : RO_SystemVariable(ID_Quad_ARG)
+   : RO_SystemVariable(ID::Quad_ARG)
 {
 }
 //-----------------------------------------------------------------------------
@@ -180,17 +182,17 @@ Quad_ARG::get_apl_value() const
 {
 const int argc = uprefs.expanded_argv.size();
 
-Value_P Z(new Value(argc, LOC));
+Value_P Z(argc, LOC);
 Cell * C = &Z->get_ravel(0);
 
    loop(a, argc)
       {
         const char * arg = uprefs.expanded_argv[a];
         const int len = strlen(arg);
-        Value_P val(new Value(len, LOC));
+        Value_P val(len, LOC);
         loop(l, len)   new (val->next_ravel())   CharCell(Unicode(arg[l]));
 
-        new (C++)   PointerCell(val);
+        new (C++)   PointerCell(val, Z.getref());
       }
 
    Z->check_value(LOC);
@@ -198,7 +200,7 @@ Cell * C = &Z->get_ravel(0);
 }
 //=============================================================================
 Quad_CT::Quad_CT()
-   : SystemVariable(ID_Quad_CT)
+   : SystemVariable(ID::Quad_CT)
 {
    Symbol::assign(FloatScalar(DEFAULT_Quad_CT, LOC), LOC);
 }
@@ -252,7 +254,7 @@ const Error * err = 0;
         // return 3 0⍴' '
         //
         Shape sh((ShapeItem)3, (ShapeItem)0);
-        Value_P Z(new Value(sh, LOC));
+        Value_P Z(sh, LOC);
         new (&Z->get_ravel(0))   CharCell(UNI_ASCII_SPACE);
         Z->check_value(LOC);
         return Z;
@@ -273,7 +275,7 @@ ShapeItem max_len = len_1;
    if (max_len < len_3)   max_len = len_3;
 
 const Shape sh(3, max_len);
-Value_P Z(new Value(sh, LOC));
+Value_P Z(sh, LOC);
    loop(l, max_len)
       {
         new (&Z->get_ravel(l))
@@ -311,7 +313,7 @@ ErrorCode ec = E_NO_ERROR;
          if (si->get_executable()->get_parse_mode() == PM_FUNCTION)   break;
        }
 
-Value_P Z(new Value(2, LOC));
+Value_P Z(2, LOC);
 
    new (&Z->get_ravel(0)) IntCell(Error::error_major(ec));
    new (&Z->get_ravel(1)) IntCell(Error::error_minor(ec));
@@ -320,9 +322,9 @@ Value_P Z(new Value(2, LOC));
    return Z;
 }
 //=============================================================================
-Quad_FC::Quad_FC() : SystemVariable(ID_Quad_FC)
+Quad_FC::Quad_FC() : SystemVariable(ID::Quad_FC)
 {
-Value_P QFC(new Value(6, LOC));
+Value_P QFC(6, LOC);
    new (QFC->next_ravel()) CharCell(UNI_ASCII_FULLSTOP);
    new (QFC->next_ravel()) CharCell(UNI_ASCII_COMMA);
    new (QFC->next_ravel()) CharCell(UNI_STAR_OPERATOR);
@@ -339,7 +341,7 @@ Quad_FC::push()
 {
    Symbol::push();
 
-Value_P QFC(new Value(6, LOC));
+Value_P QFC(6, LOC);
    new (QFC->next_ravel()) CharCell(UNI_ASCII_FULLSTOP);
    new (QFC->next_ravel()) CharCell(UNI_ASCII_COMMA);
    new (QFC->next_ravel()) CharCell(UNI_STAR_OPERATOR);
@@ -376,7 +378,7 @@ Unicode fc[6] = { UNI_ASCII_FULLSTOP, UNI_ASCII_COMMA,      UNI_STAR_OPERATOR,
       fc[4] = UNI_ASCII_SPACE;
 
 UCS_string ucs(fc, 6);
-Value_P new_val(new Value(ucs, LOC));
+Value_P new_val(ucs, LOC);
    Symbol::assign(new_val, LOC);
 }
 //-----------------------------------------------------------------------------
@@ -427,12 +429,12 @@ Unicode fc[6];
       fc[4] = UNI_ASCII_SPACE;
 
 UCS_string ucs(fc, 6);
-Value_P new_val(new Value(ucs, LOC));
+Value_P new_val(ucs, LOC);
    Symbol::assign(new_val, LOC);
 }
 //=============================================================================
 Quad_IO::Quad_IO()
-   : SystemVariable(ID_Quad_IO)
+   : SystemVariable(ID::Quad_IO)
 {
    Symbol::assign(IntScalar(1, LOC), LOC);
 }
@@ -446,14 +448,14 @@ Quad_IO::assign(Value_P value, const char * loc)
         else                         LENGTH_ERROR;
       }
 
-   if (value->get_ravel(0).get_near_bool(0.1))
+   if (value->get_ravel(0).get_near_bool())
       Symbol::assign(IntScalar(1, LOC), LOC);
    else
       Symbol::assign(IntScalar(0, LOC), LOC);
 }
 //=============================================================================
 Quad_L::Quad_L()
- : NL_SystemVariable(ID_Quad_L)
+ : NL_SystemVariable(ID::Quad_L)
 {
    Symbol::assign(IntScalar(0, LOC), LOC);
 }
@@ -485,7 +487,7 @@ StateIndicator * si = Workspace::SI_top_error();
 }
 //=============================================================================
 Quad_LC::Quad_LC()
-   : RO_SystemVariable(ID_Quad_LC)
+   : RO_SystemVariable(ID::Quad_LC)
 {
    Symbol::assign(IntScalar(0, LOC), LOC);
 }
@@ -503,7 +505,7 @@ int len = 0;
          if (si->get_executable()->get_parse_mode() == PM_FUNCTION)   ++len;
        }
 
-Value_P Z(new Value(len, LOC));
+Value_P Z(len, LOC);
 
    for (StateIndicator * si = Workspace::SI_top();
         si; si = si->get_parent())
@@ -518,7 +520,7 @@ Value_P Z(new Value(len, LOC));
 }
 //=============================================================================
 Quad_LX::Quad_LX()
-   : NL_SystemVariable(ID_Quad_LX)
+   : NL_SystemVariable(ID::Quad_LX)
 {
    Symbol::assign(Str0(LOC), LOC);
 }
@@ -533,9 +535,9 @@ Quad_LX::assign(Value_P value, const char * loc)
 }
 //=============================================================================
 Quad_PP::Quad_PP()
-   : SystemVariable(ID_Quad_PP)
+   : SystemVariable(ID::Quad_PP)
 {
-Value_P value(new Value(LOC));
+Value_P value(LOC);
 
    new (&value->get_ravel(0)) IntCell(DEFAULT_Quad_PP);
 
@@ -553,7 +555,7 @@ Quad_PP::assign(Value_P value, const char * loc)
       }
 
 const Cell & cell = value->get_ravel(0);
-APL_Integer val = cell.get_near_int(0.1);
+APL_Integer val = cell.get_near_int();
    if (val < MIN_Quad_PP)   DOMAIN_ERROR;
 
    if (val > MAX_Quad_PP)   val = MAX_Quad_PP;
@@ -562,9 +564,9 @@ APL_Integer val = cell.get_near_int(0.1);
 }
 //=============================================================================
 Quad_PR::Quad_PR()
-   : SystemVariable(ID_Quad_PR)
+   : SystemVariable(ID::Quad_PR)
 {
-   Symbol::assign(Spc(LOC), LOC);
+   Symbol::assign(CharScalar(UNI_ASCII_SPACE, LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 void
@@ -578,7 +580,7 @@ UCS_string ucs = value->get_UCS_ravel();
 }
 //=============================================================================
 Quad_PS::Quad_PS()
-   : SystemVariable(ID_Quad_PS)
+   : SystemVariable(ID::Quad_PS)
 {
    Symbol::assign(IntScalar(0, LOC), LOC);
 }
@@ -593,7 +595,7 @@ Quad_PS::assign(Value_P value, const char * loc)
       }
 
 const Cell & cell = value->get_ravel(0);
-const APL_Integer val = cell.get_near_int(0.1);
+const APL_Integer val = cell.get_near_int();
    switch(val)
       {
         case 0:
@@ -604,24 +606,8 @@ const APL_Integer val = cell.get_near_int(0.1);
       }
 }
 //=============================================================================
-Quad_PT::Quad_PT()
-   : RO_SystemVariable(ID_Quad_PT)
-{
-   Symbol::assign(IntScalar(0, LOC), LOC);
-}
-//-----------------------------------------------------------------------------
-Value_P
-Quad_PT::get_apl_value() const
-{
-Value_P Z(new Value(LOC));
-   new (&Z->get_ravel(0))   FloatCell(Workspace::get_CT());
-
-   Z->check_value(LOC);
-   return Z;
-}
-//=============================================================================
 Quad_PW::Quad_PW()
-   : SystemVariable(ID_Quad_PW)
+   : SystemVariable(ID::Quad_PW)
 {
    Symbol::assign(IntScalar(DEFAULT_Quad_PW, LOC), LOC);
 }
@@ -636,7 +622,7 @@ Quad_PW::assign(Value_P value, const char * loc)
       }
 
 const Cell & cell = value->get_ravel(0);
-const APL_Integer val = cell.get_near_int(0.1);
+const APL_Integer val = cell.get_near_int();
 
    // min val is 30. Ignore smaller values.
    if (val < 30)   return;
@@ -648,7 +634,7 @@ const APL_Integer val = cell.get_near_int(0.1);
 }
 //=============================================================================
 Quad_Quad::Quad_Quad()
- : SystemVariable(ID_Quad_Quad)
+ : SystemVariable(ID::Quad_Quad)
 {
 }
 //-----------------------------------------------------------------------------
@@ -671,17 +657,22 @@ Quad_Quad::resolve(Token & token, bool left)
 
    COUT << UNI_Quad_Quad << ":" << endl;
 
-UCS_string in = Input::get_line();
+bool eof = false;
+UCS_string line;
+   InputMux::get_line(LIM_Quad_Quad, Workspace::get_prompt(), line, eof,
+                      LineHistory::quad_quad_history);
 
-   move_2(token, Bif_F1_EXECUTE::execute_statement(in), LOC);
+   line.remove_leading_and_trailing_whitespaces();
+
+   move_2(token, Bif_F1_EXECUTE::execute_statement(line), LOC);
 }
 //=============================================================================
 Quad_QUOTE::Quad_QUOTE()
- : SystemVariable(ID_QUOTE_Quad)
+ : SystemVariable(ID::QUOTE_Quad)
 {
    // we assign a dummy value so that ⍞ is not undefined.
    //
-Value_P dummy(new Value(UCS_string(LOC), LOC));
+Value_P dummy(UCS_string(LOC), LOC);
    dummy->set_complete();
    Symbol::assign(dummy, LOC);
 }
@@ -695,7 +686,7 @@ Quad_QUOTE::done(bool with_LF, const char * loc)
 
    if (prompt.size())
       {
-        if (with_LF)   COUT << UNI_ASCII_LF;
+        if (with_LF)   COUT << endl;
         prompt.clear();
       }
 }
@@ -720,7 +711,7 @@ PrintBuffer pb(*value, pctx, 0);
       }
    else if (pb.get_height() > 0)
       {
-        COUT << pb.l1().no_pad();
+        COUT << pb.l1().no_pad() << flush;
         prompt.append(pb.l1());
       }
    else   // empty output
@@ -728,6 +719,8 @@ PrintBuffer pb(*value, pctx, 0);
         // nothing to do
       }
 
+   cout << flush;
+
    Log(LOG_cork)
       CERR << "Quad_QUOTE::assign() done, buffer = ["
            << prompt << "]" << endl;
@@ -744,28 +737,33 @@ Quad_QUOTE::get_apl_value() const
    //
 const UCS_string old_prompt(prompt);
 
-UCS_string in = Input::get_quad_cr_line(old_prompt);
+bool eof = false;
+UCS_string line;
+   InputMux::get_line(LIM_Quote_Quad, old_prompt, line, eof,
+                      LineHistory::quote_quad_history);
    done(false, LOC);   // if get_quad_cr_line() has not called it
 
+   if (interrupt_raised)   INTERRUPT;
+
 const UCS_string qpr = Workspace::get_PR();
 
    if (qpr.size() > 0)   // valid prompt replacement char
       {
-        loop(i, in.size())
+        loop(i, line.size())
            {
              if (i >= old_prompt.size())   break;
-             if (old_prompt[i] != in[i])   break;
-             in[i] = qpr[0];
+             if (old_prompt[i] != line[i])   break;
+             line[i] = qpr[0];
            }
       }
 
-Value_P Z(new Value(in, LOC));
+Value_P Z(line, LOC);
    Z->check_value(LOC);
    return Z;
 }
 //=============================================================================
 Quad_R::Quad_R()
- : NL_SystemVariable(ID_Quad_R)
+ : NL_SystemVariable(ID::Quad_R)
 {
    Symbol::assign(IntScalar(0, LOC), LOC);
 }
@@ -821,13 +819,12 @@ Quad_SYL::assign_indexed(IndexExpr & IDX, Value_P value)
    //
 Value_P X2 = IDX.extract_value(0);
 
-const APL_Float qct = Workspace::get_CT();
 const APL_Integer qio = Workspace::get_IO();
 
-   if (!X2)                                             INDEX_ERROR;
-   if (X2->element_count() != 1)                        INDEX_ERROR;
-   if (!X2->get_ravel(0).is_near_int(qct))              INDEX_ERROR;
-   if (X2->get_ravel(0).get_near_int(qct) != qio + 1)   INDEX_ERROR;
+   if (!X2)                                          INDEX_ERROR;
+   if (X2->element_count() != 1)                     INDEX_ERROR;
+   if (!X2->get_ravel(0).is_near_int())              INDEX_ERROR;
+   if (X2->get_ravel(0).get_near_int() != qio + 1)   INDEX_ERROR;
 
 Value_P X1 = IDX.extract_value(1);
 
@@ -848,8 +845,8 @@ const APL_Integer qio = Workspace::get_IO();
 
    loop(e, ec)
       {
-        const APL_Integer x = X->get_ravel(e).get_near_int(qct) - qio;
-        const APL_Integer b = B->get_ravel(e).get_near_int(qct);
+        const APL_Integer x = X->get_ravel(e).get_near_int() - qio;
+        const APL_Integer b = B->get_ravel(e).get_near_int();
 
         if (x == SYL_SI_DEPTH_LIMIT)   // SI depth limit
            {
@@ -871,12 +868,10 @@ const APL_Integer qio = Workspace::get_IO();
 
              ravel_count_limit = cells;
            }
-#if CORE_COUNT_WANTED == -3
         else if (x == SYL_CURRENT_CORES)   // number of cores
            {
-             setup_cores((CoreCount)b);
+             if (Parallel::set_core_count((CoreCount)b, false))   DOMAIN_ERROR;
            }
-#endif
         else if (x == SYL_PRINT_LIMIT)   // print length limit
            {
              if (b < 1000)   DOMAIN_ERROR;
@@ -893,12 +888,12 @@ Value_P
 Quad_SYL::get_apl_value() const
 {
 const Shape sh(SYL_MAX, 2);
-Value_P Z( new Value(sh, LOC));
+Value_P Z(sh, LOC);
 
 #define syl2(n, e, v) syl1(n, e, v)
 #define syl3(n, e, v) syl1(n, e, v)
 #define syl1(n, _e, v) \
-  new (Z->next_ravel()) PointerCell(Value_P(new Value(UCS_string(UTF8_string(n)), LOC))); \
+  new (Z->next_ravel()) PointerCell(Value_P(UCS_string(UTF8_string(n)), LOC), Z.getref()); \
   new (Z->next_ravel()) IntCell(v);
 #include "SystemLimits.def"
 
@@ -907,9 +902,9 @@ Value_P Z( new Value(sh, LOC));
 }
 //=============================================================================
 Quad_TC::Quad_TC()
-   : RO_SystemVariable(ID_Quad_TC)
+   : RO_SystemVariable(ID::Quad_TC)
 {
-Value_P QCT(new Value(3, LOC));
+Value_P QCT(3, LOC);
    new (QCT->next_ravel()) CharCell(UNI_ASCII_BS);
    new (QCT->next_ravel()) CharCell(UNI_ASCII_CR);
    new (QCT->next_ravel()) CharCell(UNI_ASCII_LF);
@@ -919,7 +914,7 @@ Value_P QCT(new Value(3, LOC));
 }
 //=============================================================================
 Quad_TS::Quad_TS()
-   : RO_SystemVariable(ID_Quad_TS)
+   : RO_SystemVariable(ID::Quad_TS)
 {
    Symbol::assign(IntScalar(0, LOC), LOC);
 }
@@ -930,7 +925,7 @@ Quad_TS::get_apl_value() const
 const APL_time_us offset_us = 1000000 * Workspace::get_v_Quad_TZ().get_offset();
 const YMDhmsu time(now() + offset_us);
 
-Value_P Z(new Value(7, LOC));
+Value_P Z(7, LOC);
    new (&Z->get_ravel(0)) IntCell(time.year);
    new (&Z->get_ravel(1)) IntCell(time.month);
    new (&Z->get_ravel(2)) IntCell(time.day);
@@ -944,7 +939,7 @@ Value_P Z(new Value(7, LOC));
 }
 //=============================================================================
 Quad_TZ::Quad_TZ()
-   : SystemVariable(ID_Quad_TZ)
+   : SystemVariable(ID::Quad_TZ)
 {
    offset_seconds = compute_offset();
 
@@ -967,8 +962,7 @@ Quad_TZ::compute_offset()
    // that localtime() and gmtime() are more portable.
    // 
 
-   // choose a reference point in time that is the same month in all time zones.
-   // we use Monday Jan 2014 in the afternoon
+   // choose a reference point in time.
    //
 const time_t ref = now()/1000000;
 
@@ -1019,7 +1013,7 @@ Cell & cell = value->get_ravel(0);
 }
 //=============================================================================
 Quad_UL::Quad_UL()
-   : RO_SystemVariable(ID_Quad_UL)
+   : RO_SystemVariable(ID::Quad_UL)
 {
    Symbol::assign(get_apl_value(), LOC);
 }
@@ -1045,14 +1039,14 @@ int user_count = 0;
    user_count = 1;
 #endif
 
-Value_P Z(new Value(LOC));
+Value_P Z(LOC);
    new (&Z->get_ravel(0))   IntCell(user_count);
    Z->check_value(LOC);
    return Z;
 }
 //=============================================================================
 Quad_WA::Quad_WA()
-   : RO_SystemVariable(ID_Quad_WA)
+   : RO_SystemVariable(ID::Quad_WA)
 {
    Symbol::assign(IntScalar(0, LOC), LOC);
 }
@@ -1060,7 +1054,7 @@ Quad_WA::Quad_WA()
 Value_P
 Quad_WA::get_apl_value() const
 {
-Value_P Z(new Value(LOC));
+Value_P Z(LOC);
 
 uint64_t total = total_memory;   // max memory as reported by RLIM_INFINITY
 uint64_t proc_mem = 0;            // memory as reported proc/mem_info
@@ -1122,7 +1116,7 @@ uint64_t proc_mem = 0;            // memory as reported proc/mem_info
 }
 //=============================================================================
 Quad_X::Quad_X()
- : NL_SystemVariable(ID_Quad_X)
+ : NL_SystemVariable(ID::Quad_X)
 {
    Symbol::assign(IntScalar(0, LOC), LOC);
 }
