@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -173,7 +173,12 @@ operator << (ostream & out, const Token & token)
         return out << token.get_Id() << token.get_int_val();
       }
 
-   if (token.get_Id() > ID_No_ID2)   return out << token.get_Id();
+   if (token.get_tag() == TOK_NOBRANCH)
+      {
+        return out << token.get_Id() << token.get_int_val();
+      }
+
+   if (token.get_Id() > ID::No_ID2)   return out << token.get_Id();
 
    switch(token.get_Class())
       {
@@ -230,12 +235,6 @@ operator << (ostream & out, const Token & token)
 }
 //-----------------------------------------------------------------------------
 void
-Token::SET_temp()
-{
-   if (is_apl_val() && !!value._apl_val())   value._apl_val()->SET_temp(); 
-}
-//-----------------------------------------------------------------------------
-void
 Token::ChangeTag(TokenTag new_tag)
 {
    Assert((tag & TV_MASK) == (new_tag & TV_MASK));
@@ -243,6 +242,19 @@ Token::ChangeTag(TokenTag new_tag)
    (TokenTag &)tag = new_tag;
 }
 //-----------------------------------------------------------------------------
+int
+Token::value_use_count() const
+{
+   if (!is_apl_val())       return 0;
+   if (!value._apl_val())   return -98;
+   return value._apl_val()->get_owner_count();
+}
+//-----------------------------------------------------------------------------
+void Token::extract_apl_val(const char * loc) const
+{
+   if (is_apl_val())   ptr_clear(value._apl_val(), loc);
+}
+//-----------------------------------------------------------------------------
 ostream &
 Token::print_function(ostream & out) const
 {
@@ -328,7 +340,8 @@ Token::print_function(ostream & out) const
         case TOK_OPER1_REDUCE1:
         case TOK_OPER1_SCAN:
         case TOK_OPER1_SCAN1:
-        case TOK_OPER2_PRODUCT: return out << get_Id();
+        case TOK_OPER2_OUTER:
+        case TOK_OPER2_INNER: return out << get_Id();
 
         case TOK_FUN0:      return get_function()->print(out << "USER-F0 ");
         case TOK_FUN2:      return get_function()->print(out << "USER-F2 ");
@@ -342,7 +355,7 @@ Token::print_function(ostream & out) const
 
    out << "{ unknown function Token " << tag;
 
-   if (get_Id() != ID_No_ID)   // tag has an Id
+   if (get_Id() != ID::No_ID)   // tag has an Id
       {
         out << ", Id = " << get_Id();
       }
@@ -357,11 +370,11 @@ Token::print_value(ostream & out) const
         case TOK_VARIABLE:  return value.sym_ptr->print(out);
         case TOK_FUN0:      return out <<  "{ fun/0 }";
         case TOK_CHARACTER: return out << value.char_val;
-        case TOK_INTEGER:   return out << value.int_val;
-        case TOK_REAL:      return out << value.flt_val;
+        case TOK_INTEGER:   return out << value.int_vals[0];
+        case TOK_REAL:      return out << value.float_vals[0];
 
-        case TOK_COMPLEX:   return out << value.complex_val.real << "J"
-                                       << value.complex_val.imag;
+        case TOK_COMPLEX:   return out << value.float_vals[0] << "J"
+                                       << value.float_vals[1];
 
         case TOK_APL_VALUE1:
         case TOK_APL_VALUE3:
@@ -417,7 +430,7 @@ Token::print_value(ostream & out) const
 
    out << "{ unknown value Token " << tag;
 
-   if (get_Id() != ID_No_ID)   // tag has an Id
+   if (get_Id() != ID::No_ID)   // tag has an Id
       {
         out << ", Id = " << get_Id();
       }
@@ -447,11 +460,16 @@ UCS_string fn = fun_name;
              out << "→" << get_int_val() << endl;
              return;
 
+        case TOK_NOBRANCH:
+             out << "→⍬" << endl;
+             return;
+
         case TOK_ESCAPE:
              out << "→" << endl;
              return;
 
         case TOK_VOID:
+        case TOK_NO_VALUE:
              out << endl;
              return;
 
@@ -555,14 +573,14 @@ UCS_string ucs;
         case TC_L_PARENT:
         case TC_R_CURLY:
         case TC_L_CURLY:
-             return id_name(get_Id());
+             return UCS_string(UTF8_string(ID::name(get_Id())));
 
         case TC_FUN0:
         case TC_FUN12:
         case TC_OPER1:
         case TC_OPER2:
-             if (get_Id() != ID_No_ID)       return id_name(get_Id());
-             return get_function()->get_name();
+             if (get_Id() == ID::No_ID)   return get_function()->get_name();
+             return UCS_string(UTF8_string(ID::name(get_Id())));
 
         default:
              CERR << "Token: " << HEX4(tag) << " " << *this
@@ -719,18 +737,20 @@ Token::copy_N(const Token & src)
    tag = src.tag;
    switch(src.get_ValueType())
       {
-        case TV_NONE:  value.int_val    = 0;                              break;
-        case TV_CHAR:  value.char_val   = src.value.char_val;             break;
-        case TV_INT:   value.int_val    = src.value.int_val;              break;
-        case TV_FLT:   value.flt_val    = src.value.flt_val;              break;
-        case TV_CPX:   value.complex_val.real = src.value.complex_val.real;
-                       value.complex_val.imag = src.value.complex_val.imag;
-                                                                          break;
-        case TV_SYM:   value.sym_ptr    = src.value.sym_ptr;              break;
-        case TV_LIN:   value.fun_line   = src.value.fun_line;             break;
-        case TV_VAL:   value._apl_val() = src.value._apl_val();           break;
-        case TV_INDEX: value.index_val  = src.value.index_val;            break;
-        case TV_FUN:   value.function   = src.value.function;             break;
+        case TV_NONE:  value.int_vals[0]   = 0;
+                       value.int_vals[1]   = src.value.int_vals[1];     break;
+
+        case TV_CHAR:  value.char_val      = src.value.char_val;        break;
+        case TV_INT:   value.int_vals[0]   = src.value.int_vals[0];
+                       value.int_vals[1]   = src.value.int_vals[1];     break;
+        case TV_FLT:   value.float_vals[0] = src.value.float_vals[0];   break;
+        case TV_CPX:   value.float_vals[0] = src.value.float_vals[0];
+                       value.float_vals[1] = src.value.float_vals[1];   break;
+        case TV_SYM:   value.sym_ptr       = src.value.sym_ptr;         break;
+        case TV_LIN:   value.fun_line      = src.value.fun_line;        break;
+        case TV_VAL:   value._apl_val()    = src.value._apl_val();      break;
+        case TV_INDEX: value.index_val     = src.value.index_val;       break;
+        case TV_FUN:   value.function      = src.value.function;        break;
         default:       FIXME;
       }
 }
