@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -21,9 +21,12 @@
 #include <string.h>
 
 #include "Bif_F12_FORMAT.hh"
+#include "Bif_F12_SORT.hh"
 #include "Bif_OPER1_COMMUTE.hh"
 #include "Bif_OPER1_EACH.hh"
-#include "Bif_OPER2_PRODUCT.hh"
+#include "Bif_OPER2_INNER.hh"
+#include "Bif_OPER2_OUTER.hh"
+#include "Bif_OPER2_POWER.hh"
 #include "Bif_OPER2_RANK.hh"
 #include "Bif_OPER1_REDUCE.hh"
 #include "Bif_OPER1_SCAN.hh"
@@ -93,14 +96,17 @@ Source<Unicode> src(input);
             {
               case TC_END:   // chars without APL meaning
                    rest_2 = src.rest();
-                   throw_parse_error(E_NO_TOKEN, LOC, loc);
+                   {
+                     UCS_string more("Tokenizer: No token for Unicode U+");
+                     more.append_hex(uni, true);
+                     throw_tokenize_error(E_NO_TOKEN, more);
+                   }
                    break;
 
               case TC_RETURN:
               case TC_LINE:
               case TC_VALUE:
               case TC_INDEX:
-              case TC_VOID:
                    CERR << "Offending token: " << tok.get_tag()
                         << " (" << tok << ")" << endl;
                    if (tok.get_tag() == TOK_CHARACTER)
@@ -109,6 +115,15 @@ Source<Unicode> src(input);
                    throw_parse_error(E_NON_APL_CHAR, LOC, loc);
                    break;
 
+              case TC_VOID:
+                   // Avec::uni_to_token returns TC_VOID for non-apl characters
+                   //
+                   rest_2 = src.rest();
+                   UERR << "Unknown APL character: " << uni
+                        << " (" << UNI(uni) << ")" << endl;
+                   throw_parse_error(E_NON_APL_CHAR, LOC, loc);
+                   break;
+
               case TC_SYMBOL:
                    if (Avec::is_quad(uni))
                       {
@@ -169,7 +184,7 @@ Source<Unicode> src(input);
                    break;
 
               case TC_OPER2:
-                   if (tok.get_tag() == TOK_OPER2_PRODUCT && src.rest())
+                   if (tok.get_tag() == TOK_OPER2_INNER && src.rest())
                       {
                         // tok is a dot. This could mean that . is either
                         //
@@ -191,6 +206,14 @@ Source<Unicode> src(input);
                       {
                         tokenize_function(src, tos);
                       }
+                   if (tos.size() >= 2 &&
+                       tos.last().get_tag() == TOK_OPER2_INNER &&
+                       tos[tos.size() - 2].get_tag() == TOK_JOT)
+                      {
+                        new (&tos.last()) Token(TOK_OPER2_OUTER,
+                                                Bif_OPER2_OUTER::fun);
+                      }
+                       
                    break;
 
               case TC_R_ARROW:
@@ -247,9 +270,9 @@ Source<Unicode> src(input);
                         << "uni:   " << uni << endl
                         << "Token = " << tok.get_tag() << endl;
 
-                   if (Id(tok.get_tag() >> 16) != ID_No_ID)
+                   if (tok.get_Id() != ID::No_ID)
                       {
-                        CERR << ", Id = " << Id(tok.get_tag() >> 16);
+                        CERR << ", Id = " << ID::Id(tok.get_tag() >> 16);
                       }
                    CERR << endl;
                    Assert(0 && "Should not happen");
@@ -266,10 +289,17 @@ Tokenizer::tokenize_function(Source<Unicode> & src, Token_string & tos)
    Log(LOG_tokenize)   CERR << "tokenize_function(" << src << ")" << endl;
 
 const Unicode uni = src.get();
-Token tok = Avec::uni_to_token(uni, LOC);
+const Token tok = tokenize_function(uni);
+   tos.append(tok, LOC);
+}
+//-----------------------------------------------------------------------------
+Token
+Tokenizer::tokenize_function(Unicode uni)
+{
+const Token tok = Avec::uni_to_token(uni, LOC);
 
 #define sys(t, f) \
-   case TOK_ ## t: new (&tok) Token(tok.get_tag(), &Bif_ ## f::fun);   break;
+   case TOK_ ## t: return Token(tok.get_tag(), Bif_ ## f::fun);   break;
 
    switch(tok.get_tag())
       {
@@ -329,22 +359,28 @@ Token tok = Avec::uni_to_token(uni, LOC);
 
         sys(JOT,           JOT)
 
-        sys(OPER1_EACH,    OPER1_EACH)
         sys(OPER1_COMMUTE, OPER1_COMMUTE)
+        sys(OPER1_EACH,    OPER1_EACH)
+        sys(OPER2_POWER,   OPER2_POWER)
         sys(OPER2_RANK,    OPER2_RANK)
         sys(OPER1_REDUCE,  OPER1_REDUCE)
         sys(OPER1_REDUCE1, OPER1_REDUCE1)
         sys(OPER1_SCAN,    OPER1_SCAN)
         sys(OPER1_SCAN1,   OPER1_SCAN1)
 
-        sys(OPER2_PRODUCT, OPER2_PRODUCT)
+        sys(OPER2_INNER,   OPER2_INNER)
 
-        default: CERR << endl << "Token = " << tok << endl;
-                 Assert(0 && "Missing Function");
+        default: break;
       }
 
+   // CAUTION: cannot print entire token here because Avec::uni_to_token()
+   // inits the token tag but not any token pointers!
+   //
+   CERR << endl << "Token = " << tok.get_tag() << endl;
+   Assert(0 && "Missing Function");
+
 #undef sys
-   tos.append(Token(tok), LOC);
+   return tok;
 }
 //-----------------------------------------------------------------------------
 void
@@ -425,8 +461,7 @@ bool got_end = false;
       }
    else
       {
-        tos.append(Token(TOK_APL_VALUE1,
-                         Value_P(new Value(string_value, LOC))), LOC);
+        tos.append(Token(TOK_APL_VALUE1, Value_P(string_value, LOC)), LOC);
       }
 }
 //-----------------------------------------------------------------------------
@@ -496,8 +531,7 @@ bool got_end = false;
 
    else
       {
-        tos.append(Token(TOK_APL_VALUE1,
-                         Value_P(new Value(string_value, LOC))), LOC);
+        tos.append(Token(TOK_APL_VALUE1, Value_P(string_value, LOC)), LOC);
       }
 }
 //-----------------------------------------------------------------------------
@@ -513,54 +547,82 @@ Tokenizer::tokenize_number(Source<Unicode> & src, Token_string & tos)
    // real 'R' real   // magnitude + angle in radian
 
 APL_Float real_flt   = 0.0;   // always valid
-APL_Integer real_int = 0.0;   // valid if real_floating is true
-const bool real_valid = tokenize_real(src, real_flt, real_int);
-bool real_floating = real_flt != real_int;
-
+APL_Integer real_int = 0;     // valid if need_float is false
+bool need_float = false;
+const bool real_valid = tokenize_real(src, need_float, real_flt, real_int);
    if (!real_valid)
       {
         rest_2 = src.rest();
         throw_parse_error(E_BAD_NUMBER, LOC, loc);
       }
-   else if (src.rest() && *src == UNI_ASCII_J)
+
+   if (src.rest() && *src == UNI_ASCII_J)
       {
         ++src;   // skip 'J'
 
-        APL_Float   imag_flt;   // always valid
-        APL_Integer imag_int;   // valid if real_floating is true
-        const bool imag_valid =  tokenize_real(src, imag_flt, imag_int);
+        APL_Float   imag_flt = 0.0;   // always valid
+        APL_Integer imag_int = 0;     // valid if need_float is false
+        bool need_float = false;
+        const bool imag_valid = tokenize_real(src, need_float,
+                                              imag_flt, imag_int);
 
         if (!imag_valid)
            {
-             --src;
-             if (real_floating)   tos.append(Token(TOK_REAL,    real_flt));
-             else                 tos.append(Token(TOK_INTEGER, real_int));
+             --src;   // undo skip 'J'
+             if (need_float)
+                {
+                  tos.append(Token(TOK_REAL,    real_flt));
+                  Log(LOG_tokenize)
+                     CERR << "  tokenize_number: real " << real_flt << endl;
+                }
+             else
+                {
+                  tos.append(Token(TOK_INTEGER, real_int));
+                  Log(LOG_tokenize)
+                     CERR << "  tokenize_number: integer " << real_int << endl;
+                }
              return;
            }
 
         tos.append(Token(TOK_COMPLEX, real_flt, imag_flt));
+        Log(LOG_tokenize)   CERR << "  tokenize_number: complex "
+                                 << real_flt << "J" << imag_flt << endl;
       }
    else if (src.rest() && *src == UNI_ASCII_D)
       {
         ++src;   // skip 'D'
 
-        APL_Float   degrees_flt;  // always valid
-        APL_Integer degrees_int;  // valid if imag_floating is true
-        const bool imag_valid = tokenize_real(src, degrees_flt, degrees_int);
+        APL_Float   degrees_flt = 0;  // always valid
+        APL_Integer degrees_int = 0;  // valid if imag_floating is true
+        bool need_float = false;
+        const bool imag_valid = tokenize_real(src, need_float,
+                                              degrees_flt, degrees_int);
 
         if (!imag_valid)
            {
-             --src;
-             if (real_floating)   tos.append(Token(TOK_REAL,    real_flt));
-             else                 tos.append(Token(TOK_INTEGER, real_int));
+             --src;   // undo skip 'D'
+             if (need_float)
+                {
+                  tos.append(Token(TOK_REAL,    real_flt));
+                  Log(LOG_tokenize)
+                     CERR << "  tokenize_number: real " << real_flt << endl;
+                }
+             else
+                {
+                  tos.append(Token(TOK_INTEGER, real_int));
+                  Log(LOG_tokenize)
+                     CERR << "  tokenize_number: integer " << real_int << endl;
+                }
              return;
            }
 
         // real_flt is the magnitude and the angle is in degrees.
         //
-        APL_Float real = cos(M_PI*degrees_flt / 180);
-        APL_Float imag = sin(M_PI*degrees_flt / 180);
-        tos.append(Token(TOK_COMPLEX, real_flt*real, real_flt*imag));
+        APL_Float real = real_flt * cos(M_PI*degrees_flt / 180);
+        APL_Float imag = real_flt * sin(M_PI*degrees_flt / 180);
+        tos.append(Token(TOK_COMPLEX, real, imag));
+        Log(LOG_tokenize)   CERR << "  tokenize_number: complex " << real
+                                 << "J" << imag << endl;
       }
    else if (src.rest() && *src == UNI_ASCII_R)
       {
@@ -568,181 +630,262 @@ bool real_floating = real_flt != real_int;
 
         APL_Float radian_flt;     // always valid
         APL_Integer radian_int;   // valid if imag_floating is true
-        const bool imag_valid = tokenize_real(src, radian_flt, radian_int);
+        bool need_float = false;
+        const bool imag_valid = tokenize_real(src, need_float,
+                                              radian_flt, radian_int);
 
         if (!imag_valid)
            {
-             --src;
-             if (real_floating)   tos.append(Token(TOK_REAL,    real_flt));
-             else                 tos.append(Token(TOK_INTEGER, real_int));
+             --src;   // undo skip 'R'
+             if (need_float)
+                {
+                  tos.append(Token(TOK_REAL,    real_flt));
+                  Log(LOG_tokenize)
+                     CERR << "  tokenize_number: real " << real_flt << endl;
+                }
+             else
+                {
+                  tos.append(Token(TOK_INTEGER, real_int));
+                  Log(LOG_tokenize)
+                     CERR << "  tokenize_number: integer " << real_int << endl;
+                }
              return;
            }
 
         // real_flt is the magnitude and the angle is in radian.
         //
-        APL_Float real = cos(radian_flt);
-        APL_Float imag = sin(radian_flt);
-        tos.append(Token(TOK_COMPLEX, real_flt*real, real_flt*imag));
+        APL_Float real = real_flt * cos(radian_flt);
+        APL_Float imag = real_flt * sin(radian_flt);
+        tos.append(Token(TOK_COMPLEX, real, imag));
+        Log(LOG_tokenize)   CERR << "  tokenize_number: complex " << real
+                                 << "J" << imag << endl;
       }
    else 
      {
-         if (real_floating)   tos.append(Token(TOK_REAL,    real_flt));
-         else                 tos.append(Token(TOK_INTEGER, real_int));
+       if (need_float)
+          {
+            tos.append(Token(TOK_REAL,    real_flt));
+            Log(LOG_tokenize)
+               CERR << "  tokenize_number: real " << real_flt << endl;
+          }
+       else
+          {
+            tos.append(Token(TOK_INTEGER, real_int));
+            Log(LOG_tokenize)
+               CERR << "  tokenize_number: integer " << real_int << endl;
+          }
       }
 }
 //-----------------------------------------------------------------------------
 bool
-Tokenizer::tokenize_real(Source<Unicode> &src, APL_Float & flt_val,
-                         APL_Integer & int_val)
+Tokenizer::tokenize_real(Source<Unicode> & src, bool & need_float,
+                         APL_Float & flt_val, APL_Integer & int_val)
 {
-bool dot_seen = false;
-int E_pos = 0;   // position of 'e' in utf
-int minus_pos = 0;   // position of (exponent-) ¯ in utf
-bool digit_seen = false;
+   flt_val = 0.0;
+   int_val = 0;
+   need_float = false;
+   
+UTF8_string digits;
+UTF8_string fract_part;
+UTF8_string expo_part;
+bool negative = false;
+bool expo_negative = false;
+bool skipped_0 = false;
 
-   // copy chars from src to utf
+   // leading sign
    //
-UTF8_string utf;
-
-   // a '¯' can occur in the mantissa and in the exponent. To make a '¯'
-   // unique, we handle a leading (i.e. mantissa) '¯' beforehand.
-   //
-   if (*src == UNI_OVERBAR)
+   if (src.rest() && *src == UNI_OVERBAR)
       {
-        utf.append('-');
+        negative = true;
         ++src;
       }
 
-   for (bool goon = true; goon && src.rest();)
-      {
-        const Unicode cc = src.get();
-        switch(cc)
-           {
-             case UNI_ASCII_0 ... UNI_ASCII_9:
-                  utf.append(cc);
-                  digit_seen = true;
-                  break;
-
-             case UNI_ASCII_FULLSTOP:
-                  if (dot_seen || E_pos)   return false;   // error
-                  utf.append('.');
-                  dot_seen = true;
-                  break;
-
-             case UNI_ASCII_E:
-             case UNI_ASCII_e:
-                  if (E_pos)   return false;   // a second e or E
-                  utf.append('e');
-                  E_pos = utf.size();
-                  break;
-
-             case UNI_OVERBAR:
-                  if (!E_pos)      return false;   // ¯ without E
-                  if (minus_pos)   return false;   // a second ¯ in exponent
-
-                  utf.append('-');
-                  minus_pos = utf.size();
-                  break;
-
-             default: --src;   goon = false;
-           }
-      }
-
-   // handle E or E¯ without exponent digits
-   //
-   if (utf.size() == E_pos)       return false;
-   if (utf.size() == minus_pos)   return false;
-   {
-     const int real_cnt = sscanf(utf.c_str(), "%lg", &flt_val);
-     long long int_val1;
-                          sscanf(utf.c_str(), "%lld", &int_val1);
-     int_val = int_val1;
-     if (digit_seen && (real_cnt > 0))   return true;   // valid real number
-   }
-
-
-   // invalid number: restore src
-   //
-   return false;
-}
-//-----------------------------------------------------------------------------
-#if 0
-int
-Tokenizer::scan_real(const char * strg, double & result,
-                     int E_pos, int minus_pos)
-{
-   // this function (which is currently not used) shall become a locale
-   // independent sscanf. Currently it gives slightly different results
-   // than sscanf (rounding errors?) and therefore breaks a few testcases.
-   // TODO!
-   //
-const bool negative = *strg == '-';
-int p = 0;
-int expo_f = 0;
-int expo_E = 0;
-
-   if (negative)   ++p;   // leading -
-
-   result = 0;
-
-   // the caller has filtered everything but 0-9 . - and e.
+   // skip leading zeros in integer part
    //
+   while (src.rest() && *src == '0')   { ++src;   skipped_0 = true; }
 
    // integer part
    //
-   for (;;)
+   while (src.rest() && *src >= '0' && *src <= '9')
       {
-        const char cc = strg[p++];
-        if (cc == 0)     goto done;
-        if (cc == '.')   goto fract;
-        if (cc == 'e')   goto expo;
-        result *= 10;
-        result += cc - '0';
+        digits.append(src.get());
       }
 
-fract:
-   for (;;)
+   // fractional part
+   //
+   if (src.rest() &&  *src == UNI_ASCII_FULLSTOP)   // fract part present
       {
-        const char cc = strg[p++];
-        if (cc == 0)     goto done;
-        if (cc == 'e')   goto expo;
-        result *= 10;
-        result += cc - '0';
-        --expo_f;
+        ++src;
+        while (src.rest() && *src >= '0' && *src <= '9')
+           {
+             fract_part.append(src.get());
+           }
       }
 
-expo:
-   {
-     if (minus_pos)   p = minus_pos;
-     else             p = E_pos;
+   // require at least one integer or fractional digit
+   //
+   if (digits.size() == 0      &&
+       fract_part.size() == 0  &&
+       !skipped_0)   return false;
 
-   for (;;)
+   // exponent part
+   //
+   if (src.rest() > 1 &&
+       (*src == UNI_ASCII_E || *src == UNI_ASCII_e))   // maybe exponent
       {
-        const char cc = strg[p++];
-        if (cc == 0)     break;
-        expo_E *= 10;
-        expo_E += cc - '0';
+        if ((src[1] == UNI_OVERBAR && src[2] >= '0' && src[2] <= '9') ||
+            (src[1] >= '0' && src[1] <= '9'))
+           {
+             need_float = true;
+             ++src;   // skip e/E
+             if (src.rest() && *src == UNI_OVERBAR)
+           {
+                  ++src;   // skip ¯
+                  expo_negative = true;
+                }
+
+             while (src.rest() && *src >= '0' && *src <= '9')
+                {
+                  expo_part.append(src.get());
+                }
+           }
       }
-   }
 
-done:
-     if (minus_pos)   expo_E = -expo_E;
-     expo_E += expo_f;
+int expo = 0;
+   loop(d, expo_part.size())   expo = 10*expo + (expo_part[d] - '0');
+   if (expo_negative)   expo = - expo;
+   expo += digits.size();
 
-     if (expo_E)
-        {
-          if (expo_E <= -308)
-             {
-               result = 0;
-               return 1;
-             }
-          result *= exp10(expo_E);
-        }
+   digits.append(fract_part);
 
-   if (negative)   result = -result;
-   return 1;
+   // at this point, digits is the fractional part ff... of 0.ff... ×10⋆expo
+   // discard leading fractional 0s and adjust expo accordingly
+   //
+   while (digits.size() && digits[0] == '0')
+      {
+        digits.drop_leading(1);   // discard '0'
+        --expo;
+      }
+
+   // at this point, digits is the fractional part ff... of 0.ff... ×10⋆expo
+   // from 0.1000... to 0.9999... Discard digits beyond integer precision.
+   //
+   // The largest 64-bit integer is 0x7FFFFFFFFFFFFFFF = 9223372036854775807
+   // which has 19 decimal digits. Discard all after first 20 digits and
+   // round according to the last digit.
+   //
+   if (digits.size() > 20)   digits.shrink(20);
+
+   if (digits.size() == 20)
+      {
+        bool carry = false;
+        if (digits.last() >= '5')   // round up
+           {
+             carry = true;
+             for (int j = digits.size() - 2; j >= 0; --j)
+                 {
+                   if (digits[j] == '9')   // propagate carry
+                      {
+                         digits[j] = '0';
+                      }
+                   else                    // eat carry
+                      {
+                         ++digits[j];
+                         carry = false;
+                         break;
+                      }
+                 }
+           }
+
+
+        // if carry survived then digits were 0.999... which was then rounded
+        // up to 1.000...
+        //
+        if (carry)
+           {
+             ++expo;   // 1.0 → 0.1
+             digits.shrink(1);
+             digits[0] = '1';
+           }
+        else
+           {
+             digits.pop();   // discard 20th digit.
+           }
+      }
+
+   // special cases: all digits 0 or very small number
+   //
+   if (digits.size() == 0)   return true;   // all digits were 0
+   if (expo <= -307)         return true;   // very small number
+
+   Assert(digits.size() <= 19);
+
+   if (expo > 19)   need_float = true;   /// large integer
+
+   // special case: integer between 9223372036854775807 and 9999999999999999999
+   //
+   if (expo == 19 && digits[0] == '9')
+      {
+        const char * maxint = "9223372036854775807";
+        loop(j, digits.size())
+            {
+               if (digits[j] < maxint[j])   break;
+               if (digits[j] == maxint[j])   continue;
+               need_float = true;
+               break;
+            }
+      }
+
+   if (digits.size() > expo)   need_float = true;
+
+   if (need_float)
+      {
+        if (digits.size() > 17)   digits.shrink(17);
+
+       int64_t v = 0;
+        loop(j, digits.size())
+          {
+            v = 10*v + (digits[j] - '0');
+            --expo;
+          }
+
+        flt_val = v;
+        if (expo > 0)
+           {
+             if (expo >= 256)   { expo -= 256;   flt_val *= 1E256; }
+             if (expo >= 128)   { expo -= 128;   flt_val *= 1E128; }
+             if (expo >=  64)   { expo -=  64;   flt_val *= 1E64;  }
+             if (expo >=  32)   { expo -=  32;   flt_val *= 1E32;  }
+             if (expo >=  16)   { expo -=  16;   flt_val *= 1E16;  }
+             if (expo >=   8)   { expo -=   8;   flt_val *= 1E8;   }
+             if (expo >=   4)   { expo -=   4;   flt_val *= 1E4;   }
+             if (expo >=   2)   { expo -=   2;   flt_val *= 1E2;   }
+             if (expo >=   1)   { expo -=   1;   flt_val *= 1E1;   }
+           }
+        else if (expo < 0)
+           {
+             if (expo <= -256)   { expo += 256;   flt_val /= 1E256; }
+             if (expo <= -128)   { expo += 128;   flt_val /= 1E128; }
+             if (expo <=  -64)   { expo +=  64;   flt_val /= 1E64;  }
+             if (expo <=  -32)   { expo +=  32;   flt_val /= 1E32;  }
+             if (expo <=  -16)   { expo +=  16;   flt_val /= 1E16;  }
+             if (expo <=   -8)   { expo +=   8;   flt_val /= 1E8;   }
+             if (expo <=   -4)   { expo +=   4;   flt_val /= 1E4;   }
+             if (expo <=   -2)   { expo +=   2;   flt_val /= 1E2;   }
+             if (expo <=   -1)   { expo +=   1;   flt_val /= 1E1;   }
+           }
+        if (negative)   flt_val = - flt_val;
+      }
+   else
+      {
+        int_val = 0;
+        loop(j, digits.size())   int_val = 10*int_val + (digits[j] - '0');
+        if (negative)   int_val = - int_val;
+        flt_val = int_val;
+      }
+
+   return true;
 }
-#endif
 //-----------------------------------------------------------------------------
 void
 Tokenizer::tokenize_symbol(Source<Unicode> & src, Token_string & tos)
@@ -767,10 +910,10 @@ UCS_string symbol;
 
         while (src.rest() && *src <= UNI_ASCII_SPACE)   src.get();   // spaces
         UCS_string symbol1(symbol, 2, symbol.size() - 2);   // without S∆/T∆
-        Value_P AB(new Value(symbol1, LOC));
+        Value_P AB(symbol1, LOC);
         Function * ST = 0;
-        if (symbol[0] == UNI_ASCII_S) ST = &Quad_STOP::fun;
-        else                          ST = &Quad_TRACE::fun;
+        if (symbol[0] == UNI_ASCII_S) ST = Quad_STOP::fun;
+        else                          ST = Quad_TRACE::fun;
 
         const bool assigned = (src.rest() && *src == UNI_LEFT_ARROW);
         if (assigned)   // dyadic: AB ∆fun
