@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -34,11 +34,10 @@
 UCS_string::UCS_string(const char * cstring)
    : Simple_string<Unicode>(0, 0)
 {
-   items_valid = strlen(cstring);
-   items_allocated = items_valid + ADD_ALLOC;
-   items = new Unicode[items_allocated];
+const size_t len = strlen(cstring);
+   extend(len + 1);
 
-   for (int l = 0; l < items_valid; ++l)   items[l] = Unicode(cstring[l]);
+   while (items_valid < len)   items[items_valid++] = Unicode(*cstring++);
    items[items_valid] = UNI_ASCII_NUL;
 }
 //-----------------------------------------------------------------------------
@@ -394,7 +393,7 @@ UCS_string::remove_trailing_padchars()
 void
 UCS_string::remove_trailing_whitespaces()
 {
-   while (size() && (last() <= UNI_ASCII_SPACE))   pop();
+   while (size() && Avec::is_white(last()))   pop();
 }
 //-----------------------------------------------------------------------------
 void
@@ -403,13 +402,31 @@ UCS_string::remove_leading_whitespaces()
 int count = 0;
    loop(s, size())
       {
-        if ((*this)[s] <= UNI_ASCII_SPACE)   ++count;
+        if (Avec::is_white((*this)[s]))   ++count;
         else                                 break;
       }
 
    drop_leading(count);
 }
 //-----------------------------------------------------------------------------
+void
+UCS_string::split_ws(UCS_string & rest)
+{
+   remove_leading_and_trailing_whitespaces();
+
+   loop(clen, size())
+       {
+         if (Avec::is_white((*this)[clen]))   // whilespace: end of command
+            {
+              ShapeItem arg = clen;
+              while (arg < size() && Avec::is_white((*this)[arg]))   ++arg;
+              while (arg < size())   rest.append((*this)[arg++]);
+              shrink(clen);
+              return;
+            }
+       }
+}
+//-----------------------------------------------------------------------------
 /// constructor
 UCS_string::UCS_string(const Value & value)
    : Simple_string<Unicode>(0, 0)
@@ -489,6 +506,13 @@ const int start_positions = 1 + size() - sub.size();
 }
 //-----------------------------------------------------------------------------
 bool 
+UCS_string::has_black() const
+{
+   loop(s, size())   if (!Avec::is_white((*this)[s]))   return true;
+   return false;
+}
+//-----------------------------------------------------------------------------
+bool 
 UCS_string::starts_with(const char * prefix) const
 {
    loop(s, size())
@@ -506,6 +530,16 @@ UCS_string::starts_with(const char * prefix) const
 }
 //-----------------------------------------------------------------------------
 bool 
+UCS_string::starts_with(const UCS_string & prefix) const
+{
+   if (prefix.size() > size())   return false;
+
+   loop(p, prefix.size())   if ((*this)[p] != prefix[p])   return false;
+
+   return true;
+}
+//-----------------------------------------------------------------------------
+bool 
 UCS_string::starts_iwith(const char * prefix) const
 {
    loop(s, size())
@@ -554,6 +588,15 @@ UCS_string ret;
    return ret;
 }
 //-----------------------------------------------------------------------------
+void
+UCS_string::map_pad()
+{
+   loop(s, size())
+      {
+        if (is_iPAD_char((*this)[s]))   (*this)[s] = UNI_ASCII_SPACE;
+      }
+}
+//-----------------------------------------------------------------------------
 UCS_string
 UCS_string::remove_pad() const
 {
@@ -653,6 +696,19 @@ char cc[40];
 }
 //-----------------------------------------------------------------------------
 void
+UCS_string::append_hex(ShapeItem num, bool uppercase)
+{
+const char * format = uppercase ? "%llX" : "%llx";
+char cc[40];
+   snprintf(cc, sizeof(cc) - 1, format, (long long)num);
+   loop(c, sizeof(cc))
+      {
+        if (cc[c])   append(Unicode(cc[c]));
+        else         break;
+      }
+}
+//-----------------------------------------------------------------------------
+void
 UCS_string::append_float(double num)
 {
 char cc[60];
@@ -707,7 +763,7 @@ int ret = 0;
       {
         const Unicode uni = (*this)[s];
 
-        if (!ret && uni <= UNI_ASCII_SPACE)   continue;   // leading whitespace
+        if (!ret && Avec::is_white(uni))   continue;   // leading whitespace
 
         if (uni < UNI_ASCII_0)                break;      // non-digit
         if (uni > UNI_ASCII_9)                break;      // non-digit
@@ -784,6 +840,18 @@ UCS_string::lexical_before(const UCS_string other) const
    return other.size() > size();
 }
 //-----------------------------------------------------------------------------
+ostream &
+UCS_string::dump(ostream & out) const
+{
+   out << right << hex << uppercase << setfill('0');
+   loop(s, size())
+      {
+        out << " U+" << setw(4) << (int)((*this)[s]);
+      }
+
+   return out << left << dec << nouppercase << setfill(' ');
+}
+//-----------------------------------------------------------------------------
 UCS_string
 UCS_string::from_int(int64_t value)
 {
@@ -1016,6 +1084,26 @@ UCS_string::sort_names(const UCS_string ** names, int count)
 }
 //----------------------------------------------------------------------------
 void
+UCS_string::sort_names(vector<UCS_string> names)
+{
+   if (names.size() < 2)   return;
+
+   for (int h = 0; h < (names.size() - 1); ++h)
+       {
+         // find smallest above (including) h
+         //
+         int smallest = h;
+         for (int j = h + 1; j < names.size(); ++j)
+             {
+               if (names[smallest].compare(names[j]) == COMP_GT)
+                  {
+                    names[smallest].swap(names[j]);
+                  }
+             }
+       }
+}
+//----------------------------------------------------------------------------
+void
 UCS_string::compute_column_width(vector<int> & result,
                                  const UCS_string ** names, int name_count,
                                  int tab_size, int quad_PW)
