@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -46,10 +46,10 @@
    0000 0800-0000 FFFF   1110xxxx 10xxxxxx 10xxxxxx
    0001 0000-001F FFFF   11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
    0020 0000-03FF FFFF   111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
-   0400 0000-7FFF FFFF   1111110x 10xxxxxx ... 10xxxxxx
+   0400 0000-7FFF FFFF   1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
 
-   In order to be more general, we follow RFC 2279 and allow 5-byte and 6-byte
-   encodings
+   In order to be more general, we follow RFC 2279 and also allow 5-byte and
+   6-byte encodings
 */
 
 //-----------------------------------------------------------------------------
@@ -133,7 +133,7 @@ UTF8_string::UTF8_string(const UCS_string & ucs)
 }
 //-----------------------------------------------------------------------------
 UTF8_string::UTF8_string(const Value & value)
-   : Simple_string((int)value.element_count(), (UTF8) 0)
+   : Simple_string<UTF8>((int)value.element_count(), (UTF8) 0)
 {
    loop(v, value.element_count())
        (*this)[v] = value.get_ravel(v).get_char_value() & 0xFF;
@@ -245,7 +245,14 @@ bool
 UTF8_string::starts_with(const char * path) const
 {
    if (path == 0)   return false;   // no path provided
-   return strncmp(path, (const char *)items, strlen(path)) == 0;
+
+const int path_len = strlen(path);
+   if (path_len > size())   return false;   // path_len longer than this string
+
+   // can't use strncmp() because this string may not be 0-terminated
+   //
+   loop(p, path_len)   if (items[p] != path[p])   return false;
+   return true;
 }
 //-----------------------------------------------------------------------------
 bool
@@ -256,7 +263,10 @@ UTF8_string::ends_with(const char * ext) const
 const int ext_len = strlen(ext);
    if (ext_len > size())   return false;   // ext longer than this string
 
-   return strcmp(ext, (const char *)items + size() - ext_len) == 0;
+   // can't use strncmp() because this string may not be 0-terminated
+   //
+   loop(e, ext_len)   if (items[size() - ext_len + e] != ext[e])   return false;
+   return true;
 }
 //-----------------------------------------------------------------------------
 ostream &
