@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -197,6 +197,7 @@ int tos_idx = 0;
 Symbol * symbols[12];
    for (; tos_idx < 12; ++tos_idx)
       {
+         if (tos_idx >= tos.size())                   break;
          if (tos[tos_idx].get_tag() == TOK_SEMICOL)   break;
          if (tos[tos_idx].get_Class() == TC_SYMBOL)
             symbols[sym_count++] = tos[tos_idx].get_sym_ptr();
@@ -247,7 +248,7 @@ Fun_signature signature = SIG_NONE;
      function_name = sym_FUN->get_name();
    }
 
-   for (;;)
+   while (tos_idx < (tos.size() - 1))
       {
         if (tos[tos_idx++].get_tag() != TOK_SEMICOL)
            {
@@ -255,7 +256,12 @@ Fun_signature signature = SIG_NONE;
              return;
            }
 
-        if  (tos_idx >= tos.size())   break;   // local vars done
+
+        if (tos_idx == tos.size())
+           {
+             error_cause = "trailing ; in function header";
+             return;
+           }
 
         const TokenTag tag = tos[tos_idx].get_tag();
         if (tag != TOK_SYMBOL && tag != TOK_Quad_CT
@@ -471,7 +477,7 @@ UserFunction::UserFunction(const UCS_string txt,
                            int & error_line, const char * & error_cause,
                            bool keep_existing, const char * loc,
                            const UTF8_string & _creator)
-  : Function(ID_USER_SYMBOL, TOK_FUN2),
+  : Function(ID::USER_SYMBOL, TOK_FUN2),
     Executable(txt, loc),
     header(txt),
     creator(_creator)
@@ -537,7 +543,7 @@ Function * old_function = header.FUN()->get_function();
 //-----------------------------------------------------------------------------
 UserFunction::UserFunction(Fun_signature sig, const UCS_string & fname,
                            const UCS_string & text, const Token_string & bdy)
-  : Function(ID_USER_SYMBOL, TOK_FUN0),
+  : Function(ID::USER_SYMBOL, TOK_FUN0),
     Executable(sig, fname, text, LOC),
     header(sig, fname),
     creator(UNI_LAMBDA)
@@ -601,10 +607,10 @@ UserFunction::eval_B(Value_P B)
              << Token(TOK_APL_VALUE1, B) << ")" << endl;
       }
 
-   Workspace::push_SI(this, LOC);
-
    if (header.LO())    SYNTAX_ERROR;   // defined as operator
 
+   Workspace::push_SI(this, LOC);
+
    if (header.B())   header.B()->push_value(B);
    if (header.X())   header.X()->push();
    if (header.A())   header.A()->push();
@@ -623,10 +629,10 @@ UserFunction::eval_XB(Value_P X, Value_P B)
              << Token(TOK_APL_VALUE1, B) << ")" << endl;
       }
 
-   Workspace::push_SI(this, LOC);
-
    if (header.LO())    SYNTAX_ERROR;   // defined as operator
 
+   Workspace::push_SI(this, LOC);
+
    if (header.B())   header.B()->push_value(B);
    if (header.X())   header.X()->push_value(X);
    if (header.A())   header.A()->push();
@@ -646,10 +652,10 @@ UserFunction::eval_AB(Value_P A, Value_P B)
              << Token(TOK_APL_VALUE1, B) << ")" << endl;
       }
 
-   Workspace::push_SI(this, LOC);
-
    if (header.LO())    SYNTAX_ERROR;   // defined as operator
 
+   Workspace::push_SI(this, LOC);
+
    if (header.B())   header.B()->push_value(B);
    if (header.X())   header.X()->push();
    if (header.A())   header.A()->push_value(A);
@@ -669,10 +675,10 @@ UserFunction::eval_AXB(Value_P A, Value_P X, Value_P B)
              << Token(TOK_APL_VALUE1, B) << ")" << endl;
       }
 
-   Workspace::push_SI(this, LOC);
-
    if (header.LO())    SYNTAX_ERROR;   // defined as operator
 
+   Workspace::push_SI(this, LOC);
+
    if (header.B())   header.B()->push_value(B);
    if (header.X())   header.X()->push_value(X);
    if (header.A())   header.A()->push_value(A);
@@ -692,10 +698,10 @@ UserFunction::eval_LB(Token & LO, Value_P B)
              << Token(TOK_APL_VALUE1, B) << ")" << endl;
       }
 
-   Workspace::push_SI(this, LOC);
-
    if (header.RO())    SYNTAX_ERROR;   // dyadic operator called monadically
 
+   Workspace::push_SI(this, LOC);
+
                 header.B()->push_value(B);
    if (header.X())   header.X()->push();
    if (header.A())   header.A()->push();
@@ -718,10 +724,10 @@ UserFunction::eval_LXB(Token & LO, Value_P X, Value_P B)
              << Token(TOK_APL_VALUE1, B) << ")" << endl;
       }
 
-   Workspace::push_SI(this, LOC);
-
    if (header.RO())    SYNTAX_ERROR;   // dyadic operator called monadically
 
+   Workspace::push_SI(this, LOC);
+
                 header.B()->push_value(B);
    if (header.X())   header.X()->push_value(X);
    if (header.A())   header.A()->push();
@@ -744,11 +750,12 @@ UserFunction::eval_ALB(Value_P A, Token & LO, Value_P B)
              << Token(TOK_APL_VALUE1, B) << ")" << endl;
       }
 
+   if (!header.A())    SYNTAX_ERROR;   // monadic function called dyadically
+   if (header.RO())    SYNTAX_ERROR;   // defined as dyadic operator
+
    Workspace::push_SI(this, LOC);
 
-   if (!header.A())    SYNTAX_ERROR;   // monadic function called dyadically
    if (header.X())   header.X()->push();
-   if (header.RO())    SYNTAX_ERROR;   // defined as dyadic operator
 
    header.B()->push_value(B);
    header.A()->push_value(A);
@@ -772,11 +779,11 @@ UserFunction::eval_ALXB(Value_P A, Token & LO, Value_P X, Value_P B)
              << Token(TOK_APL_VALUE1, B) << ")" << endl;
       }
 
-   Workspace::push_SI(this, LOC);
-
    if (!header.A())    SYNTAX_ERROR;   // monadic function called dyadically
    if (header.RO())    SYNTAX_ERROR;   // defined as dyadic operator
 
+   Workspace::push_SI(this, LOC);
+
    header.B()->push_value(B);
    if (header.X())   header.X()->push();
    header.A()->push_value(A);
@@ -800,10 +807,10 @@ UserFunction::eval_LRB(Token & LO, Token & RO, Value_P B)
              << Token(TOK_APL_VALUE1, B) << ")" << endl;
       }
 
-   Workspace::push_SI(this, LOC);
-
    if (!header.RO())    SYNTAX_ERROR;   // not defined as dyadic operator
 
+   Workspace::push_SI(this, LOC);
+
                  header.B()->push_value(B);
    if (header.X())    header.X()->push();
    if (header.A())    header.A()->push();
@@ -830,10 +837,10 @@ UserFunction::eval_LRXB(Token & LO, Token & RO, Value_P X, Value_P B)
              << Token(TOK_APL_VALUE1, B) << ")" << endl;
       }
 
-   Workspace::push_SI(this, LOC);
-
    if (!header.RO())    SYNTAX_ERROR;   // not defined as dyadic operator
 
+   Workspace::push_SI(this, LOC);
+
                 header.B()->push_value(B);
    if (header.X())   header.X()->push_value(X);
    if (header.A())   header.A()->push();
@@ -859,11 +866,12 @@ UserFunction::eval_ALRB(Value_P A, Token & LO, Token & RO, Value_P B)
              << Token(TOK_APL_VALUE1, B) << ")" << endl;
       }
 
+   if (!header.A())    SYNTAX_ERROR;    // defined monadic called dyadic
+   if (!header.RO())    SYNTAX_ERROR;   // defined monadic op called dyadic
+
    Workspace::push_SI(this, LOC);
 
-   if (!header.A())    SYNTAX_ERROR;    // defined monadic called dyadic
    if (header.X())      header.X()->push();
-   if (!header.RO())    SYNTAX_ERROR;   // defined monadic op called dyadic
 
                  header.B()->push_value(B);
                  header.A()->push_value(A);
@@ -891,11 +899,11 @@ UserFunction::eval_ALRXB(Value_P A, Token & LO, Token & RO,
              << Token(TOK_APL_VALUE1, B) << ")" << endl;
       }
 
-   Workspace::push_SI(this, LOC);
-
    if (!header.A())    SYNTAX_ERROR;    // defined monadic called dyadic
    if (!header.RO())    SYNTAX_ERROR;   // defined monadic op called dyadic
 
+   Workspace::push_SI(this, LOC);
+
                  header.B()->push_value(B);
    if (header.X())    header.X()->push_value(X);
                  header.A()->push_value(A);
@@ -1132,11 +1140,11 @@ int error_line = -1;
 }
 //-----------------------------------------------------------------------------
 Function_PC
-UserFunction::pc_for_line(int l) const
+UserFunction::pc_for_line(Function_Line line) const
 {
-   if (l < 1)                     return Function_PC(body.size() - 1);
-   if (l >= line_starts.size())   return Function_PC(body.size() - 1);
-   return line_starts[l];
+   if (line < 1)                     return Function_PC(body.size() - 1);
+   if (line >= line_starts.size())   return Function_PC(body.size() - 1);
+   return line_starts[line];
 }
 //-----------------------------------------------------------------------------
 UserFunction *
@@ -1250,12 +1258,6 @@ UCS_string ind(indent, UNI_ASCII_SPACE);
        << ind << "Body: " << body << endl;
 }
 //-----------------------------------------------------------------------------
-const UCS_string &
-UserFunction::get_name() const
-{
-   return header.get_name();
-}
-//-----------------------------------------------------------------------------
 UCS_string
 UserFunction::get_name_and_line(Function_PC pc) const
 {
@@ -1302,6 +1304,37 @@ UCS_string ucs;
    return ucs;
 }
 //-----------------------------------------------------------------------------
+void
+UserFunction::adjust_line_starts()
+{
+   // this function is called from Executable::setup_lambdas() just before
+   // Parser::remove_void_token(body) in order to adjust line_starts
+   //
+DynArray(ShapeItem, gaps, line_starts.size());   // count TOK_VOID in every line
+   loop(ls, line_starts.size())
+      {
+         gaps[ls] = 0;
+         if (ls == 0)   continue;   // function header (has no TOK_VOID)
+
+         const ShapeItem from = line_starts[ls];
+         ShapeItem to = body.size();    // end of function (for last line)
+         if (ls < (line_starts.size() - 1))   to = line_starts[ls + 1];
+
+        for (ShapeItem b = from; b < to; ++b)
+            {
+             if (body[b].get_tag() == TOK_VOID)   ++gaps[ls];
+            }
+
+      }
+
+int total_gaps = 0;
+   loop(ls, line_starts.size())
+       {
+          line_starts[ls] = Function_PC(line_starts[ls] - total_gaps);
+          total_gaps += gaps[ls];
+       }
+}
+//-----------------------------------------------------------------------------
 ostream &
 UserFunction::print_val_or_fun(ostream & out, Token & tok)
 {
