@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2014  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -28,19 +28,28 @@
 
 #include "buildtag.hh"
 static const char * build_tag[] = { BUILDTAG, 0 };
+extern const char * configure_args;
 
+#include "Bif_OPER2_INNER.hh"
+#include "Bif_OPER2_OUTER.hh"
 #include "Common.hh"
-#include "Input.hh"
 #include "IO_Files.hh"
 #include "InputFile.hh"
 #include "LibPaths.hh"
 #include "makefile.h"
 #include "Output.hh"
+#include "ScalarFunction.hh"
 #include "UserPreferences.hh"
 #include "Value.icc"
 
 UserPreferences uprefs;
 
+/// CYGWIN defines _B
+
+#ifdef _B
+#undef _B
+#endif
+
 //-----------------------------------------------------------------------------
 void
 UserPreferences::usage(const char * prog)
@@ -74,18 +83,20 @@ char cc[4000];
    snprintf(cc, sizeof(cc),
 "    --cfg                show ./configure options used and exit\n"
 "    --noCIN              do not echo input(for scripting)\n"
-"    --rawCIN             do not use the readline lib for input\n"
+"    --echoCIN            echo (final) input to COUT\n"
+"    --rawCIN             do not emit escape sequences\n"
 "    --[no]Color          start with ]XTERM ON [OFF])\n"
-"    --noCONT             do not load CONTINUE workspace on startup)\n"
+"    --noCONT             do not )LOAD CONTINUE workspace on startup)\n"
 "    --emacs              run in (classical) emacs mode\n"
 "    --emacs_arg arg      run in emacs mode with argument arg\n"
 "    --gpl                show license (GPL) and exit\n"
+"    -L wsname            )LOAD wsname on startup\n"
 "    --LX expr            execute APL expression expr first\n"
 "    -p N                 use profile N in preferences files\n"
 "    --par proc           use processor parent ID proc (default: no parent)\n"
+"    -q, --silent         do not print the welcome banner\n"
 "    -s, --script         same as --silent --noCIN --noCONT --noColor\n"
 "    --safe               safe mode (no shared vars, no native functions)\n"
-"    --silent             do not show the welcome message\n"
 "    --show_bin_dir       show binary directory and exit\n"
 "    --show_doc_dir       show documentation directory and exit\n"
 "    --show_etc_dir       show system configuration directory and exit\n"
@@ -118,7 +129,7 @@ UserPreferences::show_GPL(ostream & out)
 "    This program is GNU APL, a free implementation of the\n"
 "    ISO/IEC Standard 13751, \"Programming Language APL, Extended\"\n"
 "\n"
-"    Copyright (C) 2008-2014  Dr. Jürgen Sauermann\n"
+"    Copyright (C) 2008-2015  Dr. Jürgen Sauermann\n"
 "\n"
 "    This program is free software: you can redistribute it and/or modify\n"
 "    it under the terms of the GNU General Public License as published by\n"
@@ -142,7 +153,7 @@ UserPreferences::is_APL_script(const char * filename)
 {
    /* according to man execve():
 
-       An interpreter script is  a  text  file  that  has  execute  permission
+       An interpreter script is a text file that has execute permission
        enabled and whose first line is of the form:
 
            #! interpreter [optional-arg]
@@ -157,26 +168,53 @@ char buf[2];
 const size_t len = read(fd, buf, sizeof(buf));
    close(fd);
 
-   if (len != 2)   return false;
+   if (len != 2)        return false;
    if (buf[0] != '#')   return false;
    if (buf[1] != '!')   return false;
    return true;;
 }
 //-----------------------------------------------------------------------------
 bool
-UserPreferences::log_startup_wanted() const
+UserPreferences::parse_argv_1()
 {
-   for (int a = 2; a < expanded_argv.size(); ++a)
+bool log_startup = false;
+   for (int a = 1; a < expanded_argv.size(); )
        {
-         if (!strcmp(expanded_argv[a - 1], "-l") &&
-               atoi(expanded_argv[a]) == LID_startup)   return true;
+         const char * opt = expanded_argv[a++];
+         const char * val = (a < expanded_argv.size()) ? expanded_argv[a] : 0;
+
+         if (!strcmp(opt, "-l"))
+            {
+              ++a;
+              if (!val)
+                 {
+                   CERR << "-l without log level" << endl;
+                   exit(a);
+                 }
+
+              log_startup = (atoi(val) == LID_startup);
+              continue;
+            }
+
+         if (!strcmp(opt, "-p"))
+            {
+              ++a;
+              if (!val)
+                 {
+                   CERR << "-p without profile number" << endl;
+                   exit(a);
+                 }
+
+              user_profile = atoi(val);
+              continue;
+            }
        }
 
-   return false;
+   return log_startup;
 }
 //-----------------------------------------------------------------------------
 void
-UserPreferences::parse_argv(bool logit)
+UserPreferences::parse_argv_2(bool logit)
 {
    // execve() puts the script name (and optional args at the end of argv.
    // For example, argv might look like this:
@@ -258,7 +296,7 @@ UserPreferences::parse_argv(bool logit)
                    exit(a);
                  }
               InputFile fam(UTF8_string(val), 0, false,
-                                          !do_not_echo, true, false);
+                                        !do_not_echo, true, false);
               InputFile::files_todo.push_back(fam);
               continue;
             }
@@ -308,6 +346,19 @@ UserPreferences::parse_argv(bool logit)
               continue;
             }
 
+         if (!strcmp(opt, "-L"))
+            {
+              ++a;
+              if (!val)
+                 {
+                   CERR << "-L without workspace name" << endl;
+                   exit(a);
+                 }
+
+              initial_workspace = UTF8_string(val);
+              continue;
+            }
+
          if (!strcmp(opt, "--LX"))
             {
               ++a;
@@ -321,6 +372,14 @@ UserPreferences::parse_argv(bool logit)
               continue;
             }
 
+         if (!strcmp(opt, "--echoCIN"))
+            {
+              echo_CIN = true;
+              do_not_echo = true;
+              do_Color = false;
+              continue;
+            }
+
          if (!strcmp(opt, "--noCIN"))
             {
               do_not_echo = true;
@@ -349,14 +408,7 @@ UserPreferences::parse_argv(bool logit)
          if (!strcmp(opt, "-p"))
             {
               ++a;
-              if (!val)
-                 {
-                   CERR << "-p without profile number" << endl;
-                   exit(a);
-                 }
-
-              user_profile = atoi(val);
-              continue;
+              continue;   // -p already handled in parse_argv_1()
             }
 
          if (!strcmp(opt, "--par"))
@@ -373,7 +425,7 @@ UserPreferences::parse_argv(bool logit)
 
          if (!strcmp(opt, "--rawCIN"))
             {
-              Input::use_readline = false;
+              raw_cin = true;
               continue;
             }
 
@@ -389,7 +441,6 @@ UserPreferences::parse_argv(bool logit)
               do_CONT = false;               // --noCONT
               do_not_echo = true;            // -noCIN
               do_Color = false;              // --noColor
-              Input::use_readline = false;   // --rawCIN
               silent = true;                 // --silent
               continue;
             }
@@ -434,6 +485,12 @@ UserPreferences::parse_argv(bool logit)
               exit(0);
             }
 
+         if (!strcmp(opt, "-q"))
+            {
+              silent = true;
+              continue;
+            }
+
          if (!strcmp(opt, "--silent"))
             {
               silent = true;
@@ -651,18 +708,17 @@ UserPreferences::show_version(ostream & out)
        {
          switch(bt - build_tag)
             {
-              case 0:  out << "  Project:        ";   break;
-              case 1:  out << "  Version / SVN:  ";   break;
-              case 2:  out << "  Build Date:     ";   break;
-              case 3:  out << "  Build OS:       ";   break;
-              case 4:  out << "  config.status:  ";   break;
-              default: out << "  [" << bt - build_tag << "] ";
+              case 0:  out << "    Project:        ";   break;
+              case 1:  out << "    Version / SVN:  ";   break;
+              case 2:  out << "    Build Date:     ";   break;
+              case 3:  out << "    Build OS:       ";   break;
+              case 4:  out << "    config.status:  ";   break;
+              default: out << "    [" << bt - build_tag << "] ";
             }
          out << *bt << endl;
        }
 
-   out << "  Readline:       " << HEX4(Input::readline_version()) << endl
-       << endl;
+   out << endl;
 
    Output::set_color_mode(Output::COLM_OUTPUT);
 }
@@ -679,7 +735,34 @@ UserPreferences::show_configure_options()
                    "---------------------" << endl <<
 
    "    ASSERT_LEVEL_WANTED=" << ASSERT_LEVEL_WANTED
-        << is_default(ASSERT_LEVEL_WANTED == 1) << endl <<
+        << is_default(ASSERT_LEVEL_WANTED == 1)
+   << endl <<
+
+   "    APSERVER_PATH=" << APSERVER_PATH
+        << is_default(!strcmp(APSERVER_PATH, "/tmp/GNU-APL/APserver"))
+   << endl <<
+
+   "    APSERVER_PORT=" << APSERVER_PORT
+        << is_default(APSERVER_PORT == 16366)
+   << endl <<
+
+   "    APSERVER_TRANSPORT=" << APSERVER_TRANSPORT
+        << is_default(APSERVER_TRANSPORT == 0)
+   << endl <<
+
+   "    CORE_COUNT_WANTED=" << CORE_COUNT_WANTED <<
+#if   CORE_COUNT_WANTED == -3
+   "  (= ⎕SYL)"
+#elif CORE_COUNT_WANTED == -2
+   "  (= argv (--cc))"
+#elif CORE_COUNT_WANTED == -1
+   "  (= all)"
+#elif CORE_COUNT_WANTED == 0
+   "  (= sequential) (default)"
+#else
+   ""
+#endif
+   << endl <<
 
 #ifdef DYNAMIC_LOG_WANTED
    "    DYNAMIC_LOG_WANTED=yes"
@@ -688,13 +771,6 @@ UserPreferences::show_configure_options()
 #endif
    << endl <<
 
-#ifdef HAVE_LIBREADLINE
-   "    libreadline is used (default)"
-#else
-   "    libreadline is not used (disabled or not present)"
-#endif
-   << endl <<
-
    "    MAX_RANK_WANTED="     << MAX_RANK_WANTED
         << is_default(MAX_RANK_WANTED == 8)
    << endl <<
@@ -704,8 +780,8 @@ UserPreferences::show_configure_options()
         << ", therefore:" << endl <<
    "        sizeof(Value)       : "  << value_size  << " bytes" << endl <<
    "        sizeof(Cell)        :  " << cell_size   << " bytes" << endl <<
-   "        sizeof(Value header): "  << header_size << " bytes" << endl <<
-   endl <<
+   "        sizeof(Value header): "  << header_size << " bytes" << endl
+   << endl <<
 
 #ifdef VALUE_CHECK_WANTED
    "    VALUE_CHECK_WANTED=yes"
@@ -721,18 +797,6 @@ UserPreferences::show_configure_options()
 #endif
    << endl <<
 
-   "    CORE_COUNT_WANTED=" << CORE_COUNT_WANTED <<
-#if   CORE_COUNT_WANTED == -3
-   "  (⎕SYL)"
-#elif CORE_COUNT_WANTED == -2
-   "  (argv (--cc))"
-#elif CORE_COUNT_WANTED == -1
-   "  (all)"
-#elif CORE_COUNT_WANTED == 0
-   "  (default: (sequential))"
-#endif
-   << endl <<
-
 #ifdef VF_TRACING_WANTED
    "    VF_TRACING_WANTED=yes"
 #else
@@ -746,22 +810,23 @@ UserPreferences::show_configure_options()
    "    VISIBLE_MARKERS_WANTED=no (default)"
 #endif
    << endl
-
+   << endl
+   << "how ./configure was (probably) called:" << endl
+   << "--------------------------------------" << endl
+   << "    " << configure_args << endl
    << endl;
 
-
    show_version(CERR);
 }
 //-----------------------------------------------------------------------------
-void
-UserPreferences::read_config_file(bool sys, bool log_startup)
+FILE *
+UserPreferences::open_user_file(const char * fname, char * filename,
+                                bool sys, bool log_startup)
 {
-char filename[PATH_MAX + 1];
-
    if (sys)   // eg. /etc/gnu-apl.d/preferences
       {
         snprintf(filename, PATH_MAX,
-                 "%s/gnu-apl.d/preferences", Makefile__sysconfdir);
+                 "%s/gnu-apl.d/%s", Makefile__sysconfdir, fname);
       }
    else       // try $HOME/.gnu_apl
       {
@@ -770,10 +835,10 @@ char filename[PATH_MAX + 1];
            {
              if (log_startup)
                 CERR << "environment variable 'HOME' is not defined!" << endl;
-             return;
+             return 0;
            }
 
-        snprintf(filename, PATH_MAX, "%s/.gnu-apl/preferences", HOME);
+        snprintf(filename, PATH_MAX, "%s/.gnu-apl/%s", HOME, fname);
         filename[PATH_MAX] = 0;
 
         // check that $HOME/.gnu-apl/preferences exist and fall back to
@@ -782,8 +847,7 @@ char filename[PATH_MAX + 1];
         if (access(filename, F_OK) != 0)   // file does not exit
            {
              snprintf(filename, PATH_MAX,
-                      "%s/.config/gnu-apl/preferences", HOME);
-             filename[PATH_MAX] = 0;
+                      "%s/.config/gnu-apl/%s", HOME, fname);
            }
       }
 
@@ -793,14 +857,25 @@ FILE * f = fopen(filename, "r");
    if (f == 0)
       {
          if (log_startup)
-            CERR << "config file " << filename
-                 << " is not present/readable" << endl;
-         return;
+            CERR << "Not reading config file " << filename
+                 << " (not found/readable)" << endl;
+         return 0;
       }
 
    if (log_startup)
       CERR << "Reading config file " << filename << " ..." << endl;
 
+   return f;
+}
+//-----------------------------------------------------------------------------
+void
+UserPreferences::read_config_file(bool sys, bool log_startup)
+{
+char filename[PATH_MAX + 1];
+
+FILE * f = open_user_file("preferences", filename, sys, log_startup);
+   if (f == 0)   return;
+
 int line = 0;
 int file_profile = 0;   // the current profile in the preferences file
    for (;;)
@@ -851,7 +926,7 @@ int file_profile = 0;   // the current profile in the preferences file
                    << " of config file " << filename << " (ignored)" << endl;
               continue;
             }
-         d[0] = strtol(arg, 0, 16);
+         d[0] = strtoll(arg, 0, 16);
          const bool yes = !strcasecmp(arg, "YES"     )
                        || !strcasecmp(arg, "ENABLED" )
                        || !strcasecmp(arg, "ON"      );
@@ -890,12 +965,23 @@ int file_profile = 0;   // the current profile in the preferences file
                    do_Color = true;
                    Output::use_curses = false;
                  }
-              if (!strcasecmp(arg, "CURSES"))
+              else if (!strcasecmp(arg, "CURSES"))
                  {
                    do_Color = true;
                    Output::use_curses = true;
                  }
             }
+         else if (!strcasecmp(opt, "Keyboard"))
+            {
+              if (!strcasecmp(arg, "NOCURSES"))
+                 {
+                   Output::keys_curses = false;
+                 }
+              else if (!strcasecmp(arg, "CURSES"))
+                 {
+                   Output::keys_curses = true;
+                 }
+            }
          else if (yes_no && !strcasecmp(opt, "Welcome"))
             {
               silent = no;
@@ -939,6 +1025,59 @@ int file_profile = 0;   // the current profile in the preferences file
               loop(p, count - 1)   Output::clear_EOL[p] = (char)(d[p] & 0xFF);
               Output::clear_EOL[count - 1] = 0;
             }
+         else if (!strcasecmp(opt, "CLEAR-EOS-SEQUENCE"))
+            {
+              loop(p, count - 1)   Output::clear_EOS[p] = (char)(d[p] & 0xFF);
+              Output::clear_EOS[count - 1] = 0;
+            }
+         else if (!strcasecmp(opt, "KEY-CURSOR-UP"))
+            {
+              loop(p, count - 1)
+                  Output::ESC_CursorUp[p] = (char)(d[p] & 0xFF);
+              Output::ESC_CursorUp[count - 1] = 0;
+            }
+         else if (!strcasecmp(opt, "KEY-CURSOR-DOWN"))
+            {
+              loop(p, count - 1)
+                  Output::ESC_CursorDown[p] = (char)(d[p] & 0xFF);
+              Output::ESC_CursorDown[count - 1] = 0;
+            }
+         else if (!strcasecmp(opt, "KEY-CURSOR-RIGHT"))
+            {
+              loop(p, count - 1)
+                  Output::ESC_CursorRight[p] = (char)(d[p] & 0xFF);
+              Output::ESC_CursorRight[count - 1] = 0;
+            }
+         else if (!strcasecmp(opt, "KEY-CURSOR-LEFT"))
+            {
+              loop(p, count - 1)
+                  Output::ESC_CursorLeft[p] = (char)(d[p] & 0xFF);
+              Output::ESC_CursorLeft[count - 1] = 0;
+            }
+         else if (!strcasecmp(opt, "KEY-CURSOR-END"))
+            {
+              loop(p, count - 1)
+                  Output::ESC_CursorEnd[p] = (char)(d[p] & 0xFF);
+              Output::ESC_CursorEnd[count - 1] = 0;
+            }
+         else if (!strcasecmp(opt, "KEY-CURSOR-HOME"))
+            {
+              loop(p, count - 1)
+                  Output::ESC_CursorHome[p] = (char)(d[p] & 0xFF);
+              Output::ESC_CursorHome[count - 1] = 0;
+            }
+         else if (!strcasecmp(opt, "KEY-INSMODE"))
+            {
+              loop(p, count - 1)
+                  Output::ESC_InsertMode[p] = (char)(d[p] & 0xFF);
+              Output::ESC_InsertMode[count - 1] = 0;
+            }
+         else if (!strcasecmp(opt, "KEY-DELETE"))
+            {
+              loop(p, count - 1)
+                  Output::ESC_Delete[p] = (char)(d[p] & 0xFF);
+              Output::ESC_Delete[count - 1] = 0;
+            }
          else if (!strcasecmp(opt, "CIN-FOREGROUND"))
             {
               Output::color_CIN_foreground = atoi(arg);
@@ -983,24 +1122,128 @@ int file_profile = 0;   // the current profile in the preferences file
                  }
 
               LibPaths::set_lib_dir((LibRef)lib_ref, arg,
-                                    sys ? LibPaths::LibDir::CS_PREF_SYS
-                                        : LibPaths::LibDir::CS_PREF_HOME);
+                                    sys ? LibPaths::LibDir::CSRC_PREF_SYS
+                                        : LibPaths::LibDir::CSRC_PREF_HOME);
             }
          else if (!strcasecmp(opt, "READLINE_HISTORY_LEN"))
             {
-              Input::readline_history_len = atoi(arg);
+              line_history_len = atoi(arg);
             }
          else if (!strcasecmp(opt, "READLINE_HISTORY_PATH"))
             {
-              Input::readline_history_path = UTF8_string(arg);
+              line_history_path = UTF8_string(arg);
+            }
+         else if (!strcasecmp(opt, "NABLA-TO-HISTORY"))
+            {
+              if      (!strcasecmp(arg, "Never"))      nabla_to_history = 0;
+              else if (!strcasecmp(arg, "Modified"))   nabla_to_history = 1;
+              else if (!strcasecmp(arg, "Always"))     nabla_to_history = 2;
+              else   CERR << "bad value " << arg << " for NABLA-TO-HISTORY"
+                        << " at line " << line << " of config file "
+                        << filename << " (ignored)" << endl;
             }
          else if (yes_no && !strcasecmp(opt, "BACKUP_BEFORE_SAVE"))
             {
               backup_before_save = yes;
             }
+         else if (!strcasecmp(opt, "KEYBOARD_LAYOUT_FILE"))
+            {
+              keyboard_layout_file = UTF8_string(arg);;
+            }
+         else if (!strcasecmp(opt, "CONTROL-Ds-TO-EXIT"))
+            {
+              control_Ds_to_exit = atoi(arg);
+            }
        }
 
    fclose(f);
 }
 //-----------------------------------------------------------------------------
+void
+UserPreferences::read_threshold_file(bool sys, bool log_startup)
+{
+char filename[PATH_MAX + 1];
+
+FILE * f = open_user_file("parallel_thresholds", filename, sys, log_startup);
+   if (f == 0)   return;
+
+int line = 0;
+   for (;;)
+       {
+         enum { BUFSIZE = 200 };
+         char buffer[BUFSIZE + 1];
+         const char * s = fgets(buffer, BUFSIZE, f);
+         if (s == 0)   break;   // end of file
+
+         buffer[BUFSIZE] = 0;
+         ++line;
+
+         // skip leading spaces
+         //
+         while (*s && *s <= ' ')   ++s;
+
+         if (*s == 0)     continue;   // empty line
+         if (*s == '#')   continue;   // comment line or #undef
+
+         // eg. perfo_2(F12_MINUS, _AB, "A - B",  25 )
+         //
+         const char * p = strstr(s, "perfo_");
+         if (p == 0)   continue;
+
+         p += 6;   // skip "perfo_"
+         const int macro_type = *p++ - '0';
+         if (macro_type < 1 || macro_type > 3)
+            {
+               CERR << "Bad macro in file " << filename
+                    << " line " << line << endl;
+               continue;
+            }
+
+         // macro arguments: (param, p_ab, p_txt, p_be)
+         //
+         {
+           const char * param = strchr(p, '(');
+           if (param == 0)   goto param_error;   skip_spaces(++param);
+
+           const char * p_ab = strchr(param, ',');
+           if (p_ab == 0)   goto param_error;   skip_spaces(++p_ab);
+           const int i_ab = p_ab[1] == 'A' ? 2 : 1;
+
+           const char * p_txt = strchr(p_ab, ',');
+           if (p_txt == 0)   goto param_error;   skip_spaces(++p_txt);
+
+           const char * p_be = strchr(p_txt, ',');
+           if (p_be == 0)   goto param_error;   skip_spaces(++p_be);
+
+           const ShapeItem value = strtoll(p_be, 0, 0);
+
+         enum { _B = 1, _AB = 2 };
+#define perfo_1(bif, ab, _name, th)   if (!strncmp(param, #bif, strlen(#bif))) \
+   set_threshold(*Bif_ ## bif::fun, ab, i_ab, value);
+
+#define perfo_2(bif, ab, _name, thr)  perfo_1(bif, ab, _name, thr)
+#define perfo_3(bif, ab, _name, thr)  perfo_1(bif, ab, _name, thr)
+#define perfo_4(bif, ab, _name, thr)
+
+#include "Performance.def"
+         }
+         continue;
+
+         param_error:
+            CERR << "Bad macro format in file " << filename
+                 << " line " << line << endl;
+       }
+}
+//-----------------------------------------------------------------------------
+void
+UserPreferences::set_threshold(Function & fun, int ab, int i_ab,
+                               ShapeItem threshold)
+{
+   if (&fun == 0)    return;
+   if (ab != i_ab)   return;
+
+   if (ab == 1)   fun.set_monadic_threshold(threshold);
+   else           fun.set_dyadic_threshold(threshold);
+}
+//-----------------------------------------------------------------------------
 
