@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,6 +18,8 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+#include <sys/types.h>
+
 #include "CDR_string.hh"
 #include "CharCell.hh"
 #include "Common.hh"
@@ -26,10 +28,11 @@
 #include "IndexExpr.hh"
 #include "IndexIterator.hh"
 #include "IntCell.hh"
+#include "IO_Files.hh"
 #include "LvalCell.hh"
-#include "main.hh"
 #include "Output.hh"
 #include "PointerCell.hh"
+#include "Parallel.hh"
 #include "PrintOperator.hh"
 #include "SystemVariable.hh"
 #include "UCS_string.hh"
@@ -50,6 +53,8 @@ void
 Value::init_ravel()
 {
    owner_count = 0;
+   pointer_cell_count = 0;
+   nz_subcell_count = 0;
 
    ++value_count;
    if (Quad_SYL::value_count_limit &&
@@ -60,12 +65,11 @@ Value::init_ravel()
         // make sure that the value is properly initialized
         //
         ravel = short_value;
-        new (&shape) Shape();
-        new (ravel)   IntCell(42);
+        loop(l, SHORT_VALUE_LENGTH_WANTED)   new (ravel + l)   IntCell(42);
 
         Workspace::more_error() = UCS_string(
-        "the system limit on the APL value count (as set in ⎕SYL) was reached\n"
-        "(and to avoid lock-up, the limit in ⎕SYL was automatically cleared).");
+"the system limit on the APL value count (as set in ⎕SYL) was reached\n"
+"(and to avoid lock-up, the limit in ⎕SYL was automatically cleared).");
 
         // reset the limit so that we don't get stuck here.
         //
@@ -102,7 +106,8 @@ const ShapeItem length = shape.get_volume();
 
         try
            {
-             Cell * long_ravel =  new Cell[length];
+             Cell * long_ravel = (Cell *)(new char[length * sizeof(Cell)]);
+//           Cell * long_ravel =  new Cell[length];
              ravel = long_ravel;
            }
         catch (...)
@@ -207,21 +212,6 @@ Value::Value(const CDR_string & ui8, const char * loc)
    set_complete();
 }
 //-----------------------------------------------------------------------------
-Value::Value(const char * string, const char * loc)
-   : DynamicObject(loc, &all_values),
-     shape(strlen(string)),
-     flags(VF_NONE),
-     valid_ravel_items(0)
-{
-   ADD_EVENT(this, VHE_Create, 0, loc);
-   init_ravel();
-
-const ShapeItem length = strlen(string);
-   loop(e, length)  new (&get_ravel(e)) CharCell(Unicode(string[e]));
-
-   set_complete();
-}
-//-----------------------------------------------------------------------------
 Value::Value(const Shape & sh, const char * loc)
    : DynamicObject(loc, &all_values),
      shape(sh),
@@ -240,7 +230,7 @@ Value::Value(const Cell & cell, const char * loc)
    ADD_EVENT(this, VHE_Create, 0, loc);
    init_ravel();
 
-   get_ravel(0).init(cell);
+   get_ravel(0).init(cell, *this);
    check_value(LOC);
 }
 //-----------------------------------------------------------------------------
@@ -249,9 +239,15 @@ Value::~Value()
    ADD_EVENT(this, VHE_Destruct, 0, LOC);
    unlink();
 
-   release_sub(LOC);
-
 const ShapeItem length = nz_element_count();
+
+   if (get_pointer_cell_count() > 0)
+      {
+
+        Cell * cZ = &get_ravel(0);
+        loop(c, length)   cZ++->release(LOC);
+      }
+
    --value_count;
 
    if (ravel != short_value)
@@ -264,28 +260,21 @@ const ShapeItem length = nz_element_count();
    check_ptr = 0;
 }
 //-----------------------------------------------------------------------------
-void
-Value::release_sub(const char * loc)
-{
-const ShapeItem length = nz_element_count();
-
-Cell * cZ = &get_ravel(0);
-   loop(c, length)   cZ++->release(loc);
-}
-//-----------------------------------------------------------------------------
 Value_P
 Value::get_cellrefs(const char * loc)
 {
-Value_P ret(new Value(get_shape(), loc));
+Value_P ret(get_shape(), loc);
 
-const ShapeItem ec = nz_element_count();
+const ShapeItem ec = element_count();
 
    loop(e, ec)
       {
         Cell & cell = get_ravel(e);
-        new (ret->next_ravel())   LvalCell(&cell);
+        new (ret->next_ravel())   LvalCell(&cell, this);
       }
 
+   if (ec == 0)   new (&ret->get_ravel(0))   LvalCell(0, 0);
+
    ret->check_value(LOC);
    return ret;
 }
@@ -298,25 +287,28 @@ const ShapeItem value_count = new_value->nz_element_count();
 const Cell * src = &new_value->get_ravel(0);
 Cell * C = &get_ravel(0);
 
+   // this:      a value containing LvalCells and possibly PointerCells.
+   // cellowner: the owner of the cells that this points to
+   //
+Value * cellowner = get_lval_cellowner();
 const int src_incr = new_value->is_scalar() ? 0 : 1;
 
-   if (is_scalar())
+   // if this is a scalar and new_value is not, then enclose new_value
+   //
+   if (is_scalar() && !new_value->is_scalar())
       {
         if (!C->is_lval_cell())   LEFT_SYNTAX_ERROR;
+
         Cell * dest = C->get_lval_value();   // can be 0!
         if (dest)   dest->release(LOC);   // free sub-values etc (if any)
 
-        if (new_value->is_scalar())
-           {
-             dest->init(new_value->get_ravel(0));
-           }
-        else
-           {
-             new (dest)   PointerCell(new_value);
-           }
+        new (dest)   PointerCell(new_value, *cellowner);
         return;
       }
 
+   // at this point both this and new_value should have the same number
+   // of elements
+   //
    if (!new_value->is_scalar() && value_count != dest_count)   LENGTH_ERROR;
 
    loop(d, dest_count)
@@ -328,27 +320,26 @@ const int src_incr = new_value->is_scalar() ? 0 : 1;
 
         // erase the pointee when overriding a pointer-cell.
         //
-        if (dest)   dest->init(*src);
+        if (dest)   dest->init(*src, *this);
         src += src_incr;
       }
 }
 //-----------------------------------------------------------------------------
-bool
-Value::is_lval() const
+Value *
+Value::get_lval_cellowner() const
 {
 const ShapeItem ec = nz_element_count();
 
    loop(e, ec)
       {
         const Cell & cell = get_ravel(e);
-        if (cell.is_lval_cell())   return true;
-        if (cell.is_pointer_cell() && cell.get_pointer_value()->is_lval())
-           return true;
+        if (cell.is_lval_cell())   return cell.get_cell_owner();
+        if (!cell.is_pointer_cell())   return 0;
+        return  cell.get_pointer_value()->get_lval_cellowner();
 
-        return false;
       }
 
-   return false;
+   return 0;
 }
 //-----------------------------------------------------------------------------
 bool
@@ -424,13 +415,10 @@ Value::rollback(ShapeItem items, const char * loc)
 {
    ADD_EVENT(this, VHE_Unroll, 0, loc);
 
-   release_sub(loc);
-
-   loop(i, items)
-      {
-        Cell & cell = get_ravel(i);
-        cell.release(LOC);
-      }
+   // this value has only items < nz_element_count() valid items.
+   // init the rest...
+   //
+   while (items < nz_element_count())   new (&get_ravel(items++)) IntCell(0);
 
    ((Value *)this)->alloc_loc = loc;
 }
@@ -574,7 +562,6 @@ Value::list_one(ostream & out, bool show_owners)
         char sep = ' ';
         if (is_complete())   { out << sep << "COMPLETE";   sep = '+'; }
         if (is_marked())     { out << sep << "MARKED";     sep = '+'; }
-        if (is_temp())       { out << sep << "TEMP";       sep = '+'; }
       }
    else
       {
@@ -640,7 +627,7 @@ ShapeItem count = ec;
 }
 //-----------------------------------------------------------------------------
 void
-Value::enlist(Cell * & dest, bool left) const
+Value::enlist(Cell * & dest, Value & dest_owner, bool left) const
 {
 ShapeItem ec = element_count();
 
@@ -649,7 +636,7 @@ ShapeItem ec = element_count();
          const Cell & cell = get_ravel(c);
          if (cell.is_pointer_cell())
             {
-              cell.get_pointer_value()->enlist(dest, left);
+              cell.get_pointer_value()->enlist(dest, dest_owner, left);
             }
          else if (left && cell.is_lval_cell())
             {
@@ -660,20 +647,20 @@ ShapeItem ec = element_count();
                  }
               else if (cp->is_pointer_cell())
                  {
-                   cp->get_pointer_value()->enlist(dest, left);
+                   cp->get_pointer_value()->enlist(dest, dest_owner, left);
                  }
               else 
                  {
-                   new (dest++) LvalCell(cell.get_lval_value());
+                   new (dest++) LvalCell(cell.get_lval_value(), &dest_owner);
                  }
             }
          else if (left)
             {
-              new (dest++) LvalCell((Cell *)&cell);
+              new (dest++) LvalCell((Cell *)&cell, &dest_owner);
             }
          else
             {
-              dest++->init(cell);
+              dest++->init(cell, dest_owner);
             }
        }
 }
@@ -763,7 +750,7 @@ Value::is_int_vector(APL_Float qct) const
    loop(c, get_shape_item(0))
        {
          const Cell & cell = get_ravel(c);
-         if (!cell.is_near_int(qct))   return false;
+         if (!cell.is_near_int())   return false;
        }
 
    return true;
@@ -774,7 +761,7 @@ Value::is_int_scalar(APL_Float qct) const
 {
    if (get_rank() != 0)   return false;
 
-   return get_ravel(0).is_near_int(qct);
+   return get_ravel(0).is_near_int();
 }
 //-----------------------------------------------------------------------------
 bool
@@ -786,7 +773,7 @@ const ShapeItem ec = nz_element_count();
       {
         const Cell & cell = get_ravel(e);
         if (!cell.is_numeric())        DOMAIN_ERROR;
-        if (!cell.is_near_real(qct))   return true;
+        if (!cell.is_near_real())   return true;
       }
 
    return false;   // all cells numeric and not complex
@@ -924,7 +911,7 @@ Value::index(const IndexExpr & IX) const
 {
    Assert(IX.value_count() != 1);   // should have called index(Value_P X)
 
-   if (get_rank() != IX.value_count())   RANK_ERROR;
+   if (get_rank() != IX.value_count())   RANK_ERROR;   // ISO p. 158
 
    // Notes:
    //
@@ -965,11 +952,14 @@ Shape shape_Z;
 
    // check that all indices are valid
    //
-   if (IX.check_range(get_shape()))   INDEX_ERROR;
+   if (IX.check_range(get_shape()))
+      {
+        INDEX_ERROR;
+      }
 
-MultiIndexIterator mult(get_shape(), IX);   // deletes IDX
+MultiIndexIterator mult(get_shape(), IX);
 
-Value_P Z(new Value(shape_Z, LOC));
+Value_P Z(shape_Z, LOC);
 const ShapeItem ec_z = Z->element_count();
 
    if (ec_z == 0)   // empty result
@@ -983,7 +973,7 @@ const ShapeItem ec_z = Z->element_count();
    // We go from lower indices to higher indices in IX, which
    // means from higher indices to lower indices in this and Z
    
-   loop(z, ec_z)   Z->get_ravel(z).init(get_ravel(mult.next()));
+   loop(z, ec_z)   Z->get_ravel(z).init(get_ravel(mult.next()), Z.getref());
 
    Assert(mult.done());
 
@@ -999,23 +989,22 @@ Value::index(Value_P X) const
    if (!X)   return clone(LOC);
 
 const Shape shape_Z(X->get_shape());
-Value_P Z(new Value(shape_Z, LOC));
+Value_P Z(shape_Z, LOC);
 const ShapeItem max_idx = element_count();
 const APL_Integer qio = Workspace::get_IO();
-const APL_Float qct = Workspace::get_CT();
 
 const Cell * cI = &X->get_ravel(0);
 
    while (Z->more())
       {
-         const ShapeItem idx = cI++->get_near_int(qct) - qio;
+         const ShapeItem idx = cI++->get_near_int() - qio;
          if (idx < 0 || idx >= max_idx)
             {
               Z->rollback(Z->valid_ravel_items, LOC);
               INDEX_ERROR;
             }
 
-         Z->next_ravel()->init(get_ravel(idx));
+         Z->next_ravel()->init(get_ravel(idx), Z.getref());
       }
 
    Z->set_default(*this);
@@ -1024,6 +1013,7 @@ const Cell * cI = &X->get_ravel(0);
    return Z;
 }
 //-----------------------------------------------------------------------------
+/**
 Value_P
 Value::index(Token & IX) const
 {
@@ -1034,23 +1024,21 @@ Value::index(Token & IX) const
    //
    Q1(IX.get_tag());  FIXME;
 }
+**/
 //-----------------------------------------------------------------------------
 Rank
 Value::get_single_axis(Rank max_axis) const
 {
    if (this == 0)   AXIS_ERROR;
 
-const APL_Float   qct = Workspace::get_CT();
-const APL_Integer qio = Workspace::get_IO();
-
    if (!is_scalar_or_len1_vector())     AXIS_ERROR;
 
-   if (!get_ravel(0).is_near_int(qct))   AXIS_ERROR;
+   if (!get_ravel(0).is_near_int())   AXIS_ERROR;
 
    // if axis becomes (signed) negative then it will be (unsigned) too big.
    // Therefore we need not test for < 0.
    //
-const unsigned int axis = get_ravel(0).get_near_int(qct) - qio;
+const unsigned int axis = get_ravel(0).get_near_int() - Workspace::get_IO();
    if (axis >= max_axis)   AXIS_ERROR;
 
    return axis;
@@ -1062,12 +1050,11 @@ Value::to_shape() const
    if (this == 0)   INDEX_ERROR;   // elided index ?
 
 const ShapeItem xlen = element_count();
-const APL_Float qct = Workspace::get_CT();
 const APL_Integer qio = Workspace::get_IO();
 
 Shape shape;
      loop(x, xlen)
-        shape.add_shape_item(get_ravel(x).get_near_int(qct) - qio);
+        shape.add_shape_item(get_ravel(x).get_near_int() - qio);
 
    return shape;
 }
@@ -1113,10 +1100,10 @@ const ShapeItem len_B = B->element_count();
    Assert(A->is_scalar_or_vector());
    Assert(B->is_scalar_or_vector());
 
-Value_P Z(new Value(len_A + len_B, LOC));
+Value_P Z(len_A + len_B, LOC);
 
-   loop(a, len_A)   Z->next_ravel()->init(A->get_ravel(a));
-   loop(b, len_B)   Z->next_ravel()->init(B->get_ravel(b));
+   loop(a, len_A)   Z->next_ravel()->init(A->get_ravel(a), Z.getref());
+   loop(b, len_B)   Z->next_ravel()->init(B->get_ravel(b), Z.getref());
 
    Z->check_value(LOC);
    new (&result) Token(TOK_APL_VALUE3, Z);
@@ -1137,17 +1124,17 @@ Value::glue_strand_closed(Token & result, Value_P A, Value_P B,
    Assert(A->is_scalar_or_vector());
 
 const ShapeItem len_A = A->element_count();
-Value_P Z(new Value(len_A + 1, LOC));
+Value_P Z(len_A + 1, LOC);
 
-   loop(a, len_A)   Z->next_ravel()->init(A->get_ravel(a));
+   loop(a, len_A)   Z->next_ravel()->init(A->get_ravel(a), Z.getref());
 
    if (B->is_simple_scalar())
       {
-        Z->next_ravel()->init(B->get_ravel(0));
+        Z->next_ravel()->init(B->get_ravel(0), Z.getref());
       }
    else
       {
-        new (Z->next_ravel()) PointerCell(B);
+        new (Z->next_ravel()) PointerCell(B, Z.getref());
       }
 
    Z->check_value(LOC);
@@ -1169,18 +1156,18 @@ Value::glue_closed_strand(Token & result, Value_P A, Value_P B,
    Assert(B->is_scalar_or_vector());
 
 const ShapeItem len_B = B->element_count();
-Value_P Z(new Value(len_B + 1, LOC));
+Value_P Z(len_B + 1, LOC);
 
    if (A->is_simple_scalar())
       {
-        Z->next_ravel()->init(A->get_ravel(0));
+        Z->next_ravel()->init(A->get_ravel(0), Z.getref());
       }
    else
       {
-        new (Z->next_ravel()) PointerCell(A);
+        new (Z->next_ravel()) PointerCell(A, Z.getref());
       }
 
-   loop(b, len_B)   Z->next_ravel()->init(B->get_ravel(b));
+   loop(b, len_B)   Z->next_ravel()->init(B->get_ravel(b), Z.getref());
 
    Z->check_value(LOC);
    new (&result) Token(TOK_APL_VALUE3, Z);
@@ -1198,23 +1185,23 @@ Value::glue_closed_closed(Token & result, Value_P A, Value_P B,
              << " and " << endl << *B << endl;
       }
 
-Value_P Z(new Value(2, LOC));
+Value_P Z(2, LOC);
    if (A->is_simple_scalar())
       {
-        Z->next_ravel()->init(A->get_ravel(0));
+        Z->next_ravel()->init(A->get_ravel(0), Z.getref());
       }
    else
       {
-        new (Z->next_ravel()) PointerCell(A);
+        new (Z->next_ravel()) PointerCell(A, Z.getref());
       }
 
    if (B->is_simple_scalar())
       {
-        Z->next_ravel()->init(B->get_ravel(0));
+        Z->next_ravel()->init(B->get_ravel(0), Z.getref());
       }
    else
       {
-        new (Z->next_ravel()) PointerCell(B);
+        new (Z->next_ravel()) PointerCell(B, Z.getref());
       }
 
    Z->check_value(LOC);
@@ -1478,7 +1465,6 @@ UCS_string ind(indent, UNI_ASCII_SPACE);
        << ind << "Flags:   " << get_flags();
    if (is_complete())   out << " VF_complete";
    if (is_marked())     out << " VF_marked";
-   if (is_temp())       out << " VF_temp";
    out << endl
        << ind << "First:   " << get_ravel(0)  << endl
        << ind << "Dynamic: ";
@@ -1488,7 +1474,7 @@ UCS_string ind(indent, UNI_ASCII_SPACE);
 }
 //-----------------------------------------------------------------------------
 void
-Value::debug(const char * info)
+Value::debug(const char * info) const
 {
 const PrintContext pctx = Workspace::get_PrintContext();
 PrintBuffer pb(*this, pctx, 0);
@@ -1496,7 +1482,7 @@ PrintBuffer pb(*this, pctx, 0);
 }
 //-----------------------------------------------------------------------------
 ostream &
-Value::print_boxed(ostream & out, const char * info)
+Value::print_boxed(ostream & out, const char * info) const
 {
    if (info)   out << info << endl;
 
@@ -1527,6 +1513,7 @@ const ShapeItem name_len = get_cols();
    loop(v, var_count)
       {
         ShapeItem nidx = v*name_len;
+        const ShapeItem end = nidx + name_len;
         UCS_string & name = result[v];
         loop(n, name_len)
            {
@@ -1540,19 +1527,34 @@ const ShapeItem name_len = get_cols();
                 {
                   name.append(uni);
                 }
-             else if (last)
+             else if (uni == UNI_ASCII_SPACE)
                 {
-                  const ShapeItem end = v*name_len;
-                  while (nidx < end && get_ravel(nidx++).get_char_value()
-                         == UNI_ASCII_SPACE)   ++nidx;
-                  if (nidx == end)   break;   // remaining chars are spaces
+                  // skip space at nidx and subsequent spaces
+                  //
+                  while (nidx < end && get_ravel(nidx).get_char_value()
+                                  == UNI_ASCII_SPACE)   ++nidx;
+
+                  if (nidx == end)   break;   // only spaces (no second name)
 
-                  // another name following
                   name.clear();
-                  name.append(get_ravel(nidx++).get_char_value());
+                  if (last)
+                     {
+                       // if 'last' is true then to_varnames() was called from
+                       // ⎕SVO and the line may contains two variable names.
+                       // Return the second i.e. last) one)
+                       //
+
+
+                       // another name following
+                       //
+                       name.clear();   // discard first name
+                       name.append(get_ravel(nidx++).get_char_value());
+                       last = false;
+                     }
                 }
-             else  
+             else  // neither space nor valid variable name character
                 {
+                  name.clear();
                   break;
                 }
            }
@@ -1610,17 +1612,17 @@ const Cell & first = get_ravel(0);
    if (first.is_pointer_cell())
       {
         Value_P B0 = first.get_pointer_value();
-        Value_P Z(new Value(B0->get_shape(), loc));
+        Value_P Z(B0->get_shape(), loc);
         const ShapeItem ec_Z =  Z->nz_element_count();
 
-        loop(z, ec_Z)   Z->get_ravel(z).init_type(B0->get_ravel(z));
+        loop(z, ec_Z)   Z->get_ravel(z).init_type(B0->get_ravel(z), Z.getref());
         return Z;
       }
    else
       {
-        Value_P Z(new Value(loc));
+        Value_P Z(loc);
 
-        Z->get_ravel(0).init_type(first);
+        Z->get_ravel(0).init_type(first, Z.getref());
         return Z;
       }
 
@@ -1629,15 +1631,29 @@ const Cell & first = get_ravel(0);
 Value_P
 Value::clone(const char * loc) const
 {
-Value_P ret(new Value(get_shape(), loc));
+#ifdef PERFORMANCE_COUNTERS_WANTED
+#ifdef HAVE_RDTSC
+const uint64_t start_1 = cycle_counter();
+#endif
+#endif
+
+Value_P ret(get_shape(), loc);
 
 const Cell * src = &get_ravel(0);
 Cell * dst = &ret->get_ravel(0);
 const ShapeItem count = nz_element_count();
 
-   loop(c, count)   dst++->init(*src++);
+   loop(c, count)   dst++->init(*src++, ret.getref());
 
    ret->check_value(LOC);
+
+#ifdef PERFORMANCE_COUNTERS_WANTED
+#ifdef HAVE_RDTSC
+const uint64_t end_1 = cycle_counter();
+   Performance::fs_clone_B.add_sample(end_1 - start_1, count);
+#endif
+#endif
+
    return ret;
 }
 //-----------------------------------------------------------------------------
@@ -1763,6 +1779,7 @@ int count = 0;
         const DynamicObject * dob = stale_dobs[s];
         Value * val = stale_vals[s];
         val->print_stale_info(out, dob);
+        ++count;
        }
 
    // mark all dynamic values, and then unmark those known in the workspace
@@ -1793,6 +1810,7 @@ int count = 0;
          if (val->is_marked())
             {
               val->print_stale_info(out, dob);
+              ++count;
               val->unmark();
             }
        }
