@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2015  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -26,11 +26,9 @@ using namespace std;
 
 #include "Archive.hh"
 #include "Command.hh"
-#include "Input.hh"
 #include "InputFile.hh"
 #include "IO_Files.hh"
 #include "LibPaths.hh"
-#include "main.hh"
 #include "Output.hh"
 #include "Quad_FX.hh"
 #include "Quad_TF.hh"
@@ -38,7 +36,18 @@ using namespace std;
 #include "UserPreferences.hh"
 #include "Workspace.hh"
 
-// Workspace::the_workspace is defined in StaticObjects.cc
+/* define 'all_values' and 'all_index_exprs' here so that they will be
+   constructed BEFORE Workspace::the_workspace !
+
+   See also 3.6.2 of "ISO standard Programming Languages — C++"
+*/
+
+DynamicObject DynamicObject::all_values(LOC);
+DynamicObject DynamicObject::all_index_exprs(LOC);
+
+// now Workspace::the_workspace can be constructed
+//
+Workspace Workspace::the_workspace;
 
 //-----------------------------------------------------------------------------
 void
@@ -97,6 +106,23 @@ StateIndicator * del = SI_top();
    delete del;
 }
 //-----------------------------------------------------------------------------
+uint64_t
+Workspace::get_RL(uint64_t mod)
+{
+   // we discard random numbers >= max_rand in order to avoid a bias
+   // towards small numbers
+   //
+const uint64_t max_rand = 0xFFFFFFFFFFFFFFFFULL - (0xFFFFFFFFFFFFFFFF % mod);
+uint64_t rand = the_workspace.v_Quad_RL.get_random();
+
+   do { rand ^= rand >> 11;
+        rand ^= rand << 29;
+        rand ^= rand >> 14;
+      } while (rand >= max_rand);
+
+   return rand % mod;
+}
+//-----------------------------------------------------------------------------
 void
 Workspace::clear_error(const char * loc)
 {
@@ -237,13 +263,13 @@ Unicode av_4 = (ucs.size() > 5) ? ucs[5] : Invalid_Unicode;
    return Token(TOK_Quad_ ## t, &the_workspace.v_Quad_ ## t); }
 
 #define f0(t, l) { len = l + 1; \
-   return Token(TOK_Quad_ ## t, &Quad_ ## t::fun); }
+   return Token(TOK_Quad_ ## t, Quad_ ## t::fun); }
 
 #define f1(t, l) { len = l + 1; \
-   return Token(TOK_Quad_ ## t, &Quad_ ## t::fun); }
+   return Token(TOK_Quad_ ## t, Quad_ ## t::fun); }
 
 #define f2(t, l) { len = l + 1; \
-   return Token(TOK_Quad_ ## t, &Quad_ ## t::fun); }
+   return Token(TOK_Quad_ ## t, Quad_ ## t::fun); }
 
    switch(av_0)
       {
@@ -311,7 +337,6 @@ Unicode av_4 = (ucs.size() > 5) ? ucs[5] : Invalid_Unicode;
              if (av_1 == UNI_ASCII_P)        var(PP, 2)
              if (av_1 == UNI_ASCII_R)        var(PR, 2)
              if (av_1 == UNI_ASCII_S)        var(PS, 2)
-             if (av_1 == UNI_ASCII_T)        var(PT, 2)
              if (av_1 == UNI_ASCII_W)        var(PW, 2)
              break;
 
@@ -403,7 +428,7 @@ Workspace::is_called(const UCS_string & funname)
 Symbol * current_referent = lookup_existing_symbol(funname);
    if (current_referent == 0)   return false;   // no such symbol
 
-   Assert(current_referent->get_Id() == ID_USER_SYMBOL);
+   Assert(current_referent->get_Id() == ID::USER_SYMBOL);
 
 const NameClass nc = current_referent->get_nc();
    if (nc != NC_FUNCTION && nc != NC_OPERATOR)   return false;
@@ -455,8 +480,7 @@ Workspace::write_OUT(FILE * out, uint64_t & seq, const vector<UCS_string>
                    continue;
                  }
 
-              const Id obj_id = obj->get_Id();
-              if (obj_id == ID_USER_SYMBOL)   // user defined name
+              if (obj->get_Id() == ID::USER_SYMBOL)   // user defined name
                  {
                    const Symbol * sym = lookup_existing_symbol(objects[o]);
                    Assert(sym);
@@ -487,8 +511,8 @@ Workspace::unmark_all_values()
 
    // unmark system variables
    //
-#define rw_sv_def(x) the_workspace.v_ ## x.unmark_all_values();
-#define ro_sv_def(x) the_workspace.v_ ## x.unmark_all_values();
+#define rw_sv_def(x, txt) the_workspace.v_ ## x.unmark_all_values();
+#define ro_sv_def(x, txt) the_workspace.v_ ## x.unmark_all_values();
 #include "SystemVariable.def"
    the_workspace.v_Quad_Quad .unmark_all_values();
    the_workspace.v_Quad_QUOTE.unmark_all_values();
@@ -515,8 +539,8 @@ int count = 0;
 
    // system variabes
    //
-#define rw_sv_def(x) count += get_v_ ## x().show_owners(out, value);
-#define ro_sv_def(x) count += get_v_ ## x().show_owners(out, value);
+#define rw_sv_def(x, txt) count += get_v_ ## x().show_owners(out, value);
+#define ro_sv_def(x, txt) count += get_v_ ## x().show_owners(out, value);
 #include "SystemVariable.def"
    count += the_workspace.v_Quad_Quad .show_owners(out, value);
    count += the_workspace.v_Quad_QUOTE.show_owners(out, value);
@@ -578,8 +602,8 @@ Workspace::clear_WS(ostream & out, bool silent)
 
    // clear the value stacks of read/write system variables...
    //
-#define rw_sv_def(x) get_v_ ## x().clear_vs();
-#define ro_sv_def(x)
+#define rw_sv_def(x, txt) get_v_ ## x().destruct();
+#define ro_sv_def(x, txt)
 #include "SystemVariable.def"
 
    // at this point all values should have been gone.
@@ -587,8 +611,8 @@ Workspace::clear_WS(ostream & out, bool silent)
    //
 // Value::erase_all(out);
 
-#define rw_sv_def(x) new  (&get_v_ ##x()) x;
-#define ro_sv_def(x)
+#define rw_sv_def(x, txt) new  (&get_v_ ##x()) x;
+#define ro_sv_def(x, txt)
 #include "SystemVariable.def"
 
    get_v_Quad_RL().reset_seed();
@@ -600,6 +624,7 @@ Workspace::clear_WS(ostream & out, bool silent)
 void
 Workspace::clear_SI(ostream & out)
 {
+   // clear the SI (pops all localized symbols)
    while (SI_top())
       {
         SI_top()->escape();
@@ -631,7 +656,7 @@ bool name_from_WSID = false;
       }
    else if (lib_ws.size() > 2)   // too many arguments
       {
-        out << "BAD COMMAND" << endl;
+        out << "BAD COMMAND+" << endl;
         more_error() = UCS_string("too many parameters in command )SAVE");
         return;
       }
@@ -803,7 +828,7 @@ Workspace::dump_WS(ostream & out, vector<UCS_string> & lib_ws)
       }
    else if (lib_ws.size() > 2)   // too many arguments
       {
-        out << "BAD COMMAND" << endl;
+        out << "BAD COMMAND+" << endl;
         more_error() = UCS_string("too many parameters in command )DUMP");
         return;
       }
@@ -873,8 +898,8 @@ int variable_count = 0;
 
    // system variables
    //
-#define ro_sv_def(x)
-#define rw_sv_def(x) if (ID_ ## x != ID_Quad_SYL) { get_v_ ## x().dump(outf);   ++variable_count; }
+#define ro_sv_def(x, txt)
+#define rw_sv_def(x, txt) if (ID:: x != ID::Quad_SYL) { get_v_ ## x().dump(outf);   ++variable_count; }
 #include "SystemVariable.def"
 
    COUT << "DUMPED WORKSPACE '" << wname << "'" << endl
@@ -886,11 +911,11 @@ int variable_count = 0;
 // )LOAD WS, set ⎕LX of loaded WS on success
 void
 Workspace::load_WS(ostream & out, const vector<UCS_string> & lib_ws,
-                   UCS_string & quad_lx)
+                   UCS_string & quad_lx, bool silent)
 {
    if (lib_ws.size() < 1 || lib_ws.size() > 2)   // no or too many argument(s)
       {
-        out << "BAD COMMAND" << endl;
+        out << "BAD COMMAND+" << endl;
         more_error() = UCS_string("too many parameters in command )LOAD");
         return;
       }
@@ -906,10 +931,10 @@ XML_Loading_Archive in(filename.c_str(), dump_fd);
 
    if (dump_fd != -1)
       {
-        the_workspace.clear_WS(CERR, true);
+        the_workspace.clear_WS(out, true);
 
-        Log(LOG_command_IN)   CERR << "LOADING " << wname << " from file '"
-                                   << filename << "' ..." << endl;
+        Log(LOG_command_IN)   out << "LOADING " << wname << " from file '"
+                                  << filename << "' ..." << endl;
 
         load_DUMP(out, filename, dump_fd, true);   // closes dump_fd
 
@@ -931,19 +956,19 @@ XML_Loading_Archive in(filename.c_str(), dump_fd);
       {
         if (!in.is_open())   // open failed
            {
-             CERR << ")LOAD " << wname << " (file " << filename
-                  << ") failed: " << strerror(errno) << endl;
+             out << ")LOAD " << wname << " (file " << filename
+                 << ") failed: " << strerror(errno) << endl;
              return;
            }
 
-        Log(LOG_command_IN)   CERR << "LOADING " << wname << " from file '"
-                                   << filename << "' ..." << endl;
+        Log(LOG_command_IN)   out << "LOADING " << wname << " from file '"
+                                  << filename << "' ..." << endl;
 
         // got open file. We assume that from here on everything will be fine.
         // clear current WS and load it from file
         //
-        the_workspace.clear_WS(CERR, true);
-        in.read_Workspace();
+        the_workspace.clear_WS(out, true);
+        in.read_Workspace(silent);
       }
 
    if (Workspace::get_LX().size())  quad_lx = Workspace::get_LX();
@@ -960,7 +985,7 @@ Workspace::copy_WS(ostream & out, vector<UCS_string> & lib_ws_objects,
 
    if (lib_ws_objects.size() < 1)   // at least workspace name is required
       {
-        out << "BAD COMMAND" << endl;
+        out << "BAD COMMAND+" << endl;
         more_error() = UCS_string(
                        "missing parameter(s) in command )COPY or )PCOPY");
         return;
@@ -975,7 +1000,7 @@ LibRef libref = LIB_NONE;
 
    if (lib_ws_objects.size() < 1)   // at least workspace name is required
       {
-        out << "BAD COMMAND" << endl;
+        out << "BAD COMMAND+" << endl;
         more_error() = UCS_string(
                        "missing parameter(s) in command )COPY or )PCOPY");
         return;
@@ -1007,7 +1032,7 @@ XML_Loading_Archive in(filename.c_str(), dump_fd);
 
    in.set_protection(protection, lib_ws_objects);
    in.read_vids();
-   in.read_Workspace();
+   in.read_Workspace(false);
 }
 //-----------------------------------------------------------------------------
 void
