@@ -42,7 +42,7 @@ void DefCommand::run_command( NetworkConnection &conn, const std::vector<std::st
         stringstream out;
 
         Shape shape( content.size() );
-        Value_P function_list_value( new Value( shape, LOC ) );
+        Value_P function_list_value(shape, LOC);
         for( vector<string>::const_iterator i = content.begin() ; i != content.end() ; i++ ) {
             Value_P v;
             if( i->size() == 0 ) {
@@ -51,28 +51,28 @@ void DefCommand::run_command( NetworkConnection &conn, const std::vector<std::st
             else {
                 v = make_string_cell( *i, LOC );
             }
-            new (function_list_value->next_ravel()) PointerCell( v );
+            new (function_list_value->next_ravel()) PointerCell( v, function_list_value.getref() );
         }
         function_list_value->check_value( LOC );
 
         if( args.size() > 1 ) {
             Shape tag_shape( 2 );
-            Value_P tag( new Value( tag_shape, LOC ) );
+            Value_P tag(tag_shape, LOC);
             new (tag->next_ravel()) IntCell( 0 );
-            new (tag->next_ravel()) PointerCell( make_string_cell( args[1], LOC ) );
+            new (tag->next_ravel()) PointerCell( make_string_cell( args[1], LOC ), tag.getref());
             function_list_value->check_value( LOC );
-            Token result = Quad_FX::fun.eval_AB( tag, function_list_value );
+            Token result = Quad_FX::fun->eval_AB( tag, function_list_value );
             out << "function defined\n" << to_string( result.canonical( PST_CS_NONE ) );
         }
         else {
-            Token result = Quad_FX::fun.eval_B( function_list_value );
+            Token result = Quad_FX::fun->eval_B( function_list_value );
             if( result.is_apl_val() ) {
                 Value_P value = result.get_apl_val();
                 if( value->is_int_scalar( 0 ) ) {
                     out << "error\n"
                         << "parse error\n"
                         << "Error parsing expression\n"
-                        << value->get_ravel( 0 ).get_int_value();
+                        << value->get_ravel( 0 ).get_int_value() - Workspace::get_IO() + 1;
                 }
                 else if( value->is_char_string() ) {
                     out << "function defined\n"
