@@ -3,7 +3,7 @@
 #define Makefile__bindir     "/usr/bin"
 #define Makefile__docdir     "/usr/share/doc/apl"
 #define Makefile__sysconfdir "/etc"
-#define Makefile__pkglibdir  "/usr/lib64/apl"
+#define Makefile__pkglibdir  "/usr/lib/apl"
 #define Makefile__localedir  "/usr/share/locale"
-#define Makefile__srcdir     "/home/eedjsa/apl/apl-1.4/src"
+#define Makefile__srcdir     "/home/eedjsa/projects/juergen/apl-1.5/src"
 
